/*
 * ActionLoader.java
 *
 * Created on May 8, 2003, 10:39 PM
 */

package edu.byu.deg.OntologyEditor;

import java.awt.event.ActionEvent;
import java.io.IOException;
import java.net.URL;
import java.util.Map;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ImageIcon;
import javax.swing.KeyStroke;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 *
 * @author  Alan Wessman
 */
public class ActionLoader
{
    public static final String ACTION_ELEMENT_NAME = "Action";
    public static final String ACTION_GROUP_ELEMENT_NAME = "ActionGroup";
    public static final String ACTION_REF_ELEMENT_NAME = "ActionRef";
    
    public static final String ACTION_KEY_ATTR_NAME = "key";
    public static final String NAME_ATTR_NAME = "name";
    public static final String ACTION_MNEMONIC_ATTR_NAME = "mnemonic";
    public static final String ACTION_ACCELERATOR_ATTR_NAME = "accelerator";
    public static final String ACTION_CLASS_ATTR_NAME = "class";
    public static final String ACTION_SHORT_DESC_ATTR_NAME = "short-description";
    public static final String ACTION_ICON_ATTR_NAME = "small-icon";
    public static final String ACTION_TOGGLE_ATTR_NAME = "toggle";
    public static final String ACCELERATOR_MODIFIER_DELIMITER = "-";
    public static final String CTRL_MODIFIER = "Ctrl";
    public static final String ALT_MODIFIER = "Alt";
    public static final String SHIFT_MODIFIER = "Shift";
    
    public static final String ACTION_TOGGLE_STATE_KEY = "toggleState";
    public static final String ACTION_TOGGLE_STATE_ON  = "on";
    public static final String ACTION_TOGGLE_STATE_OFF = "off";
    
    private static final String ACTION_ATTR_NAME = "action";
    
    public static final String EDITOR_KEY_NAME = "editor";
    public static final String MENU_ITEM_KEY_NAME = "menuItem";
    
    /** Creates a new instance of ActionLoader */
    protected ActionLoader(){ /* protected to avoid public instantiation */ }
    
    private static void loadActions(OntologyEditor editor, NodeList actionList, Map actionMap)
    {   
        // read the action elements and create Actions from them
        for (int i = 0; i < actionList.getLength(); i++)
        {
            Node n = actionList.item(i);
            NamedNodeMap attrs = n.getAttributes();
            Attr className = (Attr) attrs.getNamedItem(ACTION_CLASS_ATTR_NAME);
            Attr key = (Attr) attrs.getNamedItem(ACTION_KEY_ATTR_NAME);
            Attr name = (Attr) attrs.getNamedItem(NAME_ATTR_NAME);
            Attr mnemonic = (Attr) attrs.getNamedItem(ACTION_MNEMONIC_ATTR_NAME);
            Attr shortDescription = (Attr) attrs.getNamedItem(ACTION_SHORT_DESC_ATTR_NAME);
            Attr icon = (Attr) attrs.getNamedItem(ACTION_ICON_ATTR_NAME);
            Attr accelerator = (Attr) attrs.getNamedItem(ACTION_ACCELERATOR_ATTR_NAME);
            Attr toggle = (Attr) attrs.getNamedItem(ACTION_TOGGLE_ATTR_NAME);

            if (key != null)
            {
              Action action;
              Class actionClass;
              
              if (className != null) {
                try
                {
                    actionClass = Class.forName(className.getValue());
                    action = (Action) actionClass.newInstance();
                }
                catch (ClassNotFoundException e)
                { 
                  action = new AbstractAction() { 
                    public void actionPerformed(ActionEvent e) { /* do nothing */ }
                    public boolean isEnabled() { return false; }
                  };
                    
                  actionClass = new ActionLoader().getClass();
                  
                  action.setEnabled(false);
                } // just create a dummy action if it's unknown
                catch (IllegalAccessException e)
                { 
                  action = new AbstractAction() { 
                    public void actionPerformed(ActionEvent e) { /* do nothing */ } 
                  };
                    
                  actionClass = new ActionLoader().getClass();
                  
                  action.setEnabled(false);
                } // create dummy action if instantiation exception
                catch (InstantiationException e)
                { 
                  action = new AbstractAction() { 
                    public void actionPerformed(ActionEvent e) { /* do nothing */ } 
                  };
                    
                  actionClass = new ActionLoader().getClass();
                  
                  action.setEnabled(false);
                } // create dummy action if trouble instantiating the action
              }
              else {
                action = new AbstractAction() { 
                  public void actionPerformed(ActionEvent e) { /* do nothing */ } 
                };
                  
                // we also need a Class object to load resources from
                // so, just instantiate this ActionLoader for this purpose only
                actionClass = new ActionLoader().getClass();
                
                action.setEnabled(false);
              } // null check on className
              
              action.putValue(EDITOR_KEY_NAME, editor);
              
              action.putValue(Action.ACTION_COMMAND_KEY, key.getValue());
              action.putValue(ACTION_KEY_ATTR_NAME, key.getValue());
              if (name != null) 
                action.putValue(Action.NAME, name.getValue());
              if (shortDescription != null) 
                action.putValue(Action.SHORT_DESCRIPTION, shortDescription.getValue());
              if (icon != null) {
                URL imageURL = actionClass.getResource(icon.getValue());
                ImageIcon actionIcon = null;
                if (imageURL != null) actionIcon = new ImageIcon(actionClass.getResource(icon.getValue()));
                if (actionIcon != null) action.putValue(Action.SMALL_ICON, actionIcon);
              }
              if (mnemonic != null) {
                KeyStroke ks = KeyStroke.getKeyStroke(mnemonic.getValue().charAt(0));
                action.putValue(Action.MNEMONIC_KEY, new Integer(ks.getKeyCode()));
              }
              if (accelerator != null)
                action.putValue(ACTION_ACCELERATOR_ATTR_NAME, accelerator.getValue());
              if (toggle != null) {
                action.putValue(ACTION_TOGGLE_ATTR_NAME, toggle.getValue());
                action.putValue(ACTION_TOGGLE_STATE_KEY, ACTION_TOGGLE_STATE_OFF);
              }
              actionMap.put(key.getValue(), action);
            } // null check on key
        } // for loop on actionList
    } // method loadActions
    
    private static void loadActionGroups(NodeList actionGroupList, Map actionMap, Map actionGroupMap)
    {
        // read the actionGroups
        for (int i = 0; i < actionGroupList.getLength(); i++)
        {
            Node n = actionGroupList.item(i);
            NamedNodeMap attrs = n.getAttributes();
            NodeList actionRefList = n.getChildNodes();

            if (actionRefList.getLength() > 0) 
            {
                String groupName = ((Attr) attrs.getNamedItem(NAME_ATTR_NAME)).getValue();
                ActionGroup actionGroup = new ActionGroup();
                actionGroupMap.put(groupName, actionGroup);

                for (int j = 0; j < actionRefList.getLength(); j++)
                {
                  try
                  {
                    Node actionRef = actionRefList.item(j);
                    String actionRefValue = ((Attr) actionRef.getAttributes().getNamedItem(ACTION_ATTR_NAME)).getValue();

                    // make sure it's an action we know about; ignore if not
                    if (actionMap.containsKey(actionRefValue))
                    {
                        actionGroup.add(actionMap.get(actionRefValue));
                    } // integrity check on action ref
                  }
                  catch (NullPointerException npe) { /* REVISIT: handle better? */ } // skip to the next
                } // for loop over actionRefList
            } // length check on actionRefList
        } // for loop over actionGroupList
    } // method loadActionGroups
    
    public static void load(OntologyEditor editor, URL resourceFile, Map actionMap, Map actionGroupMap) throws IOException, SAXException
    {
        if (resourceFile == null) return;
        
        try
        {
            // parse the file
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Document doc = db.parse(resourceFile.toString());

            NodeList actionList = doc.getElementsByTagName(ACTION_ELEMENT_NAME);
            NodeList actionGroupList = doc.getElementsByTagName(ACTION_GROUP_ELEMENT_NAME);
            
            loadActions(editor, actionList, actionMap);
            loadActionGroups(actionGroupList, actionMap, actionGroupMap);

        } // try block
        catch (ParserConfigurationException pce) 
        { /* REVISIT: handle better? */ }
    } // method load
    
}
