package edu.byu.deg.OntologyEditor;
import edu.byu.deg.osmx.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.JTextArea;
import javax.swing.text.Caret;
import javax.swing.text.DefaultCaret;

/**
 * <code>EditableText</code> is a simple edit control derived
 * from <code>JTextArea</code>. It is used by all derivations
 * of <code>DrawObj</code> that require text to be displayed
 * and edited.
 * <BR><BR>
 * This control modifies the behavior of JTextArea by disabling
 * the control unless it is double-clicked.
 *
 * @author Kimball Hewett
 */
public class EditableText
extends     JTextArea
implements  FocusListener, MouseListener, MouseMotionListener, TextChangeListener
{
    
    /**
     * This boolean member variable maintains the state of
     * whether or not the control is in edit mode.
     *
     * @see #isEditingText()
     */
    private boolean m_bEditing = false;
    private Color   m_selectColor;
    private Color   m_selectTextColor;
    private int     m_caretWidth = 1;

    /**
     * When the control loses focus this method will take
     * the control out of edit mode by calling editText
     * (false).
     *
     * @see #editText(boolean)
     */
    public EditableText()
    {
        initialize();
    }
    
    /**
     * EditableText constructor
     *
     * This method like the default constructor calls
     * <code>initialize()</code> to
     * prepare the control for display.
     *
     * @param text The edit control is initialized with the String
     * <code>text</code>
     * @see #initialize()
     */
    public EditableText(String text)
    {
        super(text);
        initialize();
    }
    
    /**
     * This method initializes the control for display.
     * It sets up the control as a <code>FocusListener</code>,
     * <code>MouseListener</code> and
     * <code>MouseMotionListener</code>.
     * <BR><BR>
     * By default the <code>JTextArea</code> control would
     * change the colors of the text when it is disabled.
     * This is inappropriate for the use in the
     * <code>OntologyEditor</code> so the initialize method
     * also sets the disabled colors to be the same as the
     * enabled colors.
     */
    private void initialize()
    {
        addFocusListener(this);
        addMouseListener(this);
        addMouseMotionListener(this);
        setDisabledTextColor(getForeground());
        m_selectColor = getSelectionColor();
        m_selectTextColor = getSelectedTextColor();
        setSelectionColor(getBackground());
        setSelectedTextColor(getForeground());
        setEnabled(false);
        setEditable(false);
        Caret c = getCaret();
        if (c instanceof DefaultCaret) {
            DefaultCaret dc = (DefaultCaret)c;
            if (dc.getWidth() > m_caretWidth) {
                m_caretWidth = (int)(dc.getWidth() + 0.5);
            }
        }
    }
    
    /**
     * This method may be called to determine if the control is
     * currently in editing mode.
     *
     * @return true if editing
     */
    public boolean isEditingText()
    {
        return m_bEditing;
    }
    
    /**
     * This method is called to place the control in either
     * edit or non-edit mode. If in edit mode it will enable
     * editing and will display a caret in the control.
     * <BR><BR>
     * <code>editText(boolean)</code> is typically called when
     * double-clicked or when the control loses focus.
     *
     * @param bEdit bEdit is true if the control is to be placed
     * in edit mode.<BR>
     * bEdit is false if the control is to be placed in non-edit
     * mode.
     * @see #focusLost
     * @see #mouseClicked
     */
    public void editText(boolean bEdit)
    {
        m_bEditing = bEdit;
        setEnabled(m_bEditing);
        setEditable(m_bEditing);
        if (m_bEditing) {
            // We are asking to edit this EditableText.
            requestFocusInWindow();
            setSelectionColor(m_selectColor);
            setSelectedTextColor(m_selectTextColor);
            select(0, getText().length());
        } else {
            // We are asking to stop editing this EditableText.
            // Text selected upon lost focus?  Unselect it.
            setSelectionColor(getBackground());
            setSelectedTextColor(getForeground());
            select(0, 0);
        }
    }
    
    
    public void focusGained(FocusEvent evt){ /* do nothing */ }
    
    /**
     * When the control loses focus this method will take the
     * control out of edit mode by calling
     * <code>editText(false)</code>.
     *
     * @see #editText
     */
    public void focusLost(FocusEvent evt)
    {
      editText(false);

      // Since containers will resize, let's leave something in
      // the control if it is empty
      if (getText().length() == 0)
          setText(" ");
      /*
       * NEEDSWORK: A better solution might be to let it be empty, but
       * let the container's bounds be used for enabling editing.
       */
    }

    public Dimension getPreferredSize()
    {
       Dimension textSize = super.getPreferredSize();
       textSize.setSize(textSize.getWidth() + m_caretWidth, textSize.getHeight());
       return (textSize);
    }
    
    /**
     * This method checks to see if the control has been
     * double-clicked. If so and the control is not in edit mode,
     * then the control is placed in edit mode by calling
     * <code>editText(true)</code>.
     *
     * @param evt evt provides the number of mouse clicks
     * @see #editText
     */
    public void mouseClicked(MouseEvent evt)
    {
      if (!m_bEditing && evt.getClickCount() > 1)
      {
          editText(true);
      }
    }
    public void mouseEntered(MouseEvent evt){ /* do nothing */ }
    public void mouseExited(MouseEvent evt){ /* do nothing */ }
    public void mousePressed(MouseEvent evt){ /* do nothing */ }
    public void mouseReleased(MouseEvent evt){ /* do nothing */ }
    public void mouseDragged(MouseEvent evt){ /* do nothing */ }
    public void mouseMoved(MouseEvent evt){ /* do nothing */ }

    public void setForeground(Color c)
    {
        super.setForeground(c);
        setDisabledTextColor(c);
    }

    public void setBackground(Color c)
    {
        super.setBackground(c);

        int r = c.getRed(),
            g = c.getGreen(),
            b = c.getBlue();

        if (r == g && g == b && b >= 255 && false) {
            setOpaque(false);
        } else {
            setOpaque(true);
        }
    }
    public void textChanged(String newText) {
  	   System.out.println("TextChanged::: "+newText);
    } // 
    

}
