/*
 * MenuLoader.java
 *
 * Created on May 9, 2003, 3:50 PM
 */

package edu.byu.deg.OntologyEditor;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.util.Map;
import javax.swing.Action;
import javax.swing.ButtonGroup;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.MenuElement;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;


/**
 *
 * @author  Alan Wessman
 */
public class MenuLoader {
  private static final String MENU_ITEM_ELEM_NAME = "MenuItem";
  private static final String SEPARATOR_ELEM_NAME = "Separator";
  private static final String MENU_ELEM_NAME = "Menu";
  private static final String POPUP_MENU_ELEM_NAME = "PopupMenu";
  
  private static final String NAME_ATTR_NAME = "name";
  private static final String MENU_MNEMONIC_ATTR_NAME = "mnemonic";
  private static final String ACTION_ATTR_NAME = "action";
  private static final String MUTEX_GROUP_ATTR_NAME = "mutexGroup";
  private static final String TOGGLE_ATTR_NAME = "toggle";
  private static final String KEY_ATTR_NAME = "key";
  private static final String DEFAULT_ATTR_NAME = "default";
  private static final String COMMAND_ATTR_NAME = "command";
  private static final String TOOLTIP_ATTR_NAME = "tooltip";
  
  private static final String DEFAULT_ACTION_NAME = "OntosAction";
  
  /** Creates a new instance of MenuLoader */
  protected MenuLoader() { /* protected to avoid instantiation */ }
  
  public static void loadMenus(URL menuResourceFile, Map menuMap, Map menuItemGroups, Map actions) throws IOException, SAXException {
    if (menuResourceFile == null) {
      throw new FileNotFoundException("Menu resource file not found");
    }
    
    try {
      NodeList menuNodes = getNodesFromResource(menuResourceFile);
      
      for (int i = 0; i < menuNodes.getLength(); i++) {
        Node n = menuNodes.item(i);
        if (n.getNodeType() != Node.ELEMENT_NODE) continue;
        if (n.getNodeName().equals(MENU_ELEM_NAME) || n.getNodeName().equals(POPUP_MENU_ELEM_NAME)) {
          NamedNodeMap attrs = n.getAttributes();
          String menuKey = getAttrValue(attrs, KEY_ATTR_NAME);
          if (menuKey == null) menuKey = getAttrValue(attrs, NAME_ATTR_NAME);
          if (menuKey == null) {
            System.err.println("WARNING: No key found for node " + n.getNodeName());
            continue; // skip this menu if there's no map key for it
          }
          MenuElement menuElem = createMenuItem(n, menuItemGroups, actions);
          if (menuElem != null) menuMap.put(menuKey, menuElem);
        } // check on child node == Menu element
      } // for loop on menuNodes
      
    } // try block
    catch (ParserConfigurationException pce)
    { pce.printStackTrace(); } 
    
  } // method loadMenus
  
  private static String getAttrValue(NamedNodeMap attrs, String attrName) {
    Attr attr = (Attr) attrs.getNamedItem(attrName);
    if (attr == null) return null;
    else return attr.getValue();
  }
  
  private static NodeList getNodesFromResource(URL resource) throws ParserConfigurationException, SAXException, IOException {
    DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
    DocumentBuilder db = dbf.newDocumentBuilder();
    Document doc = db.parse(resource.toString());

    return doc.getDocumentElement().getChildNodes();
  }
  
  private static MenuElement createMenu(NamedNodeMap attrs, NodeList childNodes, Map menuItemGroups, Map actions) {
    JMenu result = new JMenu();
    String mnemonicValue = getAttrValue(attrs, MENU_MNEMONIC_ATTR_NAME);
    if (mnemonicValue.length() > 0) result.setMnemonic(mnemonicValue.charAt(0));
    
    String menuName = getAttrValue(attrs, NAME_ATTR_NAME);
    if (menuName != null) {
      result.setText(menuName);
    }

    for (int i = 0; i < childNodes.getLength(); i++) {
      Node n = childNodes.item(i);
      if (n.getNodeType() != Node.ELEMENT_NODE) continue;

      if (n.getNodeName().equals(SEPARATOR_ELEM_NAME)) {
        result.addSeparator();
      }
      else if (n.getNodeName().equals(POPUP_MENU_ELEM_NAME)) {
        // do nothing; shouldn't have a PopupMenu element here
      }
      else { // the child node is not a Separator element
        JMenuItem newItem = (JMenuItem) createMenuItem(n, menuItemGroups, actions);
        if (newItem != null) result.add(newItem);
      }
    } // for loop over child nodes
    return result;
  } // method createMenu
  
  private static MenuElement createPopupMenu(NodeList childNodes, Map menuItemGroups, Map actions) {
    JPopupMenu result = new JPopupMenu();

    for (int i = 0; i < childNodes.getLength(); i++) {
      Node n = childNodes.item(i);
      if (n.getNodeName().equals(SEPARATOR_ELEM_NAME)) {
        result.addSeparator();
      }
      else if (n.getNodeName().equals(POPUP_MENU_ELEM_NAME)) {
        // do nothing; shouldn't have a PopupMenu element here
      }
      else { // the child node is not a Separator element
        JMenuItem newItem = (JMenuItem) createMenuItem(n, menuItemGroups, actions);
        if (newItem != null) {
          result.add(newItem);
        }
      }
    } // for loop over child nodes
    return result;
  } // method createPopupMenu
  
  private static MenuElement createLeafMenuItem(NamedNodeMap attrs, Map menuItemGroups, Map actions) {
    JMenuItem result = null;
    
    String actionClassName = getAttrValue(attrs, ACTION_ATTR_NAME);
    if (actionClassName == null) actionClassName = DEFAULT_ACTION_NAME;
    
    Action a = (Action) actions.get(actionClassName);
    
    if (a == null) {
      System.err.println("WARNING: Could not find action " + actionClassName);
      return null;
    }
    
    String actionKey = getAttrValue(attrs, KEY_ATTR_NAME);
    if (actionKey == null) actionKey = (String) a.getValue(Action.ACTION_COMMAND_KEY);
    
    String command = getAttrValue(attrs, COMMAND_ATTR_NAME);
    
    String itemName = getAttrValue(attrs, NAME_ATTR_NAME);
    if (itemName == null) itemName = (String) a.getValue(Action.NAME);
    if (itemName == null) {
      System.err.println("WARNING: No action name for " + actionClassName);
      return null;
    }
    
    String tooltip = getAttrValue(attrs, TOOLTIP_ATTR_NAME);
    if (tooltip == null) tooltip = (String) a.getValue(Action.SHORT_DESCRIPTION);
    
    String mutexGroup = getAttrValue(attrs, MUTEX_GROUP_ATTR_NAME);
    boolean isMutex = (mutexGroup != null);
    boolean isToggle = isMutex;
    String toggle = getAttrValue(attrs, TOGGLE_ATTR_NAME);
    isToggle = isToggle || (toggle != null && "true".equalsIgnoreCase(toggle));
    if (!isToggle) {
      toggle = (String) a.getValue(ActionLoader.ACTION_TOGGLE_ATTR_NAME);
      isToggle = (toggle != null && "true".equalsIgnoreCase(toggle));
    }
    
    if (isToggle) {
      result = new ToggleMenuItem(a);
    } else {
      result = new JMenuItem(a);
    }

    String mnemonic = getAttrValue(attrs, MENU_MNEMONIC_ATTR_NAME);
    if (mnemonic == null) mnemonic = (String) a.getValue(ActionLoader.ACTION_MNEMONIC_ATTR_NAME);

    result.setText(itemName);
    result.setToolTipText(tooltip);
    
    if (isToggle) {
      result.setActionCommand(command);
      if (isMutex) {
        ((ToggleMenuItem) result).setMutex(true);
        ButtonGroup itemGroup = (ButtonGroup) menuItemGroups.get(mutexGroup);
        if (itemGroup == null) {
          itemGroup = new ButtonGroup();
          menuItemGroups.put(mutexGroup, itemGroup);
        } // check on prior existence of the specified mutexGroup
        itemGroup.add(result);
        String defaultItem = getAttrValue(attrs, DEFAULT_ATTR_NAME);
        if (defaultItem != null && "true".equalsIgnoreCase(defaultItem)) {
          ((ToggleMenuItem) result).setDefault(true);
        }
      } // check on mutexGroup attribute
    } // check on toggle attribute
    
    return result;
  } // method createLeafMenuItem
  
  private static MenuElement createMenuItem(Node menuItemNode, Map menuItemGroups, Map actions) {
    String nodeName = menuItemNode.getNodeName();
    MenuElement result;
    NodeList childNodes = menuItemNode.getChildNodes();
    NamedNodeMap attrs = menuItemNode.getAttributes();
    
    if (nodeName.equals(MENU_ELEM_NAME)) {
      // iterate thru children, calling this method recursively if needed
      result = createMenu(attrs, childNodes, menuItemGroups, actions);
    } // node is Menu element
    else if (nodeName.equals(POPUP_MENU_ELEM_NAME)) {
      // like above case
      result = createPopupMenu(childNodes, menuItemGroups, actions);
    } // node is PopupMenu element
    else if (nodeName.equals(MENU_ITEM_ELEM_NAME)) {
      result = createLeafMenuItem(attrs, menuItemGroups, actions);
    } // node is MenuItem element
    else result = null;
    
    return result;
  } // method createMenuItem
  
}
