package edu.byu.deg.OntologyEditor;

import edu.byu.deg.OntologyEditor.epsgraphics.EpsGraphics2D;
import edu.byu.deg.OntologyEditor.shapes.*;
import edu.byu.deg.osmx.binding.impl.AggregationImpl;
import edu.byu.deg.osmx.binding.impl.ConjunctionImpl;
import edu.byu.deg.osmx.binding.impl.GenSpecImpl;
import edu.byu.deg.osmx.binding.impl.RelationshipSetImpl;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import edu.byu.deg.osmx.*;
import edu.byu.deg.osmx.binding.*;

import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;

import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import java.util.Vector;

/** Represents the "canvas" on which the visual ontology elements appear.  Handles
 * mouse and keyboard events affecting the shapes on the canvas.  Directly or
 * indirectly contains all {@link edu.byu.deg.OntologyEditor.shapes.DrawShape}s
 * that represent the ontology.
 * @author Kimball Hewitt, Alan Wessman
 */
public class OntologyCanvas extends JPanel
implements Canvas, KeyListener, MouseListener, MouseMotionListener,
DocumentChangeListener, ItemListener {

    private OSMXDocument    m_doc;
    private boolean         m_bDragging = false;
    private boolean         m_bSelecting = false;
    private boolean         m_bResizing = false;
    private Point           m_initialPoint, m_oldPoint;
    private DrawShape       m_deselectionCandidate;

    private OntologyEditor  m_editor;
    private OntologyCanvasWindow m_parentWindow;
    private DragPanel dragPanel;
    private JPanel selectionDragger;

    private Set selectedShapes;
    private Map shapesByElement;

    private DrawShape srcShape = null;
    private DrawShape destShape = null;
    private DrawShape activeShape = null;
    private MouseEvent activeMouseEvent = null;

    private final int       DPI = 96;
    private final double    PAGEWIDTH = 6.5;
    private final double    PAGEHEIGHT = 9;
    private final int       PAGESACROSS = 6;
    private final int       PAGESDOWN = 4;
    private double x_scale = 1.0, y_scale = 1.0;
    
    private static final boolean SELECT_INCLUSIVE = true;

    public static final String STATE_SHAPE_KEY = "StateAction";
    public static final String STATE_CONNECTION_KEY = "ConjunctionAction";
    public static final String TRANSITION_SHAPE_KEY = "TransitionAction";
    
    public static final String SELECTION_SHAPE_KEY = "SelectAction";
    public static final String GENERAL_CONSTRAINT_SHAPE_KEY = "ConstraintAction";
    public static final String NOTE_SHAPE_KEY = "NoteAction";
    public static final String OBJECT_SET_SHAPE_KEY = "ObjectSetAction";
    public static final String OBJECT_SHAPE_KEY = "ObjectAction";
    public static final String RELATIONSHIP_SET_SHAPE_KEY = "RelationshipSetAction";
    public static final String GEN_SPEC_SHAPE_KEY = "SpecializationAction";

    private static final String OBJECT_SET_POPUP_KEY = "ObjectSet";
    private static final String REL_SET_CONN_POPUP_KEY = "RelSetConnection";
    private static final String GEN_SPEC_CONN_POPUP_KEY = "GenSpecConnection";
    private static final String GEN_SPEC_POPUP_KEY = "GenSpec";
    private static final String RELATIONSHIP_SET_POPUP_KEY = "RelationshipSet";

    private static final String AGGREGATION_SHAPE_KEY = "AggregationAction";
    private static final String AGGREGATION_POPUP_KEY = "Aggregation";
    private static final String CHILD_REL_SET_CONN_POPUP_KEY = "ChildRelSetConnection";
    private static final String PARENT_REL_SET_CONN_POPUP_KEY = "ParentRelSetConnection";
    
    /** Initializes the canvas.
     * @param parentWindow The editor window that contains this canvas.
     */
    OntologyCanvas(OntologyCanvasWindow parentWindow) {
        dragPanel = new OntologyCanvas.DragPanel();
        dragPanel.setDragMode(DragPanel.SELECT_BOX_MODE); 
        dragPanel.setOpaque(false);
        dragPanel.addMouseListener(this);
        dragPanel.addMouseMotionListener(this);

        selectionDragger = new JPanel();
        selectionDragger.setOpaque(false);
        selectionDragger.addMouseListener(this);
        selectionDragger.addMouseMotionListener(this);

        m_parentWindow = parentWindow;
        m_doc = m_parentWindow.getDocument();
        m_doc.addDocumentChangeListener(this);

        selectedShapes = new HashSet();
        shapesByElement = new HashMap();

        setLayout(null);
        setBorder(null);
        setBackground(Color.WHITE);
        setForeground(Color.WHITE);

        // enable mouse events
        addMouseListener(this); // addMouseListener call

        addMouseMotionListener(this); // addMouseMotionListener call

        addKeyListener(this); // addKeyListener call

        m_editor = parentWindow.getEditor();

        float dash[] = new float[2];
        dash[0] = 3.0f;
        dash[1] = 3.0f;
    }

    /** Returns the set of currently selected shapes.
     * @return A set containing the currently selected shapes.  The set will be
     * non-<CODE>null</CODE> but empty if no shapes are currently selected.
     */
    public Set getSelectedShapes() {
        return selectedShapes;
    }

    /** Returns the shape that is currently active.  The active shape is the one being
     * targeted by some event (such as a mouse click) as an indication that some action
     * should be taken for that shape (such as showing a popup menu).
     * @return The currently active shape, or <CODE>null</CODE> if no shape is active.
     */
    public DrawShape getActiveShape() {
        return activeShape;
    }

    /** Returns the <CODE>MouseEvent</CODE> that caused the currently active shape to
     * become active.
     * @return The <CODE>MouseEvent</CODE> that activated the currently active shape.
     */
    public MouseEvent getActiveMouseEvent() {
        return activeMouseEvent;
    }

    /** Deletes all shapes in the selection, clears the selection, and re-draws the
     * canvas.
     */
    public void deleteSelected() {
        // delete selected objects
        Set selShapes = getSelectedShapes();
        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            shape.delete();
        }
        clearSelection();
        invalidate();
        repaint();
    } // method deleteSelected

    /** Moves the specified component to the front, so that to the user it appears
     * closer than the other components.
     * @param comp The component to move to the front.
     */
    public void toFront( Component comp)  {
        super.remove(comp);
        add(comp,0);
    }

    /** Moves the current selection to the front.
     * @see #toFront(java.awt.Component)
     */
    public void toFront() {
        Set selShapes = getSelectedShapes();
        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            super.remove(shape);
            super.add(shape,0);
        }
        super.repaint();
    }

    /** Moves the specified component to the back, so that to the user it appears
     * farther away than the other components.
     * @param comp The component to move to the back.
     */
    public void toBack(Component comp) {
        super.remove(comp);
        add(comp);
    }

    /** Moves the current selection to the back.
     * @see #toBack(java.awt.Component)
     */
    public void toBack() {
        Set selShapes = getSelectedShapes();
        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            super.remove(shape);
            super.add(shape);
        }
        super.repaint();
    }

    /** Returns the canvas' preferred size.
     * @return The preferred size of the canvas.
     */
    public Dimension getPreferredSize() {
        return new Dimension((int)(PAGESACROSS*PAGEWIDTH*DPI), (int)(PAGESDOWN*PAGEHEIGHT*DPI));
    }

    /** Returns the <CODE>OSMXDocument</CODE> that contains the ontology.
     * @return The <CODE>OSMXDocument</CODE> that forms the underlying model for this canvas.
     */
    public  OSMXDocument getMacroRootDoc() {
        return m_doc;
    }// Get the loding OSMXDocument

    /** Translates the selected objects by the specified horizontal and vertical
     * offsets.
     * @param x The distance to translate the selection horizontally (east is positive).
     * @param y The distance to translate the selection vertically (south is positive).
     */
    public void dragSelectedObjects(int x, int y) {
        for (Iterator i = selectedShapes.iterator(); i.hasNext();) {
            DrawShape ds = (DrawShape) i.next();
            ds.drag(x, y);
        }
        repaint();
    }
    
    public void resizeSelectObjectSetShape(int x, int  y) {
    	for (Iterator i = selectedShapes.iterator(); i.hasNext();) {
    		DrawShape ds = (DrawShape) i.next();
    		if (ds instanceof ObjectSetShape)
    			ds.resizeBorder(x, y);
    	}
    	this.repaint();
    }
    
    public String getResizeDragPoint() {
    	String dragPoint = "";
    	return dragPoint;
    }

    /** Required by the {@link Canvas} interface, but simply returns <CODE>null</CODE>.
     * @return <CODE>null</CODE>.
     */
    public Canvas getParentCanvas() {
        return null;
    }

    /** Causes all shapes within the bounds of the rectangle to become selected, and all
     * shapes outside the bounds of the rectangle to become de-selected.
     * @param bounds The rectangle that determines which shapes will be selected and which will not.
     */
    protected void selectAllWithinRect(Rectangle bounds) {
        Component components[] = getComponents();
        for (int i=0; i<components.length; i++) {
            if (components[i] instanceof DrawShape) {
                DrawShape shape = (DrawShape) components[i];
                shape.setSelected(bounds, SELECT_INCLUSIVE);
                if (shape.isSelected()) selectedShapes.add(shape);
                else selectedShapes.remove(shape);
            }
        }
    }

    /** Aligns all selected shapes on the left, positioned at the leftmost shape's
     * x-coordinate.
     */
    public void alignLeft() {
        int leftposition =  Integer.MAX_VALUE;
        Set selShapes = getSelectedShapes();
        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            if (! (shape instanceof ConnectorShape ||
            shape instanceof ConnectionShape)) {
                Point p = shape.getLeftPoint();
                if (p.x < leftposition)
                    leftposition = p.x;
            }
        }
        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            if (!(shape instanceof ConnectorShape || shape instanceof ConnectionShape)) {
                Point p = shape.getPosition();
                shape.setLocation(leftposition,p.y);
                StyleType st = shape.createStyle();
                if (st != null) {
                   st.setFillColor(Color.blue);
                }
            }
        }

    }

    /** Aligns all selected shapes on the right, positioned at the rightmost shape's
     * right side's x-coordinate.
     */
    public void alignRight() {
        Set selShapes = getSelectedShapes();
        int rightposition = Integer.MIN_VALUE;
        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            if (! (shape instanceof ConnectorShape ||
            shape instanceof ConnectionShape)) {
                Point p = shape.getRightPoint();
                if (p.x > rightposition)
                    rightposition = p.x;
            }
        }
        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            if (!(shape instanceof ConnectorShape || shape instanceof ConnectionShape)) {
                Point p = shape.getPosition();
                shape.setLocation(rightposition-shape.getWidth(),p.y);
            }
        }
    }


    /** Aligns all selected shapes on the top, positioned at the topmost shape's
     * y-coordinate.
     */
    public void alignTop() {
        Set selShapes = getSelectedShapes();
        int topposition = Integer.MAX_VALUE;
        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            if (! (shape instanceof ConnectorShape ||
            shape instanceof ConnectionShape)) {
                Point p = shape.getTopPoint();
                if (p.y < topposition)
                    topposition = p.y;
            }
        }
        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            if (!(shape instanceof ConnectorShape || shape instanceof ConnectionShape)) {
                Point p = shape.getPosition();
                shape.setLocation(p.x,topposition);
            }
        }
    }

    /** Aligns all selected shapes on the bottom, positioned at the bottommost shape's
     * bottom y-coordinate.
     */
    public void alignBottom() {
        int bottomposition = Integer.MIN_VALUE;
        Set selShapes = getSelectedShapes();
        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            if (! (shape instanceof ConnectorShape ||
            shape instanceof ConnectionShape)) {
                Point p = shape.getBottomPoint();
                if (p.y > bottomposition)
                    bottomposition = p.y;
            }
        }
        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            if (!(shape instanceof ConnectorShape || shape instanceof ConnectionShape)) {
                Point p = shape.getPosition();
                shape.setLocation(p.x, bottomposition-shape.getHeight());
            }
        }
    }

    /** Evenly distributes the selected shapes horizontally. */
    public void spaceAcross( ) {
        int leftposition =Integer.MAX_VALUE;
        int rightposition =  Integer.MIN_VALUE;
        int numObject = 0;
        int objectswidth = 0;
        Set selShapes = getSelectedShapes();
        Vector shapeSets =  new Vector();// To save the shapes with the "left to right" sequence

        for (Iterator i = selShapes.iterator(); i.hasNext();){
            DrawShape shape = (DrawShape)i.next();
            if (! (shape instanceof ConnectorShape ||
            shape instanceof ConnectionShape)) {
                Point pLeft =  shape.getLeftPoint();
                Point pRight = shape.getRightPoint();
                objectswidth = objectswidth + shape.getWidth();
                if (pLeft.x <leftposition)
                    leftposition = pLeft.x;
                if  (pRight.x > rightposition)
                    rightposition = pRight.x;
                numObject = numObject + 1;

                int shapeSetindex = shapeSets.size();
                for (int j =0; j < shapeSets.size(); j++){
                    DrawShape tempshape = (DrawShape) shapeSets.get(j);
                    if (shape.getLeftPoint().x < tempshape.getLeftPoint().x) {
                        shapeSetindex = j;
                        break;
                    }
                }
                shapeSets.add(shapeSetindex,shape);
            }
        }

        double space = (rightposition - leftposition + 1 - objectswidth)/(double)(numObject -1);

        //set the shapes' postion
        int nextX = leftposition;
        objectswidth = 0;
        double totalSpace = 0;
        for (int j = 0; j < shapeSets.size(); j++) {
            DrawShape tempshape = (DrawShape)shapeSets.get(j);
            Point p =  tempshape.getPosition();
            tempshape.setLocation(nextX, p.y);
            objectswidth = objectswidth + tempshape.getWidth();
            totalSpace += space;
            nextX = leftposition + objectswidth + (int)((totalSpace<0) ? totalSpace - .001 : totalSpace + .001);
        }
    }

    /** Evenly distributes the selected shapes vertically. */
    public void spaceDown() {
        int topposition =Integer.MAX_VALUE;
        int bottomposition =  Integer.MIN_VALUE;
        int numObject = 0;
        int objectsheight= 0;
        Set selShapes = getSelectedShapes();
        Vector shapeSets =  new Vector();// To save the shapes with the "top to bottom" sequence

        for (Iterator i = selShapes.iterator(); i.hasNext();){
            DrawShape shape = (DrawShape)i.next();
            if (! (shape instanceof ConnectorShape ||
            shape instanceof ConnectionShape)) {
                Point pTop =  shape.getTopPoint();
                Point pBottom = shape.getBottomPoint();
                objectsheight = objectsheight + shape.getHeight();
                if (pTop.y <topposition)
                    topposition = pTop.y;
                if  (pBottom.y > bottomposition)
                    bottomposition = pBottom.y;
                numObject = numObject + 1;

                int shapeSetindex = shapeSets.size();
                for (int j =0; j < shapeSets.size(); j++){
                    DrawShape tempshape = (DrawShape) shapeSets.get(j);
                    if (shape.getTopPoint().y < tempshape.getTopPoint().y) {
                        shapeSetindex = j;
                        break;
                    }
                }
                shapeSets.add(shapeSetindex,shape);
            }
        }

        double space = (bottomposition - topposition + 1 - objectsheight)/(double)(numObject -1);

        //set the shapes' postion
        int nextY = topposition;
        objectsheight = 0;
        double totalSpace = 0;
        for (int j = 0; j < shapeSets.size(); j++) {
            DrawShape tempshape = (DrawShape)shapeSets.get(j);
            Point p =  tempshape.getPosition();
            tempshape.setLocation(p.x, nextY);
            objectsheight = objectsheight + tempshape.getHeight();
            totalSpace += space;
            nextY = topposition + objectsheight + (int)((totalSpace<0) ? totalSpace - .001 : totalSpace + .001);
        }


     /* DrawObj is deprecated, so this code needs to be refactored for DrawShape. AW
     int farTop = Integer.MAX_VALUE;
     int farBottom = Integer.MIN_VALUE;
     int sumHeight = 0;
     int numObjects = 0;
     int i,j;

     DrawObj selObjects[] = canvas.getAllSelected(false);

     if (selObjects.length > 0) {
       int tops[] = new int[selObjects.length];
       // Find the Alignable Objects that are the farthest left and right
       for (i=0; i<selObjects.length; i++) {
         tops[i]=Integer.MAX_VALUE;
         if (selObjects[i].isAlignable()) {
           tops[i] = selObjects[i].getTop();
           if (tops[i] < farTop)
             farTop = tops[i];
           if (selObjects[i].getBottom() > farBottom)
             farBottom = selObjects[i].getBottom();
           numObjects++;
           sumHeight += selObjects[i].getHeight();
         }
       }

       double space = ((farBottom - farTop + 1) - sumHeight)/(double)(numObjects -1);
       double totalSpace = 0;
       int nextIndex=0, min;
       int nextY = farTop;
       sumHeight = 0;

       for (i=0; i<numObjects; i++) {
         // Find the object that is furthest left
         min = Integer.MAX_VALUE;
         for (j=0; j<selObjects.length; j++)
           if (min > tops[j]) {
             min = tops[j];
             nextIndex = j;
           }
         // Move the object
         selObjects[nextIndex].alignOffset(0, nextY - tops[nextIndex]);

         // Adjust
         totalSpace += space;
         sumHeight += selObjects[nextIndex].getHeight();
         nextY = farTop + sumHeight + (int)((totalSpace<0) ? totalSpace - .001 : totalSpace + .001);
         tops[nextIndex] = Integer.MAX_VALUE;
       }
     }

     Component components[] = canvas.getComponents();

     for (i=0; i<components.length; i++)
       if (components[i] instanceof Canvas)
         spaceDown((Canvas)components[i]);
      */
    }


    /** Aligns the selected shapes so that their center points share the same
     * x-coordinate.
     */
    public void alignVertCenter() {
        Set selShapes = getSelectedShapes();
        int farleftposition = Integer.MAX_VALUE;
        int farrightposition = Integer.MIN_VALUE;
        int center;
        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            if (! (shape instanceof ConnectorShape ||
            shape instanceof ConnectionShape)) {
                Point p = shape.getAbsoluteCenterPoint();
                if (p.x < farleftposition)
                    farleftposition = p.x;
                if  (p.x > farrightposition)
                    farrightposition = p.x;
            }
        }
        center = (farleftposition + farrightposition + 1)/2;

        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            if (!(shape instanceof ConnectorShape || shape instanceof ConnectionShape)) {
                Point p = shape.getPosition();
                shape.setLocation(center - shape.getWidth()/2, p.y);
            }
        }
    }

    /** Aligns the selected shapes so that their center points share the same
     * y-coordinates.
     */
    public void alignHorCenter() {
        Set selShapes = getSelectedShapes();
        int fartopposition = Integer.MAX_VALUE;
        int farbottomposition = Integer.MIN_VALUE;
        int center;
        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            if (! (shape instanceof ConnectorShape ||
            shape instanceof ConnectionShape)) {
                Point p = shape.getAbsoluteCenterPoint();
                if (p.y < fartopposition)
                    fartopposition = p.y;
                if  (p.y > farbottomposition)
                    farbottomposition = p.y;
            }
        }
        center = (fartopposition + farbottomposition + 1)/2;

        for (Iterator i = selShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            if (!(shape instanceof ConnectorShape || shape instanceof ConnectionShape)) {
                Point p = shape.getPosition();
                shape.setLocation(p.x, center - shape.getHeight()/2);
            }
        }
    }

    /** Locates and returns the shape that contains the specified point, or
     * <CODE>null</CODE> if no such shape exists.
     * @param pt The coordinate at which we seek for a shape.
     * @return The {@link edu.byu.deg.OntologyEditor.shapes.DrawShape} at the specified point, or
     * <CODE>null</CODE> if no such shape exists at that location.
     */
    public Component getComponentAt(Point pt) {
        Component comp = super.getComponentAt(pt);
        if (comp == this)
            comp = null;

        if (comp instanceof Canvas && !(comp instanceof OntologyCanvas)) {
            Point origin = ((Canvas)comp).getOrigin();
            comp = comp.getComponentAt(new Point(pt.x-origin.x,pt.y-origin.y));
        }
        return comp;
    }

    /** Returns the origin point of the component.  In this case, it is defined to be
     * the coordinate (0, 0).
     * @return The canvas' origin point.
     */
    public Point getOrigin() {
        return new Point(0,0);
    }
    
    public void canvasPrint()  {
        Rectangle bounds = null;
        Rectangle shapeBounds = null;
        Object value = null;
        
        //get the graphic area size
        for (Iterator it = shapesByElement.keySet().iterator(); it.hasNext(); ) {
            value = shapesByElement.get(it.next());
            if (value != null && value instanceof DrawShape) {
                shapeBounds = ((DrawShape)value).getBoundingBox();
                if (bounds == null) {
                    bounds = shapeBounds;
                } else {
                    bounds = bounds.union(shapeBounds);
                }
            }
        }
        		
        PrintUtilities.printComponent(this, bounds);
    }
    

    
    public void setXScale(double x) {
    	x_scale = x;
    }
    
    public double getXScale() {
    	return x_scale;
    }
    public void setYScale(double y) {
    	y_scale = y;
    }
    public double getYScale(){
    	return y_scale;
    }
    
    /** Exports the ontology diagram in encapsulated PostScript format to the specified
     * file.
     * @param file The destination file.
     * @param isPS Obsolete parameter; does nothing.
     * @throws IOException Thrown if anything goes wrong with writing to the specified file.
     * @return <CODE>true</CODE> if the method completes.
     */
    public boolean exportDocument(File file, boolean isPS) throws IOException {
        file.delete(); // clear out old contents

        FileOutputStream output = new FileOutputStream(file);

/*
     BaseExporter exporter;
     exporter = new EPSExporter();

     print(exporter);
     //writeFooter will flush and close the stream
     OutputStreamWriter writer = new OutputStreamWriter(output);
     writer.write(exporter.toString());
     writer.close();
     return true;
/*/
        Rectangle bounds = null;
        Rectangle shapeBounds = null;
        Object value = null;

        for (Iterator it = shapesByElement.keySet().iterator(); it.hasNext(); ) {
            value = shapesByElement.get(it.next());
            if (value != null && value instanceof DrawShape) {
                shapeBounds = ((DrawShape)value).getBoundingBox();
                if (bounds == null) {
                    bounds = shapeBounds;
                } else {
                    bounds = bounds.union(shapeBounds);
                }
            }
        }

        Graphics2D exporter;
        if (bounds == null) {
            exporter = new EpsGraphics2D(file.getName());
        } else {
            exporter = new EpsGraphics2D(file.getName(), file,
            bounds.x - 1, bounds.y - 1,
            bounds.x + bounds.width + 2,
            bounds.y + bounds.height + 2);
        }

        ((EpsGraphics2D)exporter).setAccurateTextMode(false);
        ((EpsGraphics2D)exporter).setUsesClip(false);

     /*
     ((EpsGraphics2D)exporter).draw(bounds);
     for (Iterator it = shapesByElement.keySet().iterator(); it.hasNext(); ) {
        value = shapesByElement.get(it.next());
        if (value != null && value instanceof DrawShape) {
            if (value instanceof RelationshipSetShape) {
                shapeBounds = ((DrawShape)value).getBoundingBox();
                ((EpsGraphics2D)exporter).draw(shapeBounds);
            }
        }
     }
      */
        
        print(exporter);
        OutputStreamWriter writer = new OutputStreamWriter(output);
        writer.write(exporter.toString());
        writer.close();
        return true;
        //*/
    }

    /** Returns the rectangle that represents the union of the selected shapes' bounding
     * rectangles.
     * @return The rectangle that tightly bounds the selected shapes.
     */
    public Rectangle getSelectionBoundingBox() {
        if (selectedShapes.size() == 0) return null;
        Rectangle result = null;

        for (Iterator i = selectedShapes.iterator(); i.hasNext();) {
            DrawShape shape = (DrawShape) i.next();
            if (!(shape instanceof ConnectorShape || shape instanceof ConnectionShape)) {
                Rectangle bnds = new Rectangle(shape.getBounds());
                Point loc = shape.getLocation();
                Point canvasLoc = SwingUtilities.convertPoint(shape.getParent(), loc, this);
                bnds.setLocation(canvasLoc);
                if (result == null) result = bnds;
                else result.add(bnds);
            }
        }
        return result;
    }

    /** Returns the rectangle that tightly bounds the shapes on the canvas, with a
     * 1-pixel addition on the bottom and right sides to correct for potential display
     * problems in graphical rendering.
     * @return The bounding rectangle.
     */
    public Rectangle getBoundingBox() {
        int top = getHeight();
        int left = getWidth();
        int bottom = 0;
        int right = 0;
        Rectangle rect;

        Component[] comps = getComponents();
        for (int i=0; i<comps.length; i++) {
            if (comps[i] instanceof DrawShape && !(comps[i] instanceof ConnectorShape || comps[i] instanceof ConnectionShape)) {
                rect = ((DrawShape) comps[i]).getBoundingBox();
                if (rect.x < left)
                    left = rect.x;
                if (rect.y < top)
                    top = rect.y;
                if (rect.x+rect.width-1 > right)
                    right = rect.x+rect.width-1;
                if (rect.y+rect.height-1 > bottom)
                    bottom = rect.y+rect.height-1;
            }
        }

        if (bottom == 0)
            top=0;
        if (right == 0)
            left=0;

        rect = new Rectangle(left,top,right-left+1,bottom-top+1);
        return rect;
    }

    /** Currently a no-op function that trivially returns <CODE>true</CODE>, but at some
     * point this will provide a way to convert part or all of the model's relationship
     * sets to hypergraph format.
     * @return Unconditionally <CODE>true</CODE> at present.
     */
    public boolean toHypergraph() {
        return true;
    }

    private boolean rsCompatible(DrawShape ds) {
        return (ds instanceof ObjectSetShape || ds instanceof ObjectShape ||
        ds instanceof RelSetConnectionShape || ds instanceof RelationshipSetShape||
        ds instanceof StateShape || ds instanceof TransitionShape); 
        //|| ds instanceof ParentRelSetConnectionShape || ds instanceof ChildRelSetConnectionShape);

    }

    private boolean gsCompatible(DrawShape ds) {
        return (ds instanceof ObjectSetShape || ds instanceof ObjectShape ||
        ds instanceof GenSpecShape || ds instanceof GenConnectionShape ||
        ds instanceof SpecConnectionShape || ds instanceof AggregationShape 
        || ds instanceof ParentRelSetConnectionShape || ds instanceof ChildRelSetConnectionShape);
    }

    /** Manages the state of the current selection in response to a mouse button being
     * pressed.  The resultant state depends on whether there is a target shape and, if
     * so, whether it already belongs to the current selection.
     * <p>
     * The following table explains what the resultant state may be:
     * <table>
     * <tr><th>Target shape null?</th>
     * <th>Target in current selection?</th>
     * <th>Result state</th></tr>
     * <tr><td>Yes</td>
     * <td>&nbsp;</td>
     * <td>Canvas gets focus; dragging and selecting flags are set; current selection
     * is cleared if neither Control nor Shift key is pressed</td></tr>
     * <tr><td>No</td>
     * <td>Yes</td>
     * <td>Shape flagged to be de-selected if the mouse button is released without
     * dragging</td></tr>
     * <tr><td>No</td>
     * <td>No</td>
     * <td>Shape added to current selection or becomes current selection if neither
     * Control nor Shift key is pressed</td></tr>
     * </table>
     * @param evt The MouseEvent that represents the pressing of the mouse button.
     * @param targetShape The shape that the mouse pointer was over when the mouse event occurred.  May be
     * <CODE>null</CODE>.
     */
    protected void setupSelection(MouseEvent evt, DrawShape targetShape) {
        if (!selectedShapes.contains(targetShape)) {
            if (!evt.isControlDown() && !evt.isShiftDown()) {
                clearSelection();
            }
            if (targetShape != null) {
            	if (targetShape instanceof ObjectSetShape)
            		this.toFront(targetShape);
                addToSelection(targetShape);
            } else {
                requestFocus();
                m_bSelecting = true;
                m_bDragging = true;
            }
        } else { // targetShape is selected
            m_deselectionCandidate = targetShape;
        }
    }

    /** Currently a no-op function; trivially returns <CODE>true</CODE>.
     * @return Unconditionally <CODE>true</CODE>, at this point.
     */
    public boolean outputScheme() {
     /* not clear what this does or how it ports to OSMX... REVISIT. AEW
      */
        return true;
    } // outputScheme

    /** Invoked when a key has been pressed.
     * See the class description for {@link KeyEvent} for a definition of
     * a key pressed event.
     * @param e The key pressed event.
     */
    public void keyPressed(KeyEvent e) { /* do nothing */ }

    /** Invoked when a key has been released.
     * See the class description for {@link KeyEvent} for a definition of
     * a key released event.
     * @param e The key released event.
     */
    public void keyReleased(KeyEvent e) {
        if (e.getKeyCode() == KeyEvent.VK_DELETE)
            deleteSelected();
    }

    /** Invoked when a key has been typed.
     * See the class description for {@link KeyEvent} for a definition of
     * a key typed event.
     * @param e The key typed event.
     */
    public void keyTyped(KeyEvent e) { /* do nothing */ }

    /** Invoked when the mouse button has been clicked (pressed
     * and released) on a component.
     * @param e The mouse clicked event.
     */
    public void mouseClicked(MouseEvent e) { /* do nothing */ }

    /** Handles drag events, both on the canvas itself and on the shapes that the canvas
     * contains.
     * @param evt The drag event.
     */
    public void mouseDragged(MouseEvent evt) {
    	
        Component evtSource = (Component) evt.getSource();
        Point pt = SwingUtilities.convertPoint(evtSource, evt.getPoint(), this);
        destShape = getTargetShape(pt);
        m_deselectionCandidate = null;
        String shapeKey = m_editor.getCurrentShape();

        if (evtSource instanceof EditableText) {
            if (((EditableText) evtSource).isEditingText()) return;
        }

        if (shapeKey.equals(SELECTION_SHAPE_KEY)) {
            if (m_bSelecting) {
                if (!isAncestorOf(dragPanel)) add(dragPanel, 0);

                // select objects in the new rectangle
                Rectangle selRect = DrawShape.normalizeRect(m_initialPoint, pt);
                selectAllWithinRect(selRect);

                // Draw new location
                Rectangle selBounds = getSelectionBoundingBox();
                if (selBounds == null) selBounds = selRect;
                selRect.add(selBounds);
                dragPanel.setDragMode(DragPanel.SELECT_BOX_MODE);
                dragPanel.setBounds(selRect);
                dragPanel.setDragSource(SwingUtilities.convertPoint(this, m_initialPoint, dragPanel));
                dragPanel.setDragDest(SwingUtilities.convertPoint(this, pt, dragPanel));
                dragPanel.repaint();
            }
            //else if (destShape.isResizing() == true) {
            else if (m_bResizing == true) {
            	resizeSelectObjectSetShape(pt.x - m_oldPoint.x,pt.y - m_oldPoint.y);
            }
            else // drag previously selected
            {
                dragSelectedObjects(pt.x - m_oldPoint.x,pt.y - m_oldPoint.y);
            }
        } // condition: selection tool is selected
        // Added by Muhammed
        else if (shapeKey.equals(STATE_CONNECTION_KEY)) {
        	//System.out.println("You are drawing a conjuction shape");
            boolean isSrcLegal = rsCompatible(srcShape);
            boolean isDestLegal = rsCompatible(destShape);
            
            isDestLegal = isDestLegal && (srcShape instanceof ObjConnectableShape ||
            destShape instanceof ObjConnectableShape);
            
           
            if (!isSrcLegal) srcShape = null;
            if (!isDestLegal) destShape = null;

            if (!isAncestorOf(dragPanel)) add(dragPanel, 0);
            dragPanel.setDragMode(DragPanel.CONNECTOR_MODE);
            Rectangle selRect = DrawShape.normalizeRect(m_initialPoint, pt);
            selRect.add(new Rectangle(pt.x - 12, pt.y - 12, 24, 24));
            if (srcShape != null) selRect.add(srcShape.getBounds());
            if (destShape != null) selRect.add(destShape.getBounds());
            if (srcShape != null && srcShape == destShape) {
                selRect.setBounds(selRect.x, selRect.y, selRect.width * 2, selRect.height * 2);
            }
            dragPanel.setBounds(selRect);
            dragPanel.setDragSource(SwingUtilities.convertPoint(this, m_initialPoint, dragPanel));
            dragPanel.setDragDest(SwingUtilities.convertPoint(this, pt, dragPanel));
            dragPanel.repaint();
        } 
        //Conjunctive
        else if (shapeKey.equals(RELATIONSHIP_SET_SHAPE_KEY)) {
        	//System.out.println("You are drawing a relationship shape");
            boolean isSrcLegal = rsCompatible(srcShape);
            boolean isDestLegal = rsCompatible(destShape);
            isDestLegal = isDestLegal && (srcShape instanceof ObjConnectableShape ||
            destShape instanceof ObjConnectableShape);
            if (!isSrcLegal) srcShape = null;
            if (!isDestLegal) destShape = null;

            if (!isAncestorOf(dragPanel)) add(dragPanel, 0);
            dragPanel.setDragMode(DragPanel.CONNECTOR_MODE);
            Rectangle selRect = DrawShape.normalizeRect(m_initialPoint, pt);
            selRect.add(new Rectangle(pt.x - 12, pt.y - 12, 24, 24));
            if (srcShape != null) selRect.add(srcShape.getBounds());
            if (destShape != null) selRect.add(destShape.getBounds());
            if (srcShape != null && srcShape == destShape) 
            {
                selRect.setBounds(selRect.x, selRect.y, selRect.width * 2, selRect.height * 2);
            }
            dragPanel.setBounds(selRect);
            dragPanel.setDragSource(SwingUtilities.convertPoint(this, m_initialPoint, dragPanel));
            dragPanel.setDragDest(SwingUtilities.convertPoint(this, pt, dragPanel));
            dragPanel.repaint();
        } // condition: relationship set tool is selected
        else if (shapeKey.equals(GEN_SPEC_SHAPE_KEY)) {
            boolean isSrcLegal = gsCompatible(srcShape);
            boolean isDestLegal = gsCompatible(destShape);
            isDestLegal = isDestLegal && (srcShape instanceof ObjConnectableShape || destShape instanceof ObjConnectableShape);
            isDestLegal = isDestLegal && (srcShape != destShape);
            if (!isSrcLegal) srcShape = null;
            if (!isDestLegal) destShape = null;

            if (!isAncestorOf(dragPanel)) add(dragPanel, 0);
            dragPanel.setDragMode(DragPanel.CONNECTOR_MODE);
            Rectangle selRect = DrawShape.normalizeRect(m_initialPoint, pt);
            selRect.add(new Rectangle(pt.x - 12, pt.y - 12, 24, 24));
            if (srcShape != null) selRect.add(srcShape.getBounds());
            if (destShape != null) selRect.add(destShape.getBounds());
            if (srcShape != null && srcShape == destShape) {
                selRect.setBounds(selRect.x, selRect.y, selRect.width * 2, selRect.height * 2);
            }
            dragPanel.setBounds(selRect);
            dragPanel.setDragSource(SwingUtilities.convertPoint(this, m_initialPoint, dragPanel));
            dragPanel.setDragDest(SwingUtilities.convertPoint(this, pt, dragPanel));
            dragPanel.repaint();
        }
        else if(shapeKey.equals(AGGREGATION_SHAPE_KEY)){
        	boolean isSrcLegal = gsCompatible(srcShape);
            boolean isDestLegal = gsCompatible(destShape);
            isDestLegal = isDestLegal && (srcShape instanceof ObjConnectableShape || destShape instanceof ObjConnectableShape);
            isDestLegal = isDestLegal && (srcShape != destShape);
            if (!isSrcLegal) srcShape = null;
            if (!isDestLegal) destShape = null;

            if (!isAncestorOf(dragPanel)) add(dragPanel, 0);
            dragPanel.setDragMode(DragPanel.CONNECTOR_MODE);
            Rectangle selRect = DrawShape.normalizeRect(m_initialPoint, pt);
            selRect.add(new Rectangle(pt.x - 12, pt.y - 12, 24, 24));
            if (srcShape != null) selRect.add(srcShape.getBounds());
            if (destShape != null) selRect.add(destShape.getBounds());
            if (srcShape != null && srcShape == destShape) {
                selRect.setBounds(selRect.x, selRect.y, selRect.width * 2, selRect.height * 2);
            }
            dragPanel.setBounds(selRect);
            dragPanel.setDragSource(SwingUtilities.convertPoint(this, m_initialPoint, dragPanel));
            dragPanel.setDragDest(SwingUtilities.convertPoint(this, pt, dragPanel));
            dragPanel.repaint();
        }

        m_bDragging = true;
        m_oldPoint = pt;
    }

    /** Invoked when the mouse enters a component.  Currently a no-op procedure.
     * @param e The mouse entered event.
     */
    public void mouseEntered(MouseEvent e) { /* do nothing */ }

    /** Invoked when the mouse exits a component.  Currently a no-op procedure.
     * @param e The mouse exited event.
     */
    public void mouseExited(MouseEvent e) { /* do nothing */ }

    /** Invoked when the mouse button has been moved on a component
     * (with no buttons down).  Currently a no-op procedure.
     * @param e The mouse moved event.
     */
    public void mouseMoved(MouseEvent e) { 
    	// Do nothing
    }

    /** Invoked when a mouse button has been pressed on a component.  The action taken
     * when the mouse button is pressed depends on several things: the currently
     * selected drawing tool, the target shape (if any), the current selection, and
     * whether the event source is an {@link EditableText} that is currently being
     * edited.
     * @param evt The mouse pressed event.
     */
    public void mousePressed(MouseEvent evt) {
        // clear transient fields that may have been set by other mouse event handlers
        // remember the coords of the event
        // find the shape at the event coords
        // determine the currently active shape tool
        // if selection shape tool active, set up selection
        // if event is context menu trigger, show context menu
        DrawShape targetShape = null;
        Component evtSource = (Component) evt.getSource();
        //System.out.println(evtSource);
        m_deselectionCandidate = null;
        m_bDragging = false;
        m_bSelecting = false;
        m_bResizing = false;
        m_initialPoint = SwingUtilities.convertPoint(evtSource, evt.getPoint(), this);
        String shapeKey = m_editor.getCurrentShape(); 
        targetShape = getTargetShape(evt);
        if (targetShape == null) {
            this.requestFocus();
        } else {
            targetShape.requestFocus();
            if (!(targetShape instanceof ParticipationConstraintShape))
            	this.toFront(targetShape);//test
        }

        m_oldPoint = m_initialPoint;
        m_editor.enableActions();

        // some OSs may catch the event on mousePressed
        // Windows catches it on mouseReleased
        if (evt.isPopupTrigger()) {
        	//System.out.println(targetShape);//test
            doPopup(targetShape, evt);
            return;
        }

        if (evtSource instanceof EditableText) { 
            if (((EditableText) evtSource).isEditingText()) return;
        }

        if (shapeKey.equals(SELECTION_SHAPE_KEY)) {
            setupSelection(evt, targetShape);
        } else if (shapeKey.equals(RELATIONSHIP_SET_SHAPE_KEY)) {
            srcShape = targetShape;
            destShape = targetShape;
        } else if (shapeKey.equals(GEN_SPEC_SHAPE_KEY)) {
            srcShape = targetShape;
            destShape = targetShape;
        }else if (shapeKey.equals(STATE_CONNECTION_KEY))
        {
        	srcShape = targetShape;
        	destShape = targetShape;
        }
        else if (shapeKey.equals(AGGREGATION_SHAPE_KEY)) {
        	srcShape = targetShape;
        	destShape = targetShape;
        }
        if (targetShape != null)
        	if (targetShape.isResizing() == true)
        		m_bResizing = true;
        
    } // mousePressed method

    /** Invoked when a mouse button has been released on a component.
     * @param evt The mouse released event.
     */
    public void mouseReleased(MouseEvent evt) {
        DrawShape targetShape = null;
        Component evtSource = (Component) evt.getSource();
        String shapeKey = m_editor.getCurrentShape();

        targetShape = getTargetShape(evt);
        Point pt = SwingUtilities.convertPoint(evtSource, evt.getPoint(), this);

        m_editor.enableActions();

        if (evt.isPopupTrigger()) {
            doPopup(targetShape, evt);
            return;
        }
        //***************Added by Muhammed to support states******************
        if (shapeKey.equals(STATE_SHAPE_KEY)) {
           try {
               //System.out.println("Mouse Released in State");
               State state = m_doc.getObjectFactory().createState();
               OSMType targetModel;
               if (targetShape == null) {
                   targetModel = m_doc.getModelRoot();
               } else {
                   targetModel = targetShape.getParentCanvas().getOSM();
               }
               targetModel.getAllModelElements().add(state);
               state.setX(pt.x);
               state.setY(pt.y);
           } catch (Exception e) {
               e.printStackTrace();
           }
       }
       else if(shapeKey.equals(TRANSITION_SHAPE_KEY)){
           try {
               //System.out.println("Mouse Released in Transition");
               //Transition transition = m_doc.getObjectFactory().createTransition();
               Transition transition = new edu.byu.deg.osmx.binding.impl.TransitionImpl();
           		//timpl.setAction("Action");
           		//timpl.setTrigger("Triggers");               
               transition.setAction("Action");
               transition.setTrigger("Triggers");
               OSMType targetModel;
               if (targetShape == null) {
                   targetModel = m_doc.getModelRoot();
               } else {
                   targetModel = targetShape.getParentCanvas().getOSM();
               }
               targetModel.getAllModelElements().add(transition);
               transition.setX(pt.x);
               transition.setY(pt.y);
           } catch (Exception e) {
               e.printStackTrace();
           }
       }
       //******************************************************
       //STATE_CONNECTION_KEY
       
      else if(shapeKey.equals(STATE_CONNECTION_KEY)) {
       // TODO: create relationship set shape
       //System.out.println("In Conjuction Action");
       boolean isSrcLegal = rsCompatible(srcShape);
       boolean isDestLegal = rsCompatible(destShape);
       //if (isDestLegal)
    	   //System.out.println("destLegal");
       isDestLegal = isDestLegal && (destShape instanceof StateConnectableContainerShape||destShape instanceof ObjConnectableShape ||
       srcShape instanceof ObjConnectableShape);
      
       if (isSrcLegal && isDestLegal) {
           if (srcShape instanceof ObjConnectableShape && destShape instanceof ObjConnectableShape) {
               // TODO: create new relationship set
           	 
               try {
                   ConjunctionConnection conn1 = m_doc.getObjectFactory().createConjunctionConnection();
                   if (srcShape instanceof StateShape) {
                       conn1.setObjectSet(((StateType) srcShape.getElement()).getId());
                   }else{
                       conn1.setObjectSet(((TransitionType) srcShape.getElement()).getId());
                   }
                  
                   ConjunctionConnection conn2 = m_doc.getObjectFactory().createConjunctionConnection();
                   if (destShape instanceof StateShape) {
                       conn2.setObjectSet(((StateType) destShape.getElement()).getId());
                   }else{
                       conn2.setObjectSet(((TransitionType) destShape.getElement()).getId());
                   }
                    OSMXConjunctionType statetype = new ConjunctionImpl();
                   //OSMXRelationshipSetType relSet = new RelationshipSetImpl();
                   //System.out.println("*** MUST BE OSMX? ***"+statetype);
                   statetype.getConjunctionConnection().add(conn1);
                   statetype.getConjunctionConnection().add(conn2);
                   
                   OSMType targetModel;
                   targetModel = m_doc.getModelRoot();
                   targetModel.getAllModelElements().add(statetype);
                   if (srcShape == destShape) {
                       AnchorType a1_1 = m_doc.getObjectFactory().createAnchor();
                       AnchorType a1_2 = m_doc.getObjectFactory().createAnchor();
                       AnchorType a2_1 = m_doc.getObjectFactory().createAnchor();
                       AnchorType a2_2 = m_doc.getObjectFactory().createAnchor();
                       a1_1.setX(srcShape.getX() + 3 * srcShape.getWidth() / 2);
                       a1_1.setY(srcShape.getY() + 3 * srcShape.getHeight() / 2);
                       a1_2.setX(srcShape.getX() + 3 * srcShape.getWidth() / 2);
                       a1_2.setY(srcShape.getY() + srcShape.getHeight() / 2);
                       a2_1.setX(a1_1.getX());
                       a2_1.setY(a1_1.getY());
                       a2_2.setX(srcShape.getX() + srcShape.getWidth() / 2);
                       a2_2.setY(srcShape.getY() + 3 * srcShape.getHeight() / 2);
                       conn1.getAnchor().add(a1_1);
                       conn1.getAnchor().add(a1_2);
                       conn2.getAnchor().add(a2_1);
                       conn2.getAnchor().add(a2_2);
                   }
                   m_doc.addElement(statetype);

               } catch (Exception e) { //TODO
                   e.printStackTrace();
               }
           } else {
               
               try {
                   ObjConnectableShape objSetShape = (ObjConnectableShape) (srcShape instanceof ObjConnectableShape ? srcShape : destShape);
                   DrawShape connectorProxy = (objSetShape == srcShape ? destShape : srcShape);
                   RelationshipSetShape relSetShape;
                   if (connectorProxy instanceof RelSetConnectionShape) {
                       relSetShape = (RelationshipSetShape) connectorProxy.getParent();
                   } else if (connectorProxy instanceof RelationshipSetShape) {
                       relSetShape = (RelationshipSetShape) connectorProxy;
                   } else {
                       // Should never happen due to legality check on shapes earlier
                       throw new Exception("Illegal connector proxy");
                   }
                   //System.out.println("YES HERE RELSETCONNECTIOn");
                   RelSetConnection conn = m_doc.getObjectFactory().createRelSetConnection();
                   if (objSetShape instanceof ObjectSetShape) {
                       conn.setObjectSet(((OSMXObjectSetType) objSetShape.getElement()).getId());
                   } else {
                       conn.setObjectSet(((OSMXObjectType) objSetShape.getElement()).getId());
                   }
                   relSetShape.addConnection((OSMXRelSetConnectionType) conn);
               } catch (Exception e) {
                   e.printStackTrace();
               }
           }
       }
       Rectangle dragPanelBounds = dragPanel.getBounds();
       remove(dragPanel);
       repaint(dragPanelBounds);
       srcShape = null;
       destShape = null;
       m_bDragging = false;
       
   }
       


        if (shapeKey.equals(SELECTION_SHAPE_KEY)) {
            // TODO
            if ((evt.isControlDown() || evt.isShiftDown())) {
                if (m_deselectionCandidate == targetShape) {
                    removeFromSelection(targetShape);
                }
            }
            m_deselectionCandidate = null;
            if (m_bSelecting && m_bDragging) {
                Point boxLoc = dragPanel.getLocation();
                Dimension boxDim = dragPanel.getSize();
                remove(dragPanel);
                repaint(boxLoc.x, boxLoc.y, boxDim.width, boxDim.height);
            }
            m_bDragging = false;
            m_bSelecting = false;
        } else if (!m_bDragging) {
            if (shapeKey.equals(GENERAL_CONSTRAINT_SHAPE_KEY)) {
                try {
                    GeneralConstraint gc = m_doc.getObjectFactory().createGeneralConstraint();
                    OSMType targetModel;
                    if (targetShape == null) {
                        targetModel = m_doc.getModelRoot();
                    } else {
                        targetModel = targetShape.getParentCanvas().getOSM();
                    }
                    targetModel.getAllModelElements().add(gc);
                    gc.setX(pt.x);
                    gc.setY(pt.y);
                } catch (Exception e) { // TODO
                    e.printStackTrace();
                }
            } else if (shapeKey.equals(NOTE_SHAPE_KEY)) {
                try {
                    Note note = m_doc.getObjectFactory().createNote();
                    OSMType targetModel;
                    if (targetShape == null) {
                        targetModel = m_doc.getModelRoot();
                    } else {
                        targetModel = targetShape.getParentCanvas().getOSM();
                    }
                    targetModel.getAllModelElements().add(note);
                    note.setX(pt.x);
                    note.setY(pt.y);
                } catch (Exception e) { //TODO
                    e.printStackTrace();
                }
            } else if (shapeKey.equals(OBJECT_SET_SHAPE_KEY)) {
                try {
                    ObjectSet objSet = m_doc.getObjectFactory().createObjectSet();
                    OSMType targetModel;
                    if (targetShape == null) {
                        targetModel = m_doc.getModelRoot();
                    } else {
                        targetModel = targetShape.getParentCanvas().getOSM();
                    }
                    targetModel.getAllModelElements().add(objSet);
                    objSet.setX(pt.x);
                    objSet.setY(pt.y);
                } catch (Exception e) { //TODO
                    e.printStackTrace();
                }
            } else if (shapeKey.equals(OBJECT_SHAPE_KEY)) {
                try {
                    OSMObject obj = m_doc.getObjectFactory().createOSMObject();
                    OSMType targetModel;
                    if (targetShape == null) {
                        targetModel = m_doc.getModelRoot();
                    } else {
                        targetModel = targetShape.getParentCanvas().getOSM();
                    }
                    targetModel.getAllModelElements().add(obj);
                    obj.setX(pt.x);
                    obj.setY(pt.y);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        } else if (shapeKey.equals(RELATIONSHIP_SET_SHAPE_KEY)) {
        	//System.out.println("MousePresed RSSK");
        	//System.out.println("relatinship set shape");
            // TODO: create relationship set shape
            boolean isSrcLegal = rsCompatible(srcShape);
            boolean isDestLegal = rsCompatible(destShape);
            isDestLegal = isDestLegal && (destShape instanceof ObjConnectableShape ||
            srcShape instanceof ObjConnectableShape);
            if (isSrcLegal && isDestLegal) {
                if (srcShape instanceof ObjConnectableShape && destShape instanceof ObjConnectableShape) {
                    // TODO: create new relationship set
                    try {
                        RelSetConnection conn1 = m_doc.getObjectFactory().createRelSetConnection();
                        ParticipationConstraintType pc1 = m_doc.getObjectFactory().createParticipationConstraintType();
                        pc1.setContent("0:*");
                        conn1.setParticipationConstraint(pc1);
                        if (srcShape instanceof ObjectSetShape) {
                            conn1.setObjectSet(((ObjectSetType) srcShape.getElement()).getId());
                        } else {
                            conn1.setObjectSet(((ObjectType) srcShape.getElement()).getId());
                        }
                        RelSetConnection conn2 = m_doc.getObjectFactory().createRelSetConnection();
                        ParticipationConstraintType pc2 = m_doc.getObjectFactory().createParticipationConstraintType();
                        pc2.setContent("0:*");                        
                        conn2.setParticipationConstraint(pc2);
                        if (destShape instanceof ObjectSetShape) {
                            conn2.setObjectSet(((ObjectSetType) destShape.getElement()).getId());
                        } else {
                            conn2.setObjectSet(((ObjectType) destShape.getElement()).getId());
                        }
                        OSMXRelationshipSetType relSet = new RelationshipSetImpl();
                        relSet.getRelSetConnection().add(conn1);
                        relSet.getRelSetConnection().add(conn2);
                        relSet.setName("has");
                        relSet.setRightArrow(true);
                        OSMType targetModel;
                        targetModel = m_doc.getModelRoot();
                        targetModel.getAllModelElements().add(relSet);
                        if (srcShape == destShape) {
                            AnchorType a1_1 = m_doc.getObjectFactory().createAnchor();
                            AnchorType a1_2 = m_doc.getObjectFactory().createAnchor();
                            AnchorType a2_1 = m_doc.getObjectFactory().createAnchor();
                            AnchorType a2_2 = m_doc.getObjectFactory().createAnchor();
                            a1_1.setX(srcShape.getX() + 3 * srcShape.getWidth() / 2);
                            a1_1.setY(srcShape.getY() + 3 * srcShape.getHeight() / 2);
                            a1_2.setX(srcShape.getX() + 3 * srcShape.getWidth() / 2);
                            a1_2.setY(srcShape.getY() + srcShape.getHeight() / 2);
                            a2_1.setX(a1_1.getX());
                            a2_1.setY(a1_1.getY());
                            a2_2.setX(srcShape.getX() + srcShape.getWidth() / 2);
                            a2_2.setY(srcShape.getY() + 3 * srcShape.getHeight() / 2);
                            conn1.getAnchor().add(a1_1);
                            conn1.getAnchor().add(a1_2);
                            conn2.getAnchor().add(a2_1);
                            conn2.getAnchor().add(a2_2);
                        }
                        m_doc.addElement(relSet);

                    } catch (Exception e) { //TODO
                        e.printStackTrace();
                    }
                } else {
                    // TODO: create new connection on existing relationship set
                    try {
                        ObjConnectableShape objSetShape = (ObjConnectableShape) (srcShape instanceof ObjConnectableShape ? srcShape : destShape);
                        DrawShape connectorProxy = (objSetShape == srcShape ? destShape : srcShape);
                        RelationshipSetShape relSetShape;
                        if (connectorProxy instanceof RelSetConnectionShape) {
                            relSetShape = (RelationshipSetShape) connectorProxy.getParent();
                        } else if (connectorProxy instanceof RelationshipSetShape) {
                            relSetShape = (RelationshipSetShape) connectorProxy;
                        } else {
                            // Should never happen due to legality check on shapes earlier
                            throw new Exception("Illegal connector proxy");
                        }
                        RelSetConnection conn = m_doc.getObjectFactory().createRelSetConnection();
                        if (objSetShape instanceof ObjectSetShape) {
                            conn.setObjectSet(((OSMXObjectSetType) objSetShape.getElement()).getId());
                        } else {
                            conn.setObjectSet(((OSMXObjectType) objSetShape.getElement()).getId());
                        }
                        relSetShape.addConnection((OSMXRelSetConnectionType) conn);
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                }
            }
            Rectangle dragPanelBounds = dragPanel.getBounds();
            remove(dragPanel);
            repaint(dragPanelBounds);
            srcShape = null;
            destShape = null;
            m_bDragging = false;
        }else if(shapeKey.equals(AGGREGATION_SHAPE_KEY)){
        	boolean isSrcLegal = gsCompatible(srcShape);
            boolean isDestLegal;
            isDestLegal = gsCompatible(destShape) &&
            (destShape instanceof ObjConnectableShape || srcShape instanceof ObjConnectableShape) &&
            (srcShape != destShape);
            if (isSrcLegal && isDestLegal) {
                if (srcShape instanceof ObjConnectableShape && destShape instanceof ObjConnectableShape) {
                    try {
                    	 ParentRelSetConnection parentConn = m_doc.getObjectFactory().createParentRelSetConnection();
                    	 if (srcShape instanceof ObjectSetShape) {
                             parentConn.setObjectSet(((ObjectSetType) srcShape.getElement()).getId());
                         } else {
                             parentConn.setObjectSet(((ObjectType) srcShape.getElement()).getId());
                         }
                         //RelSetConnection conn2 = m_doc.getObjectFactory().createRelSetConnection();
                         ChildRelSetConnection childConn = m_doc.getObjectFactory().createChildRelSetConnection();
                         ParticipationConstraintType pc1 = m_doc.getObjectFactory().createParticipationConstraintType();
                         pc1.setContent("0:*");                        
                         childConn.setParentSideParticipationConstraint(pc1);
                         ParticipationConstraintType pc2 = m_doc.getObjectFactory().createParticipationConstraintType();
                         pc2.setContent("0:*");
                         childConn.setChildSideParticipationConstraint(pc2);
                         if (destShape instanceof ObjectSetShape) {
                             childConn.setObjectSet(((ObjectSetType) destShape.getElement()).getId());
                         } else {
                             childConn.setObjectSet(((ObjectType) destShape.getElement()).getId());
                         }
                        OSMXAggregationType aggregation = new AggregationImpl();
                        //aggregataion.getRelSetConnection().add(conn1);
                        //aggregataion.getRelSetConnection().add(conn2);
                        aggregation.setParentConnection(parentConn);
                        aggregation.getChildConnection().add(childConn);
                        OSMType targetModel;
                        targetModel = m_doc.getModelRoot();
                        targetModel.getAllModelElements().add(aggregation);
                        m_doc.addElement(aggregation);

                    } catch (Exception e) { //TODO
                        e.printStackTrace();
                    }
                }else{
                	//create a new connection on existing aggregation
                	try{
                		ObjConnectableShape objSetShape = (ObjConnectableShape) (srcShape instanceof ObjConnectableShape ? srcShape : destShape);
                        DrawShape connectorProxy = (objSetShape == srcShape ? destShape : srcShape);
                        AggregationShape aggregationShape;
                        if (connectorProxy instanceof ParentRelSetConnectionShape || connectorProxy instanceof ChildRelSetConnectionShape) {
                        	aggregationShape = (AggregationShape) connectorProxy.getParent();
                        } else if (connectorProxy instanceof AggregationShape) {
                        	aggregationShape = (AggregationShape) connectorProxy;
                        } else {
                            // Should never happen due to legality check on shapes earlier
                            throw new Exception("Illegal connector proxy");
                        }
                        //RelSetConnection conn = m_doc.getObjectFactory().createRelSetConnection();
                        ChildRelSetConnection childConn = m_doc.getObjectFactory().createChildRelSetConnection();
                        ParticipationConstraintType pc1 = m_doc.getObjectFactory().createParticipationConstraintType();
                        pc1.setContent("0:*");                        
                        childConn.setParentSideParticipationConstraint(pc1);
                        ParticipationConstraintType pc2 = m_doc.getObjectFactory().createParticipationConstraintType();
                        pc2.setContent("0:*");
                        childConn.setChildSideParticipationConstraint(pc2);
                        if (objSetShape instanceof ObjectSetShape) {
                            childConn.setObjectSet(((OSMXObjectSetType) objSetShape.getElement()).getId());
                        } else {
                            childConn.setObjectSet(((OSMXObjectType) objSetShape.getElement()).getId());
                        }
                        aggregationShape.addChildConnection((OSMXChildRelSetConnection) childConn);
                	}catch(Exception e){
                		e.printStackTrace();
                	}
                }
            }
            Rectangle dragPanelBounds = dragPanel.getBounds();
            remove(dragPanel);
            repaint(dragPanelBounds);
            srcShape = null;
            destShape = null;
            m_bDragging = false;
        } 
        else if (shapeKey.equals(GEN_SPEC_SHAPE_KEY)) {
            boolean isSrcLegal = gsCompatible(srcShape);
            boolean isDestLegal;
            isDestLegal = gsCompatible(destShape) &&
            (destShape instanceof ObjConnectableShape || srcShape instanceof ObjConnectableShape) &&
            (srcShape != destShape);
            if (isSrcLegal && isDestLegal) {
                if (srcShape instanceof ObjConnectableShape && destShape instanceof ObjConnectableShape) {
                    // TODO: create new relationship set
                    try {
                        SpecializationConnection conn1 = m_doc.getObjectFactory().createSpecializationConnection();
                        if (destShape instanceof ObjectSetShape) {
                            conn1.setObjectSet(((ObjectSetType) destShape.getElement()).getId());
                        } else {
                            conn1.setObjectSet(((ObjectType) destShape.getElement()).getId());
                        }
                        GeneralizationConnection conn2 = m_doc.getObjectFactory().createGeneralizationConnection();
                        if (srcShape instanceof ObjectSetShape) {
                            conn2.setObjectSet(((ObjectSetType) srcShape.getElement()).getId());
                        } else {
                            conn2.setObjectSet(((ObjectType) srcShape.getElement()).getId());
                        }
                        OSMXGenSpecType genSpec = new GenSpecImpl();
                        genSpec.getSpecConnection().add(conn1);
                        genSpec.getGenConnection().add(conn2);
                        OSMType targetModel;
                        targetModel = m_doc.getModelRoot();
                        targetModel.getAllModelElements().add(genSpec);
                        m_doc.addElement(genSpec);

                    } catch (Exception e) { //TODO
                        e.printStackTrace();
                    }
                } else {
                    // TODO: create new connection on existing gen/spec
                    try {
                        ObjConnectableShape objSetShape = (ObjConnectableShape) (srcShape instanceof ObjConnectableShape ? srcShape : destShape);
                        DrawShape connectorProxy = (objSetShape == srcShape ? destShape : srcShape);
                        GenSpecShape genSpecShape;
                        if (connectorProxy instanceof GenConnectionShape || connectorProxy instanceof SpecConnectionShape) {
                            genSpecShape = (GenSpecShape) connectorProxy.getParent();
                        } else if (connectorProxy instanceof GenSpecShape) {
                            genSpecShape = (GenSpecShape) connectorProxy;
                        } else {
                            // Should never happen due to legality check on shapes earlier
                            throw new Exception("Illegal connector proxy");
                        }
                        BasicConnection conn;
                        if (connectorProxy instanceof SpecConnectionShape || connectorProxy instanceof GenSpecShape) {
                            conn = m_doc.getObjectFactory().createSpecializationConnection();
                            if (objSetShape instanceof ObjectSetShape) {
                                conn.setObjectSet(((ObjectSetType) objSetShape.getElement()).getId());
                            } else {
                                conn.setObjectSet(((ObjectType) objSetShape.getElement()).getId());
                            }
                            genSpecShape.addSpecConnection((OSMXSpecializationConnection) conn);
                        } else {
                            conn = m_doc.getObjectFactory().createGeneralizationConnection();
                            if (objSetShape instanceof ObjectSetShape) {
                                conn.setObjectSet(((OSMXObjectSetType) objSetShape.getElement()).getId());
                            } else {
                                conn.setObjectSet(((OSMXObjectType) objSetShape.getElement()).getId());
                            }
                            genSpecShape.addGenConnection((OSMXGeneralizationConnection) conn);
                        }
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                }
            }
            
            Rectangle dragPanelBounds = dragPanel.getBounds();
            remove(dragPanel);
            repaint(dragPanelBounds);
            srcShape = null;
            destShape = null;
            m_bDragging = false;
        }// condition: request to add new shape
        
        m_bResizing = false;
    } // mouseReleased method

    /** Invoked when the target of the listener has changed its state.
     *
     * @param e  a DocumentChangeEvent object
     *
     *
     */
    public void stateChanged(ChangeEvent e) {
        if (e instanceof DocumentChangeEvent) {
            DocumentChangeEvent evt = (DocumentChangeEvent) e;
            if (evt.getType() == DocumentChangeEvent.CREATE_ELEMENT_EVENT) {
                doElementCreated(evt);
            } // check on event type
        } // check if okay to cast to DocumentChangeEvent
    } // method stateChanged

    /** Handles element creation notifications from the
     * {@link edu.byu.deg.osmx.OSMXDocument} that belongs to this canvas.  This allows
     * the canvas to create a {@link edu.byu.deg.OntologyEditor.shapes.DrawShape} that
     * corresponds to the newly created element.
     * @param evt The event that signals the creation of a new
     * {@link edu.byu.deg.osmx.OSMXElement}.
     */
    protected void doElementCreated(DocumentChangeEvent evt) {
        OSMXElement elem = evt.getElement();
        Container container = getElementParentContainer(elem);
        DrawShape ds = createShapeFromElement(container, elem);
        if (ds != null) {
            clearSelection();
            addToSelection(ds);
        }
    } // method doElementCreated

    /** Returns the object that should contain the shape corresponding to the specified
     * element.
     * @param elem The element of interest.
     * @return The container of the specified element's parent, or this canvas if no such
     * container exists or the element has no parent.  The result is what should
     * contain the shape corresponding to the specified element, if the shape were to
     * be created.
     */
    protected Container getElementParentContainer(OSMXElement elem) {
        Container container = findShapeForElement(elem.getParent());
        if (container == null) {
            OSMXOSMType osm = elem.getParentOSM();
            if (osm != null) {
                OSMXElement osmParent = osm.getParent();
                if (osmParent != null) {
                    container = findShapeForElement(osmParent);
                    if (container == null || !(container instanceof Canvas)) {
                        container = this;
                    }
                }
            }
        }
        if (container == null) return this;
        else return container;
    }

    /** Creates the appropriate <CODE>DrawShape</CODE> for the specified element.  The
     * shape is added to the specified <CODE>Container</CODE>.
     * @param c The designated <CODE>Container</CODE> for the new shape.
     * @param elem The underlying element for the new shape.
     * @return The newly created shape.
     */
    protected DrawShape createShapeFromElement(Container c, OSMXElement elem) {
        if (elem == null) return null;
        DrawShape ds = null;
        if (elem instanceof GeneralConstraint) {
            ds = new GeneralConstraintShape(c, elem);
        } else if (elem instanceof Note) {
            ds = new NoteShape(c, elem);
        } else if (elem instanceof OSMXObjectType) {
            ds = new ObjectShape(c, elem);
        } else if (elem instanceof OSMXObjectSetType) {
            ds = new ObjectSetShape(c, elem);
        } else if (elem instanceof RelationshipSetType) {
            ds = new RelationshipSetShape(c, elem);
        } else if (elem instanceof GenSpecType) {
            ds = new GenSpecShape(c, elem);
        }else if(elem instanceof OSMXStateType) {
             ds = new StateShape(c,elem);
         }
         else if(elem instanceof OSMXTransitionType){
             ds = new TransitionShape(c,elem);
         }
         else if(elem instanceof ConjunctionType){
             ds = new ConjunctionShape(c,elem);
         }
         else if (elem instanceof AggregationType){
        	 ds = new AggregationShape(c,elem);
         }
        
        if (ds != null) {
            ds.addItemListener(this);
            shapesByElement.put(elem, ds);
        }
        return ds;
    }

    /** Removes all shapes from the selection and sets their selected state to
     * <CODE>false</CODE>.
     */
    protected void clearSelection() {
        Iterator i = selectedShapes.iterator();
        while (i.hasNext()) {
            DrawShape ds = (DrawShape) i.next();

            // Important: Must remove the item before setting selected to false.
            // Otherwise a ConcurrentModificationException will be raised due to the
            // selection event being caught by itemStateChanged.
            i.remove();
            ds.setSelected(false);
        }
    } // method clearSelection

    /** Adds the designated shape to the current selection.
     * @param ds The shape to add to the selection.
     */
    protected void addToSelection(DrawShape ds) {
        if (ds != null) {
            ds.setSelected(true);
            selectedShapes.add(ds);
        }
    } // method addToSelection

    /** Removes the designated shape from the current selection.
     * @param ds The shape to remove from the selection.
     */
    protected void removeFromSelection(DrawShape ds) {
        if (ds != null) {
            ds.setSelected(false);
            selectedShapes.remove(ds);
        }
    } // method removeFromSelection

    /** Determines the closest containing canvas for the object underneath the specified
     * point.  If the object is a canvas, it will be the return value; otherwise, the
     * return value will be the parent canvas of the targeted <CODE>DrawShape</CODE>.
     * @param pt The point under which to find the target object.
     * @return The canvas corresponding to the targeted object.
     */
    protected Canvas getTargetCanvas(Point pt) {
        Canvas canvas;
        Component obj = getComponentAt(pt);
        if (obj == null)
            canvas = this;
        else if (obj instanceof Canvas)
            canvas = (Canvas)obj;
        else
            canvas = ((DrawShape)obj).getParentCanvas();

        return canvas;
    } // method getTargetCanvas

    /** Returns the <CODE>DrawShape</CODE> located underneath the specified point.
     * @return The target shape under the given point, or <CODE>null</CODE> if no shape
     * contains that point.
     * @param pt The point under which to look for the target shape.
     */
    protected DrawShape getTargetShape(Point pt) {
        Component comp = findComponentAt(pt);
        if (comp instanceof EditableText) {
            comp = comp.getParent();
        }
        if (!(comp instanceof DrawShape))
            comp = SwingUtilities.getAncestorOfClass(DrawShape.class, comp);

        return (DrawShape) comp;
    } // method getTargetShape(Point)

    /** Returns the <CODE>DrawShape</CODE> that is the target of the given
     * <CODE>MouseEvent</CODE>.  This may not be the same as
     * {@link java.util.EventObject#getSource}, since the latter may refer to an
     * {@link EditableText} belonging to the target shape instead of the shape itself.
     * @param evt The event for which to find the affected <CODE>DrawShape</CODE>.
     * @return The shape that that is the target of the event.
     */
    protected DrawShape getTargetShape(MouseEvent evt) {
        Component eventSource = (Component) evt.getSource();
        DrawShape result = null;
        if (eventSource instanceof DrawShape) {
            result = (DrawShape) (evt.getSource());
        }
        else {
            if (eventSource instanceof EditableText) {
                EditableText text = (EditableText)(evt.getSource());
                if (text.isEditingText())
                    return null;
            }
            result = (DrawShape) SwingUtilities.getAncestorOfClass(DrawShape.class, eventSource);
        }
        return result;
    }

    /** Accessor for the root-level <CODE>OSM</CODE> element.
     * @return The root-level <CODE>OSM</CODE> element.
     */
    public OSMType getOSM() {
        return m_doc.getModelRoot();
    } // method getOSM

    /** Handles display of the appropriate popup menu.
     * @param ds The target shape whose menu is to be displayed.
     * @param evt The <CODE>MouseEvent</CODE> that originated the popup menu request; it provides
     * location and modifier key data useful for certain types of menu commands.
     */
    protected void doPopup(DrawShape ds, MouseEvent evt) {
        JPopupMenu pop = null;
        activeShape = ds;
        activeMouseEvent = evt;
        m_editor.enableActions();
        if (ds instanceof ObjectSetShape) {
            pop = (JPopupMenu) m_editor.getMenus().get(OBJECT_SET_POPUP_KEY);
        } else if (ds instanceof RelationshipSetShape) {
            pop = (JPopupMenu) m_editor.getMenus().get(RELATIONSHIP_SET_POPUP_KEY);
        } else if (ds instanceof RelSetConnectionShape) {
            pop = (JPopupMenu) m_editor.getMenus().get(REL_SET_CONN_POPUP_KEY);
        } else if (ds instanceof GenConnectionShape || ds instanceof SpecConnectionShape) {
            pop = (JPopupMenu) m_editor.getMenus().get(GEN_SPEC_CONN_POPUP_KEY);
        } else if (ds instanceof GenSpecShape) {
            pop = (JPopupMenu) m_editor.getMenus().get(GEN_SPEC_POPUP_KEY);
        } else if (ds instanceof AggregationShape){
        	pop = (JPopupMenu) m_editor.getMenus().get(AGGREGATION_POPUP_KEY);
        } else if (ds instanceof ParentRelSetConnectionShape) {
        	pop = (JPopupMenu)m_editor.getMenus().get(this.PARENT_REL_SET_CONN_POPUP_KEY);
        } else if (ds instanceof ChildRelSetConnectionShape){
        	pop = (JPopupMenu)m_editor.getMenus().get(this.CHILD_REL_SET_CONN_POPUP_KEY);
        }
        
        
        if (pop != null) {
            pop.show((Component) evt.getSource(), evt.getX(), evt.getY());
        }
    } // method doPopup

    /** Invoked when an item has been selected or deselected by the user.
     * The code written for this method performs the operations
     * that need to occur when an item is selected (or deselected).
     * @param e The event indicating selection state changed for a shape.
     */
    public void itemStateChanged(ItemEvent e) {
        DrawShape ds = (DrawShape) e.getItem();
        if (e.getStateChange() == ItemEvent.SELECTED && !selectedShapes.contains(ds)) {
            selectedShapes.add(ds);
        } else if (e.getStateChange() == ItemEvent.DESELECTED && selectedShapes.contains(ds)) {
            selectedShapes.remove(ds);
        }
    }

    /** Completely refreshes the canvas from the {@link edu.byu.deg.osmx.OSMXDocument}
     * to ensure that the display reflects the latest changes in the document.  This
     * should be used primarily after event listeners have been disabled for some
     * reason; ordinarily, the event listeners should be responsible for keeping the
     * display in synch with the underlying model.
     */
    protected void rebuildFromDocument() {
        // remove all child components
        this.removeAll();
        shapesByElement.clear();
        // call rebuildOSM() on model root
        rebuildOSM((OSMXOSMType) m_doc.getModelRoot());
        repaint();
    } // method rebuildFromDocument

    /** Re-generates the shapes for the elements contained in the specified
     * <CODE>OSM</CODE> element.
     * @param osm The <CODE>OSM</CODE> element to rebuild shapes for.
     */
    protected void rebuildOSM(OSMXOSMType osm) {
        // standalone shapes: Note, General Constraint, Object Set, State
        List relSets = new LinkedList();
        List genSpecs = new LinkedList();
        List assocs = new LinkedList();
        List aggregs = new LinkedList();
        List transitions = new LinkedList();
        List conjunctions = new LinkedList();

        DrawShape ds = null;
        Container parentContainer = getElementParentContainer(osm);
        for (Iterator i = osm.getAllModelElements().iterator(); i.hasNext();) {
            OSMXElement e = (OSMXElement) i.next();
            if (e instanceof OSMXRelationshipSetType) {
                if (((RelationshipSetType) e).isSetOSM()) {
                    ((LinkedList) relSets).addFirst(e);
                    //System.out.println("RE*:"+relSets);
                } else {
                    relSets.add(e);
                    //System.out.println("REE*:"+relSets);
                }
            } else if (e instanceof OSMXGenSpecType) {
                genSpecs.add(e);
            } else if (e instanceof OSMXAssociationType) {
                assocs.add(e);
            } else if (e instanceof OSMXAggregationType) {
                aggregs.add(e);
            //}else if (e instanceof OSMXTransitionType) {
             //   transitions.add(e);
            } else if (e instanceof Note) {
                ds = createShapeFromElement(parentContainer, e);
            } else if (e instanceof GeneralConstraint) {
                ds = createShapeFromElement(parentContainer, e);
            } else if (e instanceof OSMXObjectType) {
                ds = createShapeFromElement(parentContainer, e);
            } else if (e instanceof OSMXObjectSetType) {
                ds = rebuildObjectSet((OSMXObjectSetType) e, parentContainer);
            } else if (e instanceof OSMXStateType) {// Added by Muhammed
                ds = createShapeFromElement(parentContainer, e);
            }else if (e instanceof OSMXTransitionType){//Added by Muhammed
            	ds = createShapeFromElement(parentContainer, e);
            }else if (e instanceof OSMXConjunctionType) {//Added by Muhammed
                conjunctions.add(e);
            }
           
            
            if (ds != null) ds.addItemListener(this);
        }
        // connector shapes: Relationship Set, Gen/Spec, Association, Aggregation,
        //                   Transition
        for (Iterator i = relSets.iterator(); i.hasNext();) {
            ds = createShapeFromElement(parentContainer, (OSMXElement) i.next());
        }
        for (Iterator i = genSpecs.iterator(); i.hasNext();) {
            ds = createShapeFromElement(parentContainer, (OSMXElement) i.next());
        }
        
        /**
         * Added by Muhammed
         */
        for (Iterator i = conjunctions.iterator(); i.hasNext();) {
            ds = createShapeFromElement(parentContainer, (OSMXElement) i.next());
        }
        for (Iterator i = aggregs.iterator(); i.hasNext();){
        	ds = createShapeFromElement(parentContainer, (OSMXElement) i.next());
        }

    } // method rebuildOSM

    /** Creates the {@link edu.byu.deg.OntologyEditor.shapes.ObjectSetShape} for the
     * specified element and invokes {@link #rebuildOSM} on the nested <CODE>OSM</CODE>
     * element if the object set is a high-level element.
     * @param elem The object set element.
     * @param parentContainer The shape or canvas that will contain the new object set shape.
     * @return The newly created shape.
     */
    protected ObjectSetShape rebuildObjectSet(OSMXObjectSetType elem, Container parentContainer) {
        ObjectSetShape oss = (ObjectSetShape) createShapeFromElement(parentContainer, elem);
        if (elem.isHighLevel() && elem.isSetOSM()) {
            rebuildOSM((OSMXOSMType) elem.getOSM());
        }
        return oss;
    } // method rebuildObjectSet

    /** Returns the shape that corresponds to the specified element.
     * @param elem The element whose corresponding shape we seek.
     * @return The shape for the given element, or <CODE>null</CODE> if no such shape exists.
     */
    public DrawShape findShapeForElement(OSMXElement elem) {
        DrawShape shape = (DrawShape) shapesByElement.get(elem);
        return shape;
    }

    // This class provides an object that overlays the canvas and visually
    // displays a guide for a dragging operation.  The appearance of the guide
    // depends on the nature of the drag operation -- selection, creation of a
    // connector, or resizing a shape.
    private class DragPanel extends JPanel {
        public static final int SELECT_BOX_MODE = 0;
        public static final int CONNECTOR_MODE = 1;
        public static final int RESIZE_MODE = 2;

        private Point dragSource;
        private Point dragDest;
        private int dragMode = SELECT_BOX_MODE;

        public int getDragMode() {
            return dragMode;
        }

        public void setDragMode(int mode) {
            dragMode = mode;
        }

        public Point getDragSource() {
            return dragSource;
        }

        public void setDragSource(Point pt) {
            dragSource = pt;
        }

        public Point getDragDest() {
            return dragDest;
        }

        public void setDragDest(Point pt) {
            dragDest = pt;
        }

        public boolean contains(int x, int y) {
            return false;
        }

        public boolean contains(Point pt) {
            return false;
        }

        public void paintComponent(Graphics g) {
            super.paintComponent(g);
            if (dragMode == SELECT_BOX_MODE) paintSelectionBox(g);
            else if (dragMode == CONNECTOR_MODE) paintConnectorGuide(g);
            else if (dragMode == RESIZE_MODE) paintResizeGuide(g);
        }

        private void paintSelectionBox(Graphics g) {
            if (getDragSource() == null || getDragDest() == null) return;
            Graphics2D g2d = (Graphics2D) g.create();
            g2d.setStroke(new BasicStroke(1, BasicStroke.CAP_SQUARE, BasicStroke.JOIN_MITER, 10.0f, new float[] {2.0f}, 0.0f));
            Rectangle selBox = DrawShape.normalizeRect(dragSource, dragDest);
            g2d.setXORMode(Color.white);
            g2d.drawRect(selBox.x, selBox.y, selBox.width-1, selBox.height-1);
        }

        private void paintConnectorGuide(Graphics g) {
            // TODO
            if (getDragSource() == null || getDragDest() == null) return;
            Graphics2D g2d = (Graphics2D) g.create();
            Color endColor = (srcShape != null && destShape != null ? new Color(0, 160, 0) : Color.red);
            Color guideColor = (srcShape != null && destShape != null ? Color.black : Color.gray);
            BasicStroke dotStroke = new BasicStroke(1, BasicStroke.CAP_SQUARE, BasicStroke.JOIN_MITER, 10.0f, new float[] {1.0f, 3.0f}, 0.0f);
            BasicStroke solidStroke = new BasicStroke(1);
            BasicStroke guideStroke = (srcShape != null && destShape != null ? solidStroke : dotStroke);
            GeneralPath arrowHead = new GeneralPath();

            Point dest = (destShape == null ? dragDest : PointUtilities.lowResPoint(destShape.getConnectionPoint()));
            if (destShape != null) dest = SwingUtilities.convertPoint(destShape, dest, this);

            g2d.setColor(endColor);
            if (srcShape != null) {
                Point srcCtr = PointUtilities.lowResPoint(srcShape.getConnectionPoint());
                srcCtr = SwingUtilities.convertPoint(srcShape, srcCtr, this);
                if (destShape != null) {
                    AffineTransform af2 = new AffineTransform();
                    if (srcShape != destShape) {
                        int dy = dest.y - srcCtr.y;
                        int dx = dest.x - srcCtr.x;
                        af2.translate(dest.x, dest.y);
                        af2.rotate(Math.atan2(dy, dx));
                    } else {
                        af2.translate(srcCtr.x, srcCtr.y + srcShape.getHeight() / 2);
                        af2.rotate(-Math.PI/2);
                    }
                    //g2d.setTransform(af2);
                    arrowHead.moveTo(0, 0);
                    arrowHead.lineTo(-8, -4);
                    arrowHead.lineTo(-8, 4);
                    arrowHead.lineTo(0, 0);
                    arrowHead.transform(af2);
                }
                g2d.setColor(guideColor);
                g2d.setStroke(guideStroke);
                if (srcShape != destShape) {
                    g2d.drawLine(srcCtr.x, srcCtr.y, dest.x, dest.y);
                } else {
                    g2d.drawLine(srcCtr.x, srcCtr.y, srcCtr.x + srcShape.getWidth(), srcCtr.y);
                    g2d.drawLine(srcCtr.x + srcShape.getWidth(), srcCtr.y, srcCtr.x + srcShape.getWidth(), srcCtr.y + srcShape.getHeight());
                    g2d.drawLine(srcCtr.x + srcShape.getWidth(), srcCtr.y + srcShape.getHeight(), srcCtr.x, srcCtr.y + srcShape.getHeight());
                    g2d.drawLine(srcCtr.x, srcCtr.y + srcShape.getHeight(), srcCtr.x, srcCtr.y + srcShape.getHeight() / 2);
                }
                g2d.setColor(Color.white);
                g2d.fill(arrowHead);
                g2d.setColor(endColor);
                g2d.setStroke(new BasicStroke(2));
                g2d.draw(arrowHead);
                g2d.fillOval(srcCtr.x - 4, srcCtr.y - 4, 8, 8);
                g2d.setStroke(guideStroke);
            }
            if (srcShape == null || destShape == null) {
                g2d.setStroke(new BasicStroke(2));
                g2d.drawOval(dragDest.x - 10, dragDest.y - 10, 20, 20);
                g2d.setStroke(new BasicStroke(3));
                g2d.drawLine(dragDest.x - 6, dragDest.y - 6, dragDest.x + 6, dragDest.y + 6);
            }
        }

        private void paintResizeGuide(Graphics g) {
            // TODO
        }
    }
}

