package edu.byu.deg.OntologyEditor;

import edu.byu.deg.osmx.OSMXDocument;
import javax.xml.bind.JAXBException;

import edu.byu.deg.osmx.OSMXDocument.UnspecifiedOutputException;
import edu.byu.deg.osmx.binding.OSM;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyVetoException;
import java.io.File;
import java.io.IOException;
import javax.swing.JFileChooser;
import javax.swing.JInternalFrame;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.event.InternalFrameAdapter;
import javax.swing.event.InternalFrameEvent;


//
//
// OntologyFrame
//
//
public class OntologyCanvasWindow extends AbstractInternalFrame implements ActionListener {
  private OntologyCanvas m_canvas;
  private DataFrameEditorWindow m_dataFrame;
  private MacroLexiconEditorWindow m_macroEditor;
  private FrameInfo m_macroEditorInfo;
  private OSMXDocument ontologyDoc;

  /**
   * Constants for cascading windows
   */
  public static final int CASCADE_OFFSET = 22;
  public static final int DEFAULT_HEIGHT = 450;
  public static final int DEFAULT_WIDTH = 550;
  public static final int MIN_HEIGHT = 300;
  public static final int MIN_WIDTH = 400;
  
  private static final int HOR_SCROLL_RATE = 50;
  private static final int VERT_SCROLL_RATE = 30;

  // precondition: doc has an existing ontology model root ('new' or 'open' has been invoked)
  // REVISIT; not sure this needs to be a precondition
  public OntologyCanvasWindow(OSMXDocument doc, OntologyEditor editor) {
    super(doc.getTitle(), editor, null);
    ontologyDoc = doc;
    m_editor = editor;
    m_dataFrame = null;
    m_macroEditor = null;

    OSM modelRoot = ontologyDoc.getModelRoot();
    Point loc;
    Dimension dim;
    if (modelRoot != null) {
      int modelWidth = modelRoot.getWidth();
      if (modelWidth < MIN_WIDTH) modelWidth = MIN_WIDTH;
      int modelHeight = modelRoot.getHeight();
      if (modelHeight < MIN_HEIGHT) modelHeight = MIN_HEIGHT;
      int modelX = modelRoot.getX();
      int modelY = modelRoot.getY(); 
      loc = new Point(modelX, modelY);
      dim = new Dimension(modelWidth, modelHeight);
    } else {
      loc = new Point();
      dim = new Dimension(DEFAULT_WIDTH, DEFAULT_HEIGHT);
    }

    // Make sure the frame fits inside the desktop
    int newH = (int) dim.getHeight();
    int newW = (int) dim.getWidth();
    int newX = (int) loc.getX();
    int newY = (int) loc.getY();
    int deskH = editor.desktop.getHeight();
    int deskW = editor.desktop.getWidth();

    while (newX + MIN_WIDTH > deskW)
      newX = newX + MIN_WIDTH - deskW;
    while (newY + MIN_HEIGHT > deskH)
      newY = newY + MIN_HEIGHT - deskH;

    if (newH + newY > deskH)
      newH = deskH - newY;
    if (newW + newX > deskW)
      newW = deskW - newX;

    dim.setSize(newW,newH);
    loc.setLocation(newX,newY);
    setLocation(loc);
    
    setSize(dim);

    m_canvas = new OntologyCanvas(this);
    JScrollPane pane = new JScrollPane(m_canvas);
    pane.getHorizontalScrollBar().setUnitIncrement(HOR_SCROLL_RATE);
    pane.getVerticalScrollBar().setUnitIncrement(VERT_SCROLL_RATE);
    //pane.setViewportBorder(new SoftBevelBorder(BevelBorder.LOWERED));
    getContentPane().add(pane, "Center");
    m_canvas.setAutoscrolls(true);
    m_canvas.rebuildFromDocument();
    
    // on close do nothing by default; let internal frame listener
    // intercept close action to check if doc is modified
    this.setDefaultCloseOperation(JInternalFrame.DO_NOTHING_ON_CLOSE);
    
    this.addInternalFrameListener(new InternalFrameAdapter() {
        public void internalFrameClosing(InternalFrameEvent e) {
            if (ontologyDoc.isModified()) {
                // Prompt user whether to save changes
                Object[] options = { "Save changes", "Discard changes", "Cancel" };
                int dialogResult = JOptionPane.showOptionDialog(m_editor,
      			                  "Document is modified. Save changes?",
      			                  "Document Is Modified",
      			                  JOptionPane.YES_NO_CANCEL_OPTION,
      			                  JOptionPane.QUESTION_MESSAGE,
      			                  null,
      			                  options,
      			                  options[2]);
                switch (dialogResult) {
                	case JOptionPane.YES_OPTION:
                	    if (!saveAsDocument()) return;
                	    else break;
                	case JOptionPane.NO_OPTION:
                	    break;
                	default:
                	    return;
                }
            }
            // a bit of cleanup
            ontologyDoc = null;
            m_canvas = null;
            if (getDataFrame() != null)
                getDataFrame().doDefaultCloseAction();
            if (getMacroFrame() != null)
                getMacroFrame().doDefaultCloseAction();
            dispose();
        }
    });
  } // constructor

  public OntologyEditor getEditor() {
    return m_editor;
  } // method getEditor

  public OSMXDocument getDocument() {
    return ontologyDoc;
  } // method getDocument

  public OntologyCanvas getCanvas() {
    return m_canvas;
  } // method getCanvas

  public void setDataFrame(DataFrameEditorWindow dataFrame) {
    m_dataFrame = dataFrame;
  } // method setDataFrame

  public void dataFrameClosed() {
    m_dataFrame = null;
  } // method dataFrameClosed

  public void dataFrameOpened() {
    m_dataFrame = new DataFrameEditorWindow(this, m_editor);
  } // method dataFrameOpened

  public DataFrameEditorWindow getDataFrame() {
    return m_dataFrame;
  } // method getDataFrame

  public void showDataFrame() {
    try {
      // find selected object
      if (m_dataFrame == null) {
        dataFrameOpened();
        m_dataFrame.setVisible(false);
        m_editor.addFrame(m_dataFrame);
      }

      if (m_dataFrame.isIcon())
        m_dataFrame.setIcon(false);
      m_dataFrame.toFront();
      m_dataFrame.setVisible(true);
      m_dataFrame.setSelected(true);
    }
    catch (PropertyVetoException e) {
        /* REVISIT: handle better? */
    }
  } // method showDataFrame

  public boolean saveDocument(boolean save_local) {
    // do something about saving remote vs saving local
    try {
      ontologyDoc.saveDocument();
      setTitle(ontologyDoc.getTitle());
      return true;
    }
    catch (UnspecifiedOutputException e) {
      ontologyDoc.getModelRoot().setX((int) getLocation().getX());
      ontologyDoc.getModelRoot().setY((int) getLocation().getY());
      ontologyDoc.getModelRoot().setWidth((int) getSize().getWidth());
      ontologyDoc.getModelRoot().setHeight((int) getSize().getHeight());

      return saveAsDocument();
    }
    catch (IOException e) {
      JOptionPane.showMessageDialog(this, e, "Error Saving File", JOptionPane.ERROR_MESSAGE);
      return false;
    }
    catch (JAXBException e) {
      JOptionPane.showMessageDialog(this, e, "Configuration Error", JOptionPane.ERROR_MESSAGE);
      return false;
    }
  } // method saveDocument
  
  public boolean saveAsDocument() {
    ontologyDoc.getModelRoot().setX((int) getLocation().getX());
    ontologyDoc.getModelRoot().setY((int) getLocation().getY());
    ontologyDoc.getModelRoot().setWidth((int) getSize().getWidth());
    ontologyDoc.getModelRoot().setHeight((int) getSize().getHeight());
    File file = null;

    ExtFileChooser chooser = new ExtFileChooser();
    chooser.setFileFilter(new OSMFileFilter());

      String filename = ExtFileChooser.m_lastDirectory +
      File.separator + ontologyDoc.getTitle();
      if (filename.indexOf("Untitled") == -1) {
        file = new File(filename);
      }

    if (file != null)
      chooser.setSelectedFile(file);

    int returnVal = chooser.showSaveDialog(this);
    if (returnVal == JFileChooser.APPROVE_OPTION) {
      try {
        String name = chooser.getSelectedFile().toString();
        if (!name.regionMatches(true, name.length() - 4, ".xml", 0, 4))
          name = name.concat(".xml");

        file = new File(name);
        //m_canvas.setOrdering();
        ontologyDoc.saveAsDocument(file);
        m_menuItem.setText(ontologyDoc.getTitle());
        setTitle(ontologyDoc.getTitle());
        invalidate();
        repaint();
        return true;
      }
      catch (IOException e) {
        JOptionPane.showMessageDialog(this, e, "Error Saving File", JOptionPane.ERROR_MESSAGE);
      }
      catch (JAXBException e) {
        JOptionPane.showMessageDialog(this, e, "Configuration Error", JOptionPane.ERROR_MESSAGE);
      }
    }
    return false;
  } // method saveAsDocument

  public void dispose() {
    super.dispose();
    m_frame = null;
    m_dataFrame = null;
    m_canvas = null;
    m_editor = null;
    
    // we don't want to catch any document change events from this point
    if (ontologyDoc != null) ontologyDoc.setGeneratesDocumentChangeEvents(false);
    ontologyDoc = null;
  }

  public boolean saveOSML(boolean save_local) {
    // TODO: Figure all of this out ... use XSLT?
    return false;
  } // method saveOSML

  /**
   * print current active frame
   * @return
   */
  public void canvaswindowPrint()  {
  	m_editor.repaint();
	m_canvas.canvasPrint();
  }
  

  public boolean exportDocument() {
    m_editor.repaint();
    ExtFileChooser chooser = new ExtFileChooser();
    chooser.setFileFilter(new EPSFileFilter());
    // chooser.addChoosableFileFilter(new PSFileFilter());
    int returnVal = chooser.showSaveDialog(this);
    if (returnVal == JFileChooser.APPROVE_OPTION) {
      try {
        String name = chooser.getSelectedFile().toString();
        boolean isPS = name.regionMatches(true,name.length()-3,".ps",0,3);
        if (!(name.regionMatches(true,name.length()-4,".eps",0,4)||isPS))
          name = name.concat(".eps");

        File file = new File(name);
        if (m_canvas.exportDocument(file, isPS)) {
          /*
          JOptionPane.showMessageDialog(this, "The file has been exported as " + file.getName(),
          "Export to PostScript", JOptionPane.PLAIN_MESSAGE);
          */
          return true;
        }
        throw new IOException("blah"); // TODO: better error message!
      }
      catch (IOException e) {
        JOptionPane.showMessageDialog(this, e, "Error", JOptionPane.ERROR_MESSAGE);
      }
    }
    return false;
  } // method exportDocument

  public boolean toHypergraph() {
    String msg = "This operation is permanent.  Are you sure you wish to convert to a hypergraph?";
    int input = JOptionPane.showConfirmDialog(this, msg, "Confirmation", JOptionPane.YES_NO_OPTION);
    if(input==1)
      return false;
    return false;
  } // method toHypergraph

  public boolean outputScheme() {
    return m_canvas.outputScheme();
  } // method outputScheme

  public void actionPerformed(ActionEvent evt) {
    // trap any window-oriented events before passing the rest on to the canvas.

    String cmd = evt.getActionCommand();

    if (cmd.equals("editDataFrame")) {
      showDataFrame();
    }
  } // method actionPerformed

  public void setMacroFrame(MacroLexiconEditorWindow macroFrame) {
    m_macroEditor = macroFrame;
  } // method setMacroFrame

  public MacroLexiconEditorWindow getMacroFrame() {
    return m_macroEditor;
  } // method getMacroFrame

  public void macroFrameClosed() {
    m_macroEditor = null;
  } // method macroFrameClosed

  public void macroFrameOpened() {
    m_macroEditor = new MacroLexiconEditorWindow(m_macroEditorInfo, m_frame, m_editor);


      m_macroEditor.setRootModelnew(m_canvas.getMacroRootDoc());//Z.H
//    m_macroEditor.setRootModel(m_canvas.getMacroRootModel());
  } // method macroFrameOpened

  public void showMacroEditor() {
    try {
      if (m_macroEditor == null) {
        macroFrameOpened();
        m_macroEditor.setVisible(false);
        m_editor.addFrame(m_macroEditor);
      }

      // find selected object
      if (m_macroEditor.isIcon())
        m_macroEditor.setIcon(false);
      m_macroEditor.toFront();
      m_macroEditor.setVisible(true);
      m_macroEditor.setSelected(true);
    }
    catch (PropertyVetoException e) {
        /* REVISIT: handle better? */
    }
  } // method showMacroEditor
  
  public void updateTitle() {
    setTitle(ontologyDoc.getTitle());
  }

} /* class OntologyCanvasWindow */