package edu.byu.deg.OntologyEditor;

import edu.byu.deg.OntologyEditor.actions.GeneralAction;
import edu.byu.deg.OntologyEditor.actions.OpenAction;
import edu.byu.deg.osmx.OSMXDocument;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.beans.PropertyVetoException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.Reader;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.Stack;
import java.util.StringTokenizer;
import javax.swing.*;
import javax.swing.event.InternalFrameAdapter;
import javax.swing.event.InternalFrameEvent;
import javax.xml.bind.JAXBException;

import java.awt.BorderLayout;
import java.awt.Color; 
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;

/** The OntologyEditor is both the main executable class and main frame of the
 * Integrated Ontology Development Environment.
 * <BR>
 * The editor is designed to run either as a local Java application or as an
 * applet.  The <CODE>OntologyEditor.properties</CODE> file contains important
 * initialization parameters, including the locations of resource files for
 * actions, menus, toolbars, icons, and so forth.
 * @author Kimball Hewett
 * @author Alan Wessman
 * @version 2.0
 * @see AbstractInternalFrame
 * @see OntologyCanvasWindow
 * @see OntologyCanvas
 * @see ActionLoader
 * @see MenuLoader
 * @see ToolbarLoader
 */

public class OntologyEditor extends JApplet {
  /** */  
  public static URL baseURL = null;
  
  /**
   * This is used to tell whether the editor is run as applet or application.
   */
  public static boolean isStandalone = false;
  
  /**
   * The resource bundle represented by the OntologyEditor.properties file.
   */
  private static ResourceBundle resources;
  
  private AboutDialog m_aboutDlg;
  
  /**
   * The current state of the OntologyEditor. Represents the currently selected state button on the
   * state toolbar.
   */
  public String objectState = "";
  
  private String currentShape = "";
  
  /**
   * Identifies the name of the main tool bar identifier in the resource bundle.
   */
  private static final String MAIN_TOOLBAR_NAME = "Main";
  private static final String DRAW_TOOLBAR_NAME = "Draw";
  
  private static final String ACTIONS_RESOURCE_KEY  = "ActionsResourceFile";
  private static final String MENUS_RESOURCE_KEY    = "MenusResourceFile";
  private static final String TOOLBARS_RESOURCE_KEY = "ToolbarsResourceFile";
  
  private static final String WINDOW_MENU_NAME = "Window";
  
  private static final String SMALL_PROGRAM_ICON = "smallProgIcon";
  private static final String SMALL_DOCUMENT_ICON = "smallDocIcon";
  private static final String CLEAR_ICON = "clearIcon";
  
  /**
   * A handle to the desktop containing all of the JInternalFrames.
   */
  public JDesktopPane desktop;
  
  public static Properties props = null;
  
  private boolean initErrorOccurred = false;
  public int numOpenWindows;
  public int numWindowsInList;
  public URL daml_files[];
  public URL osm_files[];
  public int current_ont_index = 0;
  public JProgressBar progressBar;
  
  private Map actions;
  private Map actionGroups;
  private Map menus;
  private Map toolbars;
  private Map menuItemGroups;
  private Map toolbarButtonGroups;
  
  private JMenuBar mainMenu;
  
  static {
    try {
      resources = ResourceBundle.getBundle("edu.byu.deg.OntologyEditor.OntologyEditor");
    }
    catch (MissingResourceException mre) {
      mre.printStackTrace(System.err);
      System.err.println("OntologyEditor.properties not found");
    }
  }
  
  
  /**
   * The constructor loads all actions, menus and toolbars as appropriate.
   */
  public OntologyEditor() {
    try {
      // Look and Feel
      setLookAndFeel();

      // load the Actions from the resource file
      loadActions();

      // load the Menus from the resource file
      loadMenus();

      // load the Toolbars from the resource file
      loadToolbars();

      // set up visual components
      initializeEditorGUI();

      // load user-specific settings
      loadProperties();
    }
    catch (Exception e) {
      initErrorOccurred = true;
      e.printStackTrace();
    }
  }
  
  public boolean isInitError() {
    return initErrorOccurred;
  }  
  
  /** This is the main entry point into the program. It simply creates the main frame
   * to prepare for user input.
   * @param s The array of input arguments.  Currently all input arguments are ignored.
   */
  public static void main(String s[]) {
    final String BEFORE_OLDEST_JVM_CLASS_VERSION = "44.0";
    final String SUPPORTED_JVM_CLASS_VERSION = "48.0";
    final String SUPPORTED_JVM = "1.4";
    
    try {
      String vers = System.getProperty("java.class.version", BEFORE_OLDEST_JVM_CLASS_VERSION);
      //if (vers.compareTo(SUPPORTED_JVM_CLASS_VERSION) < 0) {
      if (Float.parseFloat(vers) < Float.parseFloat(SUPPORTED_JVM_CLASS_VERSION)) {
        System.err.println("!!!WARNING: This application is supported only for a " +
        SUPPORTED_JVM + " or higher version JVM!!!");
      }
      
      // main() is not called when run as applet, so isStandalone = true
      OntologyEditor.isStandalone = true;
      
      OntologyEditor editor = new OntologyEditor();
      JFrame frame = new JFrame();
      
      // Set Title
      String title = editor.getResourceString("Title");
      if (title == null)
        title = "Ontology Editor";
      frame.setTitle(title);
      
      URL frameIcon = editor.getResource(SMALL_PROGRAM_ICON);
      if (frameIcon != null) frame.setIconImage(new ImageIcon(frameIcon).getImage());
      
      // default on close is to do nothing; window listener handles close events
      // so we can detect modified documents
      frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
      frame.setBackground(Color.lightGray);
      frame.addWindowListener(new WindowCloser(editor));
      frame.getContentPane().add(editor, BorderLayout.CENTER);
      editor.init();
      editor.start();
      frame.setSize(800, 600);
      Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
      frame.setLocation((d.width - frame.getSize().width) / 2, (d.height - frame.getSize().height) / 2);
      frame.setVisible(true);
    }
    catch (Throwable t) {
      System.err.println("Uncaught exception: " + t);
      t.printStackTrace(System.err);
    }
  }
  
  /** Returns whether the editor is being run as an applet.
   * @return <CODE>true</CODE> if the editor is executed as an applet; false otherwise.
   */  
  public boolean isApplet() {
    return !isStandalone;
  }
  
  private void setLookAndFeel() {
    try {
      UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
    }
    catch (Exception exc) {
      System.err.println("Error loading look and feel: " + exc);
    }
  }
  
  private void loadActions() {
    URL actionResourceURL = getResource(ACTIONS_RESOURCE_KEY);
    if (actionResourceURL == null)
      throw new MissingResourceException("Cannot load actions", "Actions.xml", ACTIONS_RESOURCE_KEY);
    
    // initialize actions map
    if (actions == null) {
      actions = new HashMap();
    }
    else {
      actions.clear();
    }
    
    // initialize actionGroups map
    if (actionGroups == null) {
      actionGroups = new HashMap();
    }
    else {
      actionGroups.clear();
    }
    
    try {
      ActionLoader.load(this, actionResourceURL, actions, actionGroups);
    }
    catch (Exception e) { e.printStackTrace(); }
    
    enableActions();
  } // method loadActions
  
  private void loadMenus() {
    URL menuResourceURL = getResource(MENUS_RESOURCE_KEY);
    if (menuResourceURL == null)
      throw new MissingResourceException("Cannot load menus", "Menus.xml", MENUS_RESOURCE_KEY);
    
    // initialize actions map
    if (menus == null) {
      menus = new HashMap();
    }
    else {
      menus.clear();
    }
    
    if (menuItemGroups == null) {
      menuItemGroups = new HashMap();
    }
    else {
      menuItemGroups.clear();
    }
    
    try {
      MenuLoader.loadMenus(menuResourceURL, menus, menuItemGroups, actions);
    }
    catch (Exception e) { e.printStackTrace(); }
  }
  
  private void loadToolbars() {
    URL toolbarResourceURL = getResource(TOOLBARS_RESOURCE_KEY);
    if (toolbarResourceURL == null)
      throw new MissingResourceException("Cannot load toolbars", "Toolbars.xml", TOOLBARS_RESOURCE_KEY);
    
    // initialize actions map
    if (toolbars == null) {
      toolbars = new HashMap();
    }
    else {
      toolbars.clear();
    }
    
    if (toolbarButtonGroups == null) {
      toolbarButtonGroups = new HashMap();
    }
    else {
      toolbarButtonGroups.clear();
    }
    
    try {
      ToolbarLoader.loadToolbars(toolbarResourceURL, toolbars, toolbarButtonGroups, actions);
    }
    catch (Exception e) { e.printStackTrace(); }
  }
  
  private void initializeEditorGUI() {
    
    final String [] mainMenuNames = {"File", "Edit", "Shape", "Format", "Tools", WINDOW_MENU_NAME, "Help"};
    
    desktop = new JDesktopPane();
    
    if (mainMenu != null) mainMenu = null;
    mainMenu = new JMenuBar();
    for (int i = 0; i < mainMenuNames.length; i++) {
      JMenu menu = (JMenu) menus.get(mainMenuNames[i]);
      if (menu != null) mainMenu.add(menu);
    }
    setJMenuBar(mainMenu);
    
    JToolBar mainToolbar = (JToolBar) toolbars.get(MAIN_TOOLBAR_NAME);
    if (mainToolbar != null) getContentPane().add("North", mainToolbar);
    
    JPanel panel = new JPanel();
    panel.setLayout(new BorderLayout());
    JToolBar drawToolbar = (JToolBar) toolbars.get(DRAW_TOOLBAR_NAME);
    if (drawToolbar != null) panel.add("North", drawToolbar);
    
    JScrollPane desktopScroller = new JScrollPane(desktop, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    panel.add("Center", desktopScroller);
    
    getContentPane().add("Center", panel);
    
    numOpenWindows = 0;
    numWindowsInList = 0;
    
    setActionGroupEnabled(ActionGroup.ALWAYS_ENABLED_ACTIONS, true);
    setActionGroupEnabled(ActionGroup.MULTIPLE_WINDOW_ACTIONS, false);
    setActionGroupEnabled(ActionGroup.EDITOR_ACTIONS, false);
  }
  
  private void loadProperties() {
    try {
      OntologyEditor.props = new Properties();
      String filename = System.getProperty("user.home") + "/OntologyEditor.properties";
      
      File prop_file = new File(filename);
      
      if (!prop_file.exists()) {
        prop_file.createNewFile();
        props.setProperty("OntologyEditor.lastDir", System.getProperty("user.home"));
        FileOutputStream out = new FileOutputStream(filename);
        props.store(out, "Last directory used to open or save a document");
        out.close();
      }
      else {
        FileInputStream in = new FileInputStream(filename);
        props.load(in);
        in.close();
      }
    }
    catch (Exception e) {e.printStackTrace();}
  }
  
  /** Sets the enabled state of all {@link javax.swing.Action}s in the specified group
   * to the value of <CODE>enable</CODE>.
   * @param groupName Name of the group of Actions to enable or disable.
   * @param enable The desired enabled state.
   */  
  public void setActionGroupEnabled(String groupName, boolean enable) {
    ActionGroup group = (ActionGroup) actionGroups.get(groupName);
    if (group == null) return;
    for(Iterator i = group.iterator(); i.hasNext();) {
      Action a = (Action) i.next();
      a.setEnabled(enable);
    }
  }
  
  /** Invokes {@link #setActionGroupEnabled} on all action groups.
   * @param enable The desired enabled state.
   */  
  public void setAllActionGroupsEnabled(boolean enable) {
    for(Iterator i = actionGroups.keySet().iterator(); i.hasNext();) {
      setActionGroupEnabled((String) i.next(), enable);
    }
  }
  
  private URL[] makeURLArray(String files_str, String prot, String host) {
    try {
      String file;
      int cnt = 0;
      
      StringTokenizer strtok = new StringTokenizer(files_str, "^");
      URL urls[] = new URL[strtok.countTokens()];
      while (strtok.hasMoreTokens()) {
        file = strtok.nextToken();
        urls[cnt++] = new URL(prot, host, file);
      }
      return urls;
    }
    catch (MalformedURLException e) {
      e.printStackTrace();
    }
    return null;
  }
  
  public Map getMenus() {
    return menus;
  }
  
  public Map getActions() {
    return actions;
  }
  
  public Map getActionGroups() {
    return actionGroups;
  }
  
  public Map getMenuItemGroups() {
    return menuItemGroups;
  }
  
  //Initialize the applet
  public void init() {
    if (initErrorOccurred) {
      // TODO: show a message and disable everything else
      setAllActionGroupsEnabled(false);
      return;
    }
    
    OntologyEditor.baseURL = this.getCodeBase();
    
    if (!OntologyEditor.isStandalone && baseURL != null) {
      
      // get the list of loadable DAML files
      daml_files = makeURLArray(getParameter("daml_files"), baseURL.getProtocol(),
      baseURL.getHost());
      
      // get the list of loadable OSM files
      osm_files = makeURLArray(getParameter("osm_files"), baseURL.getProtocol(),
      baseURL.getHost());
      
      // determine the ontology to load by default
      String current_ont = getParameter("current_ontology");
      for (int i = 0; i < osm_files.length; i++) {
        if (osm_files[i].getFile().indexOf(current_ont) != -1) {
          current_ont_index = i;
          break;
        }
      }
      
      // load the default ontology
      OpenAction opener = (OpenAction) actions.get(OpenAction.KEY);
      opener.openDocument(osm_files[current_ont_index]);
    } // check on this editor instance == applet with successfully found baseURL
  }
  
  //Start the applet
  public void start() { /* do nothing */ }
  
  public void newOntology() throws JAXBException {
      // must happen BEFORE initializing OntologyCanvasWindow
      OSMXDocument doc = OSMXDocument.newDocument(); 
      
	  OntologyCanvasWindow frame = new OntologyCanvasWindow(doc, this);
	  frame.setDefaultCloseOperation(JInternalFrame.DO_NOTHING_ON_CLOSE);
	  addFrame(frame);
	  frame.toFront();
	  frame.setVisible(true);
	
	  try {
	    frame.setSelected(true);
	  }
	  catch (PropertyVetoException e) {
	      /* REVISIT: handle better? */
	  }
  } // method newOntology
  
  //Stop the applet
  public void stop() { /* do nothing */ }
  
  //Destroy the applet
  public void destroy() { /* do nothing */ }
  
  //Get Applet information
  public String getAppletInfo() {
    return "Applet Information";
  }
  
  //Get parameter info
  public String[][] getParameterInfo() {
    String[][] pinfo = {
      {"param_1", "String", "My first param"},
      {"param_2", "int", "My second param"},
    };
    return pinfo;
  }
  
  /** Determine the URL from which the editor as applet is being invoked.
   * @return A {@link java.net.URL URL} that represents the result of
   * {@link java.applet.Applet#getCodeBase} unless the invoking URL is via a
   * <CODE>file://</CODE> protocol (or the editor is run as a standalone app),
   * in which case the result is <CODE>null</CODE>.
   */  
  public URL getCodeBase() {
    if (OntologyEditor.isStandalone) return null;
    else {
      URL codebase = super.getCodeBase();
      if (codebase.toString().startsWith("file"))
        return null;
      else
        return codebase;
    }
  }
  
  public void insert_component_at(JComponent comp, int insert_index) {
    JToolBar toolbar = getMainToolbar();
    JComponent tmp;
    
    Stack stack = new Stack();
    
    for (int i = toolbar.getComponentCount()-1; i >= insert_index; i--) {
      tmp = (JComponent)toolbar.getComponent(i);
      stack.push(tmp);
      toolbar.remove(tmp);
    }
    stack.add(comp);
    
    while(!stack.empty()) {
      tmp = (JComponent)stack.pop();
      toolbar.add(tmp);
    }
    toolbar.updateUI();
  }

/*
  private void insert_progress_bar(int max_progress) {
    JToolBar toolbar = getMainToolbar();
    progressBar	= new JProgressBar();
    progressBar.setMinimum(0);
    progressBar.setMaximum(max_progress);
    progressBar.setValue(progressBar.getMinimum());
    progressBar.setMaximumSize(new Dimension(100, 25));
    insert_component_at(progressBar, 3);
  }
  
  private void remove_progress_bar() {
    JToolBar toolbar = getMainToolbar();
    toolbar.remove(progressBar);
    toolbar.updateUI();
  }

  private int old_priority;
  private edu.byu.deg.OntologyEditor.actions.OpenDAMLAction daml_opener;
  public void start_progressbar(edu.byu.deg.OntologyEditor.actions.OpenDAMLAction opener) {
    this.daml_opener = opener;
    //Create a progress bar and place it after the openDAML toolbar button
    insert_progress_bar(daml_opener.get_max_progress());
    
    //Create a timer.
    timer = new Timer(250,
    new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        progressBar.setValue(daml_opener.getCurrentProgress());
        if (daml_opener.done()) {
          Toolkit.getDefaultToolkit().beep();
          if (timer != null) timer.stop();
        }
      }
    });
    timer.start();
    
    old_priority = Thread.currentThread().getPriority();
    Thread.currentThread().setPriority(Thread.MIN_PRIORITY);
  }
  
  public void stop_progressbar() {
    timer.stop();
    progressBar.setValue(progressBar.getMinimum());
    timer = null;
    Thread.currentThread().setPriority(old_priority);
    remove_progress_bar();
  }
   */
  
  /**
   * Currently coded to always start the program with a width equal to the needed width of the toolbars and
   * menu. The height is computed to be five times the required height of the toolbars and menu.
   */
  public Dimension getPreferredSize() {
    Dimension size = super.getPreferredSize();
    size.height = size.height * 5;
    return size;
  }
  
  protected JToolBar getMainToolbar() {
    Component[] comps = this.getContentPane().getComponents();
    
    for (int i=0; i<comps.length; i++)
      if (comps[i] instanceof JToolBar)
        return (JToolBar) comps[i];
    
    return null;
  }
  
  protected JToolBar getObjectToolbar() {
    Component[] comps = getContentPane().getComponents();
    
    for (int i=0; i<comps.length; i++)
      if (comps[i] instanceof JPanel) {
        Component[] comps2 = ((JPanel)comps[i]).getComponents();
        for (int j=0; i<comps2.length; j++)
          if (comps2[j] instanceof JToolBar)
            return (JToolBar) comps2[j];
      }
    
    return null;
  }
  
  /**
   * A helper function that retrieves a resource string identified by <code>name</code>.
   *
   * @param name Identifies the resource string to retrieve.
   * @return String
   */
  public String getResourceString(String name) {
    String str;
    try {
      str = resources.getString(name);
    }
    catch (MissingResourceException mre) {
      str = null;
    }
    return str;
  }
  
  /**
   * A helper function to retrieve a URL identified by <code>key</code>.
   * The URL typically specifies the path to an image for use by a toolbar.
   *
   * @param key Identifies the URL to retrieve from the resource bundle.
   * @return URL
   */
  protected URL getResource(String key) {
    String name = getResourceString(key);
    if (name != null) {
      URL url = this.getClass().getResource(name);
      return url;
    }
    return null;
  }
  
  /**
   * A helper function that tokenizes a string by breaking it up so that each word becomes its
   * own string and is placed into an array of strings that is returned.
   *
   * @param input The string to tokenize
   * @return String array with each token
   */
  protected String[] tokenize(String input) {
    if (input == null)
      return null;
    StringTokenizer token = new StringTokenizer(input);
    String cmd[] = new String[token.countTokens()];
    
    int i = 0;
    while (token.hasMoreTokens())
      cmd[i++] = token.nextToken();
    
    return cmd;
  }
  
  /**
   * Shuts down the application.
   */
  public void exit() {
      /*
    Component objects[] = desktop.getComponents();
    for (int i=0; i<objects.length; i++) {
      if (objects[i] instanceof JInternalFrame) {
        JInternalFrame frame = (JInternalFrame)objects[i];
        try {
          if (!frame.isClosed())
            frame.setClosed(true);
        }
        catch(PropertyVetoException e) {
        }
        if (!frame.isClosed())
          return; // exit canceled
      }
    }
    */
    
    setVisible(false);
    
    try {
      String filename = System.getProperty("user.home") + "/OntologyEditor.properties";
      FileOutputStream out = new FileOutputStream(filename);
      props.store(out, "Last directory used to open or save a document");
      out.close();
    }catch (Exception e)
    {e.printStackTrace();}
    
    destroy();
    
    System.exit(0);
  }
  
  public void setCurrentShape(String shapeActionKey) {
    currentShape = shapeActionKey;
  }
  
  public String getCurrentShape() {
    return currentShape;
  }
  
  /**
   * A helper function to find the active internal frame.
   *
   * @return The JInternalFrame representing the active frame.
   */
  public JInternalFrame getActiveFrame() {
    if (desktop == null) return null;
    Component objects[] = desktop.getComponents();
    for (int i=0; i<objects.length; i++) {
      if (objects[i] instanceof JInternalFrame) {
        JInternalFrame frame = (JInternalFrame)objects[i];
        if (frame.isSelected())
          return frame;
      }
    }
    return null;
  }
  
  public void showAbout() {
    if (m_aboutDlg == null)
      m_aboutDlg = new AboutDialog();
    m_aboutDlg.setVisible(true);
  }
  
  public void enableActions() {
    for (Iterator i = actions.values().iterator(); i.hasNext();) {
      Action action = (Action) i.next();
      if (action instanceof GeneralAction) {
        ((GeneralAction) action).updateState();
      }
      else action.setEnabled(true);
    }
  }
  
  public void addFrame(AbstractInternalFrame frame) {
    URL frameIcon = getResource(SMALL_DOCUMENT_ICON);
    if (frameIcon != null) frame.setFrameIcon(new ImageIcon(frameIcon));
    desktop.add(frame);
    JMenu menu = (JMenu) menus.get(WINDOW_MENU_NAME);
    
    numWindowsInList++;
    
    WindowFrameAction action = new WindowFrameAction(frame);
    String actionName = "A frame";
    if (frame instanceof OntologyCanvasWindow) {
      frame.addInternalFrameListener(new FrameWindowListener(this, (OntologyCanvasWindow) frame));
      actionName = ((OntologyCanvasWindow) frame).getDocument().getTitle();
    }
    action.putValue(Action.NAME, actionName);
    URL clearIconURL = getResource(CLEAR_ICON);
    action.putValue(Action.SMALL_ICON, new ImageIcon(clearIconURL));
    JMenuItem menuItem = new JMenuItem(action);
    action.setEnabled(true);
    menu.add(menuItem);
    
    frame.setMenuItem(menuItem);
  }
  
  /** */  
  public void openDocument(Reader r) { /* REVISIT: is this obsolete? */ }
  
  public void activateFrame(JInternalFrame frame) {
    try {
      frame.moveToFront();
      frame.setSelected(true);
    }
    catch (PropertyVetoException pve) { ; } // ignore
  }
  
  public void removeFrame(AbstractInternalFrame frame) {
    JMenu menu = (JMenu) menus.get(WINDOW_MENU_NAME);
    menu.remove(frame.getMenuItem());
  }
  
  public void tileFramesVertically() {
    int windowHeight = desktop.getHeight();
    int windowWidth = desktop.getWidth() / numOpenWindows;
    
    int currWindowX = 0;
    int currWindowY = 0;
    
    JInternalFrame activeFrame = getActiveFrame();
    
    if (windowHeight >= OntologyCanvasWindow.MIN_HEIGHT && windowWidth >= OntologyCanvasWindow.MIN_WIDTH) {
      JInternalFrame frames[] = desktop.getAllFrames();
      for (int i=0; i<frames.length; i++) {
        if (frames[i].isVisible() && frames[i] instanceof OntologyCanvasWindow) {
          try {
            frames[i].setMaximum(false);
            frames[i].reshape(currWindowX,currWindowY,windowWidth,windowHeight);
            frames[i].toFront();
            frames[i].setSelected(true);
          }
          catch (PropertyVetoException e)
          {
              /* REVISIT: handle better? */
          }
          
          currWindowX += windowWidth;
        }
      }
      if (activeFrame != getActiveFrame()) activateFrame(activeFrame);
    }
    else {
      JOptionPane.showMessageDialog(null, "There are too many open windows to tile.", "Error", JOptionPane.ERROR_MESSAGE);
    }
  }
  
  public void tileFramesHorizontally() {
    int windowHeight = desktop.getHeight() / numOpenWindows;
    int windowWidth = desktop.getWidth();
    
    int currWindowX = 0;
    int currWindowY = 0;
    
    if (windowHeight >= OntologyCanvasWindow.MIN_HEIGHT && windowWidth >= OntologyCanvasWindow.MIN_WIDTH) {
      JInternalFrame frames[] = desktop.getAllFrames();
      for (int i = 0; i < frames.length; i++) {
        if (frames[i].isVisible() && frames[i] instanceof OntologyCanvasWindow) {
          try {
            frames[i].setMaximum(false);
            frames[i].reshape(currWindowX, currWindowY, windowWidth, windowHeight);
            frames[i].toFront();
            frames[i].setSelected(true);
          }
          catch (PropertyVetoException e)
          {
              /* REVISIT: handle better? */
          }
          
          currWindowY += windowHeight;
        }
      }
    }
    else {
      JOptionPane.showMessageDialog(null, "There are too many open windows to tile.", "Error", JOptionPane.ERROR_MESSAGE);
    }
  }
  
  public void cascadeFrames() {
    int i, nmWindows = numOpenWindows;
    
    // Get the height and width of the desktop
    int desktopHeight = desktop.getHeight();
    int desktopWidth = desktop.getWidth();
    
    // Calculate desktop size needed for the default window size
    int totalHeight = OntologyCanvasWindow.DEFAULT_HEIGHT + (nmWindows-1)*OntologyCanvasWindow.CASCADE_OFFSET;
    int totalWidth = OntologyCanvasWindow.DEFAULT_WIDTH + (nmWindows-1)*OntologyCanvasWindow.CASCADE_OFFSET;
    
    int windowHeight = OntologyCanvasWindow.DEFAULT_HEIGHT;
    int windowWidth = OntologyCanvasWindow.DEFAULT_WIDTH;
    
    // Adjust width and height of the cascaded windows if they go outside
    // of the desktop viewing area.
    if (totalHeight > desktopHeight) {
      windowHeight = desktopHeight - (nmWindows-1)*OntologyCanvasWindow.CASCADE_OFFSET;
      if (windowHeight < OntologyCanvasWindow.MIN_HEIGHT)
        windowHeight = desktopHeight - ((desktopHeight-OntologyCanvasWindow.MIN_HEIGHT)/OntologyCanvasWindow.CASCADE_OFFSET+1)*OntologyCanvasWindow.CASCADE_OFFSET;
    }
    if (totalWidth > desktopWidth) {
      windowWidth = desktopWidth - (nmWindows-1)*OntologyCanvasWindow.CASCADE_OFFSET;
      if (windowWidth < OntologyCanvasWindow.MIN_WIDTH)
        windowWidth = desktopWidth - ((desktopWidth-OntologyCanvasWindow.MIN_WIDTH)/OntologyCanvasWindow.CASCADE_OFFSET+1)*OntologyCanvasWindow.CASCADE_OFFSET;
    }
    
    // Finally, cascade the windows
    int currWindowX=0;
    int currWindowY=0;
    
    JInternalFrame frames[] = desktop.getAllFrames();
    for (i=0; i<frames.length; i++)
      if (frames[i].isVisible() && frames[i] instanceof OntologyCanvasWindow) {
        try {
          frames[i].setMaximum(false);
          frames[i].reshape(currWindowX,currWindowY,windowWidth,windowHeight);
          frames[i].toFront();
          frames[i].setSelected(true);
        }
        catch (PropertyVetoException e)
        {
            /* REVISIT: handle better? */
        }
        
        currWindowX+=OntologyCanvasWindow.CASCADE_OFFSET;
        currWindowY+=OntologyCanvasWindow.CASCADE_OFFSET;
        if ((currWindowX + windowWidth) > desktopWidth)
          currWindowX = 0;
        if ((currWindowY + windowHeight) > desktopHeight)
          currWindowY = 0;
      }
  }
  
  
  /**
   * Action to active an internal frame from the menu or keyboard shortcut.
   */
  class WindowFrameAction extends AbstractAction {
    private AbstractInternalFrame m_frame;
    
    public WindowFrameAction(AbstractInternalFrame frame) {
      m_frame = frame;
    }
    
    /**
     *
     * @param evt
     */
    public void actionPerformed(ActionEvent evt) {
      // invoke method in OntologyEditor
      activateFrame(m_frame);
    }
  } // class OntologyEditor.WindowFrameAction
  
  /**
   * OntosAction called from "Window|Tile Vertically".
   */
  class WindowTileVertAction extends AbstractAction {
    /**
     * Tile the windows
     *
     * @param evt
     */
    public void actionPerformed(ActionEvent evt) {
      tileFramesVertically();
    }
  } // class OntologyEditor.WindowTileVertAction
  
  /**
   * OntosAction called from "Window|Tile Horizontally".
   */
  class WindowTileHorAction extends AbstractAction {
    /**
     * Tile the windows
     *
     * @param evt
     */
    public void actionPerformed(ActionEvent evt) {
      tileFramesHorizontally();
    }
  } // class OntologyEditor.WindowTileHorAction
  
  /**
   * OntosAction called from "Window|Cascade".
   */
  class WindowCascadeAction extends AbstractAction {
    /**
     * Cascades the open windows.
     *
     * @param evt
     */
    public void actionPerformed(ActionEvent evt) {
      cascadeFrames();
    }
  } // class OntologyEditor.WindowCascadeAction
  
  /**
   * OntosAction associated with the "exit" action.
   */
  class ExitAction extends AbstractAction {
    
    /**
     * Requests that the application shutdown.
     *
     * @param evt
     */
    public void actionPerformed(ActionEvent evt) {
      exit();
    }
  } // class ExitAction
  
} // class OntologyEditor

/**
 * A helper class that is used to close down the application cleanly.
 */
class WindowCloser extends WindowAdapter {
  private OntologyEditor editor;
  
  public WindowCloser(OntologyEditor e) {
    editor = e;
  }
  
  public void windowClosing(WindowEvent e) {
      if (editor.desktop == null) return;
      JInternalFrame[] frames = editor.desktop.getAllFrames();
      for (int i=0; i<frames.length; i++) {
          frames[i].doDefaultCloseAction();
      }
      
      if (editor.numOpenWindows == 0) editor.exit();
  }
} // class WindowCloser

/**
 * A helper class that is used to close down the application cleanly.
 */
class FrameWindowListener extends InternalFrameAdapter {
  private OntologyCanvasWindow frame;
  private OntologyEditor editor;
  
  public FrameWindowListener(OntologyEditor e, OntologyCanvasWindow f) {
    frame = f;
    editor = e;
    
    if (editor.numOpenWindows==0) {
      //editor.setOpenFrameMenuEnabled(true);
      setNoOpenWindowsState();
    }
    
    editor.numOpenWindows++;
  }
  
  public void internalFrameClosed(InternalFrameEvent e) {
      editor.remove(frame);
		
      editor.numOpenWindows--;
		
      if (editor.numOpenWindows==0)
		  setNoOpenWindowsState();
  }
  
  public void internalFrameActivated(InternalFrameEvent e) {
    //editor.setActiveFrameMenuEnabled(true);
    editor.setActionGroupEnabled(ActionGroup.EDITOR_ACTIONS, true);
  }
  
  public void internalFrameDeactivated(InternalFrameEvent e) {
    //editor.setActiveFrameMenuEnabled(false);
    editor.setActionGroupEnabled(ActionGroup.EDITOR_ACTIONS, false);
  }
  
  private void setNoOpenWindowsState() {
    editor.setAllActionGroupsEnabled(false);
    editor.setActionGroupEnabled(ActionGroup.ALWAYS_ENABLED_ACTIONS, true);
  }
} // class FrameWindowListener