/*
 * DrawShape.java
 *
 * Created on June 14, 2003, 4:14 PM
 */

package edu.byu.deg.OntologyEditor.shapes;

import edu.byu.deg.OntologyEditor.Canvas;
import edu.byu.deg.OntologyEditor.CustomCoordinatePlane;
import edu.byu.deg.OntologyEditor.ObjBorder;
import edu.byu.deg.OntologyEditor.OntologyCanvas;
import edu.byu.deg.osmx.OSMXElement;
import edu.byu.deg.osmx.OSMXFontType;
import edu.byu.deg.osmx.OSMXStyleType;
import edu.byu.deg.osmx.OSMXElement.OSMXElementDeletionListener;
import edu.byu.deg.osmx.StyledElement;
import edu.byu.deg.osmx.binding.StyleType;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.ItemSelectable;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import javax.swing.JPanel;
import javax.swing.border.Border;

import java.awt.Stroke;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Iterator;

/** This abstract class is the ancestor of all other <CODE>...Shape</CODE> classes
 * that provide a graphical interface to the OSMX element classes.
 * <br>
 * <CODE>DrawShape</CODE> inherits from {@link javax.swing.JPanel}, so a great deal
 * of the supporting code for painting, events, etc. is immediately available.
 * Most of the methods defined in <CODE>DrawShape</CODE> address the following
 * problems:
 * <ul>
 * <li>Communicating with the underlying {@link edu.byu.deg.osmx.OSMXElement}
 * object ({@link #getElement}, {@link #elementPropertyChange},
 * {@link #initElement}, {@link #elementDeleted})</li>
 * <li>Translating between the {@link java.awt.Component} location coordinate
 * system and the element's coordinate system ({@link #initLocation},
 * {@link #getAbsoluteCenterPoint}, {@link #getTopPoint}, {@link #getBottomPoint},
 * {@link #getLeftPoint}, {@link #getRightPoint}, {@link #getPosition},
 * {@link #setPosition(java.awt.Point)}, {@link #setPosition(int, int)},
 * {@link #locationChanged(java.awt.Point)}, {@link #locationChanged(int, int)},
 * {@link #setLocation(java.awt.Point)}, {@link #setLocation(int, int)},
 * {@link #getOrigin}, {@link #getConnectionPoint}, {@link #getCenterPoint})</li>
 * <li>Interacting with the {@link edu.byu.deg.OntologyEditor.OntologyCanvas}
 * to which the shape belongs ({@link #getParentCanvas}, {@link #getRootCanvas},
 * {@link #drag}, {@link #delete})</li>
 * <li>Centralizing and standardizing access to style information
 * ({@link #getStyle}, {@link #getStroke}, {@link #getDefaultStyle},
 * {@link #getFillColor}, {@link #getLineColor}, {@link #getLineWidth},
 * {@link #getFontColor}, {@link #getFontSize}, {@link #getFontFamily},
 * {@link #getFontStyle})</li>
 * @author Alan Wessman
 * modified by Zonghui Lian 
 * Add the methods to get the objectset's height and width
 */
public abstract class DrawShape extends JPanel implements ItemSelectable, CustomCoordinatePlane, OSMXElementDeletionListener,  MouseListener, MouseMotionListener { 

  /** Reference to the <CODE>OntologyCanvas</CODE> that contains this shape. */
  protected OntologyCanvas rootCanvas;

  /** Reference to the nearest <CODE>Canvas</CODE> that contains this shape. */
  protected Canvas parentCanvas;

  /** Reference to the underlying element that contains state information for this
   * shape.
   */
  protected OSMXElement element;

  /** Tracks selection state (<CODE>true -></CODE> shape is selected). */
  protected boolean selected = false;

  private ArrayList itemListeners;

  /** Tracks latest location of this shape previous to a {@link #drag} operation. */
  protected Point initialLocation = null;

  /** The <CODE>Border</CODE> object used for painting this shape. */
  protected Border shapeBorder;
  /**
   * This set of <CODE>Rectangle</CODE> objects used for painting the selected shape's handle-points 
   */
  /*protected Rectangle topLeftRect, topRightRect, topCenterRect,  
  					  bottomLeftRect, bottomRightRect, bottomCenterRect,
  					  leftCenterRect, rightCenterRect;*/

  /** The default font to use in drawing this shape. */
  protected static final Font DEFAULT_FONT = new Font("SansSerif", Font.PLAIN, 12);
  
  protected static final int DEFAULT_WIDTH = 0;
  protected static final int DEFAULT_HEIGHT = 0;

  /** The default stroke (line width, style, pattern) used to draw this shape. */
  protected static final BasicStroke DEFAULT_STROKE = new BasicStroke(1.0f, BasicStroke.CAP_SQUARE, BasicStroke.JOIN_MITER, 1.0f, null, 0);
  
  /** Object used as a proxy to listen for {@link java.beans.PropertyChangeEvent}s on
   * {@link edu.byu.deg.osmx.OSMXElement}s.
   */
  protected PropertyChangeListener elementListener;

  /** Default style object for this shape. */
  protected static final OSMXStyleType DEFAULT_STYLE;
  protected String resizeDragPoint;
  protected boolean isResize;
  // initialize the DEFAULT_STYLE static object
  static {
    DEFAULT_STYLE = new OSMXStyleType();
    DEFAULT_STYLE.setFillColor(Color.WHITE);
    DEFAULT_STYLE.setLineWidth(1.0f);
    DEFAULT_STYLE.setLineColor(Color.BLACK);
    DEFAULT_STYLE.setFontFamily("SansSerif");
    DEFAULT_STYLE.setFontSize(12);
    DEFAULT_STYLE.setFontColor(Color.BLACK);
    DEFAULT_STYLE.setFontStyle(OSMXFontType.intToStyle(Font.PLAIN));
  }

  /** Creates a new instance of DrawShape, adding it to the specified
   * {@link java.awt.Container}.
   * <p>
   * Post-conditions:
   * <ol>
   * <li>The shape is added to the specified <CODE>Container</CODE></li>
   * <li>The {@link #element} field is initialized with the specified
   * <CODE>OSMXElement</CODE></li>
   * <li>The {@link #elementListener} field is initialized to listen for
   * {@link java.beans.PropertyChangeEvent}s on the <CODE>element</CODE> field and to
   * call {@link #elementPropertyChange} when it receives such an event</li>
   * <li>Any code in {@link #initElement} is performed prior to any other code being
   * invoked</li>
   * <li>The remaining instance variables are initialized</li>
   * <li>The {@link #rootCanvas} is registered to listen for
   * {@link java.awt.event.MouseEvent}s and {@link java.awt.event.KeyEvent}s fired by
   * this shape</li>
   * </ol>
   * @param c The {@link java.awt.Container} to which the shape is to be added.
   * @param elem The underlying {@link edu.byu.deg.osmx.OSMXElement} that this shape represents.
   */
  public DrawShape(Container c, OSMXElement elem) {
	/* The MouseListener and MouseMotionListener is just used to adjust the mouse cursor and check
	 * whether the select-drag point is the resize-drag-point 
	*/
	addMouseListener(this);
	addMouseMotionListener(this);
    c.add(this);
    element = elem;
    if (element != null) {
      elementListener = new PropertyChangeListener() {
        public void propertyChange(PropertyChangeEvent evt) {
          if (element != null && evt.getSource() == element) {
            elementPropertyChange(evt);
          }
        }
      };
      element.addPropertyChangeListener(elementListener);
    }
    initParentCanvas();
    initRootCanvas();
    initElement();
    shapeBorder = new ObjBorder();
    ((ObjBorder) shapeBorder).setRectColor(getLineColor());
    ((ObjBorder) shapeBorder).setFillColor(getFillColor());
    ((ObjBorder) shapeBorder).setRectWidth((int) getLineWidth());
    
    // descendant classes may invoke setBorder(shapeBorder) if necessary

    addMouseListener(rootCanvas);
    addMouseMotionListener(rootCanvas);
    addKeyListener(rootCanvas);

    super.setFont(DEFAULT_FONT);
    if (element != null) {
      element.addElementDeletionListener(this);
    }
    itemListeners = new ArrayList();
    initLocation();
    isResize = false;
    resizeDragPoint = "NONE";
  } // constructor

  /** Provides access to the underlying OSMX element that this shape represents.
   * @return The model element represented by <CODE>this</CODE>.
   */
  public OSMXElement getElement() {
    return element;
  }

  /** Handler for {@link java.beans.PropertyChangeEvent}s fired by
   * {@link edu.byu.deg.osmx.OSMXElement}s to which <CODE>this</CODE> is registered
   * as a listener.<br>
   * The source of the events is usually but need not be the same object as returned
   * by {@link #getElement}.
   * @param evt The {@link java.beans.PropertyChangeEvent} fired by an
   * {@link edu.byu.deg.osmx.OSMXElement}.
   */
  protected void elementPropertyChange(PropertyChangeEvent evt) {
    if (OSMXStyleType.STYLE_PROPERTY.equals(evt.getPropertyName())) {
      // TODO: reload all the style properties first

          ((ObjBorder) shapeBorder).setRectColor(getLineColor());
      ((ObjBorder) shapeBorder).setFillColor(getFillColor());
      ((ObjBorder) shapeBorder).setRectWidth((int) getLineWidth());
      this.setBackground(getFillColor());
      this.setForeground(getLineColor());
      this.setFont(this.getFont());
      repaint();
    }
  }

  /** Responsible for assigning the {@link edu.byu.deg.osmx.OSMXElement} object that
   * was passed into the constructor to class member variables.
   * <p>
   * This method is intended to be overridden in subclasses of
   * <CODE>DrawShape</CODE>.  It is invoked at the beginning of the
   * {@link #DrawShape(java.awt.Container, edu.byu.deg.osmx.OSMXElement) constructor}
   * for <CODE>DrawShape</CODE>, before other <CODE>DrawShape</CODE> methods are
   * called.  Thus, any member variable assignments made within this method
   * (especially overridden versions of this method within <CODE>DrawShape</CODE>
   * subclasses) are guaranteed to be in effect before any other code in the object
   * executes.
   * <p>
   * One common use of this method is to override it in a subclass, casting the
   * {@link #element} object to a narrower type (such as
   * {@link edu.byu.deg.osmx.OSMXObjectSetType}) and assigning it to a member
   * variable of that type so that other code in the subclass can reference the
   * narrowly-typed variable instead of having to cast the <CODE>element</CODE>
   * variable multiple times.
   */
  protected void initElement() {
    // allow subclasses to perform any more initialization of element references
  }

  /** Resets the value of {@link #initialLocation} to the value returned by the
   * {@link java.awt.Component#getLocation()} method.
   */
  protected void initLocation() {
    initialLocation = this.getLocation();
  }

  /** Invoked from {@link #setSelected(boolean)} or {@link #setSelected(Rectangle, boolean)} after the {@link #selected} flag has been
   * updated, this method allows subclasses to respond to the change in selection
   * state.
   */
  protected abstract void updateSelectionState();

  /** Accessor for the {@link edu.byu.deg.OntologyEditor.OntologyCanvas} to which this
   * <CODE>DrawShape</CODE> belongs.
   * @return The <CODE>OntologyCanvas</CODE> to which <CODE>this</CODE> belongs, either via
   * direct or indirect containment.
   */
  public OntologyCanvas getRootCanvas() {
    if (rootCanvas != null) return rootCanvas;

    // else...
    initRootCanvas();
    return rootCanvas;
  } // method getRootCanvas

  private void initRootCanvas() {
    Container canvas;
    canvas = (Container) getParentCanvas();
    while (canvas != null && !(canvas instanceof OntologyCanvas)) {
      canvas = canvas.getParent();
    }
    if (canvas == null || !(canvas instanceof OntologyCanvas)) return;

    rootCanvas = (OntologyCanvas) canvas;
  }

  /** Returns the nearest ancestor {@link edu.byu.deg.OntologyEditor.Canvas} that this
   * shape belongs to.  This will either be the root
   * {@link edu.byu.deg.OntologyEditor.OntologyCanvas} or a high-level shape (such as
   * {@link ObjectSetShape}) that implements <CODE>Canvas</CODE>.
   * @return The <CODE>Canvas</CODE> to which this shape belongs.
   */
  public Canvas getParentCanvas() {
    if (parentCanvas != null) return parentCanvas;

    // else...
    initParentCanvas();
    return parentCanvas;
  } // method getParentCanvas

  private void initParentCanvas() {
    Container canvas = getParent();
    while (canvas != null && !(canvas instanceof Canvas)) {
      canvas = canvas.getParent();
    }
    if (canvas == null || !(canvas instanceof Canvas)) return;

    parentCanvas = (Canvas) canvas;
  }

  /** Updates the selection state of the shape and generates an
   * {@link java.awt.event.ItemEvent} that is passed into
   * {@link java.awt.event.ItemListener#itemStateChanged} for all registered
   * {@link java.awt.event.ItemListener}s.
   * @param selectedState Indicates whether the shape is to be selected (<CODE>true</CODE>) or not
   * (<CODE>false</CODE>).
   */
  public void setSelected(boolean selectedState) {
    if (selected == selectedState) return;
    selected = selectedState;

    // delegate to subclasses for any other actions to take when selection changes
    updateSelectionState();

    // translate selection flag into ItemEvent constant value
    int selectionState = (selected ? ItemEvent.SELECTED : ItemEvent.DESELECTED);
    ItemEvent evt = new ItemEvent(this, ItemEvent.ITEM_STATE_CHANGED, this, selectionState);
    for (Iterator i = itemListeners.iterator(); i.hasNext();) {
      ItemListener l = (ItemListener) i.next();
      l.itemStateChanged(evt);
    }

    // repaint since change in selection state probably affects appearance
    repaint();
  } // method setSelected(boolean)

  /** Calls {@link #setSelected(boolean)}, passing in <CODE>true</CODE> if the shape
   * falls within the bounds of the specified {@link java.awt.Rectangle} and
   * <CODE>false</CODE> otherwise.
   * @param r The selection <CODE>Rectangle</CODE> that determines whether this shape should
   * be selected.
   * @param inclusive Specifies whether to select the shape if its bounding box intersects with the
   * input {@link java.awt.Rectangle} (when the parameter is <CODE>true</CODE>), or
   * if its bounding box is wholly contained in the input <CODE>Rectangle</CODE>
   * (when the parameter is <CODE>false</CODE>).  In other words, a <CODE>true</CODE>
   * value indicates inclusive selection, where a shape intersecting the selection
   * rectangle is selected; whereas a <CODE>false</CODE> value indicates exclusive
   * selection, where a shape must be entirely within the selection rectangle for it
   * to be selected.
   */
  public void setSelected(Rectangle r, boolean inclusive) {
    Rectangle bounds = getBounds();
    boolean select = (inclusive && r.intersects(bounds)) || r.contains(bounds);
    setSelected(select);
  } // method setSelected(Rectangle, boolean)

  /** Accessor for the <CODE>selected</CODE> property.
   * @return The current selection state of the shape.
   */
  public boolean isSelected() {
    return selected;
  } // method isSelected

  /** Adds a listener to receive item events when the selection state of the shape is
   * changed by the user. If <code>l</code> is
   * <code>null</code>, no exception is thrown and no action is performed.
   * @param l the listener to receive events
   * @see ItemEvent
   */
  public void addItemListener(ItemListener l) {
    itemListeners.add(l);
  } // method addItemListener

  /** Returns the selected items or <code>null</code> if no
   * items are selected.
   * @return The selected items or <code>null</code> if none exist.
   */
  public Object[] getSelectedObjects() {
    return (isSelected() ? new Object [] {this} : null);
  } // method getSelectedObjects

  /** Removes an item listener.
   * If <code>l</code> is <code>null</code>,
   * no exception is thrown and no action is performed.
   *
   * @param 	l the listener being removed
   * @see ItemEvent
   *
   */
  public void removeItemListener(ItemListener l) {
    itemListeners.remove(l);
  } // method removeItemListener

  /** Translates the shape by the specified amounts relative to its container.
   * @param x The number of pixels to move horizontally (right is positive).
   * @param y The number of pixels to move vertically (down is positive).
   */
  public void drag(int x, int y) {
    // note that we deal with position, not location
    Point pos = getPosition();
    pos.translate(x, y);
    setPosition(pos);
  } // method drag
  
  /**
   * Check if the mouse-press is on the resize-drag-point.
   * @return
   */
  public boolean isResizing(){
	  return isResize;
  }
  
  /**
   * Resize the shape's border
   * @param x The number of pixels to move horizontally (right is positive).
   * @param y The number of pixels to move vertically (down is positive).
   */
  public void resizeBorder(int x, int y) {
	  int oldWidth = this.getWidth(), oldHeight = this.getHeight(), l_x = 0, l_y = 0;
	  Point pos = getPosition();
	  if (resizeDragPoint.equals("TOP_LEFT")){
		  l_x = -x;
		  l_y = -y;
		  pos.translate(x, y);
	  }
	  else if (resizeDragPoint.equals("TOP_RIGHT")){
		  l_x = x;
		  l_y = -y;
		  pos.translate(0,y);
	  }
	  else if(resizeDragPoint.equals("TOP_CENTER")) {
		  l_y = -y;
		  pos.translate(0,y);
	  }
	  else if (resizeDragPoint.equals("LEFT_CENTER")){
		  l_x = -x;
		  pos.translate(x,0);
	  }
	  else if (resizeDragPoint.equals("RIGHT_CENTER")) {
		  l_x = x;
	  }
	  else if (resizeDragPoint.equals("BOTTOM_LEFT")){
		  l_x = -x;
		  l_y = y;
		  pos.translate(x, 0);
	  }
	  else if (resizeDragPoint.equals("BOTTOM_RIGHT")) {
		  l_x = x;
		  l_y = y;
	  }
	  else if (resizeDragPoint.equals("BOTTOM_CENTER")) {
		  l_y = y;
	  }
	  else return;
	  Dimension dms = new Dimension(oldWidth + l_x, oldHeight + l_y); 
	  if (isBiggerThanMinSize(dms)) {
		  ( (PlanarShape)this).setBorderSize(dms);
		  setPosition(pos);
		  ( (ConnectableContainerShape)this).updateNameTextPosition();
		  this.repaint();
		  //( (ConnectableContainerShape)this).setSize(dms);
	  }
  }

  /*
   * Based on the ConnectableShape's prefer size, check if the resized dimension is too small. 
   * 
   */
  private boolean isBiggerThanMinSize(Dimension dms) {
	  if (this instanceof ConnectableShape) {
		  Dimension preferDms = ( (ConnectableShape)this).getPreferredSize();
		  if (preferDms.width > dms.width || preferDms.height > dms.height) 
			  return false;
		  else 
			  return true;
	  }
	  else 
		  return false;
  }
  /** Passes a delete request to the internal OSMX element, which is responsible for
   * removing itself from the model and firing deletion events to registered
   * listeners (including this shape, so it can remove itself from its container).
   */
  public void delete() {
    if (element != null) {
      element.delete();
    }
  } // method delete

  /** Calculates the geometric center of the shape's bounding box.
   * @return The point at the center of the shape's bounding box, in the shape's container's
   * coordinate system.
   */
  public Point getAbsoluteCenterPoint() {
    Rectangle rect = new Rectangle(0,0,getWidth(),getHeight());
    Point loc = getPosition();
    rect.translate(loc.x,loc.y);
    Point pt = new Point(rect.x + (rect.width/2),
    rect.y + (rect.height/2));
    return pt;
  } // method getAbsoluteCenterPoint

  /** Returns the northwest corner point of the shape's bounding box, in the
   * container's coordinate system.
   * @return The upper left corner point of the shape's bounding box.
   */
  public Point getTopPoint() {
    Point pt =  getPosition();
    return pt;
  }

  /** Returns the southwest corner point of the shape's bounding box, in the
   * container's coordinate system.
   * @return The lower left corner point of the shape's bounding box.
   */
  public Point getBottomPoint(){
    Rectangle rect = new Rectangle(0,0,getWidth(),getHeight());
    Point loc = getPosition();
    rect.translate(loc.x,loc.y);
    Point pt = new Point(rect.x, rect.y + rect.height);
    return pt;
  }
  /** Returns the northwest corner point of the shape's bounding box, in the
   * container's coordinate system.
   * @return The upper left corner point of the shape's bounding box.
   */
  public Point getLeftPoint() {
    Point pt = getPosition();
    return pt;
  }
  /** Returns the northeast corner point of the shape's bounding box, in the
   * container's coordinate system.
   * @return The upper right corner of the shape's bounding box.
   */
  public Point getRightPoint() {
    Rectangle rect = new Rectangle(0,0,getWidth(),getHeight());
    Point loc = getPosition();
    rect.translate(loc.x,loc.y);
    Point pt = new Point(rect.x+rect.width, rect.y);
    return pt;
  }
  /** Returns the horizontal and vertical position of this shape's origin point
   * relative to its container's origin point.  This may be different than the value
   * returned by {@link #getLocation()} in situations where the shape's origin point
   * is not at the northwest corner of the shape's bounding box.
   * @return The position of the shape's origin point relative to its container's origin
   * point.
   * @see #getOrigin
   * @see edu.byu.deg.OntologyEditor.CustomCoordinatePlane
   */
  public Point getPosition() {
    Point loc = getLocation();
    CustomCoordinatePlane parent = (CustomCoordinatePlane) getParent();
    if (parent == null) {
        return (loc);
    }

    // At this point, loc gives the location of this shape's upper left corner
    // relative to the container's upper left corner.
    //
    // We want the location of this shape's origin point relative to the
    // container's origin point.

    Point pt = parent.getOrigin();
    loc.translate(-pt.x, -pt.y); // negative is correct: gives loc relative to parent origin
    pt = getOrigin();
    loc.translate(pt.x, pt.y); // now adjust to put in terms of this shape's origin

    return loc;
  } // method getPosition

  /** Changes the shape's position to the specified <CODE>Point</CODE>.
   * @param pt The desired new value for the shape's position.
   * @see #getPosition
   */
  protected abstract void setPosition(Point pt);

  
  /** Changes the shape's position to the specified x and y coordinates.
   * @param x The x-coordinate for the new position.
   * @param y The y-coordinate for the new position.
   */
  protected abstract void setPosition(int x, int y);

  /** Centralizes code that updates the shape's position whenever the shape's location
   * has been changed via {@link #setLocation(java.awt.Point)}.
   * @param newLoc The new value of the recently changed location.
   */
  protected void locationChanged(Point newLoc) {
    setPosition(newLoc);
  }

  /** Centralizes code that updates the shape's position whenever the shape's location
   * has been changed via {@link #setLocation(int, int)}.
   * @param x The x-coordinate of the new location.
   * @param y The y-coordinate of the new location.
   */
  protected void locationChanged(int x, int y) {
    setPosition(x, y);
  }

  /** Overrides {@link java.awt.Component#setLocation(java.awt.Point)}, calling the
   * inherited method and then {@link #locationChanged(java.awt.Point)}.
   * @param pt The desired new location for the shape.
   */
  public void setLocation(Point pt) {
    super.setLocation(pt);
    locationChanged(pt);
  }

  /** Overrides {@link java.awt.Component#setLocation(int, int)}, calling the
   * inherited method and then {@link #locationChanged(int, int)}.
   * @param x The x-coordinate for the desired new location.
   * @param y The y-coordinate for the desired new location.
   */
  public void setLocation(int x, int y) {
    super.setLocation(x, y);
    locationChanged(x, y);
  }

  /** Returns the origin point of this shape.
   * @return The origin point of the shape.
   * @see edu.byu.deg.OntologyEditor.CustomCoordinatePlane
   */
  public Point getOrigin() {
    return new Point(0, 0);
  } // method getOrigin

  /** Returns the center point of the shape as a high-precision coordinate pair.
   * @return The center point of the shape, as calculated by {@link #getCenterPoint}.
   */
  public Point2D.Double getConnectionPoint() {
    return PointUtilities.hiResPoint(getCenterPoint());
  }

  /** Calculates the center point of the shape, in the coordinate system of the shape.
   * @return The center point of the shape (width/2, height/2).
   */
  public Point getCenterPoint() {
	  //System.out.println("***");
	  int x = getWidth();
	  //System.out.println("x:* "+x);
	  int y = getHeight();
	//  System.out.println("x: "+x+" y: "+y);
    Point pt = new Point(x/2, y/2);
    return pt;
  } // method getCenterPoint

  public Point getConnectionPointAccordingToAngle(int angle) {
	  Point pt = null;
	  final double DEG_2_RAD = Math.PI/180.0;
	  int diagonal_angle;
	  double radAngle = (double) (angle * DEG_2_RAD);
	  int x = getWidth(), y = getHeight();
	  diagonal_angle = (int) ( (Math.atan( (double) ((double)y/(double)x) ))/DEG_2_RAD + 0.5);

	  if (angle <= diagonal_angle) {
		  y = (int)( ((double)y/2) - ( ((double)x/2) * Math.tan(radAngle) ) + 0.5);
		  pt = new Point(x,y);
	  }
	  else if (angle > diagonal_angle && angle <= (180- diagonal_angle)){
		  if (angle == 90){
			  pt = new Point( (int)((double)x/2 + 0.5), 0);
		  }
		  else {
			  x = (int)( (double)x/2 + ((double)y/2) / Math.tan(radAngle) + 0.5 );
			  pt = new Point(x,0);
 		  }
	  }
	  else if (angle > (180-diagonal_angle) && angle <= (180+diagonal_angle)) {
		  if(angle ==180){
			  pt = new Point(0, (int) ((double)y/2 + 0.5));
		  }
		  else {
			  y = (int) ((double)y/2 + (double)x/2 * Math.tan(radAngle) );
			  pt = new Point(0, y);
		  }
	  }
	  else if (angle <= (360-diagonal_angle) && angle > (180+diagonal_angle)) {
		  if(angle == 270){
			  pt = new Point((int)((double)x/2+0.5), y);
		  }
		  else{
			  x = (int)((double)x/2 - ((double)y/2)/Math.tan(radAngle) + 0.5);
			  pt = new Point(x,y);
		  }
	  }
	  else if (angle > (360-diagonal_angle) && angle <= 360) {
		  if (angle == 360){
			  pt = new Point(x, (int)((double)y/2+0.5));
		  }
		  else {
			  y = (int)( ((double)y/2) - ( ((double)x/2) * Math.tan(radAngle) ) + 0.5);
			  pt =  new Point(x, y);
		  }
	  }
	  //System.out.println("load width & height" + getWidth() + " " + getHeight());
	  //System.out.println("load angle: " + angle);
	  //System.out.println("pt: " + pt);
	  return pt;
  }
  //Zonghui
  public int getAngleAccordingToHandle(Component comp, Point2D.Double pt ){
	  final double DEG_2_RAD = Math.PI/180.0;
	  double width = getWidth()-3, height = getHeight();//TODO: find the reason why here the width increase 3
	  int angle = 0;
	  Point2D.Double conPt = PointUtilities.convertPoint(comp,pt,this);
	  //System.out.println("width & height: " + this.getWidth() + " " + this.getHeight());
	  //System.out.println("conPt: " + conPt);
	  double x = conPt.x, y = conPt.y;
	  if (x == width) {
		  //System.out.println("tan: "+ (height/2 - y)+ " "+ width/2 + " " + (height/2 - y)/(width/2));
		  angle = (int) (Math.atan((height/2 - y)/(width/2))/DEG_2_RAD);
		 if (angle < 0) {
			  angle = 360 + angle;
		  }
		  //System.out.println("angle " + angle);//test
	  }
	  else if (y == 0) {
		  angle = (int)( Math.atan(height/2 / (x - width/2)) /DEG_2_RAD );
		  if (angle < 0)
			  angle = 180 + angle;
	  }
	  else if (x == 0) {
		  angle = (int) ( Math.atan( ((height/2)-y)/(width/2) ) /DEG_2_RAD );
		  angle = 180 - angle;
	  }
	  else if (y == height){
		  angle = (int)( Math.atan( (height/2)/((width/2)-x )) /DEG_2_RAD );
		  if (angle >=0) {
			  angle = angle + 180;
		  }
		  else if (angle < 0){
			  angle = 360 + angle;
		  }
	  }
	  //System.out.println("angle: "+ angle);
	  return angle;
  }
  
  /** Calculates a {@link java.awt.Rectangle}, with non-negative width and height, from
   * the input {@link java.awt.Point}s.
   * @param p1 One of the points that defines the rectangle.
   * @param p2 The other point that defines the rectangle.
   * @return The rectangle defined by the input points; width and height are guaranteed to be
   * non-negative.
   */
  public static Rectangle normalizeRect(Point p1, Point p2) {
    int x,y,w,h;
    w = Math.abs(p1.x - p2.x);
    h = Math.abs(p1.y - p2.y);
    x = Math.min(p1.x, p2.x);
    y = Math.min(p1.y, p2.y);
    return new Rectangle(x,y,w,h);
  } // static method normalizeRect

  /** Delegate method that handles incoming notifications of deletion requests sent by
   * any {@link edu.byu.deg.osmx.OSMXElement} on which this shape has registered a
   * {@link edu.byu.deg.osmx.OSMXElement.OSMXElementDeletionListener}.
   * <p>
   * If the element is the one on which this shape is based, this method will act as
   * a destructor, nullifying the <CODE>element</CODE> field and removing this shape
   * from its parent container.
   * @param elem The element that is being deleted.
   * @see edu.byu.deg.osmx.OSMXElement#delete
   */
  public void elementDeleted(OSMXElement elem) {
    if (elem != null && element == elem) {
      element = null;
      Container parent = this.getParent();
      if (parent != null) parent.remove(this);
    }
  } // method elementDeleted

  /** Returns the shape's current style.
   * @return The style set on the underlying element, or the result of
   * {@link #getDefaultStyle} if the underlying element has no style set on it.
   */
  public StyleType createStyle() {
    if (element == null) return null;
    OSMXStyleType style = (OSMXStyleType)element.getStyle();
    if (style == null) {
      try {
        if (element instanceof StyledElement) {
          OSMXStyleType st = (OSMXStyleType) (element.getParentDocument().
                                              getObjectFactory().
                                              createStyleType());
          // Add a style to the XML element
          ( (StyledElement) element).setStyle(st);
          st.addPropertyChangeListener(elementListener);

          // Copy DEFAULT_STYLE as the pattern
          st.setFillColor(DEFAULT_STYLE.getFillColor());
          st.setLineWidth(DEFAULT_STYLE.getLineWidth());
          st.setLineColor(DEFAULT_STYLE.getLineColor());
          st.setFontFamily(DEFAULT_STYLE.getFontFamily());
          st.setFontSize(DEFAULT_STYLE.getFontSize());
          st.setFontColor(DEFAULT_STYLE.getFontColor());
          st.setFontStyle(DEFAULT_STYLE.getFontStyle());
          style = st;
          // Return that style
//          return (st);
        }
      } catch (Exception e) {
        return (null);
      }
    }
    else  {
      try {
      if (element instanceof StyledElement) {
        OSMXStyleType st = (OSMXStyleType) (element.getParentDocument().
                                            getObjectFactory().
                                            createStyleType());
        //copy exist style to st
        if (style.getFillColor()!= null)
          st.setFillColor(style.getFillColor());
         else
           st.setFillColor(DEFAULT_STYLE.getFillColor());

//        st.setFont(style.getFont());
         if (style.getFontColor()!=null)
            st.setFontColor(style.getFontColor());
          else
            st.setFontColor(DEFAULT_STYLE.getFontColor());

        if (style.getFontFamily()!=null)
           st.setFontFamily(style.getFontFamily());
         else
           st.setFontFamily(DEFAULT_STYLE.getFontFamily());

        Float fontsize = new  Float(style.getFontSize());
        if (fontsize.intValue()!=0)
           st.setFontSize(style.getFontSize());
        else
           st.setFontSize(DEFAULT_STYLE.getFontSize());

         if (style.getFontStyle()!= null)
            st.setFontStyle(style.getFontStyle());
          else
            st.setFontStyle(DEFAULT_STYLE.getFontStyle());

        if(style.getLineColor() != null)
          st.setLineColor(style.getLineColor());
         else
           st.setLineColor(DEFAULT_STYLE.getLineColor());

        Float linewidth = new Float(style.getLineWidth());
        if (linewidth.intValue()!= 0)
          st.setLineWidth(style.getLineWidth());
        else
          st.setLineWidth(DEFAULT_STYLE.getLineWidth());

        // Add a style to the XML element
        ( (StyledElement) element).setStyle(st);
        st.addPropertyChangeListener(elementListener);
//        return (st);
        style =  st;
      }
    } catch (Exception e) {
      return (null);
    }

    }
//    return (style);
/*    PropertyChangeListener styleListener = new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent evt) {
        firePropertyChange(OSMXStyleType.STYLE_PROPERTY,
                           evt.getOldValue(), evt.getNewValue());
      }
    };

    style.addPropertyChangeListener(styleListener);*/
    return style;
  } // method createStyle

  /** Returns the shape's current style.
   * @return The style set on the underlying element, or the result of
   * {@link #getDefaultStyle} if the underlying element has no style set on it.
   */
  public StyleType getStyle() {
    if (element == null) return getDefaultStyle();
    StyleType style = element.getStyle();
    if (style == null) return getDefaultStyle();
    return style;
  } // method getStyle

  /** Returns the standard {@link java.awt.Stroke} for drawing this shape's outline.
   * Override in subclasses if the normal outline should be drawn differently.
   * @return The <CODE>Stroke</CODE> normally used to draw the shape.
   */
  protected Stroke getStroke() {
    return new BasicStroke(getLineWidth(), BasicStroke.CAP_SQUARE, BasicStroke.JOIN_MITER, 10.0f, null, 0);
  }

  /** Returns the default style to use to paint this shape in case the underlying
   * element has no style set for it.  Override in subclasses if a different default
   * style is needed.
   * @return The default style for drawing the shape.
   */
  protected StyleType getDefaultStyle() {
    return DEFAULT_STYLE;
  }

  /** Convenience method for directly accessing the fill color of the current style.
   * @return The fill color of the current style.
   */
  protected Color getFillColor() {
    Color result = ((OSMXStyleType) getStyle()).getFillColor();
    return (result == null ? DEFAULT_STYLE.getFillColor() : result);
  }

  /** Convenience method for directly accessing the line color of the current style.
   * @return The line color of the current style.
   */
  protected Color getLineColor() {
    Color result = ((OSMXStyleType) getStyle()).getLineColor();
    return (result == null ? DEFAULT_STYLE.getLineColor() : result);
  }

  /** Convenience method for directly accessing the line width of the current style.
   * @return The line width of the current style.
   */
  protected float getLineWidth() {
    OSMXStyleType style = (OSMXStyleType) getStyle();
    if (!style.isSetLine() || !style.getLine().isSetWidth()) {
      return DEFAULT_STYLE.getLineWidth();
    } else {
      return style.getLineWidth();
    }
  }
  
  
  /** Convenience method for directly accessing the font color of the current style.
   * @return The font color of the current style.
   */
  protected Color getFontColor() {
    Color result = ((OSMXStyleType) getStyle()).getFontColor();
    return (result == null ? DEFAULT_STYLE.getFontColor() : result);
  }

  /** Convenience method for directly accessing the font size of the current style.
   * @return The font size of the current style.
   */
  protected float getFontSize() {
    OSMXStyleType style = (OSMXStyleType) getStyle();
    if (!style.isSetFont() || !style.getFont().isSetSize()) {
      return DEFAULT_STYLE.getFontSize();
    } else {
      return style.getFontSize();
    }
  }

  /** Convenience method for directly accessing the font family (typeface) of the current style.
   * @return The font family (typeface) of the current style.
   */
  protected String getFontFamily() {
    String result = ((OSMXStyleType) getStyle()).getFontFamily();
    return (result == null ? DEFAULT_STYLE.getFontFamily() : result);
  }

  /** Convenience method for directly accessing the font style (bold, italic, etc.) of the current style.
   * @return The font style (bold, italic, etc.) of the current style.
   */
  protected String getFontStyle() {
    String result = ((OSMXStyleType) getStyle()).getFontStyle();
    return (result == null ? DEFAULT_STYLE.getFontStyle() : result);
  }

  /** Return the bounds of this shape (as opposed to the panel containing
   * this shape.  Used, for example, to compute the drawing bounds for
   * EPS export purposes.  Subclasses should override this method when
   * the containing panel bounds are not the same as the shape bounds.
   * This is the case with connectors, for example.  Shapes need to be sure
   * to include text decorations (e.g., participation constraints,
   * relationship-set names) and special symbols when computing bounds.
   * @return The bounds of this shape.
   */
  public Rectangle getBoundingBox() {
    return (getBounds());
  }
  


  public void mouseClicked(MouseEvent evt) {
	  // Do nothing
  }

  public void mousePressed(MouseEvent evt) {
	  if(this.isSelected() == false) return;
	  int tempX = evt.getX(), tempY = evt.getY();
	  resizeDragPoint = ( (ObjBorder)shapeBorder ).mousePressed(tempX, tempY);
  }
		
  public void mouseReleased(MouseEvent evt) {
	  isResize = false;
  }
		
  public void mouseEntered(MouseEvent evt) {
	  // Do nothing
  }
		
  public void mouseExited(MouseEvent evt) {
	  // Do nothing				
  }
		
  public void mouseDragged(MouseEvent evt) {
	  // Do nothing
  }
		
  public void mouseMoved(MouseEvent evt) {
	  if (selected == false || this instanceof GeneralConstraintShape || this instanceof NoteShape) return;
	  int tempX = evt.getX(), tempY = evt.getY();
	  int cursorType = ( (ObjBorder)shapeBorder).updateCursor(tempX, tempY);
	  if (cursorType != Cursor.DEFAULT_CURSOR)
		  isResize = true;
	  else 
		  isResize = false;
	  Cursor cursor = new Cursor(cursorType);
	  this.setCursor(cursor);
  }
}
