/*
 * ObjectShape.java
 *
 * Created on December 30, 2003, 1:27 PM
 */

package edu.byu.deg.OntologyEditor.shapes;

import edu.byu.deg.OntologyEditor.*;
import edu.byu.deg.OntologyEditor.epsgraphics.EpsGraphics2D;
import java.awt.Color;
import java.awt.Container;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import edu.byu.deg.osmx.*;
import edu.byu.deg.osmx.binding.ObjectType;

/**
 *
 * @author  Liddle
 */
public class ObjectShape extends ObjConnectableShape
implements DocumentListener {
    protected String                objName = null;
    protected OSMXObjectType        obj = null;
    protected boolean               suppressDocEvents = false;
    protected TextShape             objNameText;
    protected ComponentAdapter      canvasSizeListener;
    protected Ellipse2D             centerRegion = null;
    protected ObjectPanel           objectPanel = null;
    protected Point                 centerPoint = null;
    
    protected static Color          m_selColor = Color.lightGray;
    
    protected static final int      OBJECT_RADIUS = 7;
    
    /** Creates a new instance of ObjectShape */
    public ObjectShape(Container c, OSMXElement elem) {
        super(c, elem);
        
        setOpaque(false);
        setLayout(null);
        setForeground(getFontColor());
        // this.setSize(this.getPreferredSize());
        initSize();
        getObject();
        obj.addPositionListener(this);
        initCenterPoint();
        initObject();
        initName();
    }
    
    protected void centerPointMoved() {
        //        super.centerPointMoved();
        repositionName();
        repositionObject();
    }
    
    public boolean contains(int x, int y) {
        Point pt = new Point(x,y);
        return contains(pt);
    }
    
    public boolean contains(Point point) {
        if (obj == null) return false;
        // if (super.contains(point)) return true;
        return masksConnectionsAt(point);
    }
    
    public void delete() {
        getParent().removeComponentListener(canvasSizeListener);
        super.delete();
    }
    
    public Rectangle getBoundingBox() {
        Rectangle bounds = objectPanel.getBounds();
        if (objNameText != null) {
            Rectangle objnt = objNameText.getBoundingBox();
            bounds = bounds.union(objnt);
        }
        return (bounds);
    }
    
    public Point getCenterPoint() {
        return centerPoint;
    }
    
    protected Ellipse2D getCenterRegion() {
        if (centerRegion == null) {
            centerRegion = new Ellipse2D.Float();
        }
        Point ctrPt = getCenterPoint();
        if (ctrPt == null) return null;
        centerRegion.setFrameFromCenter(ctrPt, new Point(ctrPt.x + OBJECT_RADIUS, ctrPt.y + OBJECT_RADIUS));
        return centerRegion;
    }

    public Point2D.Double getIntersectionPoint(Point2D.Double pt) {
        Point2D.Double centerPt = PointUtilities.hiResPoint(getCenterPoint());
        
        double theta = Math.atan2(pt.y - centerPt.y, pt.x - centerPt.x);
        double x = centerPt.x + OBJECT_RADIUS * Math.cos(theta);
        double y = centerPt.y + OBJECT_RADIUS * Math.sin(theta);
        return (new Point2D.Double(x, y));
    }

    public Point getPosition() {
        return (objectPanel.getLocation());
    }
    
    public ObjectType getObject() {
        if (obj != null) {
            return obj;
        }
        obj = (OSMXObjectType) element;
        return obj;
    }

    public Point getOrigin() {
        Point   center = getCenterPoint();
        
        if (center == null) {
            center = super.getOrigin();
        }
        return (center);
    }

    public String getShapeName() {
        if (objName == null && getObject() != null) {
            objName = obj.getName().getContent();
        }
        return objName;
    }
    
    protected void initCenterPoint() {
        if (element == null) return;
        PositionedElement posElem = (PositionedElement) element;
        if (!posElem.isSetPosition()) return;
        centerPoint = new Point((posElem.getX() + OBJECT_RADIUS), (posElem.getY() + OBJECT_RADIUS));
    }
    
    protected void initName() {
        OSMXPositionedText textElem = (OSMXPositionedText) obj.getName();
        if (textElem == null) {
            textElem = new OSMXPositionedText();
            textElem.setContent("Object");
            obj.setName(textElem);
        }
        objNameText = new TextShape(this, textElem);
        objNameText.setSize(objNameText.getPreferredSize());
        this.add(objNameText, 0);
        repositionName();
    }
    
    protected void initObject() {
        objectPanel = new ObjectPanel();
        if (obj == null) return;
        this.add(objectPanel, 0);
        repositionObject();
    }
    
    protected void initSize() {
        positionChanging = true;
        setLocation(0, 0);
        canvasSizeListener = new ComponentAdapter() {
            public void componentResized(ComponentEvent evt) {
                setSize(((Container) getParentCanvas()).getSize());
            }
            
            public void componentMoved(ComponentEvent evt) {
                positionChanging = true;
                setLocation(0, 0);
                positionChanging = false;
            }
        };
        ((Container) getParentCanvas()).addComponentListener(canvasSizeListener);
        setSize(((Container) getParentCanvas()).getSize());
        positionChanging = false;
    }
    
    public boolean masksConnectionsAt(Point pt) {
        boolean result = false;
        Point childPoint;
        Ellipse2D ctrRegion = getCenterRegion();
        if (ctrRegion != null) {
            result = result || ctrRegion.contains(pt.x, pt.y);
        }
        if (objNameText != null) {
            childPoint = SwingUtilities.convertPoint(this, pt, objNameText);
            result = result || objNameText.contains(childPoint);
        }
        return result;
    }
    
    public void paintComponent(Graphics g) {
        if (g instanceof EpsGraphics2D) {
            ((EpsGraphics2D)g).addComment("");
            ((EpsGraphics2D)g).addComment("Object<" + objName + ">");
        }
        super.paintComponent(g);
    }
    
    protected void repositionName() {
        if (obj == null) return;
        OSMXPositionedText textElem = (OSMXPositionedText) obj.getName();
        Point ctrPt = getCenterPoint();
        if (ctrPt == null) return;
        if (!textElem.isSetPosition()) {
            objNameText.setPosition(0 + OBJECT_RADIUS,
                                    -objNameText.getPreferredSize().height + OBJECT_RADIUS);
        } else {
            objNameText.setLocationFromPosition();
        }
    }
    
    protected void repositionObject() {
        Point ctrPt = getCenterPoint();
        if (ctrPt == null) return;
        int x = (ctrPt.x - OBJECT_RADIUS);
        int y = (ctrPt.y - OBJECT_RADIUS);
        objectPanel.setLocation(x, y);
    }
    
    protected void setLocationFromPosition() {
        positionChanging = true;
        if (objectPanel != null) {
            int x = posElem.getX();
            int y = posElem.getY();
            Point pt = new Point(getOrigin());
            pt.x = -pt.x;
            pt.y = -pt.y;
            pt.translate(x, y);
            Point parentOrigin = ((CustomCoordinatePlane) getParent()).getOrigin();
            if (parentOrigin != null) {
                pt.translate(parentOrigin.x, parentOrigin.y);
            }
            initCenterPoint();
            repositionObject();
            repositionName();
        }
        positionChanging = false;
    }
    
    protected void setPosition(Point pt) {
        if (pt != null) {
            setPosition(pt.x, pt.y);
        } else if (!positionChanging && element instanceof PositionedElement) {
            PositionedElement pe = (PositionedElement) element;
            pe.unsetX();
            pe.unsetY();
        }
    }
    
    protected void setPosition(int x, int y) {
        if (!positionChanging && element instanceof PositionedElement) {
            PositionedElement pe = (PositionedElement) element;
            pe.setX(x);
            pe.setY(y);
            initCenterPoint();
            repositionName();
        }
    }
    
    protected void setShapeName(String s) {
        objName = s;
        if (obj != null) {
            obj.getName().setContent(objName);
        }
    }
    
    protected void updatePosition() {
        positionChanging = true;
        if (!posElem.isSetPosition()) {
            this.setPosition(null);
        } else {
            int x = posElem.getX();
            int y = posElem.getY();
            initCenterPoint();
            this.setPosition(x, y);
        }
        positionChanging = false;
    } // method updatePosition
    
    protected void updateSelectionState() {
        super.updateSelectionState();
        objectPanel.updateSelectionState(selected);
    }
    
    protected void locationChanged() {
        if (positionChanging) return;
        Point pt = new Point(getOrigin());
        pt.translate(posElem.getX(), posElem.getY());
        setPosition(pt);
    }
    
    public void changedUpdate(DocumentEvent e) {
        repositionName();
    }
    
    public void insertUpdate(DocumentEvent e) {
        repositionName();
    }
    
    public void removeUpdate(DocumentEvent e) {
        repositionName();
    }
    
    private class ObjectPanel extends JPanel {
        private Ellipse2D dot;
        
        public ObjectPanel() {
            ObjBorder b = new ObjBorder();
            b.setRectDraw(false);
            b.setHandleWidth(2);
            b.setSelOptions(ObjBorder.TOP_CENTER | ObjBorder.RIGHT_CENTER |
                            ObjBorder.BOTTOM_CENTER | ObjBorder.LEFT_CENTER);
            setBorder(b);
            
            setOpaque(false);
            setLayout(null);
            setSize(OBJECT_RADIUS * 2, OBJECT_RADIUS * 2);
            
            double lineWidth = getLineWidth();
            dot = new Ellipse2D.Double(lineWidth / 2.0, lineWidth / 2.0,
                                       OBJECT_RADIUS * 2.0 - lineWidth,
                                       OBJECT_RADIUS * 2.0 - lineWidth);
        }

        public void paintComponent(Graphics g) {
            Graphics2D g2d = (Graphics2D)g.create();
            
            g2d.setColor(getLineColor());
            g2d.fill(dot);
            g2d.setStroke(ObjectShape.this.getStroke());
            g2d.draw(dot);
        }

        public boolean contains(int x, int y) {
            Point pt = new Point(x,y);
            return contains(pt);
        }

        public boolean contains(Point point) {
            return dot.contains(point);
        }

        protected void processMouseEvent(MouseEvent evt) {
            evt = SwingUtilities.convertMouseEvent(this, evt, ObjectShape.this);
            super.processMouseEvent(evt);
        }

        protected void updateSelectionState(boolean selected) {
            ((ObjBorder) getBorder()).setSelect(selected);
        }

    } /* nested class ObjectPanel */
    
} /* class ObjectShape */
