/*
 * AbstractValueMapper.java
 *
 * Created on March 20, 2004, 8:45 PM
 */

package edu.byu.deg.framework;

import java.util.TreeSet;
import java.util.Iterator;

/**
 * This abstract class provides convenience methods for implementing the
 * {@link ValueMapper} interface.  Sub-classes should generally need only to
 * implement the {@link #generateValueMappings(Ontology, String)} method.
 * @author Alan Wessman
 */
public abstract class AbstractValueMapper implements ValueMapper {
    
    /**
     * Reference to the ontology used in the value mapping process.
     */    
    protected Ontology ontology;
    
    /**
     * Performs the value mapping process using only information from the current
     * ontology.  This means that if the ontology does not provide candidate data for
     * mapping, no mappings can be generated.
     *
     * Note: invokes {@link #generateValueMappings(Ontology)}.
     * @throws ValueMapperConfigurationException Thrown if no ontology has been set for this mapper.
     */    
    public void generateValueMappings() throws ValueMapperConfigurationException {
        if (ontology == null)
            throw new ValueMapperConfigurationException("Ontology must be specified");
        generateValueMappings(ontology);
    }
    
    /**
     * Performs the value mapping process using the current ontology, focusing only on
     * matches belonging to the specified {@link Document}.
     *
     * Note: invokes {@link #generateValueMappings(Ontology, String)}.
     * @param doc The <CODE>Document</CODE> from which the matches to be mapped originated.
     * @throws ValueMapperConfigurationException Thrown when no ontology has been set for this mapper.
     */    
    public void generateValueMappings(Document doc) throws ValueMapperConfigurationException {
        if (ontology == null)
            throw new ValueMapperConfigurationException("Ontology must be specified");
        generateValueMappings(ontology, doc);
    }
    
    /**
     * Performs the mapping process using the specified ontology.  The input ontology
     * must provide the matches to be mapped, or else no mapping will take place.  This
     * method will ignore the mapper's default ontology.
     *
     * Note: invokes {@link #generateValueMappings(Ontology, String)}.
     * @param ont The <CODE>Ontology</CODE> that provides the matches and model structure.
     */    
    public void generateValueMappings(Ontology ont) {
        TreeSet sourceDocUris = new TreeSet(ont.getSourceDocumentURIs());
        
        // invoke generateValueMappings(Ontology, String) using root document URIs
        for (Iterator i = sourceDocUris.iterator(); i.hasNext();) {
            String docURI = (String) i.next();
            generateValueMappings(ont, docURI);
        }
    }
    
    /**
     * Performs the value mapping process for the specified ontology, using only those
     * matches provided by the ontology that came from the specified document.
     *
     * Note: invokes {@link #generateValueMappings(Ontology, String)}.
     * @param ont The ontology used in the mapping process.
     * @param doc The document from which the matches to be mapped originated.
     */    
    public void generateValueMappings(Ontology ont, Document doc) {
        // invoke generateValueMappings(Ontology, String) on URI of document
        generateValueMappings(ont, doc.getURI().toString());
    }
    
    /**
     * Performs the value mapping process on the given ontology, focusing exclusively
     * on those matches that originate from the given URI.
     * @param ont The ontology used to perform the mapping.
     * @param rootURI The URI of the root document from which all the matches to be mapped originated.
     */    
    public abstract void generateValueMappings(Ontology ont, String rootURI);
    
    /**
     * Accessor for the default ontology of the mapper.
     * @return The default ontology used by the mapper.
     */    
    public Ontology getOntology() {
        return ontology;
    }
    
    /**
     * Sets the default ontology for the mapper.
     * @param ont The ontology to use as the default when performing the value mapping process.
     */    
    public void setOntology(Ontology ont) {
        ontology = ont;
    }
    
}
