/*
 * DataExtractionEngine.java
 *
 * Created on February 8, 2003, 11:21 AM
 */

package edu.byu.deg.framework;

/** Provides a template for data extraction implementations.
 * Subclass to implement initialization and extraction routines.
 * Use {@link #setInitParameter(Object, Object)
 * setInitParameter} to store objects required for the
 * {@link #initialize() initialize } method.  Subclasses that implement initialize
 * can use {@link #getInitParameter(Object)  getInitParameter}
 * to retrieve the objects that have been stored.
 * @author Alan
 */
public abstract class DataExtractionEngine
{
  /** Contains any objects needed for the {@link #initialize initialize} method. */  
  protected java.util.Map initParameters;
  
  /**
   * Establishes namespace for the logger.
   */  
  public static final String LOGGER_NAME = "edu.byu.deg.framework";
  /**
   * Logger for producing output via the framework.
   */  
  protected static java.util.logging.Logger logger;
  
  /** Creates a new DataExtractionEngine instance. */  
  public DataExtractionEngine() {
    initParameters = new java.util.HashMap();
  }
  
  /** Performs any necessary pre-processing in anticipation of a 
   * {@link #doExtraction doExtraction } invocation.
   */  
  public abstract void initialize();
  
  /** Stores objects required for upcoming initialization process.
   * @param paramKey Used for identifying and retrieving initialization objects.
   * @param paramVal The object to use in the initialization process.
   */  
  public void setInitParameter(Object paramKey, Object paramVal) {
    initParameters.put(paramKey, paramVal);
  }
  
  /** Locates and returns an initialization object.
   * @param paramKey Object used to located the desired object.
   * @return The object previously stored for initialization.
   */  
  public Object getInitParameter(Object paramKey) {
    return initParameters.get(paramKey);
  }
  
  /**
   * Returns the number of init parameters passed to the engine; useful for iterating
   * through the list of parameters.
   * @return The number of init parameters stored by the engine.
   */  
  public int countInitParameters() {
      return initParameters.size();
  }
  
  /**
   * Removes the specified init parameter from the parameter list.
   * @param paramKey Name of the parameter to remove.
   */  
  public void removeInitParameter(Object paramKey) {
      initParameters.remove(paramKey);
  }
  
  /**
   * Determines whether a particular parameter has been stored in the parameter list.
   * @param paramKey Name of the parameter to query for.
   * @return <CODE>true</CODE> if the parameter is stored in the list; <CODE>false</CODE>
   * otherwise.
   */  
  public boolean containsInitParameter(Object paramKey) {
      return initParameters.containsKey(paramKey);
  }
  
  /** Deletes all references to previously stored initialization objects and their
   * keys.
   */  
  public void clearInitParameters() {
    initParameters.clear();
  }
  
  /**
   * Performs the primary task of data extraction.  Implementations will probably
   * need to make sure that {@link #initialize initialize} is invoked sometime
   * before this method.
   * @throws EngineConfigurationException Thrown if any required initialization has not been performed prior to this
   * method being called.
   */  
  public abstract void doExtraction() throws EngineConfigurationException;
  
  /** Returns the <CODE>Logger</CODE> associated with the data extraction engine.  If
   * subclasses need to define a different logger namespace, they may simply call
   * {@link java.util.logging.Logger#getLogger(String)} within their static
   * ininitializer and assign the result to the static {@link #logger} variable.
   * @return The root <CODE>Logger</CODE> associated with this program.
   */  
  public static java.util.logging.Logger getLogger() {
      if (logger != null) return logger;
      logger = java.util.logging.Logger.getLogger(LOGGER_NAME);
      return logger;
  }
}

/*==============================================================================
 * $Author: wessmana $
 * $Revision: 1.12 $
 * $Date: 2005/04/09 21:01:32 $
 *
 * $Log: DataExtractionEngine.java,v $
 * Revision 1.12  2005/04/09 21:01:32  wessmana
 * Merged from osmx-refactor branch.
 *
 * Revision 1.11.2.2  2005/04/01 05:15:26  wessmana
 * Import cleanup.
 *
 * Revision 1.11  2004/09/06 23:38:04  wessmana
 * Corrected Javadoc.
 *
 * Revision 1.10  2004/07/13 04:44:33  wessmana
 * Documentation.
 *
 * Revision 1.9  2004/07/02 06:45:19  wessmana
 * Added countInitParameters method.
 *
 * Revision 1.8  2004/02/13 08:18:01  wessmana
 * Refactored initialization -> init; added convenience methods for working with initParameters; execute() throws EngineConfigurationException.
 *
 * Revision 1.7  2004/02/10 16:25:15  wessmana
 * Added getLogger().
 *
 * Revision 1.6  2004/02/09 19:45:39  wessmana
 * Logger name constant.
 *
 * Revision 1.5  2004/01/06 06:51:10  wessmana
 * Fleshing out framework.
 *
 * Revision 1.4  2003/02/12 04:43:28  wessmana
 * Added Javadoc.
 *
 *
 *------------------------------------------------------------------------------
 * Copyright 2003, Brigham Young University
 */