/*
 * DataFrameMatcher.java
 *
 * Created on August 4, 2003, 11:23 AM
 */

package edu.byu.deg.ontos;

import java.net.URI;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.util.Set;
import java.util.List;
import java.util.HashSet;
import java.util.regex.*;
import edu.byu.deg.framework.Ontology;
import edu.byu.deg.framework.Document;
import edu.byu.deg.framework.ValueRecognizer;
import edu.byu.deg.framework.TextDocument;
import edu.byu.deg.framework.DataExtractionEngine;
import edu.byu.deg.osmx.binding.*;
import edu.byu.deg.osmx.binding.MatchedTextType;
import edu.byu.deg.osmx.OSMXDocument;
import edu.byu.deg.ontos.lexicon.LexiconMatcher;
import javax.xml.bind.JAXBException;

/**
 * This class implements the contract of
 * {@link edu.byu.deg.framework.ValueRecognizer}, extracting values from an input
 * text based on matching rules (value and keyword phrases) from an ontology.  It
 * stores the matched values back in the ontology according to the OSMX definition,
 * with matched values appearing as child elements of the phrase expressions they
 * matched.
 * @author Troy Walker
 * @author Alan Wessman
 */
public class DataFrameMatcher implements ValueRecognizer {
    static final int BASE_FLAGS = Pattern.MULTILINE;
    static final int CASE_INSENSITIVE = BASE_FLAGS | Pattern.CASE_INSENSITIVE;
    LexiconMatcher lexiconMatcher;
    MacroMatcher macroMatcher;
    OSMXDocument ontologyDoc;
    OSM modelRoot;
    Pattern lexMacroLabelPattern;
    Map expressionCache;
    
    
    /**
     * Creates a new instance of <CODE>DataFrameMatcher</CODE>.
     */    
    public DataFrameMatcher() {
        lexiconMatcher = new LexiconMatcher();
        expressionCache = new HashMap();
    }
    
    /**
     * Creates a new instance of DataFrameMatcher and initializes it with the given
     * ontology.
     * @param ontologyDocument The ontology from which the matcher obtains its matching rules and into which it
     * stores its matches.
     */
    public DataFrameMatcher(OSMXDocument ontologyDocument) {
        this();
        setOntology(ontologyDocument);
    }
    
    /**
     * Re-initializes the <CODE>DataFrameMatcher</CODE> with a new ontology.  The
     * method not only stores a reference to the ontology, but it also loads lexicons
     * associated with the ontology.
     * @param ontologyDocument The new ontology to use for matching.  The method will do nothing if this is the
     * same ontology as was already in use, or if the parameter is <CODE>null</CODE>.
     */    
    public void setOntology(Ontology ontologyDocument) {
        // exit early if we already have the specified document set
        if (ontologyDocument == ontologyDoc || ontologyDocument == null) return;
        
        clearExpressionCache();
        
        ontologyDoc = (OSMXDocument) ontologyDocument;
        modelRoot = ontologyDoc.getModelRoot();
        lexiconMatcher = new LexiconMatcher();
        lexiconMatcher.loadLexicons(modelRoot);
        macroMatcher = new MacroMatcher(modelRoot);
    }
    
    /**
     * Accessor for the ontology currently in use.
     * @return The ontology used by this <CODE>DataFrameMatcher</CODE>.
     */    
    public Ontology getOntology() {
        return ontologyDoc;
    }    

    /**
     * Implements the interface method, invoking
     * {@link #doMatches(edu.byu.deg.framework.TextDocument)} with the input parameter.
     * This means that the parameter must be a <CODE>TextDocument</CODE>, or else a
     * {@link ClassCastException} will result.
     * @param doc The document to search for matches.  This must be a <CODE>TextDocument</CODE>.
     */    
    public void findValues(Document doc) {
        doMatches((TextDocument) doc);
    }
    
    /**
     * Implements the interface method, invoking
     * {@link #doMatches(edu.byu.deg.framework.TextDocument)} with the input parameter.
     * This means that the parameter must be a <CODE>TextDocument</CODE>, or else a
     * {@link ClassCastException} will result.
     * @param doc The document to search for matches.  This must be a <CODE>TextDocument</CODE>.
     */    
    public void findValues(String text) {
        doMatches(text, "");
    }
    
    
    
    /**
     * Implemented interface method.  Calls {@link #setOntology} and then
     * {@link #doMatches(edu.byu.deg.framework.TextDocument)}.
     * @param ont The ontology to use for matching.
     * @param doc The <CODE>TextDocument</CODE> to supply the match values.
     */    
    public void findValues(Ontology ont, Document doc) {
        setOntology(ont);
        findValues(doc);
    }
    
    /**
     * Performs the matching process by invoking {@link #doMatches(String, String)}
     * using the document's contents and its URI as arguments.
     * @param doc The document to perform matching on.
     */    
    public void doMatches(TextDocument doc){
      try{
        doMatches(doc.getContentString(), doc.getURI().toString());
      }catch(java.io.IOException e){
        e.printStackTrace(); 
      }
    }
    
    /**
     * Extracts text matches from the given string using the stored ontology's matching
     * rules, and labels each match with the URI of the originating document.
     * @param str The contents in which to search for matches.
     * @param docURL The original location of the document we are searching for matches.
     */    
    public void doMatches(String str, String docURL){
        lexiconMatcher.findMatches(str);
        Map lexExprs = new HashMap();
        for (Iterator i = modelRoot.getAllModelElements().iterator(); i.hasNext(); ){
            Object o = i.next();
            if (o instanceof LexiconType) {
                LexiconType lex = (LexiconType) o;                
                lexExprs.put(lex.getLabel(), lexiconMatcher.toRegEx(lex, str));
                String temp = lexiconMatcher.toRegEx(lex, str);
                if(!temp.equals("([a&&[b]])")){ System.out.println("\tLexicon label: " + lex.getLabel() + ", match: " + temp + "\t\tDataFramMatcher.doMatches(String, String)"); }
                // REVISIT: check that no other lexicon has same label?
            }
        }
        
        Object o;
        for(Iterator i=ontologyDoc.getModelRoot().getAllModelElements().iterator(); i.hasNext(); ){
            o = i.next();
            if(o instanceof ObjectSet){
                DataFrameType df = ((ObjectSet)o).getDataFrame();
                if (df==null) continue;
                System.out.println("\tDataFrame: " + ((ObjectSet)o).getName() + "\t\tDataFramMatcher.doMatches(String, String)");
                matchDataFrame(str, docURL, df, lexExprs);
            }
        }
    }
    
    private void matchDataFrame(String str, String docURL, DataFrameType df, Map lexRegExMap) {
        // process value phrases for data frame
        List valPhrases = df.getValuePhrase();
        if (valPhrases != null) {
            for (Iterator i = valPhrases.iterator(); i.hasNext();) {
                ValuePhraseType valPhrase = (ValuePhraseType) i.next();
                matchValuePhrase(str, docURL, valPhrase, lexRegExMap);
            }
        }
        // process keyword phrases for data frame
        List keyPhrases = df.getKeywordPhrase();
        if (keyPhrases != null) {
            matchKeywordPhrases(str, docURL, keyPhrases, lexRegExMap);
        }
    }
    
    private void matchKeywordPhrases(String str, String docURL, List phrases, Map lexRegExMap) {
        for (Iterator i = phrases.iterator(); i.hasNext();) {
            KeywordPhraseType keyPhrase = (KeywordPhraseType) i.next();
            matchKeywordPhrase(str, docURL, keyPhrase, lexRegExMap);
        }
    }
    
    private void matchValuePhrase(String str, String docURL, ValuePhraseType vp, Map lexRegExMap) {
        // get refs to the various expressions of the value phrase
        DataFrameExpression reqContext = vp.getRequiredContextExpression();
        DataFrameExpression leftContext = vp.getLeftContextExpression();
        DataFrameExpression rightContext = vp.getRightContextExpression();
        DataFrameExpression valExpr = vp.getValueExpression();
        DataFrameExpression exceptExpr = vp.getExceptionExpression();
        DataFrameExpression subFrom = vp.getSubFromExpression();
        DataFrameExpression subTo = vp.getSubToExpression();
        
        // get cache of processed expressions
        Map cachedExprs = (Map) expressionCache.get(vp);
        if (cachedExprs == null) {
            cachedExprs = new HashMap();
            expressionCache.put(vp, cachedExprs);
        }
        String reqContextStr = getProcessedExpression(reqContext, cachedExprs);
        String leftContextStr = getProcessedExpression(leftContext, cachedExprs);
        String rightContextStr = getProcessedExpression(rightContext, cachedExprs);
        String valExprStr = getProcessedExpression(valExpr, cachedExprs);
        String exceptExprStr = getProcessedExpression(exceptExpr, cachedExprs);
        String subFromStr = getProcessedExpression(subFrom, cachedExprs);
        String subToStr = getProcessedExpression(subTo, cachedExprs);
        
        // early exit if no value expression
        if (valExprStr == null) {
            DataExtractionEngine.getLogger().warning("Null value expression for value phrase " + vp.getHint());
            return;
        }
        
        // Set up object factory for future generation of MatchedText elements
        ObjectFactory factory = ontologyDoc.getObjectFactory();
        
        // process value expression and exception expression since we know 
        // we'll need them no matter what
        Pattern valPattern = getExprPattern(valExprStr, lexRegExMap, vp.isCaseSensitive());
        if (valPattern == null) {
            DataExtractionEngine.getLogger().warning("Value expression " + valExprStr + " is invalid");
            return;
        }
        Pattern exceptPattern = getExprPattern(exceptExprStr, lexRegExMap, vp.isCaseSensitive());
                
        // if required context specified, find all instances of it
        if (reqContextStr != null) {
            Pattern rcp = getExprPattern(reqContextStr, lexRegExMap, vp.isCaseSensitive());
            if (rcp != null) {
                Matcher rcm = rcp.matcher(str);
                while (rcm.find()) {
                    String rcMatch = rcm.group();
                    //reqContext.getMatchedText().add(createMatch(factory, docURL, rcMatch, rcm.start(), rcm.end()));
                    Matcher valMatcher = valPattern.matcher(rcMatch);
                    while (valMatcher.find()) {
                        // TODO: filter out exception matches
                        String valMatch = valMatcher.group();
                        if (valMatch.length() == 0) 
                            continue;
                        int valStart = valMatcher.start() + rcm.start();
                        int valEnd = valMatcher.end() + rcm.start();
                        MatchedTextType valMatchElem = null;
                        Matcher exceptMatch = null;
                        if (exceptPattern != null) exceptMatch = exceptPattern.matcher(valMatch);
                        if (exceptMatch == null || !exceptMatch.find()) {
                            valMatchElem = createMatch(factory, docURL, valMatch, valStart, valEnd);
                            valExpr.getMatchedText().add(valMatchElem);
                        }
                        if ((subToStr != null || subFromStr != null) && valMatchElem != null) {
                            Pattern subFromPattern = null;
							if (subFromStr != null) subFromPattern = getExprPattern(subFromStr, lexRegExMap, vp.isCaseSensitive());
                            if (subFromPattern == null) {
                                subFromPattern = getExprPattern(".+", lexRegExMap, vp.isCaseSensitive());
                            }
                            if (subFromPattern != null) {
                                if (subToStr == null) subToStr = "";
                                String newValMatch = subFromPattern.matcher(valMatch).replaceAll(subToStr);
                                valMatchElem.setValue(newValMatch);
                                // REVISIT: adjust end position of string after substitution?
                            }
                        }
                    }
                }
            }
        } else {
            int valGroup = 1;
            
            // determine which capturing group represents the value expression
            // when the left context is prepended to it
            Pattern leftPattern = getExprPattern(leftContextStr, lexRegExMap, vp.isCaseSensitive());
            if (leftPattern != null) {
                // for some reason we need a matcher before we can count groups
                Matcher leftMatcher = leftPattern.matcher(str);
                valGroup += leftMatcher.groupCount();
            }
            
            StringBuffer fullExpr = new StringBuffer();
            fullExpr.append(leftContextStr == null ? "" : leftContextStr);
            fullExpr.append("(" + valExprStr + ")");
            fullExpr.append(rightContextStr == null ? "" : rightContextStr);
            Pattern fullPattern = getExprPattern(fullExpr.toString(), lexRegExMap, vp.isCaseSensitive());
            Matcher valMatcher = fullPattern.matcher(str);
            while (valMatcher.find()) {
                // TODO: filter out exception matches
                String valMatch = valMatcher.group(valGroup);
                if (valMatch.length() == 0) 
                    continue;
                int valStart = valMatcher.start(valGroup);
                int valEnd = valMatcher.end(valGroup);
                MatchedTextType valMatchElem = null;
                Matcher exceptMatcher = null;
                if (exceptPattern != null) exceptMatcher = exceptPattern.matcher(valMatch);
                if (exceptMatcher == null || !exceptMatcher.find()) {
                    valMatchElem = createMatch(factory, docURL, valMatch, valStart, valEnd);
                    valExpr.getMatchedText().add(valMatchElem);
                    System.out.println("\t\tValuePhrase match: " + valMatchElem.getValue() + " (" + valStart + ", " + valEnd + ")\t\tDataFramMatcher.matchValuePhrase(...)" );
                    
                }
                if ((subToStr != null || subFromStr != null) && valMatchElem != null) {
                    Pattern subFromPattern = null;
					if (subFromStr != null) subFromPattern = getExprPattern(subFromStr, lexRegExMap, vp.isCaseSensitive());
					if (subFromPattern == null) {
						subFromPattern = getExprPattern(".+", lexRegExMap, vp.isCaseSensitive());
                    }
                    if (subFromPattern != null) {
                        if (subToStr == null) subToStr = "";
                        String newValMatch = subFromPattern.matcher(valMatch).replaceAll(subToStr);
                        valMatchElem.setValue(newValMatch);
                        // REVISIT: adjust end position of string after substitution?
                    }
                }
            }
        }
        
        // finally, match keywords for this value phrase
        List keyPhrases = vp.getKeywordPhrase();
        if (keyPhrases != null) {
            matchKeywordPhrases(str, docURL, keyPhrases, lexRegExMap);
        }
    }
    
    private String getProcessedExpression(DataFrameExpression expr, Map exprCache) {
        // early exit if expr is null
        if (expr == null) return null;
        // try to find in cache
        String exprStr = (String) exprCache.get(expr);
        // cache miss
        if (!exprCache.containsKey(expr)) {
            // perform macro substitution; new set is for root level of recursion
            exprStr = replaceMacroRefs(expr.getExpressionText(), new HashSet());
            // cache what we found
            if (exprStr != null && exprStr.length() > 0) 
                expressionCache.put(expr, exprStr);
        }
        if (exprStr != null && exprStr.length() == 0) return null;
        return exprStr;
    }
    
    private Pattern getExprPattern(String expr, Map lexRegExMap, boolean caseSensitive) {
        String patternExpr = replaceLexiconRefs(expr, lexRegExMap);
        if (patternExpr == null) return null;
        try {
            int flags = BASE_FLAGS;
            if (!caseSensitive) patternExpr = "(?i)" + patternExpr; //flags = CASE_INSENSITIVE;
            return Pattern.compile(patternExpr, flags);
        } catch (PatternSyntaxException pse) {
            DataExtractionEngine.getLogger().warning("Invalid pattern: " + patternExpr + "\n" + pse.getMessage());
            return null;
        }
    }
    
    private void matchKeywordPhrase(String str, String docURL, KeywordPhraseType kp, Map lexRegExMap) {
        DataFrameExpression expr = kp.getKeywordExpression();
        if (expr == null) return;
        
        // get original keyword expression
        String origExpr = expr.getExpressionText();
        // realExpr will contain the final expression that we really use
        // after macro and lexicon substitution are performed
        String realExpr = null;
        
        // use the cached expression (macro substitution already done) if we have one
        // keyword phrases only have one expression so the string is the hashed value
        // note that this is different than for value phrases, so can't use getProcessedExpression()
        realExpr = (String) expressionCache.get(kp);
        if (!expressionCache.containsKey(kp)) {
            // perform macro substitution
            realExpr = replaceMacroRefs(origExpr, new HashSet());
            // cache the result for later uses
            expressionCache.put(kp, realExpr);
        }
        
        // perform lexicon substitution, which is input-dependent so can't cache
        realExpr = replaceLexiconRefs(realExpr, lexRegExMap);
        
        // early exit if expression is degenerate
        if (realExpr == null || realExpr.length() == 0) return;
        
        // compile pattern with the expression we will really use
        Pattern keyPhrasePattern = getExprPattern(realExpr, lexRegExMap, kp.isCaseSensitive());
        if (keyPhrasePattern == null) {
            DataExtractionEngine.getLogger().warning("Keyword expression " + realExpr + " is invalid");
            return;
        }
        
        // perform matching
        Matcher m = keyPhrasePattern.matcher(str);
        // get ref to object factory so we can create MatchedText elements
        ObjectFactory factory = ontologyDoc.getObjectFactory();
        while (m.find()) {
            expr.getMatchedText().add(createMatch(factory, docURL, m.group(), m.start(), m.end()));
        }
    }
    
    private MatchedTextType createMatch(ObjectFactory factory, String docURL, String match, int startPos, int endPos) {
        try {
            MatchedTextType mtt = factory.createMatchedTextType();
            mtt.setDocument(docURL);
            mtt.setValue(match);
            mtt.setStartPos(startPos);
            mtt.setEndPos(endPos);
            mtt.setLocation(startPos + ", " + endPos);
            return mtt;
        } catch (JAXBException jaxbe) {
            DataExtractionEngine.getLogger().severe("Error creating MatchedText element: " + jaxbe.getMessage());
            return null;
        }
    }
    
    private String replaceLexiconRefs(String expr, Map lexRegExMap) throws PatternSyntaxException {
        // early exit if no lexicon ref can exist due to length or nullity
        if (expr == null || expr.length() < 3) return expr;
        
        // look for any lexicon labels in the expression
        Matcher m = getLexMacroLabelPattern().matcher(expr);
        StringBuffer newExpr = new StringBuffer();
        while (m.find()) {
            String label = m.group(1);
            String sub = (String) lexRegExMap.get(label);
            if (sub == null) sub = "\\{" + m.group(1) + "\\}";
            else sub = sub.replaceAll("\\\\", "\\\\\\\\");
            m.appendReplacement(newExpr, sub);
        }
        m.appendTail(newExpr);
        return newExpr.toString();
    }
    
    private Pattern getLexMacroLabelPattern() {
        // lazy initialization of lexMacroLabelPattern
        if (lexMacroLabelPattern == null) {
            lexMacroLabelPattern = Pattern.compile("\\{([^0-9\\{\\}][^\\{\\}]*)\\}");
        }
        return lexMacroLabelPattern;
    }
    
    private String replaceMacroRefs(String expr, Set processedMacros) {
        // early exit if no macro ref can exist due to length or nullity
        if (expr == null || expr.length() < 3) return expr;

        Matcher m = getLexMacroLabelPattern().matcher(expr);
        StringBuffer newExpr = new StringBuffer();
        
        while (m.find()) {
            String label = m.group(1);
            String sub = macroMatcher.getPattern(label);
            boolean recurse = true;
            if (sub == null || processedMacros.contains(label)) {
                sub = m.group();
                recurse = false;
            }
            if (recurse) {
                processedMacros.add(label);
                sub = "(" + replaceMacroRefs(sub, processedMacros) + ")";
                processedMacros.remove(label);
            }
            m.appendReplacement(newExpr, sub.replaceAll("\\\\", "\\\\\\\\"));
        }
        m.appendTail(newExpr);
        return newExpr.toString();
    }
    
    /**
     * Removes all <CODE>MatchedText</CODE> elements from the ontology.
     */    
    public void clearMatches(){
        lexiconMatcher.clearMatches();
        Object o;
        DataFrameType df;
        List valuePhrases;
        ValuePhraseType valuePhrase;
        for(Iterator i=ontologyDoc.getModelRoot().getAllModelElements().iterator(); i.hasNext(); ){
            o = i.next();
            if(o instanceof ObjectSet){
                df = ((ObjectSet)o).getDataFrame();
                if(df==null)
                    continue;
                valuePhrases = df.getValuePhrase();
                if(valuePhrases!=null)
                    for(Iterator j=valuePhrases.iterator(); j.hasNext(); ){
                        valuePhrase = (ValuePhraseType)j.next();
                        valuePhrase.getValueExpression().getMatchedText().clear();
                        if(valuePhrase.getKeywordPhrase()!=null)
                            for(Iterator k=valuePhrase.getKeywordPhrase().iterator(); k.hasNext(); ){
                                ((KeywordPhraseType)k.next()).getKeywordExpression().getMatchedText().clear();
                            }
                    }
                if(df.getKeywordPhrase()!=null)
                    for(Iterator k=df.getKeywordPhrase().iterator(); k.hasNext(); ){
                        ((KeywordPhraseType)k.next()).getKeywordExpression().getMatchedText().clear();
                    }
            }
        }
    }
    
    /**
     * Clears the set of cached, macro-expanded expressions.  Useful if this
     * <CODE>DataFrameMatcher</CODE> is to be re-used for another ontology.
     */    
    public void clearExpressionCache() {
        expressionCache.clear();
    }
    
    /**
     * Interface method implementation.  Performs extraction on the specified string,
     * associating the matches with the given source URI.  This method invokes
     * {@link #doMatches(String, String)}.
     * @param contents The string to search for matches.
     * @param sourceURI URI of the originating document.
     */    
    public void findValues(String contents, URI sourceURI) {
        doMatches(contents, sourceURI.toString());
    }
    
    /**
     * Interface method implementation.  Stores the specified ontology and then invokes
     * {@link #findValues(String, URI)} on the remaining arguments.
     * @param ontology The ontology to use for extraction.
     * @param contents The string to search for matches.
     * @param sourceURI The URI of the source document.
     */    
    public void findValues(Ontology ontology, String contents, URI sourceURI) {
        setOntology(ontology);
        findValues(contents, sourceURI);
    }
    
    
}
