/*
 * ExtractionOntologyIndex.java
 *
 * Created on March 23, 2004, 9:18 PM
 */

package edu.byu.deg.ontos;

import edu.byu.deg.osmx.*;
import edu.byu.deg.osmx.binding.*;
import edu.byu.deg.osmx.binding.impl.DataInstanceImpl;
import edu.byu.deg.osmx.binding.impl.MatchedTextTypeImpl;
import edu.byu.deg.osmx.binding.impl.OSMObjectImpl;
import edu.byu.deg.osmx.binding.impl.ObjectTypeImpl;
import edu.byu.deg.osmx.binding.impl.SourceDocumentImpl;

import java.util.*;

import edu.byu.deg.framework.DataExtractionEngine;
import javax.xml.bind.JAXBException;

/**
 * Provides convenient accessors and some caching of access paths into parts of an
 * ontology relevant for extraction activities.
 * @author Alan Wessman
 * 
 * More functionality has been added 
 * support what is needed for Task Recognition Process and Task View Creation
 *@author Muhammed Al-Muhammed
 */
public class ExtractionOntologyIndex{
    
    OSMXDocument ontology = null;
    
    private Map objSets = new HashMap();
    private Set genSpecs = new HashSet();
    private Set relSets = new HashSet();
    private Map connectedObjSets = new HashMap();
    private Set dataInstances = new HashSet();
    private ObjectSetType primary = null;
    private Map matchExpressions = new HashMap();
    private Map expressionPhrases = new HashMap();
    private Map phraseObjSets = new HashMap();
    private Map sourceDocuments = new HashMap();
    private DataInstanceType activeDataInstance = null;
    
  
    
    /** Creates a new instance of ExtractionOntologyIndex */
    public ExtractionOntologyIndex() { /* default constructor */ }
    
    /**
     * Instantiates the index and initializes it with the given ontology.
     * @param ont The ontology to index.
     */    
    public ExtractionOntologyIndex(OSMXDocument ont) {
        setOntology(ont);
    }
    
    /**
     * Accessor for the indexed ontology.
     * @return The indexed ontology.
     */    
    public OSMXDocument getOntology() {
        return ontology;
    }
    
    /**
     * Accessor for the primary object set of the ontology. If the ontology has
     * multiple primary object sets specified, this method will return the first
     * primary object set encountered during the indexing process.  There is no
     * guaranteed algorithm for determining which object set will be encountered first.
     * @return The primary object set, or <CODE>null</CODE> if none exists.
     */    
    public ObjectSetType getPrimaryObjSet() {
        return primary;
    }
    
    /**
     * Indexes an ontology after clearing from the index any previously indexed
     * ontology.
     * @param ont The ontology to index.
     */    
    public void setOntology(OSMXDocument ont) {
        clearIndexes();
        ontology = ont;
        if (ontology == null) return;
        
        indexObjectSets();
        indexGenSpecs();
        indexRelSets();
        indexDataInstances();
        
    }
    
    private void clearIndexes() {
        objSets.clear();
        genSpecs.clear();
        relSets.clear();
        dataInstances.clear();
        sourceDocuments.clear();
        connectedObjSets.clear();
        matchExpressions.clear();
        expressionPhrases.clear();
        phraseObjSets.clear();
    }
    
    private void indexObjectSets() {
        matchExpressions.clear();
        expressionPhrases.clear();
        phraseObjSets.clear();
        for (Iterator i = ontology.getObjectSets().iterator(); i.hasNext();) {
            ObjectSetType objSet = (ObjectSetType) i.next();
            ObjSetIndex objSetIdx = new ObjSetIndex(objSet);
            objSets.put(objSet.getId(), objSetIdx);
            indexMatches(objSet);
            indexPhrases(objSetIdx);
            if (objSet.isPrimary()) {
                if (primary == null) {
                    primary = objSet;
                } else if (primary != objSet) {
                    DataExtractionEngine.getLogger().warning("Ontology has multiple primary object sets.");
                }
            }
        }
        if (primary == null) {
            DataExtractionEngine.getLogger().warning("Ontology has no primary object set.");
        }
    }
    
    private void indexPhrases(ObjSetIndex objSetIdx) {
        if (objSetIdx == null) return;
        ObjectSetType objSet = objSetIdx.objSet;
        Set valPhraseSet = objSetIdx.valPhrases;
        Set keyPhraseSet = objSetIdx.keyPhrases;
        DataFrameType datFrame = objSet.getDataFrame();
        if (datFrame == null) return;
        List vpl = datFrame.getValuePhrase();
        if (vpl != null) valPhraseSet.addAll(vpl);
        List kpl = datFrame.getKeywordPhrase();
        if (kpl != null) keyPhraseSet.addAll(kpl);
    }
    
    private void indexMatches(ObjectSetType objSet) {
        if (objSet == null) return;
        DataFrameType datFrame = objSet.getDataFrame();
        if (datFrame == null) return;
        List vpl = datFrame.getValuePhrase();
        if (vpl != null) {
            for (Iterator i = vpl.iterator(); i.hasNext();) {
                ValuePhraseType vp = (ValuePhraseType) i.next();
                phraseObjSets.put(vp, objSet);
                DataFrameExpression expr = vp.getValueExpression();
                if (expr == null) continue;
                expressionPhrases.put(expr, vp);
                for (Iterator j = expr.getMatchedText().iterator(); j.hasNext();) {
                    MatchedTextType match = (MatchedTextType) j.next();
                    matchExpressions.put(match, expr);
                }
                List kpl = vp.getKeywordPhrase();
                if (kpl != null) {
                    for (Iterator j = kpl.iterator(); j.hasNext();) {
                        KeywordPhraseType kp = (KeywordPhraseType) j.next();
                        phraseObjSets.put(kp, objSet);
                        DataFrameExpression kexpr = kp.getKeywordExpression();
                        if (kexpr == null) continue;
                        expressionPhrases.put(kexpr, kp);
                        for (Iterator k = kexpr.getMatchedText().iterator(); k.hasNext();) {
                            MatchedTextType match = (MatchedTextType) k.next();
                            matchExpressions.put(match, kexpr);
                        }
                    }
                }
            }
        }
        List kpl = datFrame.getKeywordPhrase();
        if (kpl != null) {
            for (Iterator i = kpl.iterator(); i.hasNext();) {
                KeywordPhraseType kp = (KeywordPhraseType) i.next();
                phraseObjSets.put(kp, objSet);
                DataFrameExpression expr = kp.getKeywordExpression();
                if (expr == null) continue;
                expressionPhrases.put(expr, kp);
                for (Iterator j = expr.getMatchedText().iterator(); j.hasNext();) {
                    MatchedTextType match = (MatchedTextType) j.next();
                    matchExpressions.put(match, expr);
                }
            }
        }
    }
    
    private void indexGenSpecs() {
        for (Iterator i = ontology.getGenSpecs().iterator(); i.hasNext();) {
            GenSpecType genSpec = (GenSpecType) i.next();
            genSpecs.add(genSpec);
            List genConnections = genSpec.getGenConnection();
            List specConnections = genSpec.getSpecConnection();
            if (genConnections != null) {
                for (Iterator j = genConnections.iterator(); j.hasNext();) {
                    BasicConnection conn = (BasicConnection) j.next();
                    ObjSetIndex obj = (ObjSetIndex) objSets.get(conn.getObjectSet());
                    connectedObjSets.put(conn, obj.getObjectSet().getId());
                    Set specs = obj.getSpecializations();
                    for (Iterator k = specConnections.iterator(); k.hasNext();) {
                        BasicConnection spec = (BasicConnection) k.next();
                        specs.add(spec.getObjectSet());
                    }
                }
            }
            if (specConnections != null) {
                for (Iterator j = specConnections.iterator(); j.hasNext();) {
                    BasicConnection conn = (BasicConnection) j.next();
                    ObjSetIndex obj = (ObjSetIndex) objSets.get(conn.getObjectSet());
                    connectedObjSets.put(conn, obj.getObjectSet().getId());
                    Set geners = obj.getGeneralizations();
                    for (Iterator k = genConnections.iterator(); k.hasNext();) {
                        BasicConnection gen = (BasicConnection) k.next();
                        geners.add(gen.getObjectSet());
                    }
                }
            }
        }
    }
    
    private void indexRelSets() {
        for (Iterator i = ontology.getRelSets().iterator(); i.hasNext();) {
            RelationshipSetType relSet = (RelationshipSetType) i.next();
            relSets.add(relSet);
            List relSetConnections = relSet.getRelSetConnection();
            if (relSetConnections == null) continue;
            for (Iterator j = relSetConnections.iterator(); j.hasNext();) {
                RelSetConnectionType conn = (RelSetConnectionType) j.next();
                ObjSetIndex obj = (ObjSetIndex) objSets.get(conn.getObjectSet());
                if (obj != null) connectedObjSets.put(conn, obj.getObjectSet().getId());
                obj.getRelSets().add(relSet);
            }
        }
    }
    
    private void indexDataInstances() {
        dataInstances.clear();
        sourceDocuments.clear();
        for (Iterator i = ontology.getDataInstances().iterator(); i.hasNext();) {
            DataInstanceType di = (DataInstanceType) i.next();
            dataInstances.add(di);
            for (Iterator j = di.getDataInstanceElements().iterator(); j.hasNext();) {
                Object o = j.next();
                if (!(o instanceof SourceDocumentType)) continue;
                SourceDocumentType doc = (SourceDocumentType) o;
                sourceDocuments.put(doc.getUri(), doc);
            }
        }
    }
    
    /**
     * Returns the set of <CODE>DataInstance</CODE> elements belonging to the indexed
     * ontology.
     * @return An unmodifiable set of data instances found in the ontology.
     */    
    public Set getDataInstances() {
        return Collections.unmodifiableSet(dataInstances);
    }
    
    /**
     * Returns the object set to which the given data frame expression belongs.
     * @param expr The expression whose parent object set we wish to locate.
     * @return The object set that owns the given expression.
     */    
    public ObjectSetType getObjectSet(DataFrameExpression expr) {
        return (ObjectSetType) phraseObjSets.get(expressionPhrases.get(expr));
    }
    
    /**
     * Returns the object set to which the specified <CODE>MatchedText</CODE> element
     * belongs.
     * @param match The match whose parent object set we wish to identify.
     * @return The object set that owns the given match.
     */    
    public ObjectSetType getObjectSet(MatchedTextType match) {
        DataFrameExpression expr = (DataFrameExpression) matchExpressions.get(match);
        return getObjectSet(expr);
    }
    
    /**
     * Determines whether the given <CODE>MatchedText</CODE> element is associated with a value phrase.
     * @param match The match whose phrase type we wish to know.
     * @return <CODE>true</CODE> if the match belongs to a value phrase; <CODE>false</CODE>
     * otherwise.
     */    
    public boolean isValueMatch(MatchedTextType match) {
        Object obj = expressionPhrases.get(matchExpressions.get(match));
        return (obj != null && obj instanceof ValuePhraseType);
    }
    
    /**
     * Determines whether the given <CODE>MatchedText</CODE> element is associated with
     * a keyword phrase.
     * @param match The match whose phrase type we wish to know.
     * @return <CODE>true</CODE> if the match belongs to a keyword phrase; <CODE>false</CODE>
     * otherwise.
     */    
    public boolean isKeywordMatch(MatchedTextType match) {
        Object obj = expressionPhrases.get(matchExpressions.get(match));
        return (obj != null && obj instanceof KeywordPhraseType);
    }
    
    /**
     * Accesses the <CODE>DataFrameExpression</CODE> that the given <CODE>MatchedText</CODE>
     * matched.
     * @param match The match whose originating expression we wish to identify.
     * @return The expression for which the match was extracted.
     */    
    public DataFrameExpression getExpression(MatchedTextType match) {
        return (DataFrameExpression) matchExpressions.get(match);
    }
    
    /**
     * Locates the <CODE>ValuePhrase</CODE> that owns the given <CODE>MatchedText</CODE>
     * element, or <CODE>null</CODE> if the match does not belong to a value phrase.
     * @param match The match whose parent value phrase we wish to identify.
     * @return The parent value phrase of the match, or null if no value phrase is parent of
     * the match.
     */    
    public ValuePhraseType getValuePhrase(MatchedTextType match) {
        if (!isValueMatch(match)) return null;
        return (ValuePhraseType) expressionPhrases.get(getExpression(match));
    }
    
    /**
     * Locates the <CODE>KeywordPhrase</CODE> that owns the given <CODE>MatchedText</CODE>
     * element, or <CODE>null</CODE> if the match does not belong to a keyword phrase.
     * @param match The match whose parent keyword phrase we wish to identify.
     * @return The parent keyword phrase of the match, or null if no keyword phrase is parent of
     * the match.
     */    
    public KeywordPhraseType getKeywordPhrase(MatchedTextType match) {
        if (!isKeywordMatch(match)) return null;
        return (KeywordPhraseType) expressionPhrases.get(getExpression(match));
    }
    
    /**
     * Returns a <CODE>List</CODE> of all object sets in the ontology, including those
     * that belong to a high-level element.
     * @return A <CODE>List</CODE> containing all object sets in the ontology; no order is
     * defined on this list.
     */    
    public List getObjectSets() {
        List result = new LinkedList();
        for (Iterator i = objSets.entrySet().iterator(); i.hasNext();) {
            java.util.Map.Entry entry = (java.util.Map.Entry) i.next();
            ObjSetIndex objIndex = (ObjSetIndex) entry.getValue();
            result.add(objIndex.getObjectSet());
        }
        return result;
    }
    
    /**
     * Returns a <CODE>List</CODE> of all value phrase matches in the ontology.
     * @return A list of all value phrase matches in the ontology.
     */    
    public List getValueMatches() {
        List result = new LinkedList();
        for (Iterator i = matchExpressions.keySet().iterator(); i.hasNext();) {
            MatchedTextType match = (MatchedTextType) i.next();
            if (isValueMatch(match)) result.add(match);
        }
        return result;
    }
    
    /**
     * Returns a list of all keyword phrase matches in the ontology.
     * @return List of all keyword phrase matches in the ontology.
     */    
    public List getKeywordMatches() {
        List result = new LinkedList();
        for (Iterator i = matchExpressions.keySet().iterator(); i.hasNext();) {
            MatchedTextType match = (MatchedTextType) i.next();
            if (isKeywordMatch(match)) result.add(match);
        }
        return result;
    }
    
    /**
     * Returns a list of all value phrase matches belonging to the specified object
     * set.
     * @param objSet The object set owning the desired value phrases.
     * @return List of all value phrase matches for the object set.
     */    
    public List getValueMatchesForObjSet(ObjectSetType objSet) {
        ObjSetIndex objSetIdx = (ObjSetIndex) objSets.get(objSet.getId());
        Set valPhrases = objSetIdx.getAllValuePhrases();
        List result = new ArrayList();
        for (Iterator i = valPhrases.iterator(); i.hasNext();) {
            ValuePhraseType vp = (ValuePhraseType) i.next();
            DataFrameExpression exp = vp.getValueExpression();
            if (exp == null) continue;
            result.addAll(exp.getMatchedText());
        }
        return result;
    }
    
    /**
     * Returns a list of all keyword phrase matches belonging to the specified object
     * set.
     * @param objSet The object set owning the desired keyword phrases.
     * @return List of all keyword phrase matches for the object set.
     */    
    public List getKeywordMatchesForObjSet(ObjectSetType objSet) {
        ObjSetIndex objSetIdx = (ObjSetIndex) objSets.get(objSet.getId());
        Set keyPhrases = objSetIdx.getAllKeywordPhrases();
        List result = new ArrayList();
        for (Iterator i = keyPhrases.iterator(); i.hasNext();) {
            KeywordPhraseType kp = (KeywordPhraseType) i.next();
            DataFrameExpression exp = kp.getKeywordExpression();
            if (exp == null) continue;
            result.addAll(exp.getMatchedText());
        }
        Set valPhrases = objSetIdx.getAllValuePhrases();
        for (Iterator i = valPhrases.iterator(); i.hasNext();) {
            ValuePhraseType vp = (ValuePhraseType) i.next();
            List kpl = vp.getKeywordPhrase();
            if (kpl == null) continue;
            for (Iterator j = kpl.iterator(); j.hasNext();) {
                KeywordPhraseType kp = (KeywordPhraseType) j.next();
                DataFrameExpression exp = kp.getKeywordExpression();
                if (exp == null) continue;
                result.addAll(exp.getMatchedText());
            }
        }
        return result;
    }
    
    /**
     * Returns a reference to the ontology's <CODE>SourceDocument</CODE> element corresponding to
     * the given URI.  This is a factory method of sorts; it ensures that both the
     * <CODE>SourceDocument</CODE> and its containing <CODE>DataInstance</CODE> exist
     * by creating them if it needs to before returning the result.
     * @param uri URI for the desired <CODE>SourceDocument</CODE>.
     * @return The desired <CODE>SourceDocument</CODE>.
     */    
    public OSMXSourceDocumentType getSourceDoc(String uri) {
        if (uri == null) return null;
        try {
            if (activeDataInstance == null) {
                activeDataInstance = getOntology().getObjectFactory().createDataInstance();
                getOntology().getModelRoot().getAllModelElements().add(activeDataInstance);
                dataInstances.add(activeDataInstance);
            }
            OSMXSourceDocumentType result = (OSMXSourceDocumentType) sourceDocuments.get(uri);
            if (result == null) {
                result = (OSMXSourceDocumentType) getOntology().getObjectFactory().createSourceDocument();
                result.setUri(uri);
                sourceDocuments.put(uri, result);
                activeDataInstance.getDataInstanceElements().add(result);
            }
            return result;
        } catch (JAXBException jaxbe) {
            DataExtractionEngine.getLogger().severe("JAXB exception: " + jaxbe.getMessage());
        }
        return null;
    }
    
    private class ObjSetIndex {
        Set connectedRelSets = new HashSet();
        Set singletons = new HashSet();
        Set functionalGroups = new HashSet();
        Set nAryRelSets = new HashSet();
        Map objects = new HashMap();
        Set geners = new HashSet();
        Set specs = new HashSet();
        Set valPhrases = new HashSet();
        Set keyPhrases = new HashSet();
        ObjectSetType objSet;
        
        public ObjSetIndex(ObjectSetType objectSet) {
            objSet = objectSet;
            // TODO
        }
        
        public ObjectSetType getObjectSet() {
            return objSet;
        }
        
        public Set getRelSets() {
            return connectedRelSets;
        }
        
        public Set getGeneralizations() {
            return geners;
        }
        
        public Set getSpecializations() {
            return specs;
        }
        
        public Set getAllValuePhrases() {
            Map vps = new HashMap();
            Set result = new HashSet();
            for (Iterator i = valPhrases.iterator(); i.hasNext();) {
                ValuePhraseType vp = (ValuePhraseType) i.next();
                String vpID = vp.getHint();
                if (vpID == null || vpID.length() == 0) {
                    result.add(vp);
                    continue;
                }
                if (vpID == null) continue;
                vps.put(vpID, vp);
            }
            Set inheritedValPhrases = new HashSet();
            for (Iterator i = geners.iterator(); i.hasNext();) {
                String generID = (String) i.next();
                ObjSetIndex genIdx = (ObjSetIndex) objSets.get(generID);
                inheritedValPhrases.addAll(genIdx.getAllValuePhrases());
            }
            for (Iterator i = inheritedValPhrases.iterator(); i.hasNext();) {
                ValuePhraseType vp = (ValuePhraseType) i.next();
                String vpID = vp.getHint();
                if (vpID == null || vpID.length() == 0) {
                    continue;
                }
                if (vpID == null || vps.containsKey(vpID)) continue;
                vps.put(vpID, vp);
            }
            for (Iterator i = vps.entrySet().iterator(); i.hasNext();) {
                result.add(((Map.Entry) i.next()).getValue());
            }
            return result;
        }
        
        public Set getAllKeywordPhrases() {
            Map kps = new HashMap();
            Set result = new HashSet();
            for (Iterator i = keyPhrases.iterator(); i.hasNext();) {
                KeywordPhraseType kp = (KeywordPhraseType) i.next();
                String kpID = kp.getHint();
                if (kpID == null || kpID.length() == 0) {
                    result.add(kp);
                    continue;
                }
                if (kpID == null) continue;
                kps.put(kpID, kp);
            }
            Set inheritedKeyPhrases = new HashSet();
            for (Iterator i = geners.iterator(); i.hasNext();) {
                String generID = (String) i.next();
                ObjSetIndex genIdx = (ObjSetIndex) objSets.get(generID);
                inheritedKeyPhrases.addAll(genIdx.getAllKeywordPhrases());
            }
            for (Iterator i = inheritedKeyPhrases.iterator(); i.hasNext();) {
                KeywordPhraseType kp = (KeywordPhraseType) i.next();
                String kpID = kp.getHint();
                if (kpID == null || kpID.length() == 0) {
                    continue;
                }
                if (kpID == null || kps.containsKey(kpID)) continue;
                kps.put(kpID, kp);
            }
            for (Iterator i = kps.entrySet().iterator(); i.hasNext();) {
                result.add(((Map.Entry) i.next()).getValue());
            }
            return result;
        }
    }
    
}
