/*
 * HeuristicBasedMapper.java
 *
 * Created on February 24, 2004, 12:45 AM
 */

package edu.byu.deg.ontos;

import edu.byu.deg.framework.Document;
import edu.byu.deg.framework.Ontology;
import edu.byu.deg.framework.AbstractValueMapper;
import edu.byu.deg.framework.ValueMapperConfigurationException;
import edu.byu.deg.osmx.*;
import edu.byu.deg.osmx.OSMXDocument;
import edu.byu.deg.osmx.binding.*;
import edu.byu.deg.ontos.heuristic.*;
import java.util.*;

/**
 * An implementation of <CODE>ValueMapper</CODE> that uses a heuristics-guided
 * Bernstein synthesis algorithm to map extracted values to objects and
 * relationships.
 * @author Alan Wessman
 */
public class HeuristicBasedMapper extends AbstractValueMapper {
    
    static ExtractionOntologyIndex ontIndex = null;
    
    /** Creates a new instance of HeuristicBasedMapper */
    public HeuristicBasedMapper() { /* default constructor */ }
    
    /**
     * Creates a new instance of HeuristicBasedMapper and initializes it with an
     * indexed ontology.
     * @param ontologyIndex The ontology index to use for the mapping process.
     */    
    public HeuristicBasedMapper(ExtractionOntologyIndex ontologyIndex) {
        ontIndex = ontologyIndex;
    }
    
    
    //ADDED BY MUHAMMED///////////////////////////////////////
    /**
     *returns an instance of ExtractionOntologyIndex class
     *Added by Muhammed
     */
    
    	public static ExtractionOntologyIndex getOntologyIndex(){
        return ontIndex;
    }
    //////////////////////////////////////////////////////////
    
	
	/**
     * Performs the value mapping process using the current ontology, focusing only on
     * matches belonging to the specified String of text.
     *
     * Note: invokes {@link #generateValueMappings(Ontology, String)}.
     * @param doc The <CODE>Document</CODE> from which the matches to be mapped originated.
     * @throws ValueMapperConfigurationException Thrown when no ontology has been set for this mapper.
     */    
   public void generateValueMappings() throws ValueMapperConfigurationException {
        if (ontology == null)
            throw new ValueMapperConfigurationException("Ontology must be specified");
        generateValueMappings(ontology, "");
    }
    	
    	
    
    /**
     * Performs the value mapping process by applying the mapping heuristics.
     * @param ont The ontology to use in the mapping process.
     * @param docURI URI of the document whose matches should be mapped.
     */    
    public void generateValueMappings(Ontology ont, String docURI) {
        // create the ontology index if we haven't already
        if (ontIndex == null) {
            ontIndex = new ExtractionOntologyIndex((OSMXDocument) ont);
        }
        // index the ontology
        ontIndex.setOntology((OSMXDocument) ont);
        
        // validate ontology
        // error if cycles exist
        
        // perform subsumed/overlapping value filter heuristic
        removeSubsumedOverlappingMatches(ont, docURI);
        
        // perform contextual heuristic on ontology (recurses)
        ContextualHeuristic ch = new ContextualHeuristic(ontIndex);
        OSMXObjectSetType primary = (OSMXObjectSetType) ontIndex.getPrimaryObjSet();
        MatchingContext docContext = new MatchingContext(ontIndex);
        docContext.filterOnDocURIPrefix(docURI, true);
        // discard the result of the call below
        ch.processContextualGroup(primary, docContext, new HashSet());
    }
    
    /**
     * Applies the subsumed-match and overlapping-match heuristics to matches in the
     * ontology that belong to a particular document.
     * @param ont The ontology that contains the value matches processed by the heuristic.
     * @param docURI URI of the document from which the matches must originate in order to be
     * processed by the algorithm.
     */    
    protected void removeSubsumedOverlappingMatches(Ontology ont, String docURI) {
        List objSets = ((OSMXDocument) ont).getObjectSets();
        for (Iterator i = objSets.iterator(); i.hasNext();) {
            ObjectSetType objSet = (ObjectSetType) i.next();
            DataFrameType datFrame = objSet.getDataFrame();
            if (datFrame == null) continue;
            removeSubsumedMatches(datFrame.getValuePhrase(), docURI);
            resolveOverlappingMatches(docURI);
        }
    }
    
    /**
     * Resolves conflicts caused by subsumed matches in the document.
     * @param vpl A list of value phrases containing the matches that need to be processed in
     * order to remove subsumed matches.
     * @param docURI URI of the document from which the matches to process originated.
     */    
    protected void removeSubsumedMatches(List vpl, String docURI) {
        if (vpl == null) return;
        Comparator matchComparer = new Comparator() {
            public int compare(Object o1, Object o2) {
                if (o1.equals(o2)) return 0;
                MatchedTextType m1 = (MatchedTextType) o1;
                MatchedTextType m2 = (MatchedTextType) o2;
                if (m1.getStartPos() < m2.getStartPos()) return -1;
                if (m1.getStartPos() > m2.getStartPos()) return 1;
                if (m1.getEndPos() < m2.getEndPos()) return -1;
                if (m1.getEndPos() > m2.getEndPos()) return 1;
                return 0;
            }
        };
        // must separate matches according to doc URI before subsuming
        Map docMatchMap = new HashMap();
        
        for (Iterator i = vpl.iterator(); i.hasNext();) {
            ValuePhraseType vp = (ValuePhraseType) i.next();
            if (vp.getValueExpression() == null) continue;
            for (ListIterator j = vp.getValueExpression().getMatchedText().listIterator(); j.hasNext();) {
                MatchedTextType match = (MatchedTextType) j.next();
                if (!match.getDocument().startsWith(docURI)) continue;
                List docMatches = (List) docMatchMap.get(match.getDocument());
                if (docMatches == null) {
                    docMatches = new ArrayList();
                    docMatchMap.put(match.getDocument(), docMatches);
                    docMatches.add(match);
                    continue;
                }
                // see javadoc for Collections.binarySearch for explanation of return value
                int index = Collections.binarySearch(docMatches, match, matchComparer);
                if (index >= 0) continue;
                index = -index - 2;
                // remove match from vp if this match would be subsumed by pred or succ
                MatchedTextType pred = null, succ = null;
                if (index >= 0) pred = (MatchedTextType) docMatches.get(index);
                if (index + 1 < docMatches.size()) succ = (MatchedTextType) docMatches.get(index+1);
                boolean isSubsumed = false;
                if (pred != null && pred.getStartPos() <= match.getStartPos() && pred.getEndPos() >= match.getEndPos()) isSubsumed = true;
                if (succ != null && succ.getStartPos() == match.getStartPos() && succ.getEndPos() >= match.getEndPos()) isSubsumed = true;
                if (isSubsumed) {
                    j.remove();
                    continue;
                }
                // otherwise remove predecessors or successors that this match subsumes
                int insertIndex = index;
                List toRemove = new ArrayList();
                while (index >= 0) {
                    pred = (MatchedTextType) docMatches.get(index--);
                    if (pred.getStartPos() == match.getStartPos()) {
                        toRemove.add(pred);
                        System.out.println("\t\tRemoving match: " + pred.getValue()+ ", from: " +vp.getHint().toString()+ "\t\tHeuristBMapper.removeSubsumedMatches()");
                    } else break;
                }
                index = insertIndex + 1;
                while (index < docMatches.size()) {
                    succ = (MatchedTextType) docMatches.get(index++);
                    if (succ.getEndPos() <= match.getEndPos()) {
                        toRemove.add(succ);
                        System.out.println("\t\tRemoving match: " + succ.getValue()+ ", from: " +vp.getHint().toString()+ "\t\tHeuristBMapper.removeSubsumedMatches()");
                    } else break;
                }
                if (insertIndex < 0) insertIndex = 0;
                docMatches.add(insertIndex, match);
                docMatches.removeAll(toRemove);
            }
        }
    }
    
    private void resolveOverlappingMatches(String docURI) {
        // Let the LocalOverlapHeuristic class do its job
        LocalOverlapHeuristic loh = new LocalOverlapHeuristic(ontIndex);
        loh.eliminateOverlaps(docURI);
    }
}
