/*
 * LocalDocumentRetriever.java
 *
 * Created on February 4, 2004, 9:42 PM
 */

package edu.byu.deg.ontos;

import java.io.File;
import java.io.FileFilter;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.List;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.HashMap;
import edu.byu.deg.framework.RetrieverConfigurationException;
import edu.byu.deg.framework.DocumentRetriever;
import edu.byu.deg.framework.Document;
import edu.byu.deg.framework.document.DOMDocument;

/**
 * Implements the <CODE>DocumentRetriever</CODE> interface by obtaining HTML
 * documents from specified locations on the local filesystem.
 * @author Alan Wessman
 */
public class LocalDocumentRetriever implements DocumentRetriever {
    
    protected Map docs;
    protected FileFilter fileFilter;
    
    /** Creates a new instance of LocalDocumentRetriever */
    public LocalDocumentRetriever() {
        docs = new HashMap();
        fileFilter = null;
    }
    
    /**
     * Adds a retrieval location to a list of such locations.
     * @param uri URI for the retrieval location.  This is passed to the
     * {@link java.io.File#File(URI)} constructor to obtain a local filesystem reference.
     * If the resulting <CODE>File</CODE> is a directory, all files in the directory
     * will be retrieved (provided they pass the <CODE>FileFilter</CODE> if one is
     * present).  If the <CODE>File</CODE> resolves to an individual file, the file
     * will be retrieved regardless of any filter setting.
     */    
    public void addDocumentLocation(String uri) {
        URI u = null;
        try {
            u = new URI(uri);
        } catch (URISyntaxException urise) {
            OntosEngine.getLogger().log(java.util.logging.Level.WARNING, "Bad URI for " + uri + "; " + urise.getMessage());
            return;
        }
        File f = new File(u);
        if (!f.exists()) return;
        if (f.isDirectory()) {
            File[] fileList;
            if (fileFilter != null) fileList = f.listFiles(fileFilter);
            else fileList = f.listFiles();
            
            for (int i = 0; i < fileList.length; i++) {
                docs.put(fileList[i].toURI(), fileList[i]);
            }
        }
        else {
            docs.put(uri, f);
        }
    }
    
    /**
     * Clears the list of locations from which to retrieve documents.
     */    
    public void clearDocumentLocations() {
        docs.clear();
    }
    
    /**
     * Returns the set of document locations via an {@link java.util.Iterator
     * Iterator}.
     * @return An iterator that yields all of the recorded document locations.
     */    
    public java.util.Iterator getDocumentLocations() {
        return docs.keySet().iterator();
    }
    
    /**
     * Retrieves and returns HTML documents from the specified locations.
     * @throws RetrieverConfigurationException Thrown if no locations to retrieve from have been specified prior to this call.
     * @return An iterator yielding {@link edu.byu.deg.framework.document.DOMDocument}s.
     */    
    public java.util.Iterator retrieveDocuments() throws RetrieverConfigurationException {
        if (docs.size() == 0) 
            throw new RetrieverConfigurationException("No retrieval locations specified.");
        List docLocs = new LinkedList(docs.keySet());
        List toRetrieve = new LinkedList();
        for (Iterator i = docLocs.iterator(); i.hasNext();) {
            // TODO: generalize beyond DOMDocument to handle other document types
            File docFile = (File) docs.get(i.next());
            URI docURI = docFile.toURI();
            Document doc = new DOMDocument(docURI);
            toRetrieve.add(doc);
        }
        return toRetrieve.iterator();
    }
    
    /**
     * Empty (no-op) implementation of an interface method.  This method is superseded
     * for this class by {@link #addDocumentLocation}.
     * @param uri Not used.
     */    
    public void setRetrievalLocation(String uri) { /* do nothing */ }
    
    /**
     * Performs any necessary initialization on the retriever before it performs the
     * retrieval process.
     */    
    public void initialize() { /* no initialization needed */ }
    
    /**
     * Registers a file filter with this class so that files of the wrong format are
     * not incorrectly retrieved from the document locations.
     * @param filter The filter for distinguishing extractable files from non-extractable files.
     */    
    public void setFileFilter(FileFilter filter) {
        fileFilter = filter;
    }
    
    /**
     * Removes a location from the retrieval locations list.
     * @param uri The location to remove.
     */    
    public void removeDocumentLocation(String uri) {
        docs.remove(uri);
    }
    
}
