/*
 * MappingHeuristic.java
 *
 * Created on March 31, 2004, 9:57 AM
 */

package edu.byu.deg.ontos;

import edu.byu.deg.framework.*;
import edu.byu.deg.osmx.*;
import edu.byu.deg.osmx.binding.*;
import javax.xml.bind.JAXBException;

/**
 * Superclass for the implementations of the various heuristics that govern the
 * operation of the {@link HeuristicBasedMapper}.  This class primarily provides a
 * number of convenience methods commonly used by the subclasses to achieve their
 * goals.
 * @author Alan Wessman
 */
public abstract class MappingHeuristic {
    
    protected ExtractionOntologyIndex ontIndex = null;
    protected RelSetConnectionComparator connComparer = null;
    
    public static final String ACCEPTED_MATCH = OSMXDataFrameExpression.ACCEPTED_MATCH;
    public static final String PROPOSED_MATCH = OSMXDataFrameExpression.PROPOSED_MATCH;
    public static final String REJECTED_MATCH = OSMXDataFrameExpression.REJECTED_MATCH;
    
    /**
     * Creates a new instance of the MappingHeuristic implementation.
     */
    public MappingHeuristic() { /* default constructor */ }
    
    /**
     * Creates a new instance of the MappingHeuristic subclass initializes it
     * with an index for the ontology used in the mapping process
     * @param idx An indexed view of the ontology used for the mapping process.
     */    
    protected MappingHeuristic(ExtractionOntologyIndex idx) {
        setOntologyIndex(idx);
    }
    
    /**
     * Stores a reference to the indexed ontology.
     * @param idx The ontology index to use during the mapping process.
     */    
    public void setOntologyIndex(ExtractionOntologyIndex idx) {
        ontIndex = idx;
        connComparer = new RelSetConnectionComparator(idx);
    }
    
    /**
     * Returns the index for the given ontology.  If no index has been stored yet, the
     * method creates one before returning it.  If the existing index is based on a
     * different ontology, the method re-initializes the index with the given ontology
     * before returning it.
     * @param ont The ontology whose index is required.
     * @return An index for the given ontology.
     */    
    public ExtractionOntologyIndex getOntologyIndex(OSMXDocument ont) {
        if (ontIndex == null) ontIndex = new ExtractionOntologyIndex();
        if (ontIndex.getOntology() != ont) ontIndex.setOntology(ont);
        return ontIndex;
    }
    
    /**
     * Returns the index for the given ontology.  If no index has been stored yet, the
     * method creates one before returning it.  If the existing index is based on a
     * different ontology, the method re-initializes the index with the given ontology
     * before returning it.
     * @param ont The ontology whose index is required.
     * @return An index of the given ontology.
     */    
    public ExtractionOntologyIndex getOntologyIndex(Ontology ont) {
        return getOntologyIndex((OSMXDocument) ont);
    }
    
    /**
     * Adds the given object to the identified <CODE>SourceDocument</CODE> element in the
     * ontology.  The method will log a warning but perform no action if the given URI 
     * does not identify an existing <CODE>SourceDocument</CODE> element.
     * @param obj The object to add to the ontology.
     * @param uri The URI identifying a <CODE>SourceDocument</CODE> element in the ontology.
     */    
    public void addObject(OSMXObjectType obj, String uri) {
        OSMXSourceDocumentType sourceDoc = ontIndex.getSourceDoc(uri);
        if (sourceDoc == null) {
            DataExtractionEngine.getLogger().warning("Cannot add object to unspecified document");
            return;
        }
        //System.out.println("\t\tAdding Object to Mapping Heuristic: ID=" +obj.getId()+ ", values: " + obj.getValue() + "\t\tMappingHeuristic.addObject()");
        sourceDoc.getDataInstanceElements().add(obj);
    }
    
    /**
     * Creates an <CODE>ObjectBinding</CODE> element, which connects an <CODE>Object</CODE>
     * to a <CODE>Relationship</CODE>.  The method populates the new binding with
     * references to the object and the <CODE>RelSetConnection</CODE> to which the
     * binding corresponds.  The relationship reference is not added in this method.
     * @param obj The object involved in the binding.
     * @param conn The relationship set connection that defines how the object is to be bound to a
     * relationship.
     * @return The generated object binding.
     */    
    public ObjectBindingType bindObject(ObjectType obj, RelSetConnectionType conn) {
        try {
            ObjectBindingType result = ontIndex.getOntology().getObjectFactory().createObjectBinding();
            result.setObjectRef(obj.getId());
            result.setConnectionRef(conn.getId());
            return result;
        } catch (JAXBException jaxbe) {
            DataExtractionEngine.getLogger().severe("JAXB Exception: " + jaxbe.getMessage());
        }
        return null;
    }
    
    /**
     * Adds the specified <CODE>Relationship</CODE> to the identified
     * <CODE>SourceDocument</CODE> in the ontology.
     * @param rel The relationship to add to the ontology.
     * @param uri URI identifying the source document from which the relationship was derived.
     */    
    public void addRelationship(OSMXRelationshipType rel, String uri) {
        OSMXSourceDocumentType sourceDoc = ontIndex.getSourceDoc(uri);
        if (sourceDoc == null) {
            DataExtractionEngine.getLogger().warning("Cannot add object to unspecified document");
            return;
        }
        sourceDoc.getDataInstanceElements().add(rel);
    }
    
}
