/*
 * MatchingContext.java
 *
 * Created on April 10, 2004, 1:49 PM
 */

package edu.byu.deg.ontos;

import edu.byu.deg.framework.*;
import edu.byu.deg.osmx.*;
import edu.byu.deg.osmx.binding.*;
import edu.byu.deg.osmx.binding.MatchedTextType;
import java.util.*;

/**
 * Represents a collection of value and keyword matches.  The collection may be
 * limited to a particular document, range of character positions within a
 * document, or a particular object set associated with the matches.  These
 * operations provide a way to focus extraction on a single portion or aspect (a
 * "context") of a document without irreversibly eliminating the rest of the
 * candidate matches and without requiring the document to be divided into
 * subdocuments unnecessarily.
 * @author Alan Wessman
 */
public class MatchingContext {
    
    private MatchedTextComparator matchComparer;
    private List valueList;
    private List keywordList;
    private ExtractionOntologyIndex idx;
    private String targetURI;
    
    /**
     * Creates a new instance of MatchingContext based on the information provided by
     * the specified ontology index.  The resulting instance contains all the keyword
     * and value matches present in the indexed ontology, except for those matches that
     * are marked as either "accepted" or "rejected".
     * @param index The <CODE>ExtractionOntologyIndex</CODE> that provides the set of keyword and
     * value matches to this class.
     */
    public MatchingContext(ExtractionOntologyIndex index) {
        idx = index;
        matchComparer = new MatchedTextComparator();
        valueList = new LinkedList(idx.getValueMatches());
        keywordList = new LinkedList(idx.getKeywordMatches());
        targetURI = index.getOntology().getURI().toString();
        removeDecidedMatches();
        Collections.sort(valueList, matchComparer);
        Collections.sort(keywordList, matchComparer);
    }
    
    /**
     * Copy constructor for creating a new <CODE>MatchingContext</CODE> with the same
     * set of keyword and value matches as the one passed to the constructor.
     * @param mc The <CODE>MatchingContext</CODE> we wish to copy.
     */    
    public MatchingContext(MatchingContext mc) {
        idx = mc.idx;
        matchComparer = mc.matchComparer;
        valueList = new LinkedList();
        keywordList = new LinkedList();
        valueList.addAll(mc.valueList);
        keywordList.addAll(mc.keywordList);
        removeDecidedMatches();
        Collections.sort(valueList, matchComparer);
        Collections.sort(keywordList, matchComparer);
    }
    
    /**
     * Marks the given match as accepted and marks all other matches in the context
     * that overlap with the given match as rejected.  Also removes the given match
     * from the matching context.
     * @param match The match to mark as accepted.
     */    
    public void acceptMatch(MatchedTextType match) {
        if (valueList.remove(match)) {
            match.setStatus(OSMXDataFrameExpression.ACCEPTED_MATCH);
        }
        for (Iterator i = valueList.iterator(); i.hasNext();) {
            MatchedTextType otherMatch = (MatchedTextType) i.next();
            if (otherMatch == match) continue;
            if (!match.getDocument().equals(otherMatch.getDocument())) continue;
            if (otherMatch.getEndPos() < match.getStartPos() || otherMatch.getStartPos() > match.getEndPos())
                continue;
            DataExtractionEngine.getLogger().info("Match " + matchToString(otherMatch) + " rejected due to overlap with accepted match " + matchToString(match));
            otherMatch.setStatus(OSMXDataFrameExpression.REJECTED_MATCH);
        }
    }
    
    /**
     * Marks the given match as rejected and removes it from the matching context.
     * @param match The match to reject.
     */    
    public void rejectMatch(MatchedTextType match) {
        if (valueList.remove(match)) match.setStatus(OSMXDataFrameExpression.REJECTED_MATCH);
    }
    
    /**
     * Removes the specified match from the <CODE>MatchingContext</CODE>.
     * @param match The match to remove from the context.
     */    
    public void removeMatch(MatchedTextType match) {
        if (!valueList.remove(match)) 
            keywordList.remove(match);
    }
    
    /**
     * Adds a value match to the matching context.
     * @param match The match to add to the context.
     */    
    public void addValueMatch(MatchedTextType match) {
        valueList.add(match);
        Collections.sort(valueList, matchComparer);
    }
    
    /**
     * Adds a keyword match to the context.
     * @param match The match to add to the context.
     */    
    public void addKeywordMatch(MatchedTextType match) {
        keywordList.add(match);
        Collections.sort(keywordList, matchComparer);
    }
    
    /**
     * Invokes {@link #acceptMatch} on every value match remaining in the context that
     * belongs to the specified object set.
     * @param objSet The object set whose matches are to be accepted.
     */    
    public void acceptRemaining(ObjectSetType objSet) {
        for (Iterator i = valueList.iterator(); i.hasNext();) {
            MatchedTextType match = (MatchedTextType) i.next();
            if (objSet == idx.getObjectSet(match)) {
                match.setStatus(OSMXDataFrameExpression.ACCEPTED_MATCH);
                i.remove();
            }
        }
    }
    
    /**
     * Invokes {@link #rejectMatch} for all value matches remaining in the context that
     * belong to the specified object set.
     * @param objSet The object set whose matches are to be rejected.
     */    
    public void rejectRemaining(ObjectSetType objSet) {
        for (Iterator i = valueList.iterator(); i.hasNext();) {
            MatchedTextType match = (MatchedTextType) i.next();
            if (objSet == idx.getObjectSet(match)) {
                match.setStatus(OSMXDataFrameExpression.REJECTED_MATCH);
                i.remove();
            }
        }
    }
    
    /**
     * Returns the list of keyword matches in the context, sorted according to the
     * order defined by the {@link edu.byu.deg.osmx.MatchedTextComparator} class.
     * @return A sorted list of <CODE>MatchedText</CODE> elements that belong to keyword
     * phrases and appear in this context.
     */    
    public List getKeywordMatches() {
        return keywordList;
    }
    
    /**
     * Returns the list of value matches in the context, sorted according to the
     * order defined by the {@link edu.byu.deg.osmx.MatchedTextComparator} class.
     * @return A sorted list of <CODE>MatchedText</CODE> elements that belong to value
     * phrases and appear in this context.
     */    
    public List getValueMatches() {
        return valueList;
    }
    
    /**
     * Filters out matches from the context based on the URI of the document from which
     * the matches originated.
     * @param docURI The URI on which to filter out matches.
     * @param include If <CODE>true</CODE>, matches that originated from the specified URI will be
     * retained and all others will be removed from the context.  If
     * <CODE>false</CODE>, those belonging to the document will be removed and the rest
     * will be retained.
     */    
    public void filterOnDocURI(String docURI, boolean include) {
        for (Iterator i = valueList.iterator(); i.hasNext();) {
            MatchedTextType match = (MatchedTextType) i.next();
            boolean fitsFilter = (docURI != null && docURI.equals(match.getDocument()));
            boolean toRemove = (fitsFilter && !include) || (!fitsFilter && include);
            if (toRemove) i.remove();
        }
        for (Iterator i = keywordList.iterator(); i.hasNext();) {
            MatchedTextType match = (MatchedTextType) i.next();
            boolean fitsFilter = (docURI != null && docURI.equals(match.getDocument()));
            boolean toRemove = (fitsFilter && !include) || (!fitsFilter && include);
            if (toRemove) i.remove();
        }
        if (include) targetURI = docURI;
    }
    
    /**
     * Filters out matches from the context based on the URI of the document from which
     * the matches originated.
     * @param docURI The URI (or a prefix of the URI) on which to filter out matches.
     * @param include If <CODE>true</CODE>, matches that originated from documents whose URI starts
     * with the specified value will be retained and all others will be removed from the context.  If
     * <CODE>false</CODE>, those belonging to the qualifying documents will be removed and the rest
     * will be retained.
     */    
    public void filterOnDocURIPrefix(String docURI, boolean include) {
        for (Iterator i = valueList.iterator(); i.hasNext();) {
            MatchedTextType match = (MatchedTextType) i.next();
            String matchDoc = match.getDocument();
            boolean fitsFilter = (docURI != null && matchDoc != null && matchDoc.startsWith(docURI));
            boolean toRemove = (fitsFilter && !include) || (!fitsFilter && include);
            if (toRemove) i.remove();
        }
        for (Iterator i = keywordList.iterator(); i.hasNext();) {
            MatchedTextType match = (MatchedTextType) i.next();
            String matchDoc = match.getDocument();
            boolean fitsFilter = (docURI != null && matchDoc != null && matchDoc.startsWith(docURI));
            boolean toRemove = (fitsFilter && !include) || (!fitsFilter && include);
            if (toRemove) i.remove();
        }
        if (include) targetURI = docURI;
    }
    
    /**
     * Filters out matches from the context based on a range of character offsets.
     * @param startPos The minimal starting position of qualifying matches.
     * @param endPos The maximal end position of qualifying matches.
     * @param docURI URI for the document to filter on.  All matches except those belonging to the
     * specified document will be removed from the context before the range filtering
     * occurs.  This is necessary because limiting the range
     * of matches is only meaningful in terms of a single document.
     * @param include If <CODE>true</CODE>, matches entirely contained within the specified range will be
     * retained and all others will be removed from the context.  If
     * <CODE>false</CODE>, those entirely within the range will be removed and the rest
     * will be retained.
     */    
    public void filterOnRange(int startPos, int endPos, String docURI, boolean include) {
        if (docURI == null) return;
        filterOnDocURI(docURI, true);
        filterOnRange(startPos, endPos, include);
    }
    
    /**
     * Filters out matches from the context based on a range of character offsets.
     * Programs should generally call this method only when the context has already
     * filtered out matches for all but one document, in order to avoid applying the
     * same range to matches from different documents.
     * @param startPos The minimal starting position of qualifying matches.
     * @param endPos The maximal end position of qualifying matches.
     * @param include If <CODE>true</CODE>, matches entirely contained within the specified range will be
     * retained and all others will be removed from the context.  If
     * <CODE>false</CODE>, those entirely within the range will be removed and the rest
     * will be retained.
     */    
    public void filterOnRange(int startPos, int endPos, boolean include) {
        for (Iterator i = valueList.iterator(); i.hasNext();) {
            MatchedTextType match = (MatchedTextType) i.next();
            boolean fitsFilter = (match.getStartPos() >= startPos && match.getEndPos() <= endPos);
            boolean toRemove = (fitsFilter && !include) || (!fitsFilter && include);
            if (toRemove) i.remove();
        }
        for (Iterator i = keywordList.iterator(); i.hasNext();) {
            MatchedTextType match = (MatchedTextType) i.next();
            boolean fitsFilter = (match.getStartPos() >= startPos && match.getEndPos() <= endPos);
            boolean toRemove = (fitsFilter && !include) || (!fitsFilter && include);
            if (toRemove) i.remove();
        }
    }
    
    /**
     * Filters out matches from the context based on a range of character offsets.
     * Programs should generally call this method only when the context has already
     * filtered out matches for all but one document, in order to avoid applying the
     * same range to matches from different documents.
     * @param firstStartPos The minimal starting position of qualifying matches.
     * @param lastStartPos The maximal starting position of qualifying matches.
     * @param include If <CODE>true</CODE>, matches whose starting position is contained within the specified range will be
     * retained and all others will be removed from the context.  If
     * <CODE>false</CODE>, those with starting position entirely within the range will be removed and the rest
     * will be retained.
     */    
    public void filterOnStartPosRange(int firstStartPos, int lastStartPos, boolean include) {
        for (Iterator i = valueList.iterator(); i.hasNext();) {
            MatchedTextType match = (MatchedTextType) i.next();
            boolean fitsFilter = (match.getStartPos() >= firstStartPos && match.getStartPos() <= lastStartPos);
            boolean toRemove = (fitsFilter && !include) || (!fitsFilter && include);
            if (toRemove) i.remove();
        }
        for (Iterator i = keywordList.iterator(); i.hasNext();) {
            MatchedTextType match = (MatchedTextType) i.next();
            boolean fitsFilter = (match.getStartPos() >= firstStartPos && match.getStartPos() <= lastStartPos);
            boolean toRemove = (fitsFilter && !include) || (!fitsFilter && include);
            if (toRemove) i.remove();
        }
    }
    
    /**
     * Filters out matches from the context based on the object set to which
     * the matches belong.
     * @param objSet The object set that determines which matches will be filtered out.
     * @param include If <CODE>true</CODE>, matches that belong to the specified object set will be
     * retained and all others will be removed; otherwise, those that belong to the
     * object set will be removed and the rest will be retained.
     */    
    public void filterOnObjectSet(ObjectSetType objSet, boolean include) {
        if (objSet == null) {
            if (!include) return;
            valueList.clear();
            keywordList.clear();
        } else {
            if (include) {
                valueList.retainAll(idx.getValueMatchesForObjSet(objSet));
                keywordList.retainAll(idx.getKeywordMatchesForObjSet(objSet));
            } else {
                valueList.removeAll(idx.getValueMatchesForObjSet(objSet));
                keywordList.removeAll(idx.getKeywordMatchesForObjSet(objSet));
            }
        }
    }
    
    private void removeDecidedMatches() {
        for (Iterator i = valueList.iterator(); i.hasNext();) {
            MatchedTextType match = (MatchedTextType) i.next();
            if (!match.getStatus().equals(OSMXDataFrameExpression.PROPOSED_MATCH)) i.remove();
        }
    }
    
    /**
     * Provides a non-normative string representation of a <CODE>MatchedText</CODE>
     * object.  Useful for debugging or logging purposes but not intended for other
     * program logic.
     * @param match The match object to convert to a string representation.
     * @return The string representation of a match, including its value, starting and ending
     * offsets, and the object set to which it belongs.
     */    
    public String matchToString(MatchedTextType match) {
        return match.getValue() + " [" + match.getStartPos() + "-" + match.getEndPos() + "; objSet=" + idx.getObjectSet(match).getName() + "(" + idx.getObjectSet(match).getId() + ")]";
    }
    
    /**
     * Returns the URI that best applies to the matches that belong to this matching
     * context.  This information is useful to be able to associate objects created
     * from the matches with a URI indicating their origin.
     * @return If the context has been filtered based on a document URI or URI prefix, with the
     * <CODE>include</CODE> parameter set to <CODE>true</CODE>, that URI
     * is returned.  Otherwise, the URI of the ontology is returned.
     */    
    public String getTargetURI() {
        return targetURI;
    }
    
    /**
     * Assigns a URI to the matches in the context.  This allows objects created from
     * the matches to refer back to the URI as a way to trace their origin.
     * @param uri The URI to associate with the matches in the context.
     */    
    public void setTargetURI(String uri) {
        targetURI = uri;
    }
    
    /**
     * Provides a list of all value phrase matches within the context that belong to a
     * particular object set.
     * @param objSet The object set used to determine the list of value phrase matches.
     * @return A <CODE>List</CODE> containing <CODE>MatchedText</CODE> objects that belong to
     * the specified object set.
     */    
    public List getValuesForObjectSet(ObjectSetType objSet) {
        MatchingContext mc = new MatchingContext(this);
        mc.filterOnObjectSet(objSet, true);
        return new ArrayList(mc.getValueMatches());
    }
    
    /**
     * Provides a list of all keyword phrase matches within the context that belong to a
     * particular object set.
     * @param objSet The object set used to determine the list of keyword phrase matches.
     * @return A <CODE>List</CODE> containing <CODE>MatchedText</CODE> objects that belong to
     * the specified object set.
     */    
    public List getKeywordsForObjectSet(ObjectSetType objSet) {
        MatchingContext mc = new MatchingContext(this);
        mc.filterOnObjectSet(objSet, true);
        return new ArrayList(mc.getKeywordMatches());
    }
    
}
