/*
 * ObjectRelationshipWriter.java
 *
 * Created on April 23, 2004, 12:36 PM
 */

package edu.byu.deg.ontos;

import edu.byu.deg.framework.*;
import edu.byu.deg.osmx.*;
import edu.byu.deg.osmx.binding.*;
import java.util.*;

/**
 * Implementation of <CODE>OntologyWriter</CODE> that produces a human-friendly
 * view of the objects and relationships in the data instance(s) of the ontology,
 * with output formatted as HTML for ready viewing in a browser.  NOTE: It may be
 * useful in the future to produce output in an XML format rather than HTML format,
 * to facilitate machine use of the results.
 * @author Alan Wessman
 */
public class ObjectRelationshipWriter extends OntologyWriter {
    
    private OSMXDocument ontDoc;
    private ExtractionOntologyIndex idx;
    
    /** Creates a new instance of ObjectRelationshipWriter */
    public ObjectRelationshipWriter() { /* default constructor */ }
    
    /**
     * Creates a new instance of ObjectRelationshipWriter and initializes it with the
     * ontology to produce output for.
     * @param ont The ontology from which the objects and relationships are taken and formatted
     * for output.
     */    
    public ObjectRelationshipWriter(Ontology ont) {
        super(ont);
    }
    
    /**
     * Writes the objects and relationships in the stored ontology to the designated
     * writer.  Note: This will throw {@link IllegalStateException} if no ontology has
     * been stored for this instance of ObjectRelationshipWriter.
     * @param w The output channel to which the objects and relationships are to be written.
     * @throws java.io.IOException Thrown if an I/O problem occurs during the writing process.
     */    
    public void writeDataInstance(java.io.Writer w) throws java.io.IOException {
        // TODO: make this output XML rather than HTML
        w.write("<html><head></head><body>");
        boolean pastFirst = false;
        for (Iterator i = ontDoc.getElementsOfType(OSMXDataInstanceType.class).iterator(); i.hasNext();) {
            OSMXDataInstanceType di = (OSMXDataInstanceType) i.next();
            if (pastFirst) {
                w.write("<hr />");
            }
            w.write("<h2>Data Instance " + (di.getName() == null ? "Anonymous" : di.getName()) + " (" + di.getId() + ")" + "</h2>");
            // write out the data instance
            w.write("<p>");
            // start at the primary object set and write out each member object
            OSMXObjectSetType primaryObjSet = (OSMXObjectSetType) idx.getPrimaryObjSet();
            List primaryObjs = new ArrayList(primaryObjSet.getWorkingInstance());
            Collections.sort(primaryObjs, new Comparator() {
               public int compare(Object o1, Object o2) {
                   OSMXObjectType obj1 = (OSMXObjectType) o1;
                   OSMXObjectType obj2 = (OSMXObjectType) o2;
                   OSMXSourceDocumentType sd1 = (OSMXSourceDocumentType) obj1.getParent();
                   OSMXSourceDocumentType sd2 = (OSMXSourceDocumentType) obj2.getParent();
                   return java.text.Collator.getInstance().compare(sd1.getUri(), sd2.getUri());
               }
            });
            Set visitedRels = new HashSet();
            for (Iterator j = primaryObjs.iterator(); j.hasNext();) {
                OSMXObjectType obj = (OSMXObjectType) j.next();
                w.write("<hr />");
                writeObject(w, obj, visitedRels);
            }
            w.write("</p>");
        }
        w.write("</html>");
        w.flush();
    }
    
    private void writeObject(java.io.Writer w, OSMXObjectType obj, Set visitedRels) throws java.io.IOException {
        // Get ref to object set that the object belongs to
        OSMXObjectSetType objSet = (OSMXObjectSetType) idx.getOntology().getElement(obj.getObjectSet());
        
        // Write out name and value (lexical value or nonlexical ID)
        w.write("<b>" + objSet.getName() + "</b> ");
        w.write("<i>" + (objSet.isLexical() ? obj.getValue() + "; " + obj.getLocation() : obj.getId()) + "</i>");
        
        // Find unexplored connections so child objects may be written out
        
        // Create a list of connections to explore
        List conns = new ArrayList(objSet.getRelSetConnections());
        // Sort connections by ID of connected object set so output is consistent
        Collections.sort(conns, new Comparator() {
            public int compare(Object o1, Object o2) {
                if (o1 == o2) return 0;
                String id1 = ((OSMXElement) o1).getParent().getInternalID();
                String id2 = ((OSMXElement) o2).getParent().getInternalID();
                return java.text.Collator.getInstance().compare(id1, id2);
            }
        });
        // No connections means no child objects to write; return
        if (conns == null || conns.size() == 0) return;
        
        // Loop thru the connections
        for (Iterator i = conns.iterator(); i.hasNext();) {
            // Grab refs to the connection and its rel set
            OSMXRelSetConnectionType conn = (OSMXRelSetConnectionType) i.next();
            OSMXRelationshipSetType relSet = (OSMXRelationshipSetType) conn.getParent();
            
            // Grab the object bindings for the parent object and the connection
            // (Should be only one for a max-one connection; multiple for max-many conns)
            Set bindings = conn.getObjectBindings(obj.getId());
            // No bindings means no child objects to write, so continue
            if (bindings == null || bindings.size() == 0) continue;
            
            Set unvisitedRels = new HashSet();
            for (Iterator j = bindings.iterator(); j.hasNext();) {
                Object rel = ((OSMXObjectBindingType) j.next()).getParent();
                if (!visitedRels.contains(rel)) unvisitedRels.add(rel);
            }
            visitedRels.addAll(unvisitedRels);
            
            // Remember whether we have written out a child object so we can close tags appropriately
            boolean wroteBinding = false;
            for (Iterator j = bindings.iterator(); j.hasNext();) {
                OSMXObjectBindingType binding = (OSMXObjectBindingType) j.next();
                OSMXRelationshipType rel = (OSMXRelationshipType) binding.getParent();
                if (!unvisitedRels.contains(rel)) continue;
                visitedRels.add(rel);
                List relBindings = new ArrayList(rel.getObjectBinding());
                relBindings.remove(binding);
                if (relBindings == null || relBindings.size() < 1) continue;
                if (!wroteBinding) {
                    wroteBinding = true;
                    w.write("<ul>");
                }
                w.write("<li>");
                PositionedText relSetName = relSet.getName();
                String relSetNameString = "has";
                if (relSetName != null && relSetName.getContent() != null && relSetName.getContent().length() > 0) {
                    relSetNameString = relSetName.getContent();
                }
                w.write(relSetNameString + " ");
                if (relBindings.size() == 1) {
                    OSMXObjectBindingType opposite = (OSMXObjectBindingType) relBindings.get(0);
                    OSMXObjectType oppositeObj = (OSMXObjectType) idx.getOntology().getElement(opposite.getObjectRef());
                    writeObject(w, oppositeObj, visitedRels);
                } else {
                    w.write("<ul>");
                    for (Iterator k = relBindings.iterator(); k.hasNext();) {
                        OSMXObjectBindingType opposite = (OSMXObjectBindingType) k.next();
                        OSMXObjectType oppositeObj = (OSMXObjectType) idx.getOntology().getElement(opposite.getObjectRef());
                        w.write("<li>");
                        writeObject(w, oppositeObj, visitedRels);
                        w.write("</li>");
                    }
                    w.write("</ul>");
                }
                w.write("</li>");
            }
            if (wroteBinding) w.write("</ul>");
        }
    }
    
    /**
     * Stores a reference to an ontology in the ObjectRelationshipWriter so it can be
     * written to the output in later calls.
     * @param ont The new ontology to create output for.
     */    
    public void setOntology(Ontology ont) {
        // early exit if no change
        if (ont == ontology) return;
        
        super.setOntology(ont);
        ontDoc = (OSMXDocument) ontology;
        idx = new ExtractionOntologyIndex(ontDoc);
    }
    
    /**
     * A no-op method; output of the model instance is currently not supported.
     * @param w The output channel.  Not used.
     */    
    public void writeModelInstance(java.io.Writer w) {
        /* do nothing */
    }
    
}
