/*
 * OntosEngine.java
 *
 * Created on February 4, 2004, 9:19 PM
 */

package edu.byu.deg.ontos;

import edu.byu.deg.framework.*;
import edu.byu.deg.osmx.OSMXDocument;
import javax.xml.bind.JAXBException;
import java.io.FileInputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.SimpleFormatter;
import java.util.logging.StreamHandler;
import java.util.logging.Handler;

/**
 * The main class for the Ontos data-extraction engine.  This class defines the
 * configuration parameters allowed for specifying how the engine should operate.
 * It also imposes a constraint that only OSMX ontologies are currently supported.
 *
 * Executing the main method with no arguments will cause the engine to print a
 * usage message to standard output and then exit.
 *
 * Configuration parameters may be passed in via command line arguments, by entries
 * in an external configuration file whose location is specified by an argument, or
 * by directly invoking the {@link DataExtractionEngine#setInitParameter} method.
 * @author Alan Wessman
 */
public class OntosEngine extends DataExtractionEngine {
    
    /* keys for the config file */
    public static final String RETRIEVER_KEY = "documentRetriever";
    public static final String STRUCTURE_RECOGNIZER_KEY = "documentStructureRecognizer";
    public static final String PARSER_KEY = "documentStructureParser";
    public static final String CONTENT_FILTER_KEY = "contentFilter";
    public static final String VALUE_RECOGNIZER_KEY = "valueRecognizer";
    public static final String VALUE_MAPPER_KEY = "valueMapper";
    public static final String ONTOLOGY_WRITER_KEY = "ontologyWriter";
    public static final String ONTOLOGY_KEY = "ontology";
    public static final String RETRIEVAL_PATH = "retrievalPath";
    public static final String RETRIEVAL_PATTERN = "retrievalPattern";
    public static final String OUTPUT_FILENAME = "outputFile";
    public static final String ONTOLOGY_WRITER_OUTPUT_FILENAME = "ontologyWriterOutputFile";
    public static final String EXTRACTION_PLAN_KEY = "extractionPlan";
    
    private static DataExtractionEngine engine;
    
    /* framework modules */
    protected DocumentRetriever retriever;
    protected DocumentStructureRecognizer dsr;
    protected DocumentStructureParser dsp;
    protected ContentFilter contentFilter;
    protected ValueRecognizer valueRecognizer;
    protected ValueMapper valueMapper;
    protected OSMXDocument ontology;
    protected OntologyWriter ontologyWriter;
    protected ExtractionPlan extractionPlan;
    
    /**
     * Initiates the entire data extraction process.
     * @param args The command line arguments; specify no arguments to display a usage message.
     */
    public static void main(String[] args) {
        engine = new OntosEngine();
        processArgs(args);
        try {
            engine.initialize();
            engine.doExtraction();
        } catch (EngineConfigurationException ece) {
            getLogger().severe(ece.getMessage());
            System.exit(-1);
        }
    }
    
    private static void processArgs(String[] args) {
        // process arguments and store as initialization parameters in the engine
        final String CONFIG_FILE_SWITCH = "cfg";
        final String HELP_SWITCH = "h";
        final String RETRIEVER_SWITCH = "r";
        final String STRUCTURE_RECOGNIZER_SWITCH = "dsr";
        final String PARSER_SWITCH = "dsp";
        final String CONTENT_FILTER_SWITCH = "cf";
        final String VALUE_RECOGNIZER_SWITCH = "vr";
        final String VALUE_MAPPER_SWITCH = "vm";
        final String ONTOLOGY_SWITCH = "ont";
        final String ONTOLOGY_WRITER_SWITCH = "ow";
        final String EXTRACTION_PLAN_SWITCH = "ep";
        final String ONT_OUTPUT_SWITCH = "out";
        final String ONT_WRITER_OUTPUT_SWITCH = "owt";
        
        final String[][] SWITCHES = new String[][] {
            {HELP_SWITCH,                 "None (print this help message and exit)"},
            {CONFIG_FILE_SWITCH,          "Path and filename for configuration file"},
            {EXTRACTION_PLAN_SWITCH,      "Class name of ExtractionPlan implementation"},
            {RETRIEVER_SWITCH,            "Class name of DocumentRetriever implementation"},
            {STRUCTURE_RECOGNIZER_SWITCH, "Class name of DocumentStructureRecognizer implementation"},
            {PARSER_SWITCH,               "Class name of DocumentStructureParser implementation"},
            {CONTENT_FILTER_SWITCH,       "Class name of ContentFilter implementation"},
            {VALUE_RECOGNIZER_SWITCH,     "Class name of ValueRecognizer implementation"},
            {VALUE_MAPPER_SWITCH,         "Class name of ValueMapper implementation"},
            {ONTOLOGY_WRITER_SWITCH,      "Class name of OntologyWriter implementation"},
            {ONTOLOGY_SWITCH,             "URL of the OSMX extraction ontology"},
            {ONT_OUTPUT_SWITCH,           "Path and filename of output ontology (default is to save to input ontology)"},
            {ONT_WRITER_OUTPUT_SWITCH,    "Path and filename of OntologyWriter output file"}
        };
        
        StreamHandler errHandler;
        errHandler = new StreamHandler(System.err, new SimpleFormatter());
        errHandler.setLevel(Level.INFO);
        Handler[] handlers = getLogger().getParent().getHandlers();
        for (int i = 0; i < handlers.length; i++) {
            handlers[i].setLevel(Level.OFF);
        }
        handlers = getLogger().getHandlers();
        for (int i = 0; i < handlers.length; i++) {
            handlers[i].setLevel(Level.OFF);
        }
        getLogger().addHandler(errHandler);
        
        // store the args and switches in initParameter map
        for (int i = 0; i < args.length; i++) {
           if (args[i].indexOf("-") == 0) {
               // arg val should follow switch
               String argVal = args[i+1];
               // might not have an arg val (switch only); in such a case val is null
               if (argVal.indexOf("-") == 0) argVal = null;
               
               // place the switch (with "-" trimmed) and value (if any) into map
               engine.setInitParameter(args[i].substring(1), argVal);
           }
        }
        
        if (engine.containsInitParameter(HELP_SWITCH) || engine.countInitParameters() == 0) {
            errHandler.setLevel(Level.INFO);
            // TODO: log usage message and exit
            StringBuffer expl = new StringBuffer();
            StringBuffer opts = new StringBuffer();
            expl.append("[options] are in the form of [switch] [argument] or just [switch] where no arguments are required.\n");
            opts.append("Switches include:\n");
            opts.append("\tSwitch\tArgument\n");
            for (int i = 0; i < SWITCHES.length; i++) {
                opts.append("\t-" + SWITCHES[i][0] + "\t" + SWITCHES[i][1] + "\n");
            }
            getLogger().info("Usage: OntosEngine [options]\n" + expl.toString() + opts.toString());
            System.exit(0);
        }
        
        replaceKey(RETRIEVER_SWITCH, RETRIEVER_KEY);
        replaceKey(STRUCTURE_RECOGNIZER_SWITCH, STRUCTURE_RECOGNIZER_KEY);
        replaceKey(PARSER_SWITCH, PARSER_KEY);
        replaceKey(CONTENT_FILTER_SWITCH, CONTENT_FILTER_KEY);
        replaceKey(VALUE_RECOGNIZER_SWITCH, VALUE_RECOGNIZER_KEY);
        replaceKey(VALUE_MAPPER_SWITCH, VALUE_MAPPER_KEY);
        replaceKey(ONTOLOGY_SWITCH, ONTOLOGY_KEY);
        replaceKey(ONTOLOGY_WRITER_SWITCH, ONTOLOGY_WRITER_KEY);
        replaceKey(EXTRACTION_PLAN_SWITCH, EXTRACTION_PLAN_KEY);
        
        if (engine.containsInitParameter(CONFIG_FILE_SWITCH)) {
            loadConfigFile((String) engine.getInitParameter(CONFIG_FILE_SWITCH));
        }
    } // static method processArgs
    
    private static void loadConfigFile(String filename) {
        try {
            FileInputStream fis = new FileInputStream(filename);
            Properties props = new Properties();
            props.load(fis);
            for (Iterator i = props.keySet().iterator(); i.hasNext();) {
                Object key = i.next();
                if (!engine.containsInitParameter(key)) {
                    engine.setInitParameter(key, props.get(key));
                }
            }
        } catch (java.io.FileNotFoundException fnfe) {
            getLogger().severe("Cannot find file " + filename);
            getLogger().info("Exception: " + fnfe.getMessage());
            getLogger().severe("Exiting...");
            Handler[] handlers = getLogger().getHandlers();
            for (int i = 0; i < handlers.length; i++) {
                handlers[i].flush();
                handlers[i].close();
            }
            System.exit(-1);
        } catch (IOException ioe) {
            getLogger().severe("Cannot read configuration file " + filename);
            getLogger().info("Exception: " + ioe.getMessage());
            getLogger().severe("Exiting...");
            Handler[] handlers = getLogger().getHandlers();
            for (int i = 0; i < handlers.length; i++) {
                handlers[i].flush();
                handlers[i].close();
            }
            System.exit(-1);
        }
    }
    
    private static void replaceKey(String oldKey, String newKey) {
        if (!engine.containsInitParameter(oldKey)) return;
        engine.setInitParameter(newKey, engine.getInitParameter(oldKey));
        engine.removeInitParameter(oldKey);
    }
    
    private void loadClass(String classKey, Class[] formalParams, Object[] actualParams) {
        Object classVal = getInitParameter(classKey);
        if (!(classVal instanceof String)) return;
        
        String className = (String) classVal;
        if (className == null) return;
        try {
            Constructor c = Class.forName(className).getConstructor(formalParams);
            Object obj = c.newInstance(actualParams);
            engine.setInitParameter(classKey, obj);
        } catch (ClassNotFoundException cnfe) {
            getLogger().warning("Class not found: " + className);
            engine.setInitParameter(classKey, null);
        } catch (InstantiationException inste) {
            getLogger().warning("Could not instantiate object of class " + className);
            engine.setInitParameter(classKey, null);
        } catch (NoSuchMethodException nsme) {
            StringBuffer params = new StringBuffer();
            for (int i = 0; i < formalParams.length; i++) {
                if (i > 0) params.append(", ");
                params.append(formalParams[i].getName());
            }
            getLogger().warning("Could not find constructor " + className + "(" + params.toString() + ")");
        } catch (InvocationTargetException ite) {
            StringBuffer params = new StringBuffer();
            for (int i = 0; i < formalParams.length; i++) {
                if (i > 0) params.append(", ");
                params.append(formalParams[i].getName());
            }
            getLogger().warning("Exception invoking constructor for " + className);
            engine.setInitParameter(classKey, null);
        } catch (IllegalAccessException ille) {
            getLogger().warning("Access not allowed to " + className);
            engine.setInitParameter(classKey, null);
        }
    }
    
    private void loadClass(String classKey) {
        loadClass(classKey, new Class[] {}, new Object[] {});
    }
    
    /**
     * Performs the extraction step by executing the extraction plan.
     * @throws EngineConfigurationException Thrown if required objects have not been instantiated or properly initialized
     * via the {@link #initialize} method before this method is invoked.
     */    
    public void doExtraction() throws EngineConfigurationException {
        boolean mayExtract = true;
        StringBuffer errs = new StringBuffer();
        if (retriever == null) {
            mayExtract = false;
            errs.append("Cannot extract without a document retriever.\n");
        }
        if (valueRecognizer == null && valueMapper == null) {
            mayExtract = false;
            errs.append("Cannot extract without either a value recognizer or a value mapper.\n");
        }
        if (ontology == null) {
            mayExtract = false;
            errs.append("Cannot extract without an extraction ontology.\n");
        }
        
        if (!mayExtract) {
            getLogger().severe("Fatal Error:\n" + errs.toString());
           throw new EngineConfigurationException("A required module was not found.");
        }
        
        try {
            extractionPlan.execute();
        } catch (PlanExecutionException pce) {
            getLogger().severe("Fatal error executing extraction plan: " + pce.getMessage());
            throw new EngineConfigurationException("Could not perform extraction.");
        }
        
    }
    
    private void loadOntology(String ontKey) throws IOException, JAXBException {
        String ontFile = (String) getInitParameter(ONTOLOGY_KEY);
        ontology = OSMXDocument.openDocument(ontFile);
        setInitParameter(ONTOLOGY_KEY, ontology);
    }
    
    /**
     * Loads classes and instantiates objects based on the configuration parameters, in
     * preparation for the data extraction process.
     */    
    public void initialize() {
        loadClass(RETRIEVER_KEY);
        loadClass(STRUCTURE_RECOGNIZER_KEY);
        loadClass(PARSER_KEY);
        loadClass(CONTENT_FILTER_KEY);
        loadClass(VALUE_RECOGNIZER_KEY);
        loadClass(VALUE_MAPPER_KEY);
        loadClass(ONTOLOGY_WRITER_KEY);
        try {
            loadOntology(ONTOLOGY_KEY);
        } catch (Exception e) {
            // TODO: Handle exception
            e.printStackTrace();
            ontology = null;
        }
        loadClass(EXTRACTION_PLAN_KEY, new Class[] {this.getClass()}, new Object[] {this});
        
        retriever = (DocumentRetriever) getInitParameter(RETRIEVER_KEY);
        dsr = (DocumentStructureRecognizer) getInitParameter(STRUCTURE_RECOGNIZER_KEY);
        dsp = (DocumentStructureParser) getInitParameter(PARSER_KEY);
        contentFilter = (ContentFilter) getInitParameter(CONTENT_FILTER_KEY);
        valueRecognizer = (ValueRecognizer) getInitParameter(VALUE_RECOGNIZER_KEY);
        valueMapper = (ValueMapper) getInitParameter(VALUE_MAPPER_KEY);
        ontologyWriter = (OntologyWriter) getInitParameter(ONTOLOGY_WRITER_KEY);
        extractionPlan = (ExtractionPlan) getInitParameter(EXTRACTION_PLAN_KEY);
        
        if (extractionPlan == null) extractionPlan = new OntosExtractionPlan(this);
        
        if (ontology != null) {
            ontology.setOutputFilename((String) getInitParameter(OUTPUT_FILENAME));
            
            for (Iterator i = initParameters.entrySet().iterator(); i.hasNext();) {
                Object o = ((Map.Entry) i.next()).getValue();
                if (o instanceof OntologySubscriber) {
                    ((OntologySubscriber) o).setOntology(ontology);
                }
            }
        }
    }
    
}
