/*
 * OntosExtractionPlan.java
 *
 * Created on May 25, 2004, 9:36 PM
 */

package edu.byu.deg.ontos;

import edu.byu.deg.framework.*;
import edu.byu.deg.framework.retriever.DOMRetriever;
import edu.byu.deg.osmx.OSMXDocument;
import edu.byu.deg.util.Timer;

import java.io.File;
import java.io.FileFilter;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Pattern;

/**
 * An implementation of <CODE>ExtractionPlan</CODE> that attempts to replicate the
 * functionality and fundamental algorithm of the Ontos system described in
 * <a href="http://www.deg.byu.edu/papers/dke99.pdf">this paper</a>, in which a set of heuristics
 * are used to govern the execution of a Bernstein synthesis algorithm for deriving
 * objects and relationships given a particular ontology structure.
 * @author Alan Wessman
 */
public class OntosExtractionPlan extends ExtractionPlan {
    
    protected DocumentRetriever retriever;
    protected DocumentStructureRecognizer dsr;
    protected DocumentStructureParser dsp;
    protected ContentFilter contentFilter;
    protected ValueRecognizer valueRecognizer;
    protected ValueMapper valueMapper;
    protected OntologyWriter ontologyWriter;
    protected OSMXDocument ontology;
    protected OntosEngine eng;
    
    public static final String ONTOLOGY_WRITER_OUTPUT_FILENAME = "ontologyWriterOutputFile";
    
    /**
     * Creates a new instance of OntosExtractionPlan
     * @param e The <CODE>DataExtractionEngine</CODE> that will invoke this extraction plan.
     */
    public OntosExtractionPlan(OntosEngine e) {
        super(e);
        eng = e;
        
        retriever = (DocumentRetriever) eng.getInitParameter(OntosEngine.RETRIEVER_KEY);
        dsr = (DocumentStructureRecognizer) eng.getInitParameter(OntosEngine.STRUCTURE_RECOGNIZER_KEY);
        dsp = (DocumentStructureParser) eng.getInitParameter(OntosEngine.PARSER_KEY);
        contentFilter = (ContentFilter) eng.getInitParameter(OntosEngine.CONTENT_FILTER_KEY);
        valueRecognizer = (ValueRecognizer) eng.getInitParameter(OntosEngine.VALUE_RECOGNIZER_KEY);
        valueMapper = (ValueMapper) eng.getInitParameter(OntosEngine.VALUE_MAPPER_KEY);
        ontologyWriter = (OntologyWriter) eng.getInitParameter(OntosEngine.ONTOLOGY_WRITER_KEY);
        ontology = (OSMXDocument) eng.getInitParameter(OntosEngine.ONTOLOGY_KEY);
    }
    
    /**
     * Executes the extraction plan.  The plan is hard-coded to perform extraction
     * on each retrieved document sequentially.
     * @throws PlanExecutionException Thrown whenever a fatal error occurs during the extraction process.
     */    
    public void execute() throws PlanExecutionException {
        Iterator documents = null;
        try {
            retriever.initialize();
            String retPath = (String) engine.getInitParameter(OntosEngine.RETRIEVAL_PATH);
            String retPattern = (String) engine.getInitParameter(OntosEngine.RETRIEVAL_PATTERN);
            if (retPath != null) {
                if (retPattern == null) retPattern = "*";
                final String filterString = retPattern;                
                ((LocalDocumentRetriever) retriever).setFileFilter(new FileFilter() {     
                    private Pattern pat = Pattern.compile(filterString);                    
                    public boolean accept(File pathname) {
                        String filename = pathname.getName();
                        //System.out.println("Matches: File: " + filename+ "     pattern: " + pat.pattern()+ "     matches: " + pat.matcher(filename).matches());
                        if(filename.indexOf(pat.pattern()) >= 0)
                        		return true;
                        return false;
                        //return (pat.matcher(filename).matches());
                    }
                });
                try {
                    retriever.addDocumentLocation(retPath);
                }
                catch (URISyntaxException urise) {
                    OntosEngine.getLogger().warning("Location " + retPath + " is not a valid URI");
                }
            }
            documents = retriever.retrieveDocuments();
        } catch (RetrieverConfigurationException rce) {
            throw new PlanExecutionException(rce);
        }
        
        while (documents.hasNext()) {
            Document doc = (Document) documents.next();
            processDocument(doc);
        }
    }
    
    private void processDocument(Document doc) {
        if (dsr != null) {
            dsp = dsr.getDocumentParser(doc);
        }
        if (dsp != null) {
            try {
                doc = dsp.parse(doc);
            } catch (UnrecognizedFormatException ufe) {
                // TODO: log exception
                return;
            }
        }
        
        if (contentFilter != null) {
            doc = contentFilter.filterDocument(doc);
        }
        
        // if subdocuments exist, assume we have a multi-record document
        // so process the subdocuments instead of the root document
        List docsToProcess = new ArrayList();
        for (Iterator i = doc.getSubDocuments(); i.hasNext();) {
            docsToProcess.add(i.next());
        }
        // add the root document if it's a single record document (no subdocuments)
        if (docsToProcess.isEmpty()) docsToProcess.add(doc);
        
        for (Iterator i = docsToProcess.iterator(); i.hasNext();) {
            Document toProcess = (Document) i.next();
            DataExtractionEngine.getLogger().info("Processing document " + toProcess.getURI());
            try {
    //            System.out.println(doc.toString());
                if (valueRecognizer != null) valueRecognizer.findValues(toProcess);
                if (valueMapper != null) valueMapper.generateValueMappings(toProcess);
                ontology.saveDocument();
                edu.byu.deg.query.DataPrinter printer = new edu.byu.deg.query.DataPrinter();
                //printer.printToFile("result.html", ontology);
                
            } catch (Exception e) {
                // TODO: handle exception
                DataExtractionEngine.getLogger().severe("Could not process document: " + e.getMessage());
                e.printStackTrace();
            }
        }
        try {
            if (ontologyWriter != null && !docsToProcess.isEmpty()) {
                ontologyWriter.setOntology(ontology);
                ontologyWriter.writeDataInstance(new java.io.FileWriter((String) engine.getInitParameter(ONTOLOGY_WRITER_OUTPUT_FILENAME)));
            }
        } catch (Exception e) {
            DataExtractionEngine.getLogger().warning("Error writing data instance: " + e.getMessage());
            e.printStackTrace();
        }
    }
}
