/*
 * RelSetConnectionComparator.java
 *
 * Created on May 10, 2004, 11:13 AM
 */

package edu.byu.deg.ontos;

import java.util.*;
import edu.byu.deg.osmx.*;
import edu.byu.deg.osmx.binding.MatchedTextType;

/**
 * A <CODE>Comparator</CODE> that establishes an ordering function over a set of
 * relationship set connections.  The function attempts to rank the connections
 * based on their likelihood of yielding accurate value mappings within a
 * particular <CODE>MappingContext</CODE>.
 * @author aew2
 */
public class RelSetConnectionComparator implements Comparator {
    
    private ExtractionOntologyIndex idx;
    private MatchingContext ctxt = null;
    
    /**
     * Creates a new instance of RelSetConnectionComparator and initializes it with an
     * indexed ontology.
     * @param index The ontology index used during the comparison process.
     */
    public RelSetConnectionComparator(ExtractionOntologyIndex index) {
        idx = index;
    }
    
    /**
     * Creates a new instance of RelSetConnectionComparator and initializes it with an
     * indexed ontology and the matching context against which the connections are to
     * be ranked.
     * @param index The ontology index used in the comparison process.
     * @param context The matching context for which the connections will be ranked according to
     * accurate mapping generation likelihood.
     */    
    public RelSetConnectionComparator(ExtractionOntologyIndex index, MatchingContext context) {
        this(index);
        setContext(context);
    }
    
    /**
     * Establishes a different context for the comparison than previously existed.
     * @param context The context under which subsequent comparisons will be conducted.
     */    
    public void setContext(MatchingContext context) {
        ctxt = context;
    }
    
    /**
     * Attempts to compare two relationship set connections
     * based on their likelihood of yielding accurate value mappings within a
     * particular <CODE>MappingContext</CODE>.
     * @param o1 One of the relationship set connections to compare.
     * @param o2 The other relationship set connection to compare.
     * @return -1 if the first relationship set connection is deemed to be more likely than the
     * second of generating accurate mappings for the present context; 1 if the second
     * is more likely than the first.  If no distinction can be made regarding this
     * likelihood, the connection with the ID value that appears earliest in dictionary
     * order is considered to be "best", so zero is never returned from this method.
     * This guards against nondeterministic behavior of the system due to random
     * permutations of relationship set connections stemming from the ranking induced
     * by this method.
     */    
    public int compare(Object o1, Object o2) {
        if (ctxt == null) throw new IllegalStateException();
        
        OSMXRelSetConnectionType c1 = (OSMXRelSetConnectionType) o1;
        OSMXRelSetConnectionType c2 = (OSMXRelSetConnectionType) o2;
        
        // get the default result in case we need to exit early
        int defaultResult = java.text.Collator.getInstance().compare(c1.getId(), c2.getId());
        
        // quick equality check
        if (c1.getId().equals(c2.getId())) return 0;
        
        // comparisons...
        
        OSMXObjectSetType objSet1 = (OSMXObjectSetType) idx.getOntology().getElement(c1.getObjectSet());
        OSMXObjectSetType objSet2 = (OSMXObjectSetType) idx.getOntology().getElement(c2.getObjectSet());
        
        // early exit if we're looking at the same object set twice
        if (objSet1 == objSet2) return defaultResult;
        
        MatchingContext ctxt1 = new MatchingContext(ctxt);
        MatchingContext ctxt2 = new MatchingContext(ctxt);
        
        ctxt1.filterOnObjectSet(objSet1, true);
        ctxt2.filterOnObjectSet(objSet2, true);
        
        // lexical precedes nonlexical
        if (objSet1.isLexical() != objSet2.isLexical()) {
            if (objSet1.isLexical()) return -1;
            else return 1;
        }
        
        // if nonlexical...
        if (!objSet1.isLexical()) {
            List kw1 = ctxt1.getKeywordMatches();
            List kw2 = ctxt2.getKeywordMatches();
            
            // keywords present in at least one of the object sets
            if (!(kw1.isEmpty() && kw2.isEmpty())) {
                // if one of the object sets has no keyword matches, prefer the other
                if (kw2.isEmpty()) return -1;
                if (kw1.isEmpty()) return 1;
                
                // keyword matches exist for both obj sets; return the one with the earliest kw match
                MatchedTextType m1 = (MatchedTextType) kw1.get(0);
                MatchedTextType m2 = (MatchedTextType) kw2.get(0);
                if (m1.getStartPos() < m2.getStartPos()) return -1;
                else if (m2.getStartPos() < m1.getStartPos()) return 1;
                else return defaultResult;
            } else return defaultResult;
        }
        
        // lexical...
        List kw1 = ctxt1.getKeywordMatches(), kw2 = ctxt2.getKeywordMatches();
        List v1 = ctxt1.getValueMatches(), v2 = ctxt2.getValueMatches();
        
        // prefer the one with value matches if the other doesn't have any
        if (!v1.isEmpty() && v2.isEmpty()) return -1;
        else if (v1.isEmpty() && !v2.isEmpty()) return 1;
        
        // prefer the one with keywords if the other doesn't have any
        if (!kw1.isEmpty() && kw2.isEmpty()) return -1;
        else if (kw1.isEmpty() && !kw2.isEmpty()) return 1;
        
        // prefer the one with the smallest distance between a val match and a kw match
        double bestDist1 = Double.MAX_VALUE, bestDist2 = Double.MAX_VALUE;
        for (Iterator i = kw1.iterator(); i.hasNext();) {
            MatchedTextType kwm = (MatchedTextType) i.next();
            OSMXKeywordPhraseType kp = (OSMXKeywordPhraseType) idx.getKeywordPhrase(kwm);
            double leftAffinity = kp.getLeftAffinity(), rightAffinity = kp.getRightAffinity();
            for (Iterator j = v1.iterator(); j.hasNext();) {
                MatchedTextType val = (MatchedTextType) j.next();
                int rawDist = 0;
                if (kwm.getStartPos() > val.getEndPos()) rawDist = kwm.getStartPos() - val.getEndPos();
                else if (kwm.getEndPos() < val.getStartPos()) rawDist = val.getStartPos() - kwm.getEndPos();
                if (kwm.getStartPos() > val.getEndPos()) {
                    if (leftAffinity == 0.0) continue;
                    else if ((rawDist/leftAffinity) < bestDist1) bestDist1 = (rawDist/leftAffinity);
                }
                else if (kwm.getEndPos() < val.getStartPos()) {
                    if (rightAffinity == 0.0) continue;
                    else if ((rawDist/rightAffinity) < bestDist1) bestDist1 = (rawDist/rightAffinity);
                }
                else bestDist1 = 0.0;
            }
        }
        for (Iterator i = kw2.iterator(); i.hasNext();) {
            MatchedTextType kwm = (MatchedTextType) i.next();
            OSMXKeywordPhraseType kp = (OSMXKeywordPhraseType) idx.getKeywordPhrase(kwm);
            double leftAffinity = kp.getLeftAffinity(), rightAffinity = kp.getRightAffinity();
            for (Iterator j = v2.iterator(); j.hasNext();) {
                MatchedTextType val = (MatchedTextType) j.next();
                int rawDist = 0;
                if (kwm.getStartPos() > val.getEndPos()) rawDist = kwm.getStartPos() - val.getEndPos();
                else if (kwm.getEndPos() < val.getStartPos()) rawDist = val.getStartPos() - kwm.getEndPos();
                if (kwm.getStartPos() > val.getEndPos()) {
                    if (leftAffinity == 0.0) continue;
                    else if ((rawDist/leftAffinity) < bestDist2) bestDist2 = (rawDist/leftAffinity);
                }
                else if (kwm.getEndPos() < val.getStartPos()) {
                    if (rightAffinity == 0.0) continue;
                    else if ((rawDist/rightAffinity) < bestDist2) bestDist2 = (rawDist/rightAffinity);
                }
                else bestDist2 = 0.0;
            }
        }
        // now compare bestDist1 to bestDist2
        if (bestDist1 < bestDist2) return -1;
        else if (bestDist2 < bestDist1) return 1;
        
        // default: compare based on lexical comparison of connections' IDs.
        return defaultResult;
    }
    
}
