/*
 * ContextualHeuristic.java
 *
 * Created on April 10, 2004, 1:26 PM
 */

package edu.byu.deg.ontos.heuristic;

import edu.byu.deg.ontos.*;
import edu.byu.deg.osmx.*;
import edu.byu.deg.osmx.binding.*;
import edu.byu.deg.osmx.binding.MatchedTextType;
import java.util.*;
import javax.xml.bind.JAXBException;
import java.text.Collator;

/**
 * This class is the central part of the heuristics-based extraction algorithm.
 * This algorithm is recursive; at each level of recursion, the algorithm accepts a
 * designated object set as the "primary" object set for a particular matching
 * context.
 * Responsibilities of <CODE>ContextualHeuristic</CODE> include:
 * <ul>
 * <li>Generating an object for the primary object set from the matching
 * context</li>
 * <li>Choosing the next part of the ontology structure to recurse upon based on
 * the relationship set connections associated with the primary object set</li>
 * <li>Binding relationships generated by the connection-based heuristics
 * (singleton, nonfunctional, functional group, and nested group) to the object
 * generated by this heuristic</li>
 * </ul>
 *
 * Note that <CODE>ContextualHeuristic</CODE> implements a
 * substantial portion of the Bernstein synthesis algorithm as well as the
 * contextual object-generation heuristic for which the class is named.
 * @author Alan Wessman
 */
public class ContextualHeuristic extends MappingHeuristic {
    
    /** Creates a new instance of ContextualHeuristic */
    public ContextualHeuristic() { /* default constructor */ }
    
    /**
     * Creates a new instance of ContextualHeuristic and initializes it with an indexed
     * ontology.
     * @param idx Index for the ontology to be used by this class.
     */    
    public ContextualHeuristic(ExtractionOntologyIndex idx) {
        super(idx);
    }
    
    /**
     * The main algorithm of the class, this method searches the given context to find
     * the best candidate match from which to generate an object for the primary object
     * set of the context if the object set is lexical; if nonlexical, an object is
     * always generated.  If a nonlexical object is generated, the method then invokes the
     * appropriate connection-based heuristic (singleton, nonfunctional, functional
     * group, nested group) on each of the unvisited relationship set connections
     * associated with the primary object set.  From these, a set of relationships is
     * generated, and each is bound to the generated object.  The generated object is
     * finally returned to the code that invoked this method.
     * @param primary The object set that is the "primary" set for the given matching context.  This
     * is not necessarily the same as the main-level primary object set of the ontology
     * (of which there can be at most one), but rather represents the object set for
     * which we wish to generate an object from the matches in the context at this time.
     * @param ctxt The matching context from which the object for the primary object set is to be
     * genearted.
     * @param visitedRelSets A collection of relationship sets that have previously been visited in the
     * recursion, and that therefore should not be re-visited so that infinite
     * recursion is avoided.
     * @return The object generated by this algorithm.  This will be <CODE>null</CODE> under
     * three common scenarios:
     * <ol>
     * <li>The specified matching context contains no value matches (so it contains no
     * evidence sufficient to generate objects and relationships)</li>
     * <li>The primary object set is lexical but the matching context yields no value
     * matches from which an object can be generated for the object set</li>
     * <li>The primary object set is nonlexical but no relationships were generated by
     * the recursion over the associated relationship set connections</li>
     * </ol>
     */    
    public OSMXObjectType processContextualGroup(OSMXObjectSetType primary, final MatchingContext ctxt, Set visitedRelSets) {
        OSMXObjectType result = null;
        
        // exit early if there are no value matches in the context; we have nothing to extract
        if (ctxt.getValueMatches().size() == 0) return null;
        
        // if obj set is lexical, try to find the best matching value; if not found, exit early
        if (primary.isLexical()) {
            // get the best match for the object set
            MatchedTextType bestMatch = getBestLexicalMatch(primary, ctxt);
            // if no match found, return null
            if (bestMatch == null) {
                ctxt.rejectRemaining(primary);
                return null;
            }
            
            // else 
            // accept match
            ctxt.acceptMatch(bestMatch);
            // create object with match value
            try {
                result = (OSMXObjectType) ontIndex.getOntology().getObjectFactory().createOSMObject();
            } catch (JAXBException jaxbe) {
                return null;
            }
            result.setValue(bestMatch.getValue());
            result.setObjectSet(primary.getId());
            result.setLocation(bestMatch.getLocation());
            addObject(result, bestMatch.getDocument());
        } else {
            // primary is nonlexical; generate a new object
            try {
                result = (OSMXObjectType) ontIndex.getOntology().getObjectFactory().createOSMObject();
                result.setObjectSet(primary.getId());
            } catch (JAXBException jaxbe) {
                return null;
            }
        }
        
        // identify previously unvisited relsets for primary object set
        // then mark those relsets as visited to avoid infinite recursion
        Set unvisitedRelSets = new HashSet();
        for (Iterator i = primary.getRelSetConnections().iterator(); i.hasNext();) {
            Object relSetConn = i.next();
            OSMXRelationshipSetType relSet = ((OSMXRelSetConnectionType) relSetConn).getRelSet();
            if (visitedRelSets.contains(relSet)) continue;
            unvisitedRelSets.add(relSetConn);
            visitedRelSets.add(relSet);
        }
        
        // partition newly marked relsets into singleton, functional group, nested group
        Set singletons = new HashSet(), funcGroups = new HashSet();
        Set nestedGroups = new HashSet(), nonfuncs = new HashSet();
        groupRelSets(unvisitedRelSets, singletons, funcGroups, nestedGroups, nonfuncs);
        
        // track whether the traversal generates any child objects so we know whether to prune nonlexical
        boolean didGenerateObject = false;
        
        // process singleton heuristic on each singleton relset
        final SingletonHeuristic sh = new SingletonHeuristic(ontIndex);
        
        // sort singleton rel sets based on comparison function defined in SingletonHeuristic
        Set rankedSingletons = new TreeSet(new Comparator() {
            public int compare(Object o1, Object o2) {
                return sh.compareRelSets((OSMXRelSetConnectionType) o1, (OSMXRelSetConnectionType) o2, ctxt);
            }
        });
        rankedSingletons.addAll(singletons);
        
        // recurses into singleton relsets to get contextual boundary
        for (Iterator i = rankedSingletons.iterator(); i.hasNext();) {
            OSMXRelSetConnectionType relSetConn = (OSMXRelSetConnectionType) i.next();
            OSMXRelationshipType singletonRel = sh.inferRelationship(relSetConn, ctxt, visitedRelSets);
            if (singletonRel == null) continue;
            didGenerateObject = true;
            ObjectBindingType objBinding = bindObject(result, relSetConn);
            singletonRel.getObjectBinding().add(objBinding);
            addRelationship(singletonRel, ctxt.getTargetURI());
        }
        
        // process nested group heuristic on each nested group relset
        final NestedGroupHeuristic ngh = new NestedGroupHeuristic(ontIndex);
        
        // sort nested group rel sets based on comparison function defined in NestedGroupHeuristic
        Set rankedNestedGroups = new TreeSet(new Comparator() {
            public int compare(Object o1, Object o2) {
                return ngh.compareRelSets((OSMXRelSetConnectionType) o1, (OSMXRelSetConnectionType) o2, ctxt);
            }
        });
        rankedNestedGroups.addAll(nestedGroups);
        
        for (Iterator i = rankedNestedGroups.iterator(); i.hasNext();) {
            OSMXRelSetConnectionType relSetConn = (OSMXRelSetConnectionType) i.next();
            Set nestedGroupRels = ngh.inferRelationships(relSetConn, ctxt, visitedRelSets);
            if (nestedGroupRels.size() == 0) continue;
            didGenerateObject = true;
            for (Iterator j = nestedGroupRels.iterator(); j.hasNext();) {
                ObjectBindingType objBinding = bindObject(result, relSetConn);
                OSMXRelationshipType rel = (OSMXRelationshipType) j.next();
                rel.getObjectBinding().add(objBinding);
                addRelationship(rel, ctxt.getTargetURI());
            }
        }
        
        // process functional group heuristic on each functional group
        final FunctionalGroupHeuristic fgh = new FunctionalGroupHeuristic(ontIndex);
        
        // sort singleton rel sets based on comparison function defined in SingletonHeuristic
        Set rankedFuncGroups = new TreeSet(new Comparator() {
            public int compare(Object o1, Object o2) {
                return fgh.compareRelSets((OSMXRelSetConnectionType) o1, (OSMXRelSetConnectionType) o2, ctxt);
            }
        });
        rankedFuncGroups.addAll(funcGroups);
        
        for (Iterator i = rankedFuncGroups.iterator(); i.hasNext();) {
            OSMXRelSetConnectionType relSetConn = (OSMXRelSetConnectionType) i.next();
            Set funcGroupRels = fgh.inferRelationships(relSetConn, ctxt, visitedRelSets);
            if (funcGroupRels.size() == 0) continue;
            didGenerateObject = true;
            for (Iterator j = funcGroupRels.iterator(); j.hasNext();) {
                ObjectBindingType objBinding = bindObject(result, relSetConn);
                OSMXRelationshipType rel = (OSMXRelationshipType) j.next();
                rel.getObjectBinding().add(objBinding);
                addRelationship(rel, ctxt.getTargetURI());
            }
        }
        
        // pick up any remaining lexical nonfunctionals
        final NonfunctionalHeuristic nfh = new NonfunctionalHeuristic(ontIndex);
        
        // sort singleton rel sets based on comparison function defined in SingletonHeuristic
        Set rankedNonfunctionals = new TreeSet(new Comparator() {
            public int compare(Object o1, Object o2) {
                return nfh.compareRelSets((OSMXRelSetConnectionType) o1, (OSMXRelSetConnectionType) o2, ctxt);
            }
        });
        rankedNonfunctionals.addAll(nonfuncs);
        for (Iterator i = rankedNonfunctionals.iterator(); i.hasNext();) {
            OSMXRelSetConnectionType relSetConn = (OSMXRelSetConnectionType) i.next();
            Set nonfuncRels = nfh.inferRelationships(relSetConn, ctxt, visitedRelSets);
            if (nonfuncRels.size() == 0) continue;
            didGenerateObject = true;
            for (Iterator j = nonfuncRels.iterator(); j.hasNext();) {
                ObjectBindingType objBinding = bindObject(result, relSetConn);
                OSMXRelationshipType rel = (OSMXRelationshipType) j.next();
                rel.getObjectBinding().add(objBinding);
                addRelationship(rel, ctxt.getTargetURI());
            }
        }
        
        // if not enough evidence to support generated nonlexical primary, delete it
        if (!primary.isLexical()) {
            if (!didGenerateObject) {
                result.delete();
                result = null;
            } else {
                addObject(result, ctxt.getTargetURI());
            }
        }
        
        // return the created object, or null if we didn't create one
        return result;
    }
    
    /**
     * Partitions the set of unvisited relationship set connections for a particular
     * object set into the four different possible types of connections: singleton,
     * nonfunctional, functional-group, and nested-group.
     * @param unvisitedRelSets The set of <CODE>RelSetConnection</CODE> elements to sort into the different
     * categories of connection.
     * @param singletons The set into which singleton connections should be sorted.
     * @param funcGroups The set into which functional groups should be sorted.
     * @param nestedGroups The set into which nested groups should be sorted.
     * @param nonfuncs The set into which nonfunctional connections should be sorted.
     */    
    protected void groupRelSets(Set unvisitedRelSets, Set singletons, Set funcGroups, Set nestedGroups, Set nonfuncs) {
        for (Iterator i = unvisitedRelSets.iterator(); i.hasNext();) {
            OSMXRelSetConnectionType relSetConn = (OSMXRelSetConnectionType) i.next();
            
            if (isNestedGroup(relSetConn)) nestedGroups.add(relSetConn);
            else if (isSingleton(relSetConn)) singletons.add(relSetConn);
            else if (isFunctionalGroup(relSetConn)) funcGroups.add(relSetConn);
            else if (isNonfunctional(relSetConn)) nonfuncs.add(relSetConn);
        }
    }
    
    /**
     * Determines whether a particular <CODE>RelSetConnection</CODE> qualifies as a
     * singleton connection.
     * @param conn The connection to classify.
     * @return <CODE>true</CODE> if the connection qualifies as a singleton; <CODE>false</CODE>
     * otherwise.  A singleton connection is a max-one connection for a binary
     * relationship set.
     */    
    protected boolean isSingleton(OSMXRelSetConnectionType conn) {
        if (isNestedGroup(conn)) return false;
        return (conn.isOneMax() || conn.isRefinedOneMax());
    }
    
    /**
     * Determines whether a particular <CODE>RelSetConnection</CODE> qualifies as a
     * nested group connection.
     * @param conn The connection to classify.
     * @return <CODE>true</CODE> if the connection qualifies as a nested group connection;
     * <CODE>false</CODE> otherwise.  A connection is a nested group connection if it
     * belongs to a relationship set with arity greater than 2.
     */    
    protected boolean isNestedGroup(OSMXRelSetConnectionType conn) {
        int arity = conn.getRelSet().getRelSetConnection().size();
        return (arity > 2);
    }
    
    /**
     * Determines whether a particular <CODE>RelSetConnection</CODE> qualifies as a
     * functional group connection.
     * @param conn The connection to classify.
     * @return <CODE>true</CODE> if the connection qualifies as a functional group connection;
     * <CODE>false</CODE> otherwise.  A connection is a functional group connection if
     * it is a max-many connection belonging to a binary relationship set that connects
     * to a non-leaf object set on the other end.
     */    
    protected boolean isFunctionalGroup(OSMXRelSetConnectionType conn) {
        if (isNestedGroup(conn) || isSingleton(conn)) return false;
        // determine opposite rel set connection
        List oppConns = new ArrayList(conn.getRelSet().getRelSetConnection());
        oppConns.remove(conn);
        OSMXRelSetConnectionType oppConn = (OSMXRelSetConnectionType) oppConns.get(0);
        
        // determine opposite obj set
        OSMXObjectSetType oppObjSet = (OSMXObjectSetType) ontIndex.getOntology().getElement(oppConn.getObjectSet());
        
        // determine how many other rel sets are connected to opposite obj set
        // return (number > 0)
        return (oppObjSet.getRelSetConnections().size() > 1);
    }
    
    /**
     * Determines whether a particular <CODE>RelSetConnection</CODE> qualifies as a
     * nonfunctional connection.
     * @param conn The connection to classify.
     * @return <CODE>true</CODE> if the connection qualifies as a nonfunctional connection;
     * <CODE>false</CODE> otherwise.  A connection is nonfunctional if it is a max-many
     * connection belonging to a binary relationship set that connects to a leaf object
     * set on the other end.
     */    
    protected boolean isNonfunctional(OSMXRelSetConnectionType conn) {
        if (isNestedGroup(conn) || isSingleton(conn) || isFunctionalGroup(conn)) return false;
        return true;
    }
    
    /**
     * Determines the best value match within a MatchingContext for the given object set.
     * @param objSet The object set to find the best value match for within the given context.
     * @param context The context containing candidate matches.
     * @return The "best" value match belonging to the specified object set and that appears in
     * the given context; or <CODE>null</CODE> if no value matches for the object set
     * exist in the context.  If no keyword matches are present, the best match is defined as
     * the earliest occurring match.  If keyword matchess are present, the best match is the
     * one with the smallest weighted distance from an applicable keyword match.
     */
    protected MatchedTextType getBestLexicalMatch(OSMXObjectSetType objSet, MatchingContext context) {
        // Clone the context so we can filter it without losing information from the original
        MatchingContext localContext = new MatchingContext(context);
        localContext.filterOnObjectSet(objSet, true);
        
        MatchedTextType result = null;
        
        // Early exit if no value matches exist
        if (localContext.getValueMatches().size() == 0) return null;
        
        // Which match is the best depends on whether keywords are present or not.
        if (localContext.getKeywordMatches().size() == 0) {
            /* No keywords; best match is earliest in string or longest if there's a tie */
            
            // Re-sort value matches based on start position (ascending) then length (descending)
            List valMatches = new ArrayList(localContext.getValueMatches());
            Collections.sort(valMatches, new Comparator() {
                public int compare(Object o1, Object o2) { 
                    MatchedTextType m1 = (MatchedTextType) o1;
                    MatchedTextType m2 = (MatchedTextType) o2;
                    if (m1.getStartPos() < m2.getStartPos()) return -1;
                    else if (m2.getStartPos() < m1.getStartPos()) return 1;
                    else if (m1.getEndPos() > m2.getEndPos()) return -1;
                    else if (m2.getEndPos() > m1.getEndPos()) return 1;
                    else return 0;
                }
            });
            
            // Pick the first item in the list as the best match
            result = (MatchedTextType) valMatches.get(0);
        } else {
            /* Find value match closest to a keyword match.
             * "Closest" = (# of characters between keyword and value) * (1 / affinity)
             * Affinity is the leftAffinity or rightAffinity specified for the keyword phrase
             * Use left affinity if value is to the left of the keyword; right affinity otherwise.
             * Affinity of zero means act as if the keyword didn't exist.
             */
            
            // variables to store best match and the best value for the "closest" calculation
            MatchedTextType bestValMatch = null;
            double bestDist = Double.MAX_VALUE;
            boolean isBestDueToKeywordForValue = false;
            
            // Iterate through matches, comparing to best match found so far to see if we found a better one
            for (Iterator i = localContext.getValueMatches().iterator(); i.hasNext();) {
                MatchedTextType match = (MatchedTextType) i.next();
                ValuePhraseType vp = ontIndex.getValuePhrase(match);
                
                // Iterate over keywords to find the closest keyword to this match
                for (Iterator j = localContext.getKeywordMatches().iterator(); j.hasNext();) {
                    MatchedTextType keyMatch = (MatchedTextType) j.next();
                    OSMXKeywordPhraseType kp = (OSMXKeywordPhraseType) ontIndex.getKeywordPhrase(keyMatch);
                    
                    // Get affinity values for the keyword
                    double leftAff = kp.getLeftAffinity(), rightAff = kp.getRightAffinity();
                    
                    // Determine the actual distance between the match and keyword
                    int rawBestDist = match.getStartPos() - keyMatch.getEndPos();
                    if (rawBestDist < 0) rawBestDist = keyMatch.getStartPos() - match.getEndPos();
                    if (rawBestDist < 0) rawBestDist = 0;
                    
                    // Calculate the weighted distance
                    double weightedDist = Double.MAX_VALUE;
                    if (match.getStartPos() > keyMatch.getEndPos()) {
                        if (rightAff != 0.0) weightedDist = rawBestDist/rightAff;
                    }
                    else if (match.getEndPos() < keyMatch.getStartPos()) {
                        if (leftAff != 0.0) weightedDist = rawBestDist/leftAff;
                    }
                    else weightedDist = 0.0;
                    
                    if (vp.getKeywordPhrase() != null) {
                        if (vp.getKeywordPhrase().contains(keyMatch)) {
                            if (!isBestDueToKeywordForValue && weightedDist != Double.MAX_VALUE) {
                                isBestDueToKeywordForValue = true;
                                bestValMatch = match;
                                bestDist = weightedDist;
                            }
                            continue;
                        } else {
                            if (isBestDueToKeywordForValue) continue;
                        }
                    }
                    
                    // get number of characters between key match and value match, or zero if overlap
                    if (weightedDist < bestDist) {
                        bestDist = weightedDist;
                        bestValMatch = match;
                    }
                }
                if (isBestDueToKeywordForValue) continue;
                if (bestValMatch == null) bestValMatch = match;
            }
            result = bestValMatch;
        }
        return result;
    }
    
    /**
     * Defines a total order on a set of relationship set connections by determining,
     * for a pair of such connections, which should precede the other in a sorted list.
     * @param r1 The first relationship set connection to compare.
     * @param r2 The second relationship set connection to compare.
     * @param ctxt The current matching context, which provides information useful in assessing
     * which connection should precede the other.
     * @return -1 if the first connection should precede the second; or 1 if the second should
     * precede the first.  This function does not return 0 unless
     * <CODE>r1 == r2</CODE>.
     *
     * Precedence is determined through a complex set of rules.  The general goal is to
     * order a set of connections such that the first in the resulting list is the most
     * likely to yield an accurately extracted object from the given context if the
     * heuristics algorithm is recursively applied to the connection, and the last in
     * the list is the least promising.
     */    
    protected int compareRelSets(OSMXRelSetConnectionType r1, OSMXRelSetConnectionType r2, MatchingContext ctxt) {
        
        // prefer the required relationship over the optional relationship
        int pcMin1 = r1.getMin(), pcMin2 = r2.getMin();
        if (pcMin1 > pcMin2) return -1;
        else if (pcMin2 > pcMin1) return 1;
        
        // prefer the rel set with the largest difference between actual obj participation and expected obj participation
        
        // prefer rel set with higher arity
        int arity1 = r1.getRelSet().getArity(), arity2 = r2.getRelSet().getArity();
        if (arity1 > arity2) return -1;
        else if (arity2 > arity1) return 1;
        
        // prefer rel set with best opposite connection
        List allConns = new ArrayList();
        allConns.addAll(r1.getOppositeConnections());
        allConns.addAll(r2.getOppositeConnections());
        
        this.connComparer.setContext(ctxt);
        Collections.sort(allConns, this.connComparer);
        OSMXRelSetConnectionType conn = (OSMXRelSetConnectionType) allConns.get(0);
        if (r1.getRelSet() != r2.getRelSet()) {
            if (conn.getRelSet() == r1.getRelSet()) return -1;
            else return 1;
        } else {
            for (int i = 0; i < allConns.size(); i++) {
                if (allConns.get(i) == r1) return -1;
                if (allConns.get(i) == r2) return 1;
            }
        }
        
        // if all else is equal, just compare on the internal ID
        return Collator.getInstance().compare(r1.getId(), r2.getId());
    }
    
}
