/*
 * FunctionalGroupHeuristic.java
 *
 * Created on April 7, 2004, 6:46 PM
 */

package edu.byu.deg.ontos.heuristic;

import edu.byu.deg.ontos.*;
import edu.byu.deg.framework.*;
import edu.byu.deg.osmx.*;
import edu.byu.deg.osmx.binding.*;
import java.util.*;
import javax.xml.bind.JAXBException;

/**
 * A mapping heuristic that generates a relationship based on a relationship set
 * connection that qualifies as a functional group in the traversal of the ontology
 * structure.
 * @author Alan Wessman
 */
public class FunctionalGroupHeuristic extends MappingHeuristic {
    
    /** Creates a new instance of FunctionalGroupHeuristic */
    public FunctionalGroupHeuristic() { /* default constructor */ }
    
    /**
     * Creates a new instance of FunctionalGroupHeuristic and initializes it with an
     * indexed ontology.
     * @param idx The index for the ontology used by this heuristic.
     */    
    public FunctionalGroupHeuristic(ExtractionOntologyIndex idx) {
        super(idx);
    }
    
    /**
     * The main algorithm for this heuristic.  The method generates relationships from
     * the context by greedily invoking {@link ContextualHeuristic} recursively
     * until no more objects are produced.  A relationship is created for each object
     * returned by <CODE>ContextualHeuristic</CODE>, and the set of relationships is
     * returned.
     * @param primaryConn The relationship set connection from which the traversal is originating; in
     * other words, the "side" of the relationship set that the invoking code acted
     * upon to invoke this heuristic.  This parameter identifies both the relationship
     * set for which the relationships should be generated, as well as in which
     * direction to continue the recursion.
     * @param ctxt The matching context from which relationships should be generated.
     * @param visitedRelSets A collection of relationship sets previously visited during the recursion;
     * allows us to avoid infinite recursion.
     * @return A <CODE>Set</CODE> of <CODE>Relationship</CODE> elements.
     */    
    public Set inferRelationships(OSMXRelSetConnectionType primaryConn, MatchingContext ctxt, Set visitedRelSets) {
        Set result = new HashSet();
        int numEliminatedMatches, numRemainingMatches;
        do {
            numRemainingMatches = ctxt.getValueMatches().size();
            OSMXRelationshipType rel = inferRelationship(primaryConn, ctxt, visitedRelSets);
            if (rel == null) break;
            result.add(rel);
            numEliminatedMatches = numRemainingMatches - ctxt.getValueMatches().size();
        } while (numEliminatedMatches > 0);
        return result;
    }
    
    /**
     * Generates a single relationship from the given context.
     * @param primaryConn The connection that identifies the relationship set for which the relationship
     * should be generated.  It also indicates from which direction the ontology
     * traversal is proceeding, so that infinite recursion can be avoided.
     * @param ctxt The matching context from which the relationship is to be derived.
     * @param visitedRelSets A set of previously visited relationship sets, used to remember where the
     * traversal has already been performed so that infinite recursion can be avoided.
     * @return The generated relationship, or <CODE>null</CODE> if none could be successfully
     * generated from the context.
     */    
    protected OSMXRelationshipType inferRelationship(OSMXRelSetConnectionType primaryConn, MatchingContext ctxt, Set visitedRelSets) {
        OSMXRelationshipType result = null;
        OSMXRelSetConnectionType oppConn = (OSMXRelSetConnectionType) primaryConn.getOppositeConnections().toArray()[0];
        OSMXObjectSetType objSet = (OSMXObjectSetType) ontIndex.getOntology().getElement(oppConn.getObjectSet());
        ContextualHeuristic ch = new ContextualHeuristic(ontIndex);
        OSMXObjectType obj = ch.processContextualGroup(objSet, ctxt, visitedRelSets);
        if (obj == null) return null;
        try {
            ObjectBindingType objBinding = bindObject(obj, oppConn);
            result = (OSMXRelationshipType) ontIndex.getOntology().getObjectFactory().createRelationship();
            result.setRelationshipSet(primaryConn.getRelSet().getId());
            result.getObjectBinding().add(objBinding);
        } catch (JAXBException jaxbe) {
            DataExtractionEngine.getLogger().severe("JAXB exception: " + jaxbe.getMessage());
        }
        return result;
    }
    
    /**
     * Defines a total order on a set of relationship set connections by determining,
     * for a pair of such connections, which should precede the other in a sorted list.
     * @param r1 The first relationship set connection to compare.
     * @param r2 The second relationship set connection to compare.
     * @param ctxt The current matching context, which provides information useful in assessing
     * which connection should precede the other.
     * @return -1 if the first connection should precede the second; or 1 if the second should
     * precede the first.  This function does not return 0 unless
     * <CODE>r1 == r2</CODE>.
     *
     * This method bases the order only on a dictionary ordering of the IDs of the
     * relationship sets being compared.
     */      
    protected int compareRelSets(OSMXRelSetConnectionType r1, OSMXRelSetConnectionType r2, MatchingContext ctxt) {
        return java.text.Collator.getInstance().compare(r1.getId(), r2.getId());
    }
    
    /**
     * Determines the likely outer boundaries of the functional group from the given context
     * and returns the result as a matching context.
     * @param conn The connection that identifies the functional group.
     * @param ctxt The context from which to derive the functional group context.
     * @return A matching context that is guaranteed to be a subset of the given context.  The
     * method currently returns the same context as is passed in, effectively rendering
     * the method a no-op.
     */    
    protected MatchingContext findBestSubContext(OSMXRelSetConnectionType conn, MatchingContext ctxt) {
        return ctxt;
    }
    
    /**
     * Divides the given context into multiple matching contexts whose boundaries
     * should approximate the contextual boundaries of the individual objects to be
     * generated for the functional group.  In other words, if the original context
     * would cause multiple objects to be generated, it would be divided into that
     * number of sub-contexts, from each of which an object may be derived.
     * @param ctxt The context to subdivide.
     * @return An array of matching contexts, each corresponding to an object expected to be
     * generated in the recursive invocation of <CODE>ContextualHeuristic</CODE>.
     * Currently this method is not implemented and will throw
     * <CODE>UnsupportedOperationException</CODE> if called.
     */    
    protected MatchingContext[] subdivideContext(MatchingContext ctxt) {
        throw new UnsupportedOperationException("Context subdivision is not yet implemented");
    }
    
}
