/*
 * GlobalOverlapHeuristic.java
 *
 * Created on April 1, 2004, 2:22 PM
 */

package edu.byu.deg.ontos.heuristic;

import java.util.*;
import edu.byu.deg.framework.*;
import edu.byu.deg.osmx.*;
import edu.byu.deg.osmx.binding.*;

/**
 * Implements the heuristic that eliminates conflicting claims caused by
 * overlapping matches.  This implementation resolves overlapping claims even when
 * the claims are made by different object sets -- hence the "global" nature of the
 * class.  Note that since all overlapping claims are decided by this
 * implementation, the heuristic effectively maps each character of the document to
 * at most one object set, and so very little remains to decide in the value
 * mapping process.  The local version of this heuristic is more likely to yield
 * accurate results.
 * @author Alan Wessman
 */
public class GlobalOverlapHeuristic extends edu.byu.deg.ontos.MappingHeuristic {
    
    Comparator matchComparer;
    
    /** Creates a new instance of GlobalOverlapHeuristic */
    public GlobalOverlapHeuristic() {
        matchComparer = new MatchedTextComparator();
    }
    
    /**
     * Resolves overlapping value match claims, rejecting all but one of the matches
     * that lay claim to a particular region of the document.
     * @param ont The ontology in which the value matches are stored.
     * @param rootURI The URI of the root of the document tree to which the value matches must belong
     * to be considered during the overlap elimination process.
     */    
    public void eliminateOverlaps(edu.byu.deg.framework.Ontology ont, String rootURI) {
        Map matchesByDocument = new HashMap();
        
        // separate matches by document
        List objSets = ((OSMXDocument) ont).getObjectSets();
        for (Iterator i = objSets.iterator(); i.hasNext();) {
            ObjectSetType objSet = (ObjectSetType) i.next();
            DataFrameType datFrame = objSet.getDataFrame();
            if (datFrame == null) continue;
            for (Iterator j = datFrame.getValuePhrase().iterator(); j.hasNext();) {
                ValuePhraseType vp = (ValuePhraseType) j.next();
                DataFrameExpression expr = vp.getValueExpression();
                if (expr == null) continue;
                List matches = expr.getMatchedText();
                if (matches == null) continue;
                for (Iterator k = matches.iterator(); k.hasNext();) {
                    MatchedTextType match = (MatchedTextType) k.next();
                    if (!match.getDocument().startsWith(rootURI)) continue;
                    if (!match.getStatus().equalsIgnoreCase(PROPOSED_MATCH)) continue;
                    List docMatches = (List) matchesByDocument.get(match.getDocument());
                    if (docMatches == null) {
                        docMatches = new LinkedList();
                        matchesByDocument.put(match.getDocument(), docMatches);
                    }
                    docMatches.add(match);
                }
            }
        }
        
        
        
        // for each document, get the match list
        for (Iterator i = matchesByDocument.keySet().iterator(); i.hasNext();) {
            List docMatches = (List) matchesByDocument.get(i.next());
            
            // sort the list
            Collections.sort(docMatches, matchComparer);
            
            // iterate over each match
            for (Iterator j = docMatches.iterator(); j.hasNext();) {
                MatchedTextType match = (MatchedTextType) j.next();
                
                // may have been rejected by earlier comparison; skip if so
                if (match.getStatus().equalsIgnoreCase(REJECTED_MATCH)) continue;
                
                // look forward to see if matches ahead overlap
                for (int k = docMatches.indexOf(match) + 1; k < docMatches.size(); k++) {
                    MatchedTextType compare = (MatchedTextType) docMatches.get(k);
                    
                    // if start pos of forward match is past the end of the present match, cannot overlap; exit loop
                    if (compare.getStartPos() >= match.getEndPos()) break;
                    
                    
                    int matchLength = match.getEndPos() - match.getStartPos();
                    int compareLength = compare.getEndPos() - compare.getStartPos();
                    
                    // if start pos equal, the forward match must be longer due to sorting, so reject present match
                    if (compareLength > matchLength) {
                        match.setStatus(REJECTED_MATCH);
                        String matchInfo = match.getValue() + "(" + match.getStartPos() + "-" + match.getEndPos() + ")";
                        DataExtractionEngine.getLogger().info("Rejected match due to overlap; size won: " + matchInfo);
                    } else {
                        compare.setStatus(REJECTED_MATCH);
                        String matchInfo = compare.getValue() + "(" + compare.getStartPos() + "-" + compare.getEndPos() + ")";
                        if (compareLength == matchLength) {
                            DataExtractionEngine.getLogger().info("Rejected match due to overlap; position won: " + matchInfo);
                        } else {
                            DataExtractionEngine.getLogger().info("Rejected match due to overlap; size won: " + matchInfo);
                        }
                    }
                }
            }
        }
    }
    
}
