/*
 * LocalSubsumeHeuristic.java
 *
 * Created on April 1, 2004, 2:01 PM
 */

package edu.byu.deg.ontos.heuristic;

import java.util.*;
import edu.byu.deg.osmx.*;
import edu.byu.deg.osmx.binding.*;
import edu.byu.deg.framework.*;

/**
 * Resolves conflicts over value match pairs in which one match subsumes the other.
 * This implementation of the heuristic operates only on pairs in which both
 * matches belong to the same object set.
 * @author Alan Wessman
 */
public class LocalSubsumeHeuristic extends edu.byu.deg.ontos.MappingHeuristic {
    
    Comparator matchComparer;
    
    /** Creates a new instance of LocalSubsumeHeuristic */
    public LocalSubsumeHeuristic() {
        matchComparer = new MatchedTextComparator();
    }
    
    /**
     * Resolves conflicts caused by one match subsuming another match.
     * @param ont The ontology in which the value matches are stored.
     * @param rootURI The URI of the root of the document tree to which the value matches must belong
     * to be considered during the subsumption elimination process.
     */    
    public void eliminateSubsumptions(edu.byu.deg.framework.Ontology ont, String rootURI) {
        // separate matches by document
        List objSets = ((OSMXDocument) ont).getObjectSets();
        for (Iterator i = objSets.iterator(); i.hasNext();) {
            ObjectSetType objSet = (ObjectSetType) i.next();
            eliminateSubsumptions(objSet, rootURI);
        }
    }
    
    private void eliminateSubsumptions(ObjectSetType objSet, String rootURI) {
        
        Map matchesByDocument = new HashMap();
        
        // separate matches by document
        if (objSet == null) return;
        DataFrameType datFrame = objSet.getDataFrame();
        if (datFrame == null) return;
        for (Iterator j = datFrame.getValuePhrase().iterator(); j.hasNext();) {
            ValuePhraseType vp = (ValuePhraseType) j.next();
            DataFrameExpression expr = vp.getValueExpression();
            if (expr == null) continue;
            List matches = expr.getMatchedText();
            if (matches == null) continue;
            for (Iterator k = matches.iterator(); k.hasNext();) {
                MatchedTextType match = (MatchedTextType) k.next();
                if (!match.getDocument().startsWith(rootURI)) continue;
                if (!match.getStatus().equalsIgnoreCase(PROPOSED_MATCH)) continue;
                List docMatches = (List) matchesByDocument.get(match.getDocument());
                if (docMatches == null) {
                    docMatches = new LinkedList();
                    matchesByDocument.put(match.getDocument(), docMatches);
                }
                docMatches.add(match);
            }
        }
        
        // for each document, get the match list
        for (Iterator i = matchesByDocument.keySet().iterator(); i.hasNext();) {
            List docMatches = (List) matchesByDocument.get(i.next());
            
            // sort the list
            Collections.sort(docMatches, matchComparer);
            
            // iterate over each match
            for (Iterator j = docMatches.iterator(); j.hasNext();) {
                MatchedTextType match = (MatchedTextType) j.next();
                
                // may have been rejected by earlier comparison; skip if so
                if (match.getStatus().equalsIgnoreCase(REJECTED_MATCH)) continue;
                
                // look forward to see if matches ahead subsume or are subsumed
                for (int k = docMatches.indexOf(match) + 1; k < docMatches.size(); k++) {
                    MatchedTextType compare = (MatchedTextType) docMatches.get(k);
                    
                    // if start pos of forward match is past the end of the present match, cannot subsume; exit loop
                    if (compare.getStartPos() >= match.getEndPos()) break;
                    
                    // if start pos equal, the forward match must be longer due to sorting, so reject present match
                    if (compare.getStartPos() == match.getStartPos()) {
                        match.setStatus(REJECTED_MATCH);
                        String matchInfo = match.getValue() + "(" + match.getStartPos() + "-" + match.getEndPos() + ")";
                        DataExtractionEngine.getLogger().info("Rejected match due to subsumption: " + matchInfo);
                    }
                    
                    // otherwise if forward match end pos doesn't surpass present match end pos, reject forward match
                    else if (compare.getEndPos() <= match.getEndPos()) {
                        compare.setStatus(REJECTED_MATCH);
                        String matchInfo = compare.getValue() + "(" + compare.getStartPos() + "-" + compare.getEndPos() + ")";
                        DataExtractionEngine.getLogger().info("Rejected match due to subsumption: " + matchInfo);
                    }
                }
            }
        }
    }
    
}
