 /*
 * NestedGroupHeuristic.java
 *
 * Created on April 17, 2004, 11:17 AM
 */

package edu.byu.deg.ontos.heuristic;

import edu.byu.deg.framework.*;
import edu.byu.deg.ontos.*;
import edu.byu.deg.osmx.*;
import edu.byu.deg.osmx.binding.*;
import edu.byu.deg.osmx.binding.MatchedTextType;
import java.util.*;
import javax.xml.bind.JAXBException;

/**
 * A mapping heuristic that generates a relationship based on a relationship set
 * connection that qualifies as a nested group in the traversal of the ontology
 * structure.
 * @author Alan Wessman
 */
public class NestedGroupHeuristic extends MappingHeuristic {
    
    protected ContextualHeuristic ch;
    
    /** Creates a new instance of NestedGroupHeuristic */
    public NestedGroupHeuristic() {
        ch = null;
    }
    
    /**
     * Creates a new instance of NestedGroupHeuristic and initializes it with an
     * indexed ontology.
     * @param idx An index for the ontology used by the heuristic.
     */    
    public NestedGroupHeuristic(ExtractionOntologyIndex idx) {
        super(idx);
    }
    
    /**
     * The main algorithm for this heuristic.  The method generates relationships from
     * the context by greedily invoking {@link ContextualHeuristic} recursively
     * until no more objects are produced.  A relationship is created for each object
     * returned by <CODE>ContextualHeuristic</CODE>, and the set of relationships is
     * returned.
     * @param primaryConn The relationship set connection from which the traversal is originating; in
     * other words, the "side" of the relationship set that the invoking code acted
     * upon to invoke this heuristic.  This parameter identifies both the relationship
     * set for which the relationships should be generated, as well as in which
     * direction to continue the recursion.
     * @param ctxt The matching context from which relationships should be generated.
     * @param visitedRelSets A collection of relationship sets previously visited during the recursion;
     * allows us to avoid infinite recursion.
     * @return A <CODE>Set</CODE> of <CODE>Relationship</CODE> elements.
     */    
    public Set inferRelationships(OSMXRelSetConnectionType primaryConn, MatchingContext ctxt, Set visitedRelSets) {
        Set result = new HashSet();
                
        // if primary conn is max one, perform contextual heuristic on each child obj set and try to form relationship
        if (primaryConn.isOneMax()) {
            OSMXRelationshipType rel = inferOneMaxRelationship(primaryConn, ctxt, visitedRelSets);
            if (rel != null) result.add(rel);
        } else {
            Set rels = inferManyMaxRelationships(primaryConn, ctxt, visitedRelSets);
            result.addAll(rels);
        }
        return result;
    }
    
    /**
     * Stores a reference to the indexed ontology.
     * @param idx The ontology index to use during the mapping process.
     */  
    public void setOntologyIndex(ExtractionOntologyIndex idx) {
        super.setOntologyIndex(idx);
        if (ch == null) ch = new ContextualHeuristic();
        ch.setOntologyIndex(idx);
    }
    
    /**
     * Generates a single relationship for a nested group based on the assumption that
     * the primary connection has a max-one participation constraint.  Because of the
     * max-one constraint, only one object will be generated for each of the other
     * connections in the relationship set.  Each connection must generate an object,
     * though, in order to permit a relationship to be generated.
     * @param primaryConn The connection that defines the relationship set as a max-one nested group.
     * Also indicates from which direction the ontology structure traversal is
     * proceeding.
     * @param ctxt The matching context from which to generate the relationship.
     * @param visitedRelSets Set of relationship sets that have previously been visited.  Useful for avoiding
     * infinite recursion.
     * @return The generated relationship, or <CODE>null</CODE> if a relationship could not be
     * generated because objects could not be generated for each of the relationship
     * set connections.
     */    
    protected OSMXRelationshipType inferOneMaxRelationship(OSMXRelSetConnectionType primaryConn, MatchingContext ctxt, Set visitedRelSets) {
        try {
            // this set will collect object bindings generated as objects are inferred
            Set objBindings = new HashSet();
            
            // if primary conn is max one, perform contextual heuristic on each child obj set and try to form relationship
            List childConns = new ArrayList(primaryConn.getOppositeConnections());
            
            // all elements in the rel set tuple must be filled to be a valid relationship
            boolean satisfiesRelSet = true;
            while (childConns.size() > 0) {
                // sort the list of remaining connections to get the best prospect
                Collections.sort(childConns, connComparer);
                OSMXRelSetConnectionType conn = (OSMXRelSetConnectionType) childConns.remove(0);
                OSMXObjectSetType objSet = (OSMXObjectSetType) ontIndex.getOntology().getElement(conn.getObjectSet());
                
                // call contextual heuristic to get the object for the connection
                // Note: This call may consume matches from the context; future work
                //       might be needed to revert back to the original if the call returns null
                OSMXObjectType obj = ch.processContextualGroup(objSet, ctxt, visitedRelSets);
                
                // if object is null then we cannot satisfy the relationship
                if (obj == null) {
                    satisfiesRelSet = false;
                    break;
                }
                
                // create a new object binding
                OSMXObjectBindingType binding = (OSMXObjectBindingType) ontIndex.getOntology().getObjectFactory().createObjectBinding();
                binding.setConnectionRef(conn.getId());
                binding.setObjectRef(obj.getId());
                objBindings.add(binding);
            }
            // not enough objects to satisfy the relationship
            if (!satisfiesRelSet) {
                return null;
            } else {
                OSMXRelationshipType rel = (OSMXRelationshipType) ontIndex.getOntology().getObjectFactory().createRelationship();
                rel.setRelationshipSet(primaryConn.getRelSet().getId());
                rel.getObjectBinding().addAll(objBindings);
                return rel;
            }
        } catch (javax.xml.bind.JAXBException jaxbe) {
            DataExtractionEngine.getLogger().warning(jaxbe.getMessage());
            return null;
        }
    }
    
    /**
     * Generates a set of relationships for a nested group based on the assumption
     * that the primary connection has a max-many participation constraint.  If all
     * other connections for the relationship set have max-one constraints, the set
     * behaves like an <i>n</i>-ary functional group.  In such a case, a relationship
     * is generated each time objects are successfully generated for every non-primary
     * connection.   If any other connection is max-many, though, the {@link
     * #inferNestedRelationships} method will be invoked, which is more complex in its
     * relationship generation rules.
     * @param primaryConn The relationship set connection that defines the relationship set as a max-many
     * nested group.  It also indicates from which direction the traversal of the
     * ontology structure is proceeding.
     * @param ctxt The matching context from which relationships are generated.
     * @param visitedRelSets Set of relationship sets that have previously been visited.  Useful for avoiding
     * infinite recursion.
     * @return A set of <CODE>Relationship</CODE> elements that were successfully generated
     * from the given context.  The set will be non-<CODE>null</CODE> but empty if no
     * relationships were generated.
     */    
    protected Set inferManyMaxRelationships(OSMXRelSetConnectionType primaryConn, MatchingContext ctxt, Set visitedRelSets) {
        Set result = new HashSet();
        
        // divide the child rel set connections into max-many and max-one lists
        List maxManyChildren = new ArrayList(primaryConn.getOppositeConnections());
        List maxOneChildren = new ArrayList();
        for (Iterator i = maxManyChildren.iterator(); i.hasNext();) {
            OSMXRelSetConnectionType conn = (OSMXRelSetConnectionType) i.next();
            if (conn.isOneMax() || conn.isRefinedOneMax()) {
                i.remove();
                maxOneChildren.add(conn);
            }
        }
        
        // if we only have max-one obj sets then just grab relationships until there are no more
        if (maxManyChildren.size() == 0) {
            OSMXRelationshipType rel;
            do {
                rel = inferOneMaxRelationship(primaryConn, ctxt, visitedRelSets);
                if (rel != null) result.add(rel);
            } while (rel != null);
        } else {
            try {
                result.addAll(inferNestedRelationships(primaryConn, maxManyChildren, maxOneChildren, ctxt, visitedRelSets));
            } catch (JAXBException jaxbe) {
                DataExtractionEngine.getLogger().warning("JAXB Exception: " + jaxbe.getMessage());
            }
        }
        
        return result;
    }
    
    /**
     * Generates a set of relationships for a nested group that contains multiple
     * max-many relationship set connections (one being the primary connection).  In
     * such a case, we assume that concepts represented in the matching context, that
     * will become objects, are expressed in a "nesting" structure.  The outermost
     * level is the primary object; it may be associated with multiple objects
     * belonging to a different connection.  Each of these objects, in turn, defines
     * its own sub-context containing multiple objects for another
     * connection, and so forth until the lowest level.  At this level, if any
     * connections for the nested group relationship set are still unfulfilled by
     * generated objects, the connections must necessarily have max-one constraints and
     * objects are generated from this innermost context&mdash;one for each max-one
     * connection.
     *
     * The effect is something like this:
     *
     * <PRE>
     * |--------------- A1 ---------------||--------------- A2 ------------------|
     * |--- B1 ---||--- B2 ---||--- B3 ---||-------- B4 ------||------ B5 -------|
     * |--- C1 ---|| C2 || C3 || C4 || C5 ||-- C6 --||-- C7 --||------ C8 -------|
     * </PRE>
     *
     * In this diagram, A is the primary object set, B is an object set at the end of a
     * max-many connection in the nested group, and C is an object set for the
     * remaining connection in the relationship set.  Pipes (|) mark context
     * boundaries within which objects (A1, C5, etc.) appear.  The overall diagram
     * represents a matching context, and this method's algorithm induces a nesting
     * structure such as the one above upon the input context in order to
     * determine which objects to generate and how to satisfy constraints.
     *
     * The following relationships are generated from the above example:
     * <ul>
     * <li>A1-B1-C1</li>
     * <li>A1-B2-C2</li>
     * <li>A1-B2-C3</li>
     * <li>A1-B3-C4</li>
     * <li>A1-B3-C5</li>
     * <li>A2-B4-C6</li>
     * <li>A2-B4-C7</li>
     * <li>A2-B5-C8</li>
     * </ul>
     *
     * This method invokes itself recursively since the nesting can in theory be
     * arbitrarily deep.
     * @param primaryConn The relationship set connection that defines the relationship set as a max-many
     * nested group.  It also indicates from which direction the traversal of the
     * ontology structure is proceeding.
     * @param maxMany A list of the <CODE>RelSetConnection</CODE> elements for the nested group which
     * have max-many participation constraints.
     * @param maxOne A list of the <CODE>RelSetConnection</CODE> elements for the nested group which
     * have max-one participation constraints.
     * @param ctxt The input context from which relationships are to be inferred.
     * @param visitedRelSets Set of relationship sets that have previously been visited.  Useful for avoiding
     * infinite recursion.
     * @throws JAXBException Thrown if a relationship cannot be generated for some reason.
     * @return A set of <CODE>Relationship</CODE> elements generated by this method.
     */    
    protected Set inferNestedRelationships(OSMXRelSetConnectionType primaryConn, List maxMany, List maxOne, MatchingContext ctxt, Set visitedRelSets) throws JAXBException {
        Set result = new HashSet();
        
        // find the most likely max-many that defines the outer nesting level
        OSMXRelSetConnectionType outerMaxMany = findOuterMaxMany(maxMany, ctxt);
        
        // exit early if outerMaxMany is null
        if (outerMaxMany == null) return result;
        
        // we have at least one max-many to deal with and a non-empty context
        
        // get the object set for the outerMaxMany connection
        OSMXObjectSetType outerObjSet = (OSMXObjectSetType) ontIndex.getOntology().getElement(outerMaxMany.getObjectSet());
        
        // remember the beginning position of the parent context
        int contextBegin = 0;
        MatchedTextType earliestMatch = null;
        if (ctxt.getValueMatches().size() > 0) {
            earliestMatch = (MatchedTextType) ctxt.getValueMatches().get(0);
            contextBegin = earliestMatch.getStartPos();
        }
        if (ctxt.getKeywordMatches().size() > 0) {
            earliestMatch = (MatchedTextType) ctxt.getKeywordMatches().get(0);
            if (earliestMatch.getStartPos() < contextBegin) contextBegin = earliestMatch.getStartPos();
        }
        
        // loop will terminate by break statements inside
        while (true) {
            // find the context boundary end (next probable instance of outerMaxMany, or parent context end)
            MatchingContext objSetCtxt = new MatchingContext(ctxt);
            objSetCtxt.filterOnObjectSet(outerObjSet, true);
            int contextEnd = 0;
            if (outerObjSet.isLexical()) {
                // find the second value match for the object set, if it exists
                if (objSetCtxt.getValueMatches().size() > 1) {
                    List valMatches = objSetCtxt.getValueMatches();
                    contextEnd = ((MatchedTextType) valMatches.get(1)).getStartPos();
                }
            } else {
                // find the second keyword match for the object set, if it exists
                if (objSetCtxt.getKeywordMatches().size() > 1) {
                    List keyMatches = objSetCtxt.getKeywordMatches();
                    contextEnd = ((MatchedTextType) keyMatches.get(1)).getStartPos();
                }
            }
            // context end should actually be right before the next match
            contextEnd--;
            
            // track whether this is the last subcontext
            boolean isLastSubcontext = false;
            // create new subcontext using begin and end positions
            MatchingContext subCtxt = new MatchingContext(ctxt);
            // filter subCtxt on begin and end positions
            if (contextEnd > contextBegin) {
                subCtxt.filterOnStartPosRange(contextBegin, contextEnd, true);
            } else {
                isLastSubcontext = true;
            }
            
            // remove outerMaxMany from maxMany list
            List remainingMaxMany = new ArrayList(maxMany);
            remainingMaxMany.remove(outerMaxMany);
            
            // invoke contextual heuristic to get object from subcontext
            OSMXObjectType obj = ch.processContextualGroup(outerObjSet, subCtxt, visitedRelSets);
            
            // if no object resulted, break out of loop if last subcontext or continue otherwise
            if (obj == null) {
                if (isLastSubcontext) break;
                else continue;
            }
            
            // find the next context begin point (necessary to do this?)
            contextBegin = contextEnd + 1;
            
            // if no more elements in maxMany list, process maxOne connections
            if (remainingMaxMany.size() == 0) {
                // this loop allows us to generate several relationships within the innermost max-many level
                while (true) {
                    boolean satisfiesRelSet = true;
                    Set maxOneObjs = new HashSet();
                    for (Iterator i = maxOne.iterator(); i.hasNext();) {
                        OSMXRelSetConnectionType conn = (OSMXRelSetConnectionType) i.next();
                        OSMXObjectSetType innerObjSet = (OSMXObjectSetType) ontIndex.getOntology().getElement(conn.getObjectSet());
                        // invoke contextual heuristic to get object from subcontext
                        OSMXObjectType innerObj = ch.processContextualGroup(innerObjSet, subCtxt, visitedRelSets);
                        // if no object, set flag and break out of loop
                        if (innerObj == null) {
                            satisfiesRelSet = false;
                            break;
                        }

                        // create tentative object binding
                        ObjectBindingType objBinding = bindObject(innerObj, conn);

                        // add object binding to local set to create relationship later
                        maxOneObjs.add(objBinding);
                    }
                    // break out of loop if we cannot satisfy the rel set
                    if (!satisfiesRelSet) break;

                    // create relationship
                    OSMXRelationshipType rel = (OSMXRelationshipType) ontIndex.getOntology().getObjectFactory().createRelationship();
                    rel.setRelationshipSet(primaryConn.getRelSet().getId());

                    // bind all maxOne objects to the relationship
                    for (Iterator i = maxOneObjs.iterator(); i.hasNext();) {
                        ObjectBindingType objBinding = (ObjectBindingType) i.next();
                        // add the object binding to the relationship
                        rel.getObjectBinding().add(objBinding);
                    }
                    // bind the maxMany object to the relationship
                    ObjectBindingType objBinding = bindObject(obj, outerMaxMany);
                    // add the object binding to the relationship
                    rel.getObjectBinding().add(objBinding);

                    // add the relationship to the result set
                    result.add(rel);
                    
                    // break out of loop if no maxOne connections existed
                    if (maxOne.isEmpty()) break;
                }
            } else {
                // recurse using decreased maxMany list and local context
                Set recurseResults = inferNestedRelationships(primaryConn, remainingMaxMany, maxOne, subCtxt, visitedRelSets);
                
                // for each relationship in the recursed result set
                for (Iterator i = recurseResults.iterator(); i.hasNext();) {
                    // bind the maxMany object to the relationship
                    OSMXRelationshipType rel = (OSMXRelationshipType) i.next();
                    ObjectBindingType objBinding = bindObject(obj, outerMaxMany);
                    rel.getObjectBinding().add(objBinding);
                }
                // add all relationships in the recursed result set to the current result set
                result.addAll(recurseResults);
            }
        }
        
        return result;
    }
    
    /**
     * Determines which relationship set connection from a given list best corresponds
     * to the outermost nesting level of the given matching context.
     * @param maxMany A list of max-many relationship set connections from which the connection for
     * the outermost nesting level should be selected.  The list need not contain all
     * max-many connections for the nested group relationship set; rather, it should
     * only contain those max-many connections for which objects still need to be generated from
     * the input matching context.
     * @param ctxt The context over which to induce the nesting structure in order to determine the
     * outermost nesting connection.
     * @return The connection from the list that best corresponds with the outermost nesting
     * level of the given context.
     */    
    protected OSMXRelSetConnectionType findOuterMaxMany(List maxMany, MatchingContext ctxt) {
        OSMXRelSetConnectionType result = null;
        if (maxMany.size() == 0) return null;
        if (ctxt.getKeywordMatches().isEmpty() && ctxt.getValueMatches().isEmpty()) return null;
        
        int earliestMatchPos = -1;
        for (Iterator i = maxMany.iterator(); i.hasNext();) {
            OSMXRelSetConnectionType conn = (OSMXRelSetConnectionType) i.next();
            OSMXObjectSetType objSet = (OSMXObjectSetType) ontIndex.getOntology().getElement(conn.getObjectSet());
            if (objSet.isLexical()) {
                List valMatches = ctxt.getValuesForObjectSet(objSet);
                if (valMatches.size() > 0) {
                    MatchedTextType match = (MatchedTextType) valMatches.get(0);
                    if (match.getStartPos() < earliestMatchPos || earliestMatchPos == -1) {
                        earliestMatchPos = match.getStartPos();
                        result = conn;
                    }
                }
            } else {
                List keyMatches = ctxt.getKeywordsForObjectSet(objSet);
                if (keyMatches.size() > 0) {
                    MatchedTextType match = (MatchedTextType) keyMatches.get(0);
                    if (match.getStartPos() < earliestMatchPos || earliestMatchPos == -1) {
                        earliestMatchPos = match.getStartPos();
                        result = conn;
                    }
                }
            }
        }
        
        return result;
    }
    
    /**
     * Defines a total order on a set of relationship set connections by determining,
     * for a pair of such connections, which should precede the other in a sorted list.
     * @param r1 The first relationship set connection to compare.
     * @param r2 The second relationship set connection to compare.
     * @param ctxt The current matching context, which provides information useful in assessing
     * which connection should precede the other.
     * @return -1 if the first connection should precede the second; or 1 if the second should
     * precede the first.  This function does not return 0 unless
     * <CODE>r1 == r2</CODE>.
     *
     * This method bases the order only on a dictionary ordering of the IDs of the
     * relationship sets being compared.
     */     
    protected int compareRelSets(OSMXRelSetConnectionType r1, OSMXRelSetConnectionType r2, MatchingContext ctxt) {
        return java.text.Collator.getInstance().compare(r1.getId(), r2.getId());
    }
    
    /**
     * Divides the given context into multiple matching contexts whose boundaries
     * should approximate the contextual boundaries of the individual objects to be
     * generated for the nested group.  In other words, if the original context
     * would cause multiple objects to be generated, it would be divided into that
     * number of sub-contexts, from each of which an object may be derived.
     * @param ctxt The context to subdivide.
     * @return An array of matching contexts, each corresponding to an object expected to be
     * generated in the recursive invocation of <CODE>ContextualHeuristic</CODE>.
     * Currently this method is not implemented and will throw
     * <CODE>UnsupportedOperationException</CODE> if called.
     */     
    protected MatchingContext[] subdivideContext(MatchingContext ctxt) {
        throw new UnsupportedOperationException("Context subdivision is not yet implemented");
    }
    
    /**
     * Determines the likely outer boundaries of the nested group from the given context
     * and returns the result as a matching context.
     * @param conn The connection that identifies the nested group.
     * @param ctxt The context from which to derive the nested group context.
     * @return A matching context that is guaranteed to be a subset of the given context.  The
     * method currently returns the same context as is passed in, effectively rendering
     * the method a no-op.
     */    
    protected MatchingContext findBestSubContext(OSMXRelSetConnectionType conn, MatchingContext ctxt) {
        return ctxt;
    }
    
}
