/*
 * SingletonHeuristic.java
 *
 * Created on April 16, 2004, 10:28 AM
 */

package edu.byu.deg.ontos.heuristic;

import edu.byu.deg.ontos.*;
import edu.byu.deg.framework.*;
import edu.byu.deg.osmx.*;
import edu.byu.deg.osmx.binding.*;
import java.util.*;
import javax.xml.bind.JAXBException;

/**
 * A mapping heuristic that generates a relationship based on a relationship set
 * connection that qualifies as a singleton in the traversal of the ontology
 * structure.
 * @author Alan Wessman
 */
public class SingletonHeuristic extends MappingHeuristic {
    
    /** Creates a new instance of SingletonHeuristic */
    public SingletonHeuristic() { /* default constructor */ }
    
    /**
     * Creates a new instance of SingletonHeuristic and initializes it with an
     * indexed ontology.
     * @param idx An index for the ontology used by the heuristic.
     */    
    public SingletonHeuristic(ExtractionOntologyIndex idx) {
        super(idx);
    }
    
    /**
     * The main algorithm for this heuristic.  The method generates a relationship from
     * the context by invoking {@link ContextualHeuristic} exactly one time.  A
     * relationship is created and returned if an object is returned by
     * <CODE>ContextualHeuristic</CODE>.
     * @param primaryConn The relationship set connection from which the traversal is originating; in
     * other words, the "side" of the relationship set that the invoking code acted
     * upon to invoke this heuristic.  This parameter identifies both the relationship
     * set for which the relationship should be generated, as well as in which
     * direction to continue the recursion.
     * @param ctxt The matching context from which a relationship should be generated.
     * @param visitedRelSets A collection of relationship sets previously visited during the recursion;
     * allows us to avoid infinite recursion.
     * @return A <CODE>Relationship</CODE> element if an <CODE>Object</CODE> was generated, or
     * <CODE>null</CODE> otherwise.
     */  
    public OSMXRelationshipType inferRelationship(OSMXRelSetConnectionType primaryConn, MatchingContext ctxt, Set visitedRelSets) {
        OSMXRelationshipType result = null;
        OSMXRelSetConnectionType singletonConn = (OSMXRelSetConnectionType) primaryConn.getOppositeConnections().toArray()[0];
        OSMXObjectSetType objSet = (OSMXObjectSetType) ontIndex.getOntology().getElement(singletonConn.getObjectSet());
        ContextualHeuristic ch = new ContextualHeuristic(ontIndex);
        MatchingContext subContext = findBestSubContext(primaryConn, ctxt);
        OSMXObjectType obj = ch.processContextualGroup(objSet, subContext, visitedRelSets);
        if (obj == null) return null;
        try {
            ObjectBindingType objBinding = bindObject(obj, singletonConn);
            result = (OSMXRelationshipType) ontIndex.getOntology().getObjectFactory().createRelationship();
            result.setRelationshipSet(primaryConn.getRelSet().getId());
            result.getObjectBinding().add(objBinding);
        } catch (JAXBException jaxbe) {
            DataExtractionEngine.getLogger().severe("JAXB exception: " + jaxbe.getMessage());
        }
        return result;
    }
    
    /**
     * Defines an ordering over a set of <CODE>RelSetConnection</CODE>s with respect to
     * a particular matching context.  Lower-ordered connections are interpreted as
     * better candidates for yielding extracted objects and relationships from the
     * given context than higher-ordered connections.
     * @param r1 The first <CODE>RelSetConnection</CODE> to compare.
     * @param r2 The second <CODE>RelSetConnection</CODE> to compare.
     * @param ctxt The context in which to compare the connections.  For example, a connection to
     * an <CODE>Age</CODE> object set is more likely to generate an extracted object if
     * the context has a one- or two-digit number match in it than if it did not, so
     * the <CODE>Age</CODE> object set would likely be ranked lower in this ordering
     * in the former case.
     * @return -1 if the first connection should precede the second; 1 if the second should
     * precede the first; 0 if they are equal (which is possible only if they are the
     * same connection).
     */    
    protected int compareRelSets(OSMXRelSetConnectionType r1, OSMXRelSetConnectionType r2, MatchingContext ctxt) {
        OSMXObjectSetType o1, o2;
        String objSetRef1, objSetRef2;
        objSetRef1 = ((OSMXRelSetConnectionType) r1.getOppositeConnections().toArray()[0]).getObjectSet();
        objSetRef2 = ((OSMXRelSetConnectionType) r2.getOppositeConnections().toArray()[0]).getObjectSet();
        o1 = (OSMXObjectSetType) ontIndex.getOntology().getElement(objSetRef1);
        o2 = (OSMXObjectSetType) ontIndex.getOntology().getElement(objSetRef2);
        
        int pcMin1 = r1.getMin(), pcMin2 = r2.getMin();
        if (pcMin1 != pcMin2) {
            // prefer the required singleton over the optional singleton
            if (pcMin1 > pcMin2) return -1;
            else return 1;
        }
        
        MatchingContext ctxt1 = new MatchingContext(ctxt), ctxt2 = new MatchingContext(ctxt);
        ctxt1.filterOnObjectSet(o1, true);
        ctxt2.filterOnObjectSet(o2, true);
        
        if (ctxt1.getKeywordMatches().size() != ctxt2.getKeywordMatches().size()) {
            // prefer the one with keyword matches if the other has no keyword matches
            if (ctxt1.getKeywordMatches().size() == 0) return 1;
            if (ctxt2.getKeywordMatches().size() == 0) return -1;
        }
        
        // score value matches and rank them for each object set
        // (we cache the results so we don't have to re-rank them every time we compare)
        // then return a result based on which object set has the best value match
        // (the better object set should precede the worse, so -1 if o1 better than o2)
        return java.text.Collator.getInstance().compare(r1.getId(), r2.getId());
    }
    
    /**
     * Intended as a way to narrow down the input context in order to exclude
     * extraneous matches from the context that should generate objects for this
     * Singleton.  Currently this method simply returns the input context.
     * @param conn The singleton connection for which the context should be narrowed.
     * @param ctxt The context to narrow down for more accurate extraction.
     * @return A matching context that has had keyword and value entries removed to improve
     * extraction accuracy.  If no entries were removed, this method might return the
     * originally passed-in object.
     */    
    protected MatchingContext findBestSubContext(OSMXRelSetConnectionType conn, MatchingContext ctxt) {
        return ctxt;
    }
    
}
