/*
 * BinarySearchLexicon.java
 *
 * Created on July 29, 2003, 2:01 PM
 */

package edu.byu.deg.ontos.lexicon;

import edu.byu.deg.framework.LexiconAccessor;
import java.util.ArrayList;
import java.net.URL;

/** <p>This {@link edu.byu.deg.framework.LexiconAccessor LexiconAccessor} stores words
 * in an ArrayList on which binary searches are done for lookups.  Memory use is
 * kept to a minimum.</p>
 * <p>Inserts are done in O(1) time as long as the input is sorted, O(n) otherwise.</p>
 * <p>Lookups are accomplished in O(log n) time.</p>
 * @author troywalk
 */
public class BinarySearchLexicon implements LexiconAccessor{
  ArrayList words;
  boolean caseSensitive;
  
  /** Creates a new instance of BinarySearchLexicon */
  public BinarySearchLexicon(){
    caseSensitive = false;
    words = new ArrayList();
  }
  
  /** Creates a new instance of BinarySearchLexicon
   * @param caseSensitive whether the lexicon is case sensitive
   */  
  public BinarySearchLexicon(boolean caseSensitive){
    this.caseSensitive = caseSensitive;
    words = new ArrayList();
  }

  /** Adds a word to the lexicon
   * @param word the word to be added
   */  
  public void addWord(String word){
    if(!caseSensitive)
      word = word.toLowerCase();
    if((words.size()==0) || (word.compareTo(words.get(words.size()-1))>0)){
      words.add(word);
    }
    else {
      for(int i=words.size()-2;i>=0;i--){
        if(word.compareTo(words.get(i))>0){
          words.add(i+1, word);
          return;
        }
      }
      words.add(0, word);
      System.err.println(word + " out of order");
    }
  }
  
  private int getIndex(String word){
    int start = 0;
    int end = words.size()-1;
    int current = 0;
    while(start!=end){
      current = (start + end) / 2;
      if(word.compareTo(words.get(current))>0){
        start = current+1;
      }
      else{
        end = current;
      }
    }
    if(word.compareTo(words.get(current))>0)
      current++;
    return current;
  }
  
  /** Looks for a string in the lexicon.
   * @param input the string to look up.  Must be lower case if not case sensitive
   * @param offset position at which to start
   * @return length of the longest match
   * @param inputToken first token in input.  This is the minimum match.
   */  
  public int lookup(String inputToken, String input, int offset){
    input = input.substring(offset);
    int current = getIndex(input);
    String word;
    word = (String)words.get(current);
    while((inputToken.compareTo(word)<=0)){
      if(input.startsWith(word)){
        return word.length();
      }
      if(current==0)
        return 0;
      current--;
      word = (String)words.get(current);
    }
    return 0;
  }
  
  public int getSize() {
      return words.size();
  }  
  
  public java.util.List getEntries() {
      return new ArrayList(words);
  }
  
  public void load(URL lexURL) {
      throw new UnsupportedOperationException();
  }
  
  public java.util.Set findContainedEntries(String input, boolean caseSensitive) {
      throw new UnsupportedOperationException();
  }
  
}
