/*
 * LexiconMatcher.java
 *
 * Created on July 23, 2003, 11:07 AM
 */

package edu.byu.deg.ontos.lexicon;

import java.util.*;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.File;
import java.net.URI;
import edu.byu.deg.osmx.binding.LexiconType;
import edu.byu.deg.osmx.binding.OSM;
import edu.byu.deg.osmx.OSMXOSMType;
import edu.byu.deg.framework.LexiconAccessor;


/** <p>LexiconMatcher reads all of the lexicons referenced by an ontology and stores
 * them in a LexiconAccessor.  Matches are hashed by lexicon and originating string
 * for later use.</p>
 * @author troywalk
 */

public class LexiconMatcher {
    private final String DEFAULT_DELIMITER_BASE = " \t\r\n";
    private List lexicons; //lists (LexiconType, LexiconAccessor) pairs
    private HashMap matchStrings; //hash on Lexicon label then on string to match to get set of match strings
        
    private void resetDelimiterList(boolean[] foundChars){
        for(int i=0;i<127;i++){
            foundChars[i] = false;
        }
    }
    
    private void updateDelimiterList(boolean[] foundChars, String str){
        for(int i=0;i<str.length();i++){
            char c = str.charAt(i);
            if (!Character.isSpaceChar(c) && c > 32 && c < 127 && !foundChars[c]){
                if (Character.isLetter(c)){
                    foundChars[Character.toLowerCase(c)] = foundChars[Character.toUpperCase(c)] = true;
                } else {
                    foundChars[c] = true;
                }
            }
        }
    }
    
    private String makeDelimiterString(boolean[] foundChars){
        StringBuffer b = new StringBuffer(DEFAULT_DELIMITER_BASE);
        for (int i=33; i<127; i++) {
            if (!foundChars[i]) {
                b.append((char)i);
            }
        }
        return b.toString();
    }
    
    // temporarily here to preserve signature; deprecated.
    public void loadLexicons(OSM ontology, String workingDir) {
        loadLexicons(ontology);
    }
    
    /** Loads the lexicons required for an ontology using the given path.
     * @param ontology the ontology to be loaded
     */
    public void loadLexicons(OSM ontology){
        //Create LexiconAccessor for each lexicon in the ontology and add to hashmap
        List l  = ontology.getAllModelElements();
        boolean calculateDelimiters;
        boolean[] foundChars = new boolean[127];
        String line;
        lexicons = new ArrayList();
        matchStrings = new HashMap();
        for(Iterator i=l.iterator();i.hasNext();){
            Object o = i.next();
            if(o instanceof LexiconType){
                LexiconType lex = (LexiconType)o;
                String lexuriString = lex.getUri();
                URI onturi = ((OSMXOSMType) ontology).getParentDocument().getURI();
                URI lexuri = onturi.resolve(lexuriString);
                
                // clear out list of delimiters
                resetDelimiterList(foundChars);
                
                LexiconAccessor lexData = new BinarySearchLexicon(lex.isCaseSensitive());
                calculateDelimiters = (lex.getDelimiters()==null)||(lex.getDelimiters().length()==0);
                
                try{
                    BufferedReader r = new BufferedReader(new FileReader(new File(lexuri)));
                    while(r.ready()){
                        line = r.readLine().trim();
                        if (line == null || line.length() == 0) continue;
                        if(calculateDelimiters){
                            updateDelimiterList(foundChars, line);
                        }
                        lexData.addWord(line);
                    }
                } catch(Exception e) {
                    e.printStackTrace();
                    System.err.println("Error reading lexicon " + lex.getLabel());
                }
                matchStrings.put(lex, new HashMap());
                lexicons.add(lex);
                lexicons.add(lexData);
                if(calculateDelimiters){
                    lex.setDelimiters(makeDelimiterString(foundChars));
                }
            }
        }
    }
    
    
    
    /** Clears the match cache. */
    public void clearMatches(){
        for(Iterator i = lexicons.iterator();i.hasNext();){
            Object lexicon = i.next();
            i.next();
            HashMap m;
            m = (HashMap)matchStrings.get(lexicon);
            m.clear();
        }
    }
    
    /** Finds and caches all of the matches for a given string
     * @param input the string to be matched
     */
    public void findMatches(String input){
        String lowerStr = null;
        String str;
        for(Iterator i = lexicons.iterator();i.hasNext();){
            LexiconType lexicon = (LexiconType)i.next();
            LexiconAccessor lexAccessor = (LexiconAccessor)i.next();
            Set matchstrings = new TreeSet();
            if(lexicon.isCaseSensitive()){
                str = input;
            } else {
                if(lowerStr==null)
                    lowerStr = input.toLowerCase();
                str = lowerStr;
            }
            
            for (Iterator j = lexAccessor.getEntries().iterator(); j.hasNext();) {
                String entry = (String) j.next();
                if (!lexicon.isCaseSensitive()) entry = entry.toLowerCase();
                if (str.indexOf(entry) >= 0){ matchstrings.add(entry); }// System.out.println("LEXICON: " + lexicon.getLabel() + ", found match: " + entry); }
            }
            /*
            String delimiters = lexicon.getDelimiters();
            StringTokenizer st = new StringTokenizer(str, delimiters, true);
            
            String word;
            int len;
            int pos = 0;
            int matchLength;
            while (st.hasMoreTokens()){
                word = st.nextToken();
                len = word.length();
                // Separator tokens are returned as 1-character strings.
                if ((len != 1 || delimiters.indexOf(word) < 0)){
                    int j=0;
                    //while(word.length()>0){
                        matchLength = lexAccessor.lookup(word, str, pos+j);
                        if(matchLength>0){
                            matchstrings.add(input.substring(pos+j, matchLength+pos+j));
                        } else System.out.println(word + "***");
                        //word = word.substring(1);
                        //j++;
                    //}
                }
                pos += len;
            }
             */
            ((HashMap)matchStrings.get(lexicon)).put(input, matchstrings);
        }
        
    }
    
    
    /** Determines whether the given label names a lexicon loaded by the
     * <CODE>LexiconMatcher</CODE>
     * @param label The label of the lexicon to find
     * @return whether the lexicon exists
     */
    public boolean exists(String label){
        for(Iterator i = lexicons.iterator();i.hasNext();){
            LexiconType lexicon = (LexiconType)i.next();
            i.next(); //skip the LexiconAccessor
            if(lexicon.getLabel().equals(label)){
                return true;
            }
        }
        //lexicon not found
        return false;
    }
    
    public String toRegEx(LexiconType lex, String input) {
        HashMap m = (HashMap)matchStrings.get(lex);
        if (m == null) return "";
        Set set = (Set)m.get(input);
        if (set == null) return "";
        StringBuffer expr = new StringBuffer();
        expr.append("(");
        boolean pastFirst = false;
        for (Iterator i = set.iterator(); i.hasNext();) {
                if (pastFirst) expr.append("|");
                pastFirst = true;
                String matchVal = (String) i.next();
                // quote string after escaping any accidental '\E' within the string
                // this way each match won't be treated as a regular expression
                expr.append("\\Q" + matchVal.replaceAll("\\\\E", "\\\\\\\\E") + "\\E");
        }
        expr.append(")");
        
        // if no lexicon matches, return an expression that guarantees no matches
        // we do this instead of returning "" because substitution might make
        // an empty string more meaningful than it should be: 
        // "({LexLabel})" would become "()", which matches everything
        if (!pastFirst) return "([a&&[b]])";
        
        return expr.toString();
    }
}
