/*
 * SuffixTreeLexicon.java
 *
 * Created on July 23, 2003, 2:01 PM
 */

package edu.byu.deg.ontos.lexicon;

import edu.byu.deg.framework.LexiconAccessor;
import java.util.HashMap;

/** Node of a dictionary tree representing a lexicon.  Insertion and lookup take O(n).
 * where n is the length of the word.  This implementation of {@link
 * edu.byu.deg.framework.LexiconAccessor LexiconAccessor} is faster than using a
 * binary tree, but uses twice as much memory.
 * @author troywalk
 */
public class SuffixTreeLexicon implements LexiconAccessor{
  private final int MAP_CAPACITY = 1;
  boolean isWord;
  HashMap nextNodes;
  char[] fragment;
  int indexChar;
  int start;
  int length;
  boolean caseSensitive;
  
  static SuffixTreeLexicon tempNode = new SuffixTreeLexicon();
  
  /** Creates a new instance of SuffixTreeLexicon
   */
  public SuffixTreeLexicon(){
    caseSensitive = false;
    isWord = false;
    nextNodes = null;
    length = 0;
    fragment = new char[0];
    indexChar = '\0';
  }
  
  /** Creates a new instance of SuffixTreeLexicon
   * @param word fragment in the dictionary
   * @param caseSensitive whether the lexicon is case sensitive
   */  
  private SuffixTreeLexicon(char[] word, int sta, int len) {
    isWord = true;
    nextNodes = null;
    this.caseSensitive = false;
    length = len;
    fragment = word;
    start = sta;
    indexChar = fragment[start];
  }
    
  /** Creates a new instance of SuffixTreeLexicon
   * @param caseSensitive whether the lexicon is case sensitive
   */  
  public SuffixTreeLexicon(boolean caseSensitive){
    this.caseSensitive = caseSensitive;
    isWord = false;
    nextNodes = null;
    length = 0;
    fragment = new char[0];
    indexChar = '\0';
    start = 0;
  }

  /** Adds a word to the current subtree
   * @param word the word to be added
   */  
  public void addWord(String word){
    if(!caseSensitive)
      word = word.toLowerCase();
    addWord(word.toCharArray(), 0, word.length());
  }

  public void addWord(char[] word, int wordoffset, int wordlength){
    int newLen;
    if(wordlength==0){
      return;
    }
    int matchLength = 0;
    while((matchLength<wordlength)&&(matchLength<length)){
      if(word[wordoffset+matchLength]!=fragment[start+matchLength])
        break;
      matchLength++;
    }
    if(matchLength==wordlength&&matchLength==length){ //the words are equal
      return;                                       //nothing to do
    }
    if(matchLength==wordlength){ //case 1, the new word is a prefix of the old
      newLen = length-matchLength;
      SuffixTreeLexicon newChild = new SuffixTreeLexicon(fragment, matchLength+start, newLen);
      newChild.isWord = isWord;
      newChild.nextNodes = nextNodes;

      nextNodes = new HashMap(MAP_CAPACITY, 1);
      nextNodes.put(newChild, newChild);
      isWord = true;
      
      length = matchLength;
      return;
    }
    if(matchLength==length){  //case 2, the old word is a prefix of the new
      if(nextNodes==null){
        nextNodes = new HashMap(MAP_CAPACITY, 1);
      }
      tempNode.indexChar = word[matchLength+wordoffset];
      SuffixTreeLexicon next = (SuffixTreeLexicon)nextNodes.get(tempNode);
      newLen = wordlength-matchLength;
      if(next==null){
        next = new SuffixTreeLexicon(word, matchLength+wordoffset, newLen);
        nextNodes.put(next, next);
        return;
      }
      next.addWord(word, matchLength+wordoffset, wordlength-matchLength);
      return;
    }
    //otherwise, matchLength<fragment.length and word.length
    newLen = length-matchLength;
    SuffixTreeLexicon newChild = new SuffixTreeLexicon(fragment, matchLength+start, newLen);
    newChild.isWord = isWord;
    newChild.nextNodes = nextNodes;

    nextNodes = new HashMap(MAP_CAPACITY, 1);
    nextNodes.put(newChild, newChild);
    isWord = false;

    length = matchLength;
      
    newLen = wordlength-matchLength;
    newChild = new SuffixTreeLexicon(word, matchLength+wordoffset, newLen);
    nextNodes.put(newChild, newChild);
    return;
  }
  
  /** Looks for a string in the lexicon.
   * @param inputToken minimum word to match
   * @param input the string to look up
   * @param offset position at which to start
   * @return length of the longest match
   */  
  public int lookup(String inputToken, String input, int offset){
    if(!caseSensitive)
      input = input.toLowerCase();
    int matchLength = lookup(this, input.toCharArray());
    return matchLength>=inputToken.length()?matchLength:0;
  }
  
  private static int lookup(SuffixTreeLexicon next, char[] input){
    SuffixTreeLexicon nextnext;
    int inputlength = input.length;
    int offset = 0;
    int bestlength = 0;
    int lengthToNow = 0;
    int inputlen = inputlength;
    while(true){
      if(inputlen<next.length)
        return bestlength;
      for(int i=0;i<next.length;i++)
        if(input[offset+i]!=next.fragment[next.start+i])
          return bestlength;
      lengthToNow += next.length;
      if(next.length==inputlen){
        return next.isWord?lengthToNow:bestlength;
      }
      if(next.nextNodes==null){
        return next.isWord?lengthToNow:bestlength;
      }
      if(next.isWord){
        bestlength = lengthToNow;
      }
      tempNode.indexChar = input[offset+next.length];
      nextnext = (SuffixTreeLexicon)next.nextNodes.get(tempNode);
      if(nextnext==null){
        return next.isWord?lengthToNow:bestlength;
      }
      offset = offset + next.length;
      inputlen = inputlen - next.length;
      next = nextnext;
    }
  }

/*
  private int lookuprecursive(char[] input, int inputlength, int offset){
    int inputlen = inputlength-offset;
    if(inputlen<length)
      return 0;
    for(int i=0;i<length;i++)
      if(input[offset+i]!=fragment[start+i])
        return 0;

    if(length==inputlen){
      return isWord?length:0;
    }
    
    if(nextNodes==null){
      return isWord?length:0;
    }
      
    tempNode.indexChar = input[offset+length];
    SuffixTreeLexicon next = (SuffixTreeLexicon)nextNodes.get(tempNode);

    if(next==null){
      return isWord?length:0;
    }
    int nextLength = next.lookuprecursive(input, inputlength, offset+length);
    if(nextLength==0){
      return isWord?length:0;
    }
    return nextLength + length;
  }
*/

  /** overrides Object equals
   * @param o the object to compare to
   * @return whether it is equal
   */  
  public boolean equals(Object o){
    SuffixTreeLexicon n = (SuffixTreeLexicon)o;
    return n.indexChar==indexChar;
  }
  /** builds hashcode based on index character 
   * @return the hashCode
   */  
  public int hashCode(){
    return indexChar;
  }
  
  public java.util.List getEntries() {
      throw new UnsupportedOperationException();
  }
  
  public int getSize() {
      throw new UnsupportedOperationException();
  }
  
  public java.util.Set findContainedEntries(String input, boolean caseSensitive) {
      throw new UnsupportedOperationException();
  }
  
  public void load(java.net.URL lexURL) {
      throw new UnsupportedOperationException();
  }
  
}
