/*
 * OSMXBasicConnection.java
 *
 * Created on June 11, 2003, 11:43 AM
 */

package edu.byu.deg.osmx;

import edu.byu.deg.osmx.binding.StyleType;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
/**
 * Represents a <CODE>BasicConnection</CODE> element in OSMX.
 * @author Alan Wessman
 */
public class OSMXBasicConnection extends edu.byu.deg.osmx.binding.impl.BasicConnectionImpl implements StyledElement {

    /**
     * Constant to use as a key for the <CODE>PropertyChangeEvent</CODE>.
     */    
  public static final String CONNECTION_PATH_PROPERTY = "connectionPath";
  /**
   * Listener that this element registers on other elements so that this element can
   * react to position changes on the other elements.
   */  
  protected PositionListener posListener;
  /**
   * Listener that this element registers on its child <CODE>Style</CODE> element
   * (if such exists) so it can wrap the <CODE>PropertyChangeEvent</CODE> before
   * passing it along to its listeners.
   */  
  protected PropertyChangeListener styleListener;

  // this is a bit of a hack to get object set deletion events noticed
  /**
   * Flags whether this {@link #setObjectSet} has been invoked for this element.
   */  
  boolean knownObjectSet = false;

  /**
   * Creates a new instance of OSMXBasicConnection.
   */  
  public OSMXBasicConnection() {
    super();
    styleListener = null;
    posListener = new PositionListener() {
      public void updateX(OSMXElement src, int value) {
        firePropertyChange(CONNECTION_PATH_PROPERTY, src, src);
      }

      public void updateY(OSMXElement src, int value) {
        firePropertyChange(CONNECTION_PATH_PROPERTY, src, src);
      }

      public void updateOrder(OSMXElement src, int value) {
        firePropertyChange(CONNECTION_PATH_PROPERTY, src, src);
      }
    };
    OSMXElementList anchorList = this.getInnerList(this.getAnchor());
    anchorList.addListObserver(
      new OSMXElementList.ListObserver() {
        public void elementAdded(OSMXElement elem) {
          elem.addElementDeletionListener(deleteListener);
          ((PositionedElement) elem).addPositionListener(posListener);
          firePropertyChange(CONNECTION_PATH_PROPERTY, null, elem);
        }

        public void elementRemoved(OSMXElement elem) {
          elem.removeElementDeletionListener(deleteListener);
          ((PositionedElement) elem).removePositionListener(posListener);
          firePropertyChange(CONNECTION_PATH_PROPERTY, elem, null);
        }
      }
    );
  }

  /**
   * Overrides the inherited method in order to pass the call recursively down to its
   * child elements.
   * @param elem The new parent of this element.
   */  
  void setParent(OSMXElement elem) {
    parent = elem;
    addToParentDocument();
    this.setAsParentOf((OSMXElement) getStyle());
    this.setAsParentOf(getAnchor());
  }

  /**
   * Gives this element a new <CODE>Style</CODE> element and notifies listeners of a
   * style change.
   * @param value The new <CODE>Style</CODE> element.
   */  
  public void setStyle(StyleType value) {
    StyleType oldVal = getStyle();
    this.replaceChild((OSMXElement) getStyle(), (OSMXElement) value);
    super.setStyle(value);
    if (value instanceof OSMXStyleType) {
      if (styleListener == null) {
        styleListener = new PropertyChangeListener() {
          public void propertyChange(PropertyChangeEvent evt) {
            firePropertyChange(OSMXStyleType.STYLE_PROPERTY,
                               evt.getOldValue(), evt.getNewValue());
          }
        };
          }
      ((OSMXStyleType)value).addPropertyChangeListener(styleListener);
    }
    firePropertyChange(OSMXStyleType.STYLE_PROPERTY, oldVal, value);

  }

  /**
   * Returns the reference to the connected object set or object.  This method also invokes {@link #setObjectSet} if
   * it has not previously been invoked; this is a hack to allow deletion of the object set
   * or object to be detected correctly.
   * @return The <CODE>id</CODE> attribute of the connected object set or object.
   */  
  public String getObjectSet() {
    String result = super.getObjectSet();
    if (!knownObjectSet) {
      knownObjectSet = true;
      setObjectSet(result);
    }
    return result;
  }

  /**
   * Detaches the existing connected object set or object and attaches a new object set or object to this
   * connection.
   * @param value The ID of the new object set or object to connect.
   */  
  public void setObjectSet(String value) {
    OSMXDocument parentDoc = getParentDocument();
    if (parentDoc != null) {
      OSMXElement oldObjSet = parentDoc.getElement(getObjectSet());
      if (oldObjSet != null) {
          if (oldObjSet instanceof OSMXObjectSetType) {
              ((OSMXObjectSetType)oldObjSet).removeConnection(this);
              ((OSMXObjectSetType)oldObjSet).removeElementDeletionListener(deleteListener);
          } else if (oldObjSet instanceof OSMXObjectType) {
              ((OSMXObjectType)oldObjSet).removeConnection(this);
              ((OSMXObjectType)oldObjSet).removeElementDeletionListener(deleteListener);
          }
      }
      OSMXElement objSet = getParentDocument().getElement(value);
      if (objSet != null) {
          if (objSet instanceof OSMXObjectSetType) {
              ((OSMXObjectSetType)objSet).addElementDeletionListener(deleteListener);
              ((OSMXObjectSetType)objSet).addConnection(this);
          } else if (objSet instanceof OSMXObjectType) {
              ((OSMXObjectType)objSet).addElementDeletionListener(deleteListener);
              ((OSMXObjectType)objSet).addConnection(this);
          }
      }
    } else {
      knownObjectSet = false;
    }
    super.setObjectSet(value);
  }

  /**
   * Handles an element deletion event, deleting the associated element if it is a
   * child of this connection (i.e., it is an <CODE>Anchor</CODE> or an
   * <CODE>ObjectSet</CODE> or <CODE>Object</CODE>).
   * @param deletedElement The element associated with the deletion event.
   */  
  protected void handleDeletionEvent(OSMXElement deletedElement) {
    super.handleDeletionEvent(deletedElement);
    if (deletedElement instanceof OSMXAnchorType) {
      getAnchor().remove(deletedElement);
    } else if (deletedElement instanceof OSMXObjectSetType) {
      if (deletedElement.getInternalID().equals(getObjectSet())) {
        delete();
      }
    }
  }

}
