package edu.byu.deg.osmx;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Set;

import edu.byu.deg.osmx.binding.ParticipationConstraintType;
import edu.byu.deg.osmx.binding.StyleType;
import edu.byu.deg.osmx.binding.impl.ChildRelSetConnectionImpl;

public class OSMXChildRelSetConnection extends ChildRelSetConnectionImpl implements StyledElement{
	public static final String PARENT_SIDE_FUNCTIONAL_PROPERTY = "parentSideFunctional";
	public static final String CHILD_SIDE_FUNCTIONAL_PROPERTY = "childSideFunctional";	
	public static final String PARENT_SIDE_OPTIONAL_PROPERTY = "parentSideOptional";
	public static final String CHILD_SIDE_OPTIONAL_PROPERTY = "childSideOptional";	
	public static final String PARENT_SIDE_PARTICIPATION_CONSTRAINT_PROPERTY = "parentSideParticipationConstraint";
	public static final String CHILD_SIDE_PARTICIPATION_CONSTRAINT_PROPERTY = "childSideParticipationConstraint";	
	
	private ParticipationConstraintType parentSideSuppressedParticipationConstraint;
	private ParticipationConstraintType childSideSuppressedParticipationConstraint;
	
	public OSMXChildRelSetConnection(){
		parentSideSuppressedParticipationConstraint = null;
		childSideSuppressedParticipationConstraint = null;
	    styleListener = null;
	}
	private static String[] getParticipationValues(String constraint) {
		String[] result = constraint.split(":");
		return result;
	}
	public String getParentSideUnrefinedParticipationConstraint() {
	      String[] constraints = splitParentSideParticipationRefinements();
	      return constraints[0].trim();
	}
	public String getChildSideUnrefinedParticipationConstraint() {
	      String[] constraints = splitChildSideParticipationRefinements();
	      return constraints[0].trim();
	}	
	private String[] splitParentSideParticipationRefinements() {
		ParticipationConstraintType pc = getParentSideParticipationConstraint();
		if (pc == null) return new String[] {""};
		String rawConstraint = pc.getContent();
		if (rawConstraint == null) return new String[] {""};
		String[] constraints = rawConstraint.trim().split("->");
		return constraints;
	}	
	private String[] splitChildSideParticipationRefinements() {
		ParticipationConstraintType pc = getChildSideParticipationConstraint();
		if (pc == null) return new String[] {""};
		String rawConstraint = pc.getContent();
		if (rawConstraint == null) return new String[] {""};
		String[] constraints = rawConstraint.trim().split("->");
		return constraints;
	}	
	  /** Determines whether the connection represents a functional dependancy.
	   * @return the participation constraints indicate a one-max relationship
	   */
	public boolean isParentSideOneMax(){
		return getParentSideMax()==1;
	}	
	public boolean isChildSideOneMax(){
		return getChildSideMax()==1;
	}	
	
	  public String getParentSideRefinedParticipationConstraint() {
	      String[] constraints = splitParentSideParticipationRefinements();
	      return constraints[constraints.length - 1].trim();
	  }	
	  public String getChildSideRefinedParticipationConstraint() {
	      String[] constraints = splitChildSideParticipationRefinements();
	      return constraints[constraints.length - 1].trim();
	  }		
	  /**
	   * Determines whether the refined participation constraint (or unrefined if no
	   * refined constraint is provided) has a maximum participation of one.
	   * @return <CODE>true</CODE> if the constraint is a max-one; <CODE>false</CODE> otherwise.
	   */  
	public boolean isParentSideRefinedOneMax() {
		return getParentSideRefinedMax() == 1;
	}
	public boolean isChildSideRefinedOneMax() {
		return getChildSideRefinedMax() == 1;
	}
	  /**
	   * Returns the maximum of this connection's participation constraint
	   * @return The maximum number of times an object may be bound to this connection.
	   */
	public int getParentSideMax(){
		String[] pcVals = getParticipationValues(getParentSideUnrefinedParticipationConstraint());
		String maxString = pcVals[pcVals.length - 1];
		try{
			return Integer.parseInt(maxString);
		} catch(Exception e){
			return Integer.MAX_VALUE;
		}
	}	
	public int getChildSideMax(){
		String[] pcVals = getParticipationValues(getChildSideUnrefinedParticipationConstraint());
		String maxString = pcVals[pcVals.length - 1];
		try{
			return Integer.parseInt(maxString);
		} catch(Exception e){
			return Integer.MAX_VALUE;
		}
	}	
	  /**
	   * Returns the minimum participation constraint for this connection.
	   * @return The minimum number of times an object participating in this relationship via
	   * this connection must be bound under the connection.
	   */  
	  public int getParentSideMin() {
	      String[] pcVals = getParticipationValues(getParentSideUnrefinedParticipationConstraint());
	      String minString = pcVals[0];
	      try {
	          return Integer.parseInt(minString);
	      } catch (Exception e) {
	          return Integer.MAX_VALUE;
	      }
	  }
	  public int getChildSideMin() {
	      String[] pcVals = getParticipationValues(getChildSideUnrefinedParticipationConstraint());
	      String minString = pcVals[0];
	      try {
	          return Integer.parseInt(minString);
	      } catch (Exception e) {
	          return Integer.MAX_VALUE;
	      }
	  }
	  
	  
	  /**
	   * Returns the refined maximum participation constraint for this connection,
	   * defaulting to the unrefined constraint if no refinement exists.
	   * @return The maximum number of times that an object may be bound to this connection
	   * together with a unique combination of other objects bound to different
	   * connections in the relationship set.
	   */  
	  public int getParentSideRefinedMax() {
	      String[] pcVals = getParticipationValues(getParentSideRefinedParticipationConstraint());
	      String maxString = pcVals[pcVals.length - 1];
	      try{
	          return Integer.parseInt(maxString);
	      } catch(Exception e){
	          return Integer.MAX_VALUE;
	      }
	  }
	  public int getChildSideRefinedMax() {
	      String[] pcVals = getParticipationValues(getChildSideRefinedParticipationConstraint());
	      String maxString = pcVals[pcVals.length - 1];
	      try{
	          return Integer.parseInt(maxString);
	      } catch(Exception e){
	          return Integer.MAX_VALUE;
	      }
	  }	  
	  /**
	   * Returns the refined minimum participation constraint for this connection,
	   * defaulting to the unrefined minimum if no refinement exists.
	   * @return The minimum number of times an object participating in the relationship set must be
	   * bound to this connection when the relationship is refined by a unique
	   * combination of bound objects in the relationship's other connections.
	   */  
	  public int getParentSideRefinedMin() {
	      String[] pcVals = getParticipationValues(getParentSideRefinedParticipationConstraint());
	      String minString = pcVals[0];
	      try {
	          return Integer.parseInt(minString);
	      } catch (Exception e) {
	          return Integer.MAX_VALUE;
	      }
	  }
	  public int getChildSideRefinedMin() {
	      String[] pcVals = getParticipationValues(getChildSideRefinedParticipationConstraint());
	      String minString = pcVals[0];
	      try {
	          return Integer.parseInt(minString);
	      } catch (Exception e) {
	          return Integer.MAX_VALUE;
	      }
	  }	  
	  /**
	   * Determines whether a refined participation constraint exists for this
	   * connection.
	   * @return <CODE>true</CODE> if there is a <CODE>-></CODE> symbol in the constraint string
	   * followed by a constraint number or range; <CODE>false</CODE> otherwise.
	   */  
	  public boolean hasParentSideRefinedParticipationConstraint() {
	      return (splitParentSideParticipationRefinements().length > 1);
	  }
	  public boolean hasChildSideRefinedParticipationConstraint() {
	      return (splitChildSideParticipationRefinements().length > 1);
	  }
	  /**
	   * Overrides the inherited method in order to pass the call recursively down to its
	   * child elements.
	   * @param elem The new parent of this element.
	   */  
	  void setParent(OSMXElement elem) {
	    parent = elem;
	    addToParentDocument();
	    this.setAsParentOf((OSMXElement) getParentSideParticipationConstraint());
	    this.setAsParentOf((OSMXElement) getChildSideParticipationConstraint());
	    this.setAsParentOf(getAnchor());
	    this.setAsParentOf((OSMXElement) getStyle());
	  }

	  /**
	   * Gives this element a new <CODE>Style</CODE> element and notifies listeners of a
	   * style change.
	   * @param value The new <CODE>Style</CODE> element.
	   */  
	  public void setStyle(StyleType value) {
	    StyleType oldVal = getStyle();
	    this.replaceChild((OSMXElement) getStyle(), (OSMXElement) value);
	    super.setStyle(value);
	    if (value instanceof OSMXStyleType) {
	      if (styleListener == null) {
	        styleListener = new PropertyChangeListener() {
	          public void propertyChange(PropertyChangeEvent evt) {
	            firePropertyChange(OSMXStyleType.STYLE_PROPERTY,
	                               evt.getOldValue(), evt.getNewValue());
	          }
	        };
	          }
	      ((OSMXStyleType)value).addPropertyChangeListener(styleListener);
	    }
	    firePropertyChange(OSMXStyleType.STYLE_PROPERTY, oldVal, value);

	  }

	  /**
	   * Assigns a <CODE>ParticipationConstraint</CODE> element to this connection.
	   * @param value The <CODE>ParticipationConstraint</CODE> to assign.
	   */  
	  public void setParentSideParticipationConstraint(ParticipationConstraintType value) {
	    Object oldVal = getParentSideParticipationConstraint();
	    this.replaceChild((OSMXElement) oldVal, (OSMXElement) value);
	    super.setParentSideParticipationConstraint(value);
	    //Reema modification
	    /*if (value != null) {
	      setFunctional(false);
	      setOptional(false);
	    }
	     */
	    firePropertyChange(PARENT_SIDE_PARTICIPATION_CONSTRAINT_PROPERTY, oldVal, value);
	  }
	  public void setChildSideParticipationConstraint(ParticipationConstraintType value) {
		    Object oldVal = getChildSideParticipationConstraint();
		    this.replaceChild((OSMXElement) oldVal, (OSMXElement) value);
		    super.setChildSideParticipationConstraint(value);
		    //Reema modification
		    /*if (value != null) {
		      setFunctional(false);
		      setOptional(false);
		    }
		     */
		    firePropertyChange(CHILD_SIDE_PARTICIPATION_CONSTRAINT_PROPERTY, oldVal, value);
		  }	  

	  /**
	   * Removes the <CODE>ParticipationConstraint</CODE> element.
	   */  
	  public void unsetParentSideParticipationConstraint() {
	    Object oldVal = getParentSideParticipationConstraint();
	    super.unsetParentSideParticipationConstraint();
	    firePropertyChange(PARENT_SIDE_PARTICIPATION_CONSTRAINT_PROPERTY, oldVal, null);
	  }
	  public void unsetChildSideParticipationConstraint() {
		    Object oldVal = getChildSideParticipationConstraint();
		    super.unsetChildSideParticipationConstraint();
		    firePropertyChange(CHILD_SIDE_PARTICIPATION_CONSTRAINT_PROPERTY, oldVal, null);
	  }
	  /**
	   * Removes the current <CODE>ParticipationConstraint</CODE> element from the
	   * connection but keeps it in memory so it can be restored later if desired.
	   */  
	  public void suppressParentSideParticipationConstraint() {
	    /*if (suppressedParticipationConstraint == null) {
	      suppressedParticipationConstraint = getParticipationConstraint();
	      setParticipationConstraint(null);
	      firePropertyChange(PARTICIPATION_CONSTRAINT_PROPERTY, suppressedParticipationConstraint, null);
	    }*/
	    //if (suppressedParticipationConstraint == null) {
	        parentSideSuppressedParticipationConstraint = getParentSideParticipationConstraint();
	        setParentSideParticipationConstraint(null);
	        firePropertyChange(PARENT_SIDE_PARTICIPATION_CONSTRAINT_PROPERTY, parentSideSuppressedParticipationConstraint, null);
	      //}
	  } // method suppressParticipationConstraint
	  public void suppressChildSideParticipationConstraint() {
		    /*if (suppressedParticipationConstraint == null) {
		      suppressedParticipationConstraint = getParticipationConstraint();
		      setParticipationConstraint(null);
		      firePropertyChange(PARTICIPATION_CONSTRAINT_PROPERTY, suppressedParticipationConstraint, null);
		    }*/
		    //if (suppressedParticipationConstraint == null) {
		        childSideSuppressedParticipationConstraint = getChildSideParticipationConstraint();
		        setChildSideParticipationConstraint(null);
		        firePropertyChange(CHILD_SIDE_PARTICIPATION_CONSTRAINT_PROPERTY, childSideSuppressedParticipationConstraint, null);
		      //}
		  } // method suppressParticipationConstraint	  

	  /**
	   * Restores a <CODE>ParticipationConstraint</CODE> element that was previously
	   * suppressed.
	   */  
	  public void restoreParentSideParticipationConstraint() {
	    if (parentSideSuppressedParticipationConstraint != null && !isSetParentSideParticipationConstraint()) {
	      setParentSideParticipationConstraint(parentSideSuppressedParticipationConstraint);
	      parentSideSuppressedParticipationConstraint = null;
	      firePropertyChange(PARENT_SIDE_PARTICIPATION_CONSTRAINT_PROPERTY, null, getParentSideParticipationConstraint());
	    }
	  } // method restoreParticipationConstraint
	  public void restoreChildSideParticipationConstraint() {
		    if (childSideSuppressedParticipationConstraint != null && !isSetChildSideParticipationConstraint()) {
		      setChildSideParticipationConstraint(childSideSuppressedParticipationConstraint);
		      childSideSuppressedParticipationConstraint = null;
		      firePropertyChange(CHILD_SIDE_PARTICIPATION_CONSTRAINT_PROPERTY, null, getChildSideParticipationConstraint());
		    }
		  }
	  /**
	   * Sets the state of the <CODE>functional</CODE> flag for this connection.
	   * @param value The new value to set: <CODE>true</CODE> if the connection should be set to
	   * functional; <CODE>false</CODE> if it should not.
	   */  
	  public void setParentSideFunctional(boolean value) {
	    /*
	    if (value) {
	      suppressParticipationConstraint();
	    } else if (!isOptional()) {
	      restoreParticipationConstraint();
	    }
	    */
	    // REEMA MODIFICATION   
	    if (value) {
	      restoreParentSideParticipationConstraint();
	    //} else if (!isOptional()) {
	    } else if (!isParentSideOptional()) {
	      restoreParentSideParticipationConstraint();
	    } 
	    //
	    Boolean oldVal = new Boolean(isSetParentSideFunctional() && isParentSideFunctional());
	    Boolean newVal = new Boolean(value);
	    super.setParentSideFunctional(value);
	    firePropertyChange(PARENT_SIDE_FUNCTIONAL_PROPERTY, oldVal, newVal);
	  }
	  public void setChildSideFunctional(boolean value) {
		    /*
		    if (value) {
		      suppressParticipationConstraint();
		    } else if (!isOptional()) {
		      restoreParticipationConstraint();
		    }
		    */
		    // REEMA MODIFICATION   
		    if (value) {
		      restoreChildSideParticipationConstraint();
		    //} else if (!isOptional()) {
		    } else if (!isChildSideOptional()) {		      
		      restoreChildSideParticipationConstraint();
		    } 
		    //
		    Boolean oldVal = new Boolean(isSetChildSideFunctional() && isChildSideFunctional());
		    Boolean newVal = new Boolean(value);
		    super.setChildSideFunctional(value);
		    firePropertyChange(CHILD_SIDE_FUNCTIONAL_PROPERTY, oldVal, newVal);
		  }

	  /**
	   * Clears the current setting for the <CODE>functional</CODE> attribute.
	   */  
	  public void unsetParentSideFunctional() {
	    /*
	    restoreParticipationConstraint();
	    */
	    //REEMA MODIFICATION
	    restoreParentSideParticipationConstraint(); 
	    //***************
	    Boolean oldVal = new Boolean(isSetParentSideFunctional() && isParentSideFunctional());
	    super.unsetParentSideFunctional();
	    firePropertyChange(PARENT_SIDE_FUNCTIONAL_PROPERTY, oldVal, Boolean.TRUE);
	  }
	  public void unsetChildSideFunctional() {
		    /*
		    restoreParticipationConstraint();
		    */
		    //REEMA MODIFICATION
		    restoreChildSideParticipationConstraint(); 
		    //***************
		    Boolean oldVal = new Boolean(isSetChildSideFunctional() && isChildSideFunctional());
		    super.unsetChildSideFunctional();
		    firePropertyChange(CHILD_SIDE_FUNCTIONAL_PROPERTY, oldVal, Boolean.TRUE);
		  }

	  /**
	   * Sets the state of the <CODE>optional</CODE> attribute for this connection.
	   * @param value The new value to set: <CODE>true</CODE> if the connection should be set to
	   * optional; <CODE>false</CODE> if it should not.
	   */  
	  public void setParentSideOptional(boolean value) {
	    /*
	    if (value) {
	      suppressParticipationConstraint();
	    } else if (!isFunctional()) {
	      restoreParticipationConstraint();
	    }
	    */
	    //REEMA MODIFICATION
	    if (value) {
	      restoreParentSideParticipationConstraint(); 
	    //} else if (!isFunctional()) {
	    } else if (!isParentSideFunctional()) {
	      restoreParentSideParticipationConstraint();
	    }  
	      
	    //****************  
	    Boolean oldVal = new Boolean(isSetParentSideOptional() && isParentSideOptional());
	    Boolean newVal = new Boolean(value);
	    super.setParentSideOptional(value);
	    firePropertyChange(PARENT_SIDE_OPTIONAL_PROPERTY, oldVal, newVal);
	  }
	  public void setChildSideOptional(boolean value) {
		    /*
		    if (value) {
		      suppressParticipationConstraint();
		    } else if (!isFunctional()) {
		      restoreParticipationConstraint();
		    }
		    */
		    //REEMA MODIFICATION
		    if (value) {
		      restoreChildSideParticipationConstraint(); 
		    //} else if (!isFunctional()) {
		    } else if (!isChildSideFunctional()) {
		      restoreChildSideParticipationConstraint();
		    }  
		      
		    //****************  
		    Boolean oldVal = new Boolean(isSetChildSideOptional() && isChildSideOptional());
		    Boolean newVal = new Boolean(value);
		    super.setChildSideOptional(value);
		    firePropertyChange(CHILD_SIDE_OPTIONAL_PROPERTY, oldVal, newVal);
		  }

	  /**
	   * Clears the current setting for the <CODE>optional</CODE> attribute.
	   */  
	  public void unsetParentSideOptional() {
	    /*
	    restoreParticipationConstraint();
	    */
	    //REEMA MODIFICATION
	    restoreParentSideParticipationConstraint();  
	    //************************  
	    Boolean oldVal = new Boolean(isSetParentSideOptional() && isParentSideOptional());
	    super.unsetParentSideOptional();
	    firePropertyChange(PARENT_SIDE_OPTIONAL_PROPERTY, oldVal, Boolean.FALSE);
	  }
	  public void unsetChildSideOptional() {
		    /*
		    restoreParticipationConstraint();
		    */
		    //REEMA MODIFICATION
		    restoreChildSideParticipationConstraint();  
		    //************************  
		    Boolean oldVal = new Boolean(isSetChildSideOptional() && isChildSideOptional());
		    super.unsetChildSideOptional();
		    firePropertyChange(CHILD_SIDE_OPTIONAL_PROPERTY, oldVal, Boolean.FALSE);
		  }
		  	  
	  
	  /**
	   * Provides access to the relationship set to which this connection belongs.
	   * @return The relationship set that owns this connection.
	   */  
	  public OSMXAggregationType getRelSet() {
	      return (OSMXAggregationType) getParent();
	  }
	  
	  /**
	   * Returns all connections beside this one that belong to this connection's
	   * relationship set.
	   * @return A <CODE>Set</CODE> containing all <CODE>RelSetConnection</CODE> elements that belong
	   * to the parent relationship set, except this connection element.
	   */  
	  /*public Set getOppositeConnections() {
	      Set result = new HashSet(getRelSet().getRelSetConnection());
	      result.remove(this);
	      return result;
	  }*/
	
}
