/*
 * OSMXDatatypeConverter.java
 *
 * Created on April 12, 2003, 11:58 AM
 */

package edu.byu.deg.osmx;

import java.math.BigInteger;
import javax.xml.bind.DatatypeConverter;
import java.awt.Color;

/** Utility class for JAXB generation of XML binding classes.
 * @author Alan
 */
public class OSMXDatatypeConverter {
  
  /** Converts a String representation of the ExtendedBoolean
   * OSM-X datatype into a Java <CODE>boolean</CODE> primitive.
   * @param s String to convert to a <CODE>boolean</CODE>.
   * @return <CODE>true</CODE> if the input String contains one
   * of the true values for the ExtendedBoolean datatype
   * (comparison is case-insensitive), or false otherwise
   * (including if the String is not a legal ExtendedBoolean
   * value).
   */
  public static boolean parseExtendedBoolean(String s) {
    return (s.equalsIgnoreCase("true") || s.equalsIgnoreCase("y") ||
    s.equals("1") ? true : false );
  }
  
  /** Converts a <CODE>boolean</CODE> primitive into an OSM-X ExtendedBoolean
   * string value.
   * @param b The input value.
   * @return <CODE>&quot;true&quot;</CODE> or <CODE>&quot;false&quot;</CODE> according to the
   * input value.
   */
  public static String printExtendedBoolean(boolean b) {
    return b ? "true" : "false";
  }
  
  /**
   * Converts an XML <CODE>integer</CODE> value to a Java <CODE>int</CODE> value.
   * @param value The XML integer to convert.
   * @return The same value as an <CODE>int</CODE>.
   */  
  public static int parseIntegerToInt(String value) {
    BigInteger result = DatatypeConverter.parseInteger(value);
    return result.intValue();
  }
  
  /**
   * Converts a Java <CODE>int</CODE> to a string satisfying the XML <CODE>integer</CODE>
   * definition.
   * @param value The <CODE>int</CODE> value to convert.
   * @return A string satisfying the XML <CODE>integer</CODE> type definition.
   */  
  public static String printIntToInteger(int value) {
    BigInteger result = BigInteger.valueOf(value);
    return DatatypeConverter.printInteger(result);
  }
  
  private static String toHexString(int value) {
      String hex = Integer.toHexString(value);
      if (hex.length() < 2) {
          hex = "0" + hex;
      }
      return (hex);
  }
  
  /**
   * Converts a {@link java.awt.Color} object to a string representation for the XML
   * storage format of OSMX.
   * @param value The <CODE>Color</CODE> object to obtain the string value for.
   * @return The string representation of the color.  This is a six-character hexadecimal
   * code: RRGGBB.  Thus, alpha transparency is not supported.
   */  
  public static String printColor(Color value) {
    StringBuffer result = new StringBuffer();
    result.append(toHexString(value.getRed()));
    result.append(toHexString(value.getGreen()));
    result.append(toHexString(value.getBlue()));
    return result.toString();
  }
  
  /**
   * Converts a string value into a {@link java.awt.Color} object.
   * @param value The string representation of the <CODE>Color</CODE>.
   * @return A <CODE>Color</CODE> represented by the string (which must be encoded as six
   * hexadecimal values: RRGGBB), or <CODE>null</CODE> if the string cannot be
   * converted into an allowed <CODE>Color</CODE>.
   */  
  public static Color parseColor(String value) {
    try {
      return Color.decode("#" + value);
    } catch (NumberFormatException e) {
      return null;
    }
  }
}
