/*
 * OSMXDocument.java
 *
 * Created on May 31, 2003, 1:13 PM
 */

package edu.byu.deg.osmx;

import java.util.*;
import java.net.URL;
import java.net.URI;
import java.net.URISyntaxException;
import java.io.*;
import edu.byu.deg.osmx.binding.*;

import javax.xml.bind.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import edu.byu.deg.framework.Ontology;

/** Encapsulates the behavior of an XML document that adheres to the OSMX schema.
 * Features include creation of an OSMX document, opening an existing OSMX
 * document, and saving changes.
 * @author Alan Wessman
 */
public class OSMXDocument implements Ontology {
    
    // ctxt: used in JAXB marshalling/unmarshalling
    private static JAXBContext ctxt;
    // marshaller: used to serialize (marshal) Java objects to XML via JAXB
    private static Marshaller marshaller;
    // unmarshaller: used to de-serialize (unmarshal) Java objects from XML via JAXB
    private static Unmarshaller unmarshaller;
    // objectFactory: creates instances of OSMX elements
    private ObjectFactory objectFactory;
    // idGen: generates unique IDs (within this document) for new OSMX elements
    private IDGenerator idGen;
    // docURI: uri of this OSMX document
    private URI docURI;
    
    // string constants used to create JAXB context
    private static final String OSMX_BINDING_PKG = "edu.byu.deg.osmx.binding";
    private static final String OSMX_SCHEMA_LOCATION = "http://www.deg.byu.edu/xml/osmx.xsd http://www.deg.byu.edu/xml/osmx.xsd";
    
    // counter for naming new documents: Untitled-1, Untitled-2, etc.
    private static int newDocNumber = 0;
    
    /** The {@link edu.byu.deg.osmx.binding.OSMType OSM} element that serves as the root
     * of the document hierarchy.
     */
    protected OSM modelRoot;
    /** A mapping of element IDs to the actual elements.  Used as an index to locate
     * elements by their IDs.
     */
    protected Map modelElements;
    /** Flag indicating that a document's state has changed since it was last created,
     * opened, or saved.
     */
    protected boolean modified;
    /** Remembers the file name and path after document has been opened or saved so
     * the document can be saved to the same location later.
     */
    protected String defaultOutputFilename;
    
    //  object registered to receive DocumentChangeEvents
    private List docChangeListeners; 
    // name or title of document (not filename)
    private String title;
    // controls whether to fire DocumentChangeEvent notifications
    // when changes occur; useful when document is unmarshalling
    // and in an unstable state
    // REVISIT: there may be a better strategy for handling this problem
    private boolean generateDocumentChangeEvents = true;
    
    // static initializer just guarantees the current context is null
    // We instantiate the context in the constructor because the factory
    // method requires a classloader, which we reference via the 'this' object.
    static {
        ctxt = null;
    }
    
    /** Creates a new instance of OSMXDocument */
    public OSMXDocument() {
        // guarantee existence of the JAXB context object
        if(ctxt==null){
              try {
                  ctxt = JAXBContext.newInstance(OSMX_BINDING_PKG, this.getClass().getClassLoader());
                  marshaller = ctxt.createMarshaller();
                  unmarshaller = ctxt.createUnmarshaller();
                  //fix for classloader bug in jaxb see
                  // http://forum.java.sun.com/thread.jsp?forum=34&thread=320813&start=45&range=15&hilite=false&q=
              }
              catch (Exception e) {
                  e.printStackTrace();
                  // TODO: better error handling here (this should be a fatal error)
              }
        }
        // create ID generator before doing any JAXB stuff
        idGen = new IDGenerator();
        
        // note custom ObjectFactory subclass - it keeps track of created elements
        // by adding them to the document's id/element index
        objectFactory = new OSMXObjectFactory();
        
        // no content in the document at this point
        modelRoot = null;
        
        // create the id/element index
        modelElements = new HashMap();
        
        // we don't know where on the filesystem the document belongs yet
        defaultOutputFilename = null;
        
        // use LinkedList: primary uses are iteration and adding to list;
        // random access not as important here so LinkedList should perform better
        docChangeListeners = new LinkedList();
        
        // never modified to begin with
        setModified(false);
    } // constructor
    
    /** Locates an {@link OSMXElement } given its internal ID.
     * @param id The ID of the desired element.
     * @return The <CODE>OSMXElement</CODE> with the given ID, or <CODE>null</CODE> if no such
     * object exists in the document's element index.
     */
    public OSMXElement getElement(String id) {
        return (OSMXElement) modelElements.get(id);
    } // method getElement
    
    /** Provides access to the document's index of elements.
     * @return An unmodifiable {@link java.util.Map} of elements indexed by their internal ID.
     */
    public Map getModelElements() {
        return Collections.unmodifiableMap(modelElements);
    } // method getModelElements
    
    /** Returns whether the document has been modified since it was last created,
     * opened, or saved.
     * @return <CODE>true</CODE> if the document has been modified; <CODE>false</CODE>
     * otherwise.
     */
    public boolean isModified() {
        return modified;
    } // method isModified
    
    /** Sets the "modified" flag for the document.
     * @param doModify The intended value of the document modification state.
     */
    protected void setModified(boolean doModify) {
        modified = doModify;
    } // method setModified
    
    /** Adds the specified {@link OSMXElement } to the document to allow later lookup by
     * element ID as well as detection of element property changes and element deletion
     * events.
     * If <CODE>elem</CODE> is null, the method is a no-op.
     * If <CODE>elem</CODE> has no internal ID, a unique one is generated.
     * If another element with the same ID already belongs to the document, <CODE>elem</CODE>
     * keeps its internal ID and a new ID is generated for the other element.
     * The document is registered as an ID change listener, property change listener, and
     * element deletion listener on the added element.
     * @param elem The OSMXElement to add to the document.  The method will perform no action if
     * <CODE>elem</CODE> is <CODE>null</CODE>.
     */
    public void addElement(OSMXElement elem) {
        // can't add a null element
        if (elem == null) return;
        
        // set reference in element back to this document
        if (elem.getParentDocument() != this) elem.setParentDocument(this);
        
        /* This next part is concerned with indexing the element by its ID.
         * If another element already exists in the index under the ID of the
         * added element, we have to give one of them a new ID.
         */
        // determine key for Map
        String id = elem.getInternalID();
        
        // find any existing entry in the Map for the key
        Object idOwner = modelElements.get(id);
        
        // how we place the element in the Map depends on any existing entry
        if (idOwner == elem) {
            // early exit if the element already exists for that ID
            return;
        } else if (idOwner == null) {
            // no entry exists in the map for the given key
            
            // generate an ID if the element doesn't have one already
            if (id == null || id.length() == 0) {
                id = idGen.generateID();
                
                // element gets the generated ID
                elem.setInternalID(id);
            }
            // place the element in the map for the ID
            modelElements.put(id, elem);
        } else {
            // id cannot be null; else we would have a null idOwner
            
            /* We now have an element with the ID attribute set, trying to get the
             * spot in the Map for its ID.  But there's already a different element
             * there.  We assume that elem is the rightful owner of that spot and
             * therefore ought to displace the one that we found in the Map,
             * giving it a new ID and place in the Map.
             * 
             * It seems backwards to make the existing element
             * change its ID instead of the new one, but we do it this way
             * because changing the new element's ID might break any references
             * that another to-be-added element might have to it.
             */
            
            // generate new ID for idOwner to receive; reuse id variable since we
            // have access to the original id value thru elem.
            id = idGen.generateID();
            
            // remove idOwner from the Map
            modelElements.remove(id);
            
            // insert elem into the Map in its rightful place
            modelElements.put(elem.getInternalID(), elem);
            
            // change the ID of idOwner to the new ID and put it in the Map
            ((OSMXElement) idOwner).setInternalID(id);
            modelElements.put(id, idOwner);
        }
        
        /* At this point we are guaranteed that elem is found in the Map under
         * its internalID.  If elem's internalID should change for any reason,
         * this OSMXDocument needs to know about it.  The next part sets up the
         * necessary listeners to handle this.
         */
        
        // register as internal ID change listener on the newly created element
        // we declare this listener as a final local variable so we can refer to it
        // later on in the OSMXElementDeletionListener
        final OSMXElement.InternalIDChangeListener iidcl = new OSMXElement.InternalIDChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                String oldID = (String) evt.getOldValue();
                String newID = (String) evt.getNewValue();
                modelElements.remove(oldID);
                modelElements.put(newID, evt.getSource());
            }
        };
        elem.addInternalIDChangeListener(iidcl);
        
        // register as property change listener so any change to the element is
        // detected as a change to the document, so we can set the modified
        // flag accordingly.
        final PropertyChangeListener pcl = new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                // set modified to true on any property change from elem
                setModified(true);
            }
        };
        elem.addPropertyChangeListener(pcl);
        
        // Having added the element to the document, we need to provide a way to
        // unhook all listeners and remove the element when it is deleted.  The
        // code below performs this step.
        
        // register as deletion listener on the newly created element
        elem.addElementDeletionListener(new OSMXElement.OSMXElementDeletionListener() {
            public void elementDeleted(OSMXElement elem) {
                String id = elem.getInternalID();
                modelElements.remove(id);
                
                // iidcl is the OSMXElement.InternalIDChangeListener declared locally above
                elem.removeInternalIDChangeListener(iidcl);
                
                // pcl is the PropertyChangeListener declared locally
                elem.removePropertyChangeListener(pcl);
                
                // element was deleted so set dirty flag
                setModified(true);
            }
        }); // addElementDeletionListener call
        
        // At this point the element has been added to the document with all
        // necessary listeners to detect ID changes or deletions.  Now we just need
        // to let any DocumentChangeListeners know that the document state has
        // changed.
        
        // fire DocumentChangeEvent unless such notification is disabled
        if (generateDocumentChangeEvents) {
            DocumentChangeEvent dce = new DocumentChangeEvent(OSMXDocument.this, DocumentChangeEvent.CREATE_ELEMENT_EVENT);
            dce.setElement(elem);
            // copy list in case original list changes during event processing
            List dclList = new LinkedList(docChangeListeners);
            for (Iterator iter = dclList.iterator(); iter.hasNext();) {
                DocumentChangeListener dcl = (DocumentChangeListener) iter.next();
                dcl.stateChanged(dce);
            }
        }
        
    } // end method addElement
    
    /** Sets whether {@link DocumentChangeEvent }s are generated when changes occur in
     * the document.  A new or fully opened document is set to generate these events by
     * default; set to <CODE>false</CODE> to turn off events if they need to be
     * ignored.
     * @param generates <CODE>true</CODE> to allow <CODE>DocumentChangeEvent</CODE>s to be generated;
     * <CODE>false</CODE> otherwise.
     */
    public void setGeneratesDocumentChangeEvents(boolean generates) {
        generateDocumentChangeEvents = generates;
    } // method setGeneratesDocumentChangeEvents
    
    /** Loads the data from the OSMX document at the URL specified.  This version of the
     * method invokes {@link #openDocument(java.io.Reader) } internally.
     * @param uri The location of the OSMX document, either local or on the Internet.
     * @throws IOException Thrown in the event of an I/O error when opening or reading the file.
     * @throws JAXBException Thrown in the event of invalid XML or some other JAXB processing error.
     * @return An <code>OSMXDocument</code> with content unmarshalled from the resource identified by the input URL.
     */
    public static OSMXDocument openDocument(URL uri) throws IOException, JAXBException {
        OSMXDocument doc = openDocument(new InputStreamReader(uri.openStream()));
        try {
            doc.docURI = new URI(uri.toString());
        } catch (URISyntaxException urise) {
            throw new IOException("Invalid URI: " + uri.toString());
        }
        return doc;
    } // method openDocument(URL)
    
    /** Opens an OSMX document given a <CODE>String</CODE> representing a local file
     * path and name.  This version of the method invokes
     * {@link #openDocument(java.io.Reader) } internally.
     * @param filename The local file path and name.  May be relative or absolute.
     * @throws IOException Thrown in the event of an I/O error when opening or reading the file.
     * @throws JAXBException Thrown in the event of invalid XML or some other JAXB processing error.
     * @return An <code>OSMXDocument</code> with content unmarshalled from the file or resource identified by the input string.
     */
    public static OSMXDocument openDocument(String filename) throws IOException, JAXBException {
        OSMXDocument doc = null;
        try {
            doc = openDocument(new URL(filename));
        } catch (java.io.IOException e) {
            doc = openDocument(new File(filename));
        }
        return doc;
    } // method openDocument(String)
    
    /** Factory method that opens an OSMX document from a {@link java.io.Reader }.
     * @param r The Reader from which to read the OSMX data.
     * @throws JAXBException Thrown in the event of invalid XML or some other JAXB processing error.
     * @return A new <code>OSMXDocument</code> object with content unmarshalled from the <code>Reader</code>.
     */
    public static OSMXDocument openDocument(Reader r) throws JAXBException {
        OSMXDocument doc = new OSMXDocument();
        
        // don't bother firing events since nobody's listening
        doc.setGeneratesDocumentChangeEvents(false);
        
        // the u.unmarshal call does the work of turning XML into Java objects
        // REVISIT: we should explicitly check that the object returned
        // by the unmarshaller is an 'OSM' element, since the schema
        // itself does not specify a particular element as root.
        // If the unmarshaller does not return an OSM element, we should
        // throw a custom exception.
        doc.setModelRoot((OSM) unmarshaller.unmarshal(new org.xml.sax.InputSource(r)));
        
        // get the document title from the name stored in the document
        doc.setTitle(doc.modelRoot.getOntologyName());
        
        doc.docURI = null;
        
        // re-enable document change events
        doc.setGeneratesDocumentChangeEvents(true);
        
        return doc;
    } // method openDocument(Reader)
    
    /** Factory method that opens an OSMX document from the specified {@link java.io.File}.  This version of the method invokes
     * {@link #openDocument(java.io.Reader) } internally.
     * @param f The <CODE>File</CODE> containing the OSMX document.
     * @throws IOException Thrown in the event of an I/O error when opening or reading the file.
     * @throws JAXBException Thrown in the event of invalid XML or some other JAXB processing error.
     * @return An <code>OSMXDocument</code> with content unmarshalled from the input <code>File</code>.
     */
    public static OSMXDocument openDocument(File f) throws IOException, JAXBException {
        OSMXDocument doc = openDocument(new FileReader(f));
        // remember where we opened the document from so we can save to it later
        doc.setOutputFilename(f.getAbsolutePath());
        doc.docURI = f.toURI();
        return doc;
    } // method openDocument(File)
    
    /** Creates a valid but (mostly) empty OSMX document, ready for adding new elements.
     * The new document contains only a root {@link edu.byu.deg.osmx.binding.OSM }
     * element.  Calling {@link #isModified} after this call will return
     * <CODE>false</CODE>.
     * @throws JAXBException Thrown if there is a problem creating the root <CODE>OSM</CODE> element.
     * @return A new <code>OSMXDocument</code> with no content except for the root element.
     */
    public static OSMXDocument newDocument() throws JAXBException {
        OSMXDocument doc = new OSMXDocument();
        
        // don't bother firing events because nobody's listening yet
        doc.setGeneratesDocumentChangeEvents(false);
        
        // satisfy postcondition that the document is valid OSMX if saved
        doc.setModelRoot(doc.objectFactory.createOSM());
        
        // we don't want users to be prompted to save if an untouched document is closed
        doc.setModified(false);
        
        // no URI for a newly created document
        doc.docURI = null;
        
        // enable document change event firing
        doc.setGeneratesDocumentChangeEvents(true);
        
        return doc;
    } // method newDocument
    
    /**
     * Stores the name of the file to which the OSMX document should be saved.
     * @param filename The destination filename for the document.
     */    
    public void setOutputFilename(String filename) {
        defaultOutputFilename = filename;
    }
    
    /** Marshals the in-memory OSMX objects and writes them to a file.
     * @throws UnspecifiedOutputException Thrown when the document has not been opened from or saved to a local file yet.
     * In such a case, there is no way to know where to save to and no way to prompt
     * (since there is no UI at this level).
     * @throws JAXBException Thrown when a JAXB marshalling error occurs.  The XML object tree may be
     * invalid, or an unusual processing error could have occurred.
     * @throws IOException Thrown when the output file cannot be written for some reason.
     */
    public void saveDocument() throws UnspecifiedOutputException, JAXBException, IOException {
        // we can't save without knowing the destination
        if (defaultOutputFilename == null || defaultOutputFilename.equals("")) {
            throw new UnspecifiedOutputException("No output destination specified.");
        }
        // set marshaller to indent the XML properly
        marshaller.setProperty( Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE );
        // set marshaller to write XML to reference online schema to enable validation
        marshaller.setProperty( Marshaller.JAXB_SCHEMA_LOCATION, OSMX_SCHEMA_LOCATION );
        File outFile = new File(defaultOutputFilename);
        marshaller.marshal(modelRoot, new FileWriter(outFile));
        
        // not modified anymore
        setModified(false);
    } // method saveDocument
    
    /** Writes the OSMX document to the specified {@link java.io.Writer}.
     * @param w The output <CODE>Writer</CODE>.
     * @throws JAXBException Thrown if the JAXB marshalling process fails for some reason.
     */
    public void saveAsDocument(Writer w) throws JAXBException {
        // set marshaller to indent the XML properly
        marshaller.setProperty( Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE );
        // set marshaller to write XML to reference online schema to enable validation
        marshaller.setProperty( Marshaller.JAXB_SCHEMA_LOCATION, OSMX_SCHEMA_LOCATION );
        marshaller.marshal(modelRoot, w);
        
        // not modified anymore
        setModified(false);
    } // saveAsDocument(Writer)
    
    /** Writes the OSMX document to the specified {@link java.io.File}.
     * @param f The <CODE>File</CODE> to save the document to.
     * @throws IOException Thrown if the <CODE>Writer</CODE> encounters an error performing the output.
     * @throws JAXBException Thrown if the JAXB marshalling process fails for some reason.
     */
    public void saveAsDocument(File f) throws IOException, JAXBException {
        setTitle(f.getName());
        saveAsDocument(new FileWriter(f));
        setOutputFilename(f.getAbsolutePath());
        docURI = f.toURI();
    } // saveAsDocument(File)
    
    /** Writes the OSMX document to the specified file path and name.
     * @param filename Represents the path and name of the file to save the document to.
     * @throws IOException Thrown if the <CODE>Writer</CODE> encounters an error performing the output.
     * @throws JAXBException Thrown if the JAXB marshalling process fails for some reason.
     */
    public void saveAsDocument(String filename) throws IOException, JAXBException {
        saveAsDocument(new File(filename));
    } // saveAsDocument(String)
    
    /** Returns an {@link edu.byu.deg.osmx.binding.ObjectFactory } instance that can be
     * used to create OSMX elements; the factory also automatically adds the elements
     * to the <CODE>OSMXDocument</CODE>.
     * @return The document's <CODE>ObjectFactory</CODE> instance.
     */
    public ObjectFactory getObjectFactory() {
        return objectFactory;
    } // method getObjectFactory
    
    /** Returns the title of the document.
     * @return The title of the document; will be <CODE>"Untitled-<I>n</I>"</CODE> if no title has been specified
     * (where <I>n</I> is the number of untitled documents that have been created).
     */
    public String getTitle() {
        if (title == null) {
            // create an "Untitled-n" type of title
            newDocNumber++;
            setTitle("Untitled-" + newDocNumber);
        }
        return title;
    }
    
    /**
     * Returns the location of this OSMX document as a URI.  This will be
     * <CODE>null</CODE> if the document has been newly created but not yet saved, or
     * was opened from a stream rather than a local or remote file source.
     * @return The location of this document, or <CODE>null</CODE> if the location is not known
     * or has not been specified.
     */    
    public URI getURI() {
        return docURI;
    }
    
    /** Sets the title of the document.
     * @param docTitle The new title of the document.
     */
    public void setTitle(String docTitle) {
        // default to current title if param is null
        if (docTitle == null) docTitle = getTitle();
        
        // set the instance variable
        title = docTitle;
        
        // set the ontology title
        OSM root = getModelRoot();
        if (root != null) {
            String ontologyName = root.getOntologyName();
            if (!title.equals(ontologyName)) root.setOntologyName(title);
        }
    }
    
    /** Accessor for the root element of the document.
     * @return The root-level element of the OSMX document.
     */
    public OSM getModelRoot() {
        return modelRoot;
    } // method getModelRoot
    
    /** Establishes an {@link edu.byu.deg.osmx.binding.OSMType} as the root element of
     * the document.  If the document already has a root element, this method does nothing.
     * @param root The intended root element of the document.
     */
    private void setModelRoot(OSM root) {
        // do nothing if modelRoot has already been set
        if (modelRoot != null) return;
        
        // set the instance variable
        modelRoot = root;
        
        // hook up new model root to document
        if (root != null && root instanceof OSMXOSMType) {
            ((OSMXOSMType) root).setParentDocument(this);
        }
    }
    
    /** Registers an object to receive notification of changes to the document.
     * @param dcl The object to receive document change notifications.
     */
    public void addDocumentChangeListener(DocumentChangeListener dcl) {
        if (!docChangeListeners.contains(dcl)) docChangeListeners.add(dcl);
    } // method addDocumentChangeListener
    
    /** Removes an object from the document change notification system.
     * @param dcl The object that should no longer receive document change notifications.
     */
    public void removeDocumentChangeListener(DocumentChangeListener dcl) {
        docChangeListeners.remove(dcl);
    } // method removeDocumentChangeListener
    
    private void notifyDocumentChangeListeners(int eventType) {
        // copy listener list to avoid concurrent modification problems
        Iterator i = new LinkedList(docChangeListeners).iterator();
        
        // REVISIT: sharing an event object with all listeners... is this good practice?
        DocumentChangeEvent evt = new DocumentChangeEvent(this, eventType);
        while (i.hasNext()) {
            DocumentChangeListener dcl = (DocumentChangeListener) i.next();
            dcl.stateChanged(evt);
        }
    } // method notifyDocumentChangeListeners
    
    /** Forces a {@link DocumentChangeEvent} to be sent to all registered
     * {@link DocumentChangeListener}s in order to guarantee that all listeners have an
     * updated view of the document.
     */
    public void updateModel() {
        notifyDocumentChangeListeners(DocumentChangeEvent.GENERAL_EVENT);
    } // method updateModel
    
    /**
     * Returns the first primary object set in the document if one is found
     * @return The primary object set of the ontology, or <CODE>null</CODE> if no object sets
     * are marked as primary.
     */
    public OSMXObjectSetType getPrimaryObjectSet(){
        // REVISIT: we could cache this result for better performance
        List l = getObjectSets();
        for(Iterator i=l.iterator();i.hasNext();){
            OSMXObjectSetType objset = (OSMXObjectSetType) i.next();
            if (objset.isPrimary()) return objset;
        }
        return null;
    }
    
    /**
     * Convenience method for accessing all of the ontology's object sets in one list.
     * @return A <CODE>List</CODE> of <CODE>ObjectSet</CODE> elements.
     */    
    public List getObjectSets() {
        return getElementsOfType(OSMXObjectSetType.class);
    }
    
    /**
     * Convenience method for accessing all of the ontology's
     * generalization/specialization relationships in one list.
     * @return A <CODE>List</CODE> of all <CODE>GenSpec</CODE> elements.
     */    
    public List getGenSpecs() {
        return getElementsOfType(OSMXGenSpecType.class);
    }
    
    /**
     * Convenience method for accessing all of the ontology's relationship sets in one list.
     * @return A <CODE>List</CODE> of all <CODE>RelationshipSet</CODE> elements.
     */    
    public List getRelSets() {
        return getElementsOfType(OSMXRelationshipSetType.class);
    }
    
    /**
     * Convenience method for accessing all of the ontology's data instances in one list.
     * @return A <CODE>List</CODE> of all <CODE>DataInstance</CODE> elements.
     */    
    public List getDataInstances() {
        return getElementsOfType(OSMXDataInstanceType.class);
    }
    
    private List getElementsOfType(Class type, List modelElements) {
        // REVISIT: cache results in a Map indexed by class for better performance?
        LinkedList result = new LinkedList();
        if (modelElements == null) return Collections.unmodifiableList(result);
        for (Iterator i = modelElements.iterator(); i.hasNext();) {
            Object o = i.next();
            if (type.isInstance(o)) {
                result.add(o);
            }
            
            if (o instanceof ModelContainer) {
                OSMType osm = ((ModelContainer) o).getOSM();
                if (osm != null) {
                    result.addAll(getElementsOfType(type, osm.getAllModelElements()));
                }
            }
        }
        return Collections.unmodifiableList(result);
        
    }
    
    /**
     * Convenience method for accessing all elements of a particular type that belong
     * to the ontology, in one list.
     * @param type The <CODE>Class</CODE> that determines which elements of the ontology to return.
     * @return All elements of the ontology that are an <CODE>instanceof</CODE> the given
     * class.
     */    
    public List getElementsOfType(Class type) {
        if (!OSMXElement.class.isAssignableFrom(type)) return null;
        List allElems = getModelRoot().getAllModelElements();
        List result = getElementsOfType(type, allElems);
        
        return result;
    }
    
    /**
     * Removes all <CODE>DataInstance</CODE> elements from the ontology.
     */    
    public void clearDataInstances(){
      List l = getModelRoot().getAllModelElements();
      for(int i=0;i<l.size();i++){
        if(l.get(i) instanceof DataInstance){
          l.remove(i);
          i--;
        }
      }
    }
    
    public List getSourceDocumentURIs() {
        SortedSet rootURIs = new TreeSet();
        
        // determine set of root documents
        List objSets = getObjectSets();
        for (Iterator i = objSets.iterator(); i.hasNext();) {
            DataFrameType datFr = ((ObjectSetType) i.next()).getDataFrame();
            if (datFr == null || datFr.getValuePhrase() == null) continue;
            List valPhrases = datFr.getValuePhrase();
            if (valPhrases == null) continue;
            for (Iterator j = valPhrases.iterator(); j.hasNext();) {
                DataFrameExpression dfe = ((ValuePhrase) j.next()).getValueExpression();
                if (dfe == null) continue;
                for (Iterator k = dfe.getMatchedText().iterator(); k.hasNext();) {
                    rootURIs.add(((MatchedTextType) k.next()).getDocument());
                }
            }
        }
        
        // remove URIs that have another doc URI as the prefix
        // we can do this in one pass because the set is sorted
        String rootURI = (String) rootURIs.first();
        for (Iterator i = rootURIs.iterator(); i.hasNext();) {
            String docURI = (String) i.next();
            if (!rootURI.equals(docURI)) {
                if (docURI.startsWith(rootURI)) {
                    i.remove();
                } else {
                    rootURI = docURI;
                }
            }
        }   
        return new ArrayList(rootURIs);
    }
    
    /* Inner classes */
    
    /** Exception indicating that an output destination has not been specified for the
     * document when output is attempted.
     */
    public class UnspecifiedOutputException extends Exception {
        /** Standard exception constructor.
         * @param msg Message to associate with the exception.
         */
        public UnspecifiedOutputException(String msg) {
            super(msg);
        }
    } // class UnspecifiedOutputException
    
    private class OSMXObjectFactory extends ObjectFactory {
        
        private void addToDocument(Object obj) {
            OSMXDocument.this.addElement((OSMXElement) obj);
        }
        
        public Relationship createRelationship() throws JAXBException {
            Relationship result = super.createRelationship();
            addToDocument(result);
            return result;
        }
        
        public DataFrameType createDataFrameType() throws JAXBException {
            DataFrameType result = super.createDataFrameType();
            addToDocument(result);
            return result;
        }
        
        public SpecializationConnection createSpecializationConnection() throws JAXBException {
            SpecializationConnection result = super.createSpecializationConnection();
            addToDocument(result);
            return result;
        }
        
        public Macro createMacro() throws JAXBException {
            Macro result = super.createMacro();
            addToDocument(result);
            return result;
        }
        
        public GeneralConstraint createGeneralConstraint() throws JAXBException {
            GeneralConstraint result = super.createGeneralConstraint();
            addToDocument(result);
            return result;
        }
        
        public RealTimeConstraint createRealTimeConstraint() throws JAXBException {
            RealTimeConstraint result = super.createRealTimeConstraint();
            addToDocument(result);
            return result;
        }
        
        public PositionedText createPositionedText() throws JAXBException {
            PositionedText result = super.createPositionedText();
            addToDocument(result);
            return result;
        }
        
        public Parameter createParameter() throws JAXBException {
            Parameter result = super.createParameter();
            addToDocument(result);
            return result;
        }
        
        public KeywordPhraseType createKeywordPhraseType() throws JAXBException {
            KeywordPhraseType result = super.createKeywordPhraseType();
            addToDocument(result);
            return result;
        }
        
        public Transition createTransition() throws JAXBException {
            Transition result = super.createTransition();
            addToDocument(result);
            return result;
        }
        
        public StateType createStateType() throws JAXBException {
            StateType result = super.createStateType();
            addToDocument(result);
            return result;
        }
        
        public ObjectType createObjectType() throws JAXBException {
            ObjectType result = super.createObjectType();
            addToDocument(result);
            return result;
        }
        
        public GenSpecType createGenSpecType() throws JAXBException {
            GenSpecType result = super.createGenSpecType();
            addToDocument(result);
            return result;
        }
        
        public ValuePhraseType createValuePhraseType() throws JAXBException {
            ValuePhraseType result = super.createValuePhraseType();
            addToDocument(result);
            return result;
        }
        
        public MethodType createMethodType() throws JAXBException {
            MethodType result = super.createMethodType();
            addToDocument(result);
            return result;
        }
        
        public StyleType createStyleType() throws JAXBException {
            StyleType result = super.createStyleType();
            addToDocument(result);
            return result;
        }
        
        public RealTimeConstraintType createRealTimeConstraintType() throws JAXBException {
            RealTimeConstraintType result = super.createRealTimeConstraintType();
            addToDocument(result);
            return result;
        }
        
        public RelationshipSetType createRelationshipSetType() throws JAXBException {
            RelationshipSetType result = super.createRelationshipSetType();
            addToDocument(result);
            return result;
        }
        
        public GeneralizationConnection createGeneralizationConnection() throws JAXBException {
            GeneralizationConnection result = super.createGeneralizationConnection();
            addToDocument(result);
            return result;
        }
        
        public TransitionType createTransitionType() throws JAXBException {
            TransitionType result = super.createTransitionType();
            addToDocument(result);
            return result;
        }
        
        public ObjectSet createObjectSet() throws JAXBException {
            ObjectSet result = super.createObjectSet();
            addToDocument(result);
            return result;
        }
        
        public Style createStyle() throws JAXBException {
            Style result = super.createStyle();
            addToDocument(result);
            return result;
        }
        
        public ParameterType createParameterType() throws JAXBException {
            ParameterType result = super.createParameterType();
            addToDocument(result);
            return result;
        }
        
        public KeywordPhrase createKeywordPhrase() throws JAXBException {
            KeywordPhrase result = super.createKeywordPhrase();
            addToDocument(result);
            return result;
        }
        
        public CoOccurrenceConstraintType createCoOccurrenceConstraintType() throws JAXBException {
            CoOccurrenceConstraintType result = super.createCoOccurrenceConstraintType();
            addToDocument(result);
            return result;
        }
        
        public AnchorType createAnchorType() throws JAXBException {
            AnchorType result = super.createAnchorType();
            addToDocument(result);
            return result;
        }
        
        public DataType createDataType() throws JAXBException {
            DataType result = super.createDataType();
            addToDocument(result);
            return result;
        }
        
        public ObjectBinding createObjectBinding() throws JAXBException {
            ObjectBinding result = super.createObjectBinding();
            addToDocument(result);
            return result;
        }
        
        public ObjectBindingType createObjectBindingType() throws JAXBException {
            ObjectBindingType result = super.createObjectBindingType();
            addToDocument(result);
            return result;
        }
        
        public ObjectSetReference createObjectSetReference() throws JAXBException {
            ObjectSetReference result = super.createObjectSetReference();
            addToDocument(result);
            return result;
        }
        
        public ModelElement createModelElement() throws JAXBException {
            ModelElement result = super.createModelElement();
            addToDocument(result);
            return result;
        }
        
        public LexiconType createLexiconType() throws JAXBException {
            LexiconType result = super.createLexiconType();
            addToDocument(result);
            return result;
        }
        
        public RelSetConnectionType createRelSetConnectionType() throws JAXBException {
            RelSetConnectionType result = super.createRelSetConnectionType();
            addToDocument(result);
            return result;
        }
        
        public Note createNote() throws JAXBException {
            Note result = super.createNote();
            addToDocument(result);
            return result;
        }
        
        public RelationshipSet createRelationshipSet() throws JAXBException {
            RelationshipSet result = super.createRelationshipSet();
            addToDocument(result);
            return result;
        }
        
        public TypeSpecification createTypeSpecification() throws JAXBException {
            TypeSpecification result = super.createTypeSpecification();
            addToDocument(result);
            return result;
        }
        
        public ObjectSetReferenceType createObjectSetReferenceType() throws JAXBException {
            ObjectSetReferenceType result = super.createObjectSetReferenceType();
            addToDocument(result);
            return result;
        }
        
        public ConjunctionConnectionType createConjunctionConnectionType() throws JAXBException {
            ConjunctionConnectionType result = super.createConjunctionConnectionType();
            addToDocument(result);
            return result;
        }
        
        public Association createAssociation() throws JAXBException {
            Association result = super.createAssociation();
            addToDocument(result);
            return result;
        }
        
        public OSM createOSM() throws JAXBException {
            OSM result = super.createOSM();
            addToDocument(result);
            return result;
        }
        
        public Method createMethod() throws JAXBException {
            Method result = super.createMethod();
            addToDocument(result);
            return result;
        }
        
        public ConjunctionType createConjunctionType() throws JAXBException {
            ConjunctionType result = super.createConjunctionType();
            addToDocument(result);
            return result;
        }
        
        public ConjunctionConnection createConjunctionConnection() throws JAXBException {
            ConjunctionConnection result = super.createConjunctionConnection();
            addToDocument(result);
            return result;
        }
        
        public SourceDocument createSourceDocument() throws JAXBException {
            SourceDocument result = super.createSourceDocument();
            addToDocument(result);
            return result;
        }
        
        public Anchor createAnchor() throws JAXBException {
            Anchor result = super.createAnchor();
            addToDocument(result);
            return result;
        }
        
        public DataFrameExpression createDataFrameExpression() throws JAXBException {
            DataFrameExpression result = super.createDataFrameExpression();
            addToDocument(result);
            return result;
        }
        
        public Conjunction createConjunction() throws JAXBException {
            Conjunction result = super.createConjunction();
            addToDocument(result);
            return result;
        }
        
        public DataInstance createDataInstance() throws JAXBException {
            DataInstance result = super.createDataInstance();
            addToDocument(result);
            return result;
        }
        
        public ConditionType createConditionType() throws JAXBException {
            ConditionType result = super.createConditionType();
            addToDocument(result);
            return result;
        }
        
        public MacroType createMacroType() throws JAXBException {
            MacroType result = super.createMacroType();
            addToDocument(result);
            return result;
        }
        
        public AggregationType createAggregationType() throws JAXBException {
            AggregationType result = super.createAggregationType();
            addToDocument(result);
            return result;
        }
        
        public CoOccurrenceConstraint createCoOccurrenceConstraint() throws JAXBException {
            CoOccurrenceConstraint result = super.createCoOccurrenceConstraint();
            addToDocument(result);
            return result;
        }
        
        public DataTypeType createDataTypeType() throws JAXBException {
            DataTypeType result = super.createDataTypeType();
            addToDocument(result);
            return result;
        }
        
        public SourceDocumentType createSourceDocumentType() throws JAXBException {
            SourceDocumentType result = super.createSourceDocumentType();
            addToDocument(result);
            return result;
        }
        
        public AssociationType createAssociationType() throws JAXBException {
            AssociationType result = super.createAssociationType();
            addToDocument(result);
            return result;
        }
        
        public Lexicon createLexicon() throws JAXBException {
            Lexicon result = super.createLexicon();
            addToDocument(result);
            return result;
        }
        
        public State createState() throws JAXBException {
            State result = super.createState();
            addToDocument(result);
            return result;
        }
        
        public RelSetConnection createRelSetConnection() throws JAXBException {
            RelSetConnection result = super.createRelSetConnection();
            addToDocument(result);
            return result;
        }
        
        public ParticipationConstraintType createParticipationConstraint() throws JAXBException {
            ParticipationConstraintType result = super.createParticipationConstraintType();
            addToDocument(result);
            return result;
        }
        
        public DataFrame createDataFrame() throws JAXBException {
            DataFrame result = super.createDataFrame();
            addToDocument(result);
            return result;
        }
        
        public GenSpec createGenSpec() throws JAXBException {
            GenSpec result = super.createGenSpec();
            addToDocument(result);
            return result;
        }
        
        public Aggregation createAggregation() throws JAXBException {
            Aggregation result = super.createAggregation();
            addToDocument(result);
            return result;
        }
        
        public OSMType createOSMType() throws JAXBException {
            OSMType result = super.createOSMType();
            addToDocument(result);
            return result;
        }
        
        public DataInstanceType createDataInstanceType() throws JAXBException {
            DataInstanceType result = super.createDataInstanceType();
            addToDocument(result);
            return result;
        }
        
        public OSMObject createOSMObject() throws JAXBException {
            OSMObject result = super.createOSMObject();
            addToDocument(result);
            return result;
        }
        
        public ObjectSetType createObjectSetType() throws JAXBException {
            ObjectSetType result = super.createObjectSetType();
            addToDocument(result);
            return result;
        }
        
        public RelationshipType createRelationshipType() throws JAXBException {
            RelationshipType result = super.createRelationshipType();
            addToDocument(result);
            return result;
        }
        
        public ValuePhrase createValuePhrase() throws JAXBException {
            ValuePhrase result = super.createValuePhrase();
            addToDocument(result);
            return result;
        }
        
    } // class OSMXObjectFactory
    
    private class IDGenerator {
        private int lastID = 0;
        public IDGenerator() {
            /* default constructor */
        }
        
        public String generateID() {
            while (modelElements.containsKey(constructID(++lastID)));
            return constructID(lastID);
        }
        
        private String constructID(int idNum) {
            return "osmx" + idNum;
        }
    } // class IDGenerator
    
}
