/*
 * OSMXElementList.java
 *
 * Created on June 12, 2003, 10:34 PM
 */

package edu.byu.deg.osmx;

import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Collection;

/**
 * A custom <CODE>List</CODE> implementation used in the OSMX element classes to
 * manage collections of child elements of the same type.  The class provides hooks
 * for receiving notifications when elements are added to or removed from the list.
 * It also provides a way for the child elements to navigate upward to their
 * parent element.
 * @author Alan Wessman
 */
public class OSMXElementList extends java.util.ArrayList {
  
  protected List listObservers;
  protected OSMXElement parent;
  protected OSMXElement.OSMXElementDeletionListener delListener;
  
  /** Creates a new instance of OSMXElementList */
  public OSMXElementList() {
    listObservers = new LinkedList();
    delListener = new OSMXElement.OSMXElementDeletionListener() {
      public void elementDeleted(OSMXElement elem) {
        if (contains(elem)) remove(elem);
      }
    };
  }
  
  /**
   * Creates a new instance of OSMXElementList, initializing it with the contents of
   * an existing collection.
   * @param c The <CODE>Collection</CODE> containing elements used to initialize this list.
   */  
  public OSMXElementList(Collection c) {
    this();
    super.addAll(c);
  }
  
  /**
   * Invokes {@link OSMXElement#setParent} on each element in the list.
   * @param elem The element to set as parent of each element in the list.
   */  
  protected void setParent(OSMXElement elem) {
    parent = elem;
    for (Iterator i = iterator(); i.hasNext();) {
      ((OSMXElement) i.next()).setParent(parent);
    }
  }
  
  /**
   * Returns the parent element of the list's elements.
   * @return The parent of the elements in the list.
   */  
  public OSMXElement getParent() {
    return parent;
  }
  
  /** Appends the specified element to the end of this list.
   *
   * @param o element to be appended to this list.
   * @return <tt>true</tt> (as per the general contract of Collection.add).
   *
   */
  public boolean add(Object o) {
    boolean retValue;
    OSMXElement elem = (OSMXElement) o;
    
    retValue = super.add(elem);
    hookupElement(elem);
    return retValue;
  } // method add(Object)
  
  /**
   * Adds the specified element at the specified index in the list and sets it up to
   * fire notifications when later removed or deleted.  Also fires a notification
   * that an element was added to the list.
   * @param index The index (0-based) at which to add the specified element.
   * @param element The element to add to the list.
   */  
  public void add(int index, Object element) {
    super.add(index, element);
    hookupElement((OSMXElement) element);
  } // method add(int, Object)
  
  /**
   * Centralizes code that sets up the specified element to fire notifications when
   * added, removed, or deleted.
   * @param elem The element that should notify others of the events listed above.
   */  
  protected void hookupElement(OSMXElement elem) {
    elem.setParent(parent);
    elem.addElementDeletionListener(delListener);
    fireAddNotification(elem);
  }
  
  protected void unhookElement(OSMXElement elem) {
      elem.setParent(null);
      elem.removeElementDeletionListener(delListener);
      fireRemoveNotification(elem);
  }
  
  /**
   * Removes the element at the specified position in this list.
   * Shifts any subsequent elements to the left (subtracts one from their
   * indices).
   *
   * @return the element that was removed from the list.
   * @param index the index of the element to removed.
   */
  public Object remove(int index) {
    Object retValue;
    
    retValue = super.remove(index);
    
    unhookElement((OSMXElement) retValue);
    
    return retValue;
  }
  
  /**
   * Removes the element from this list as well as the parent element, and fires a
   * removal notification.
   * @param o The element to remove from the parent element.
   * @return <CODE>true</CODE> if the element was successfully removed; <CODE>false</CODE>
   * otherwise.
   */  
  public boolean remove(Object o) {
    boolean retValue = super.remove(o);
    if (!retValue) return retValue;
    
    unhookElement((OSMXElement) o);
    
    return retValue;
  }
  
  /**
   * Registers a <CODE>ListObserver</CODE> to receive add and remove notifications
   * from this list.
   * @param obs The object to add as a list observer.
   */  
  public void addListObserver(OSMXElementList.ListObserver obs) {
    listObservers.add(obs);
  } // method addListObserver
  
  /**
   * Un-registers a <CODE>ListObserver</CODE> so it no longer receives notifications
   * from this list.
   * @param obs The list observer to un-register.
   */  
  public void removeListObserver(OSMXElementList.ListObserver obs) {
    listObservers.remove(obs);
  } // method removeListObserver
  
  /**
   * Notifies all registered <CODE>ListObserver</CODE>s that an element was added
   * to the list.
   * @param elem The element added to the list.
   */  
  protected void fireAddNotification(OSMXElement elem) {
    // notify listObservers
    for (Iterator i = listObservers.iterator(); i.hasNext(); ) {
      OSMXElementList.ListObserver obs = (OSMXElementList.ListObserver) i.next();
      obs.elementAdded(elem);
    }
  } // method fireAddNotification
  
  /**
   * Notifies all registered <CODE>ListObserver</CODE>s that an element was removed
   * from the list.
   * @param elem The element removed from the list.
   */  
  protected void fireRemoveNotification(OSMXElement elem) {
    // notify listObservers
    for (Iterator i = listObservers.iterator(); i.hasNext(); ) {
      OSMXElementList.ListObserver obs = (OSMXElementList.ListObserver) i.next();
      obs.elementRemoved(elem);
    }
  } // method fireRemoveNotification
  
  /**
   * Interface for classes that need to monitor an <CODE>OSMXElementList</CODE> for
   * additions to or removals from the list.
   */  
  public interface ListObserver {
      /**
       * Invoked by an <CODE>OSMXElementList</CODE> that this <CODE>ListObserver</CODE>
       * is registered on, when an element is added to the list.
       * @param elem The element added to the list.
       */      
    public void elementAdded(OSMXElement elem);
    /**
     * Invoked by an <CODE>OSMXElementList</CODE> that this <CODE>ListObserver</CODE>
     * is registered on, when an element is removed from the list.
     * @param elem The element removed from the list.
     */    
    public void elementRemoved(OSMXElement elem);
  } // interface ListObserver
  
}
