/*
 * OSMXGenSpecType.java
 *
 * Created on June 11, 2003, 11:43 AM
 */

package edu.byu.deg.osmx;

import edu.byu.deg.osmx.binding.StyleType;
import java.util.*;

/**
 * Represents a <CODE>GenSpec</CODE> element in OSMX.
 * @author Alan Wessman
 */
public class OSMXGenSpecType extends edu.byu.deg.osmx.binding.impl.GenSpecTypeImpl
implements StyledElement, PositionedElement {
  
  public static final String UNION_TYPE = "UNION";
  public static final String MUTEX_TYPE = "MUTEX";
  public static final String INTERSECTION_TYPE = "INTERSECTION";
  public static final String PARTITION_TYPE = "PARTITION";
  public static final String NONE_TYPE = "NONE";
  
  public static final String CONSTRAINT_PROPERTY = "genSpecConstraint";
  
  protected List positionListeners;
  
  /**
   * Creates a new instance of OSMXGenSpecType
   */  
  public OSMXGenSpecType() {
    super();
    positionListeners = new ArrayList();
  } // constructor
  
  /**
   * Registers a <CODE>PositionListener</CODE> with this element to detect changes to
   * the <CODE>x</CODE>, <CODE>y</CODE>, and <CODE>order</CODE> attributes.
   * @param pl The <CODE>PositionListener</CODE> to register to receive notifications.
   */  
  public void addPositionListener(PositionListener pl) {
    positionListeners.add(pl);
  } // method addPositionListener
  
  /**
   * Un-registers a <CODE>PositionListener</CODE> so it no longer receives
   * notifications from this element regarding position changes.
   * @param pl The <CODE>PositionListener</CODE> to un-register.
   */  
  public void removePositionListener(PositionListener pl) {
    positionListeners.remove(pl);
  } // method removePositionListener
  
  /**
   * Overrides the inherited method in order to pass the call recursively down to its
   * child elements.
   * @param elem The new parent of this element.
   */  
  void setParent(OSMXElement elem) {
    parent = elem;
    addToParentDocument();
    this.setAsParentOf(getGenConnection());
    this.setAsParentOf(getSpecConnection());
    this.setAsParentOf((OSMXElement) getStyle());
  }

  /**
   * Determines whether the position of this element has been defined.
   * @return <CODE>true</CODE> if the element's position has been defined; <CODE>false</CODE>
   * otherwise.
   */  
  public boolean isSetPosition() {
    return (isSetX() && isSetY());
  }

  /**
   * Gives this element a new <CODE>Style</CODE> element and notifies listeners of a
   * style change.
   * @param value The new <CODE>Style</CODE> element.
   */  
  public void setStyle(StyleType value) {
    this.replaceChild((OSMXElement) getStyle(), (OSMXElement) value);
    super.setStyle(value);
  }
  
  /**
   * Stores a new value for the generalization/specialization constraint for this
   * element.
   * @param value The new value for the constraint.  Permitted values are defined in the OSMX
   * specification.  They are:
   * <pre>
   * UNION
   * INTERSECTION
   * PARTITION
   * MUTEX
   * NONE
   * </pre>
   */  
  public void setGenSpecConstraint(String value) {
    String oldVal = getGenSpecConstraint();
    super.setGenSpecConstraint(value);
    firePropertyChange(CONSTRAINT_PROPERTY, oldVal, value);
  }
  
  /**
   * Removes the current generalization/specialization constraint.
   */  
  public void unsetGenSpecConstraint() {
    String oldVal = getGenSpecConstraint();
    super.unsetGenSpecConstraint();
    firePropertyChange(CONSTRAINT_PROPERTY, oldVal, null);
  }
}
