/*
 * OSMXMethodType.java
 *
 * Created on June 11, 2003, 11:37 AM
 */

package edu.byu.deg.osmx;

import java.util.Iterator;
import edu.byu.deg.osmx.binding.TypeSpecification;

/**
 * Represents a <CODE>Method</CODE> element in OSMX.
 * @author Alan Wessman
 */
public class OSMXMethodType extends edu.byu.deg.osmx.binding.impl.MethodTypeImpl
implements KeywordListOwner {
  public static final String HINT_PROPERTY = "hint";
  public static final String NAME_PROPERTY = "name";
  public static final String LANGUAGE_PROPERTY = "language";
  
  private TypeSpecification suppressedReturnType = null;
  
  /**
   * Constructs and returns a string representation of the method's signature.
   * @return The signature for the method, in the following format:
   * <CODE>method_name([formal_parameter[, formal_parameter]*]): [return_type]</CODE>
   */  
  public String getSignature() {
    String retType = (isSetReturnType() ? getReturnType().toString() : "");
    String dispName = getName();
    String params = "";
    if (isSetParameter()) {
      int j = 0;
      for (Iterator i = getParameter().iterator(); i.hasNext();) {
        OSMXParameterType param = (OSMXParameterType) i.next();
        String paramSig;
        if (param.isSetName()) {
          paramSig = param.toString();
        } else {
          paramSig = param.toString("param" + j);
        }
        if (++j > 1) {
          params += ", ";
        }
        params += paramSig;
      }
    }
    String result = dispName + "(" + params + ")";
    if (retType != null && retType.length() > 0) result += ": " + retType;
    return result;
  }
  
  /**
   * Removes the return type from the method signature but remembers what was removed
   * so it can be restored later if desired, as long as the object remains in memory.
   */  
  public void suppressReturnType() {
    if (suppressedReturnType == null) {
      suppressedReturnType = getReturnType();
      setReturnType(null);
    }
  }
  
  /**
   * Restores the return type to the method signature after it had been suppressed
   * and held in memory.
   */  
  public void restoreReturnType() {
    if (suppressedReturnType != null && !isSetReturnType()) {
      setReturnType(suppressedReturnType);
      suppressedReturnType = null;
    }
  }
  
  /**
   * Assigns a new hint (display-only label) to the method.
   * @param value The new value to assign.
   */  
  public void setHint(String value) {
    String oldVal = getHint();
    String newVal = new String(value); // in case value == null
    if (!newVal.equals(oldVal)) {
      super.setHint(value);
      firePropertyChange(HINT_PROPERTY, oldVal, value);
    }
  } // method setHint
  
  /**
   * Assigns a new name to the method.
   * @param value The new value to assign.
   */  
  public void setName(String value) {
    String oldVal = getName();
    String newVal = new String(value); // in case value == null
    if (!newVal.equals(oldVal)) {
      super.setName(value);
      firePropertyChange(NAME_PROPERTY, oldVal, value);
    }
  } // method setName
  
  /**
   * Assigns a new value for the implementation language of the method.
   * @param value The new value to assign.
   */  
  public void setLanguage(String value) {
    String oldVal = getLanguage();
    String newVal = new String(value); // in case value == null
    if (!newVal.equals(oldVal)) {
      super.setLanguage(value);
      firePropertyChange(LANGUAGE_PROPERTY, oldVal, value);
    }
  } // method setLanguage
  
  /**
   * Overrides the inherited method in order to pass the call recursively down to its
   * child elements.
   * @param elem The new parent of this element.
   */  
  void setParent(OSMXElement elem) {
    parent = elem;
    addToParentDocument();
    this.setAsParentOf((OSMXElement) getReturnType());
    this.setAsParentOf(this.getInnerList(getParameter()));
    this.setAsParentOf(this.getInnerList(getKeywordPhrase()));
  }
  
  /**
   * Assigns a new <CODE>TypeSpecification</CODE> element as the return type for this
   * method.
   * @param value The element to assign as the return type.
   */  
  public void setReturnType(TypeSpecification value) {
    this.replaceChild((OSMXElement) getReturnType(), (OSMXElement) value);
    super.setReturnType(value);
  }
  
}
