/*
 * OSMXModelElement.java
 *
 * Created on June 11, 2003, 11:38 AM
 */

package edu.byu.deg.osmx;

import java.beans.*;

/**
 * Represents any OSMX element that may be a direct child of an <CODE>OSM</CODE>
 * element.
 * @author Alan Wessman
 */
public class OSMXModelElement extends edu.byu.deg.osmx.binding.impl.ModelElementImpl {
  
  private PropertyChangeSupport idChangeListeners;
  
  public static final String ID_PROPERTY = "id";
  
  /** Creates a new instance of OSMXModelElement */
  public OSMXModelElement() {
    idChangeListeners = new PropertyChangeSupport(this);
  } // constructor
  
  /**
   * Accessor for the <CODE>id</CODE> attribute of the element.
   * @return The element's ID.
   */  
  public String getId() {
    if (!this.isSetId()) {
      setId(this.getInternalID());
    }
    return super.getId();
  } // method getId
  
  /**
   * Gives the element a new <CODE>ID</CODE> and fires a notification in case other
   * elements refer to this element via its ID.
   * @param value The new ID to store.
   */  
  public void setId(String value) {
    String oldId = (this.isSetId() ? this.getId() : null);
    super.setId(value);
    setInternalID(value);
    idChangeListeners.firePropertyChange(ID_PROPERTY, oldId, value);
  } // method setId
  
  /**
   * Stores a new value for the internal ID of the element.  This will also set the
   * model ID of the element if the new value is different.
   * @param s The new value to store.
   */  
  public void setInternalID(String s) {
    if (s == null) return;
    if (s.equals(getId())) {
      super.setInternalID(s);
      return;
    }
    setId(s);
  }
  
  /**
   * Accessor for the element's internal ID value, which is guaranteed to be unique
   * in the model.
   * @return The element's internal ID.
   */  
  public String getInternalID() {
    if (this.isSetId()) {
      return getId();
    }
    return super.getInternalID();
  }
  
  /**
   * Registers an <CODE>IdChangeListener</CODE> to receive notifications whenever
   * this element's ID changes.
   * @param idcl The object to receive ID change notifications.
   */  
  public void addIdChangeListener(IdChangeListener idcl) {
    idChangeListeners.addPropertyChangeListener(idcl);
  } // method addIdChangeListener
  
  /**
   * Un-registers an object from receiving ID change notifications from this element.
   * @param idcl The <CODE>IdChangeListener</CODE> to remove from the notification list.
   */  
  public void removeIdChangeListener(IdChangeListener idcl) {
    idChangeListeners.removePropertyChangeListener(idcl);
  } // method removeIdChangeListener
  
  /**
   * Returns the list of <CODE>IdChangeListener</CODE>s registered with this element.
   * @return An array of listeners registered with the element.
   */  
  public IdChangeListener[] getIdChangeListeners() {
    return (IdChangeListener[]) idChangeListeners.getPropertyChangeListeners();
  } // method getIdChangeListeners
  
  /**
   * Represents objects that can register to receive notifications of changes to the
   * ID attribute of an <CODE>OSMXModelElement</CODE>.
   */  
  public interface IdChangeListener extends PropertyChangeListener {
      /* marker interface; no new methods defined */
  } // interface IdChangeListener
  
}
