/*
 * OSMXObjectSetType.java
 *
 * Created on June 11, 2003, 11:43 AM
 */

package edu.byu.deg.osmx;

import java.util.*;
import edu.byu.deg.osmx.binding.OSMType;
import edu.byu.deg.osmx.binding.DataFrameType;
import edu.byu.deg.osmx.binding.StyleType;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

/**
 * Represents an <CODE>ObjectSet</CODE> element in OSMX.
 * @author Alan Wessman
 */
public class OSMXObjectSetType extends edu.byu.deg.osmx.binding.impl.ObjectSetTypeImpl
implements PositionedElement, ScalableElement, StyledElement, TextAlignableElement, ModelContainer {
  
  public static final String DATA_FRAME_PROPERTY = "dataFrame";
  public static final String LEXICAL_PROPERTY = "lexical";
  public static final String CENTER_NAME_PROPERTY = "centerName";
  public static final String READ_ONLY_PROPERTY = "readOnly";
  public static final String PRIMARY_OBJECT_SET_PROPERTY = "primaryObjectSet";
  public static final String CARDINALITY_CONSTRAINT_PROPERTY = "cardinalityConstraint";
  
  protected List positionListeners = new ArrayList();
  protected PropertyChangeListener styleListener;
  private DataFrameType suppressedDataFrame;
  protected Set relSetConnections = new HashSet();
  protected Set genConnections = new HashSet();
  protected Set specConnections = new HashSet();
  protected Set workingInstance = new HashSet();
  
  /** Creates a new instance of OSMXObjectSetType */
  public OSMXObjectSetType() {
    styleListener = null;
    suppressedDataFrame = null;
  } // constructor
  
  /**
   * Registers an object to listen for position changes by this element.
   * @param pl The object to register for position change event notification.
   */  
  public void addPositionListener(PositionListener pl) {
    positionListeners.add(pl);
  } // method addPositionListener
  
  /**
   * Un-registers an object currently listening for position change events.
   * @param pl The object to remove from the notification list for position changes.
   */  
  public void removePositionListener(PositionListener pl) {
    positionListeners.remove(pl);
  } // method removePositionListener
  
  /**
   * Changes the x-coordinate position and causes the position listeners to be
   * notified.
   * @param value The new x-coordinate position of this element.
   */  
  public void setX(int value) {
    super.setX(value);
    for (Iterator i = positionListeners.iterator(); i.hasNext();) {
      PositionListener pl = (PositionListener) i.next();
      pl.updateX(this, value);
    }
  } // method setX
  
  /**
   * Changes the y-coordinate position and causes the position listeners to be
   * notified.
   * @param value The new y-coordinate position of this element.
   */
  public void setY(int value) {
    super.setY(value);
    for (Iterator i = positionListeners.iterator(); i.hasNext();) {
      PositionListener pl = (PositionListener) i.next();
      pl.updateY(this, value);
    }
  } // method setY
  
  public void setWidth(int w) {
	  super.setWidth(w);
	  // TODO notify a listener
  }
  
  public void setHeight(int h){
	  super.setHeight(h);
	  //TODO notify a listener
  }
  /**
   * Changes the value of the <CODE>order</CODE> attribute and notifies the position
   * change listeners.
   * @param value The new value for the z-order of this element.
   */  
  public void setOrder(int value) {
    super.setOrder(value);
    for (Iterator i = positionListeners.iterator(); i.hasNext();) {
      PositionListener pl = (PositionListener) i.next();
      pl.updateOrder(this, value);
    }
  } // method setOrder
  
  /** Returns a list of IDs that correspond to object sets that are directly funtionally
   * determined by this object set.
   * @param osm The OSM context of the object set for finding the relationship sets
   * @return List of object set IDs
   */  
  public List getFDObjectSets(OSMType osm){
    ArrayList oss = new ArrayList();
    List l = osm.getAllModelElements();
    for (Iterator i = l.iterator(); i.hasNext();) {
      Object o = i.next();
      if(o instanceof OSMXRelationshipSetType){
        OSMXRelationshipSetType relset = (OSMXRelationshipSetType)o;
        OSMXRelSetConnectionType c = relset.getConnectionFor(this);
        if ((c!=null) && (c.isOneMax())) {
          for (Iterator j = relset.getRelSetConnection().iterator(); j.hasNext(); ) {
            o = j.next();
            if ((c!=o) && (o instanceof OSMXRelSetConnectionType)) {
              OSMXRelSetConnectionType con = (OSMXRelSetConnectionType) o;
              oss.add(con.getObjectSet());
            }
          }
        }
      }
    }
    return oss;
  }

  /** Returns a list of IDs that correspond to object sets that are directly non-functionally
   * determined by this object set.
   * @param osm The OSM context of the object set for finding the relationship sets
   * @return List of object set IDs
   */  
  public List getNonFDObjectSets(OSMType osm){
    ArrayList oss = new ArrayList();
    List l = osm.getAllModelElements();
    for (Iterator i = l.iterator(); i.hasNext();) {
      Object o = i.next();
      if(o instanceof OSMXRelationshipSetType){
        OSMXRelationshipSetType relset = (OSMXRelationshipSetType)o;
        OSMXRelSetConnectionType c = relset.getConnectionFor(this);
        if((c!=null) && (!c.isOneMax())){
          for(Iterator j = relset.getRelSetConnection().iterator(); j.hasNext(); ) {
            o = j.next();
            if((c!=o) && (o instanceof OSMXRelSetConnectionType)){
              OSMXRelSetConnectionType con = (OSMXRelSetConnectionType)o;
              oss.add(con.getObjectSet());
            }
          }
        }
      }
    }
    return oss;
  }
  
  /**
   * Returns the value of the <CODE>name</CODE> attribute of the object set, if set;
   * otherwise, sets the <CODE>name</CODE> attribute value to the object set's ID and
   * returns it.
   * @return The name of the object set.
   */  
  public String getPrimaryName() {
    String result = this.getId();
    String name = getName();
    if (name == null || name.length() == 0) {
      setName(result);
      return result;
    }
    return name;
  }
  
  /**
   * Removes the currently stored <CODE>DataFrame</CODE> element from the object set
   * but keeps it in memory so it can be restored later if desired.
   */  
  public void suppressDataFrame() {
    if (suppressedDataFrame == null) {
      suppressedDataFrame = getDataFrame();
      setDataFrame(null);
    }
  } // method suppressDataFrame
  
  /**
   * Restores a previously suppressed <CODE>DataFrame</CODE> element to the object
   * set to which it originally belonged.
   */  
  public void restoreDataFrame() {
    if (suppressedDataFrame != null && !isSetDataFrame()) {
      setDataFrame(suppressedDataFrame);
      suppressedDataFrame = null;
    }
  } // method restoreDataFrame
  
  /**
   * Assigns a <CODE>DataFrame</CODE> element to this object set.
   * @param value The data frame element to assign.
   */  
  public void setDataFrame(DataFrameType value) {
    DataFrameType oldVal = getDataFrame();
    this.replaceChild((OSMXElement) oldVal, (OSMXElement) value);
    super.setDataFrame(value);
    firePropertyChange(DATA_FRAME_PROPERTY, oldVal, value);
  } // method setDataFrame
  
  /**
   * Expresses the object set as a string.
   * @return The string representation of the object set, which is the string returned by
   * {@link #getPrimaryName}.
   */  
  public String toString() {
    return getPrimaryName();
  } // method toString
  
  /**
   * Sets whether the object set is lexical or nonlexical.
   * @param value The value to set: <CODE>true</CODE> if the object set should be lexical;
   * <CODE>false</CODE> if not.
   */  
  public void setLexical(boolean value) {
    Boolean oldVal = new Boolean(isSetLexical() && isLexical());
    Boolean newVal = new Boolean(value);
    super.setLexical(value);
    firePropertyChange(LEXICAL_PROPERTY, oldVal, newVal);
  }
  
  /**
   * Clears the current setting for the <CODE>lexical</CODE> attribute.
   */  
  public void unsetLexical() {
    Boolean oldVal = new Boolean(isSetLexical() && isLexical());
    super.unsetLexical();
    if (oldVal.booleanValue()) {
      firePropertyChange(LEXICAL_PROPERTY, oldVal, Boolean.FALSE);
    }
  }

  public void setCentered(boolean value) {
	  Boolean oldVal = new Boolean(isSetCentered() && isCentered());
	  Boolean newVal = new Boolean(value);
	  super.setCentered(value);
	  firePropertyChange(CENTER_NAME_PROPERTY, oldVal, newVal);
  }
  
  public void unsetCentered(boolean value) {
	Boolean oldVal = new Boolean(isSetCentered() && isCentered());
	super.unsetCentered();
	if (oldVal.booleanValue()) {
		firePropertyChange(CENTER_NAME_PROPERTY, oldVal, Boolean.FALSE);
	}
  }
  
  /**
   * Sets whether the object set is read-only.
   * @param value If <CODE>true</CODE>, the object set is read-only; if <CODE>false</CODE>, it is
   * not.
   */  
  public void setReadOnly(boolean value) {
    Boolean oldVal = new Boolean(isSetReadOnly() && isReadOnly());
    Boolean newVal = new Boolean(value);
    super.setReadOnly(value);
    firePropertyChange(READ_ONLY_PROPERTY, oldVal, newVal);
  }
  
  /**
   * Clears the current setting for the <CODE>readOnly</CODE> attribute.
   */  
  public void unsetReadOnly() {
    Boolean oldVal = new Boolean(isSetReadOnly() && isReadOnly());
    super.unsetReadOnly();
    if (oldVal.booleanValue()) {
      firePropertyChange(READ_ONLY_PROPERTY, oldVal, Boolean.FALSE);
    }
  }
  
  /**
   * Stores a new value for the <CODE>cardinalityConstraint</CODE> attribute of the
   * object set.
   * @param value The new cardinality constraint to set for the object set.
   */  
  public void setCardinalityConstraint(String value) {
    Comparable oldVal = getCardinalityConstraint();
    super.setCardinalityConstraint(value.toString());
    firePropertyChange(CARDINALITY_CONSTRAINT_PROPERTY, oldVal, value);
  }
  
  /**
   * Clears the current setting for the <CODE>cardinalityConstraint</CODE> attribute.
   */  
  public void unsetCardinalityConstraint() {
    Comparable oldVal = getCardinalityConstraint();
    super.unsetCardinalityConstraint();
    firePropertyChange(CARDINALITY_CONSTRAINT_PROPERTY, oldVal, null);
  }
  
  /**
   * Indicates whether this element has explicitly been given an x and y coordinate
   * position.
   * @return <CODE>true</CODE> if both the <CODE>x</CODE> and <CODE>y</CODE> coordinate
   * attributes have been set; <CODE>false</CODE> otherwise.
   */  
  public boolean isSetPosition() {
    return (isSetX() && isSetY());
  }
  
  /**
   * Overrides the inherited method in order to pass the call recursively down to its
   * child elements.
   * @param elem The new parent of this element.
   */  
  void setParent(OSMXElement elem) {
    parent = elem;
    if (elem != null) addToParentDocument();
    this.setAsParentOf((OSMXElement) getDataFrame());
    this.setAsParentOf((OSMXElement) getOSM());
    this.setAsParentOf((OSMXElement) getStyle());
    initWorkingInstance();
  }

  /**
   * Gives this element a new <CODE>Style</CODE> element and notifies listeners of a
   * style change.
   * @param value The new <CODE>Style</CODE> element.
   */  
  public void setStyle(StyleType value) {
    StyleType oldVal = getStyle();
    this.replaceChild((OSMXElement) getStyle(), (OSMXElement) value);
    super.setStyle(value);
    if (value instanceof OSMXStyleType) {
      if (styleListener == null) {
        styleListener = new PropertyChangeListener() {
          public void propertyChange(PropertyChangeEvent evt) {
            firePropertyChange(OSMXStyleType.STYLE_PROPERTY,
                               evt.getOldValue(), evt.getNewValue());
          }
        };
          }
      ((OSMXStyleType)value).addPropertyChangeListener(styleListener);
    }
    firePropertyChange(OSMXStyleType.STYLE_PROPERTY, oldVal, value);
  }
  
  /**
   * Caches the set of objects belonging to this object set that are stored in the
   * ontology's data instances, for convenience in accessing them via this object
   * set.
   */  
  protected void initWorkingInstance() {
      if (getParentDocument() == null) return;
      List dataInstances = getParentDocument().getElementsOfType(OSMXDataInstanceType.class);
      for (Iterator i = dataInstances.iterator(); i.hasNext();) {
          OSMXDataInstanceType di = (OSMXDataInstanceType) i.next();
          if (di.getDataInstanceElements() == null) continue;
          loadMemberObjects(di.getDataInstanceElements());
      }
  }
  
  /**
   * Caches the objects from the ontology's data instances that belong to this object
   * set, so that they can be accessed via this element.
   * @param dataInstanceElements The <CODE>DataInstance</CODE> elements in the ontology.
   */  
  protected void loadMemberObjects(List dataInstanceElements) {
      for (Iterator i = dataInstanceElements.iterator(); i.hasNext();) {
          Object o = i.next();
          if (o instanceof OSMXDataInstanceType) {
              OSMXDataInstanceType di = (OSMXDataInstanceType) o;
              if (di.getDataInstanceElements() == null) continue;
              loadMemberObjects(di.getDataInstanceElements());
          } else if (o instanceof OSMXObjectType) {
              OSMXObjectType obj = (OSMXObjectType) o;
              if (obj.getObjectSet() == this.getId()) workingInstance.add(obj);
          }
      }
  }
  
  /**
   * Assigns a sub-model to the object set.  This does not depend on or affect the
   * state of the object set's <CODE>highLevel</CODE> attribute.
   * @param value The <CODE>OSM</CODE> element that is the root of the sub-model to add to the
   * object set.
   */  
  public void setOSM(OSMType value) {
    this.replaceChild((OSMXElement) getOSM(), (OSMXElement) value);
    super.setOSM(value);
  }
  
  /**
   * Sets whether the object set is the primary object set of the ontology.
   * @param value If <CODE>true</CODE>, the object set is marked as the primary object set; if
   * <CODE>false</CODE>, it is a non-primary object set.  Note that if other object
   * sets have also been marked as the primary object set, setting this value may not
   * necessarily mean that querying the ontology for "the" primary object set will
   * return this one.
   */  
  public void setPrimary(boolean value) {
    Boolean oldVal = new Boolean(isSetPrimary() && isPrimary());
    Boolean newVal = new Boolean(value);
    super.setPrimary(value);
    firePropertyChange(PRIMARY_OBJECT_SET_PROPERTY, oldVal, newVal);
  }
  
  /**
   * Clears the current setting for the <CODE>primary</CODE> attribute.
   */  
  public void unsetPrimary() {
    Boolean oldVal = new Boolean(isSetPrimary() && isPrimary());
    super.unsetPrimary();
    if (oldVal.booleanValue()) {
      firePropertyChange(PRIMARY_OBJECT_SET_PROPERTY, oldVal, Boolean.FALSE);
    }
  }
  
  /**
   * Stores a reference to a connection attached to this object set, in order to
   * provide access via this object set to connected elements such as relationship
   * sets, generalization/specialization relationships, and other object sets.
   * @param conn The connection to add to the connection cache.
   */  
  public void addConnection(OSMXBasicConnection conn) {
      if (conn instanceof OSMXRelSetConnectionType) {
          relSetConnections.add(conn);
      } else if (conn instanceof OSMXGeneralizationConnection) {
          genConnections.add(conn);
      } else if (conn instanceof OSMXSpecializationConnection) {
          specConnections.add(conn);
      }
  }
  
  /**
   * Removes a connection from the connection cache, usually because it is no longer
   * attached to the object set.
   * @param conn The connection to remove.
   */  
  public void removeConnection(OSMXBasicConnection conn) {
      if (conn instanceof OSMXRelSetConnectionType) {
          relSetConnections.remove(conn);
      } else if (conn instanceof OSMXGeneralizationConnection) {
          genConnections.remove(conn);
      } else if (conn instanceof OSMXSpecializationConnection) {
          specConnections.remove(conn);
      }
  }
  
  /**
   * Returns the set of objects that currently belong to this object set.
   * @return A <CODE>Set</CODE> of <CODE>Object</CODE> elements.
   */  
  public Set getWorkingInstance() {
      return workingInstance;
  }
  
  /**
   * Determines whether the object set's cardinality constraint is satisfied or
   * violated by the number of objects that currently exist in the ontology and that
   * belong to the object set.
   * @return <CODE>true</CODE> if the number of objects for this object set is less than or
   * equal to the number expressed by the cardinality constraint, or if the
   * cardinality constraint cannot be interpreted as a number.  The method returns
   * <CODE>false</CODE> if the cardinality is exceeded by the number of objects for
   * the object set.
   */  
  public boolean isCardinalityConstraintSatisfied() {
      String cardinalityStr = getCardinalityConstraint();
      
      // constraint trivially satisfied if it doesn't exist or is infinite
      if (cardinalityStr == null) return true;
      if (cardinalityStr.equals("*")) return true;
      
      try {
          int cardinality = Integer.parseInt(cardinalityStr);
          return (getCardinality() <= cardinality);
      } catch (NumberFormatException e) {
          return true;
      }
  }
  
  /**
   * Counts the number of objects that belong to the object set and are currently
   * stored in the ontology.
   * @return The number of objects for this object set.
   */  
  public int getCardinality() {
      return getAllMemberObjects().size();
  }
  
  /**
   * Provides access to the collection of objects that belong to this object set and
   * that are stored in the ontology.
   * @return The set of objects belonging to this object set.
   */  
  public Set getAllMemberObjects() {
      Set result = new HashSet();
      result.addAll(getWorkingInstance());
      for (Iterator i = getSpecializations().iterator(); i.hasNext();) {
          OSMXObjectSetType child = (OSMXObjectSetType) i.next();
          result.addAll(child.getAllMemberObjects());
      }
      return Collections.unmodifiableSet(result);
  }
  
  /**
   * Returns the set of all relationship sets that are connected to object sets in
   * the generalization ancestry of this object set.
   * @return A <CODE>Set</CODE> of <CODE>RelationshipSet</CODE> elements that connect to this
   * object set's ancestors.
   */  
  public Set getInheritedRelSets() {
      Set result = new HashSet();
      for (Iterator i = relSetConnections.iterator(); i.hasNext();) {
          OSMXRelSetConnectionType conn = (OSMXRelSetConnectionType) i.next();
          result.add(conn.getParent());
      }
      for (Iterator i = getGeneralizations().iterator(); i.hasNext();) {
          OSMXObjectSetType ancestor = (OSMXObjectSetType) i.next();
          result.addAll(ancestor.getInheritedRelSets());
      }
      return Collections.unmodifiableSet(result);
  }
  
  /**
   * Accessor function for the relationship sets connected to this object set.
   * @return A <CODE>Set</CODE> of relationship sets connected to this object set.
   */  
  public Set getRelSetConnections() {
      return Collections.unmodifiableSet(relSetConnections);
  }
  
  /**
   * Returns the object sets that are direct generalizations of this object set.
   * @return A <CODE>Set</CODE> of <CODE>ObjectSet</CODE> elements that directly generalize
   * this object set.
   */  
  public Set getGeneralizations() {
      Set result = new HashSet();
      for (Iterator i = specConnections.iterator(); i.hasNext();) {
          OSMXSpecializationConnection specConn = (OSMXSpecializationConnection) i.next();
          OSMXGenSpecType genSpec = (OSMXGenSpecType) specConn.getParent();
          List genConns = genSpec.getGenConnection();
          if (genConns == null) continue;
          for (Iterator j = genConns.iterator(); j.hasNext();) {
              OSMXGeneralizationConnection genConn = (OSMXGeneralizationConnection) j.next();
              String objSetRef = genConn.getObjectSet();
              if (objSetRef == null) continue;
              OSMXObjectSetType child = (OSMXObjectSetType) getParentDocument().getElement(objSetRef);
              if (child == null) continue;
              result.add(child);
          }
      }
      return Collections.unmodifiableSet(result);
  }
  
  /**
   * Returns the object sets that are direct specializations of this object set.
   * @return A <CODE>Set</CODE> of <CODE>ObjectSet</CODE> elements that directly specialize
   * this object set.
   */  
  public Set getSpecializations() {
      Set result = new HashSet();
      for (Iterator i = genConnections.iterator(); i.hasNext();) {
          OSMXGeneralizationConnection genConn = (OSMXGeneralizationConnection) i.next();
          OSMXGenSpecType genSpec = (OSMXGenSpecType) genConn.getParent();
          List specConns = genSpec.getSpecConnection();
          if (specConns == null) continue;
          for (Iterator j = specConns.iterator(); j.hasNext();) {
              OSMXSpecializationConnection specConn = (OSMXSpecializationConnection) j.next();
              String objSetRef = specConn.getObjectSet();
              if (objSetRef == null) continue;
              OSMXObjectSetType child = (OSMXObjectSetType) getParentDocument().getElement(objSetRef);
              if (child == null) continue;
              result.add(child);
          }
      }
      return Collections.unmodifiableSet(result);
  }

  private static final String NEWLINE = "\n";
  private static final String ENCODED_NEWLINE = "\\\\n";
  
  /**
   * Returns the name of this object set.
   * @return The name of this object set.
   */  
  public java.lang.String getName() {
      java.lang.String retValue;
      
      retValue = super.getName();
      if (retValue != null) {
          retValue = retValue.replaceAll(ENCODED_NEWLINE, NEWLINE);
      }
      return retValue;
  }
  
  /**
   * Stores a new value for the <CODE>name</CODE> attribute of this object set.
   * @param value The new value to store.
   */  
  public void setName(java.lang.String value) {
      if (value != null) {
          value = value.replaceAll(NEWLINE, ENCODED_NEWLINE);
      }
      super.setName(value);
  }
  
}
