/*
 * OSMXObjectType.java
 *
 * Created on June 11, 2003, 11:43 AM
 */

package edu.byu.deg.osmx;

import java.util.*;
import edu.byu.deg.osmx.binding.PositionedText;

/**
 * Represents an <CODE>Object</CODE> element in OSMX.
 * @author Alan Wessman
 */
public class OSMXObjectType extends edu.byu.deg.osmx.binding.impl.ObjectTypeImpl
implements PositionedElement {
    
    protected List positionListeners;
    protected Set relSetConnections = new HashSet();
    protected Set genConnections = new HashSet();
    protected Set specConnections = new HashSet();
    protected Set workingInstance = new HashSet();
    
    /** Creates a new instance of OSMXObjectType */
    public OSMXObjectType() {
        positionListeners = new ArrayList();
    } // constructor
    
    /**
     * Adds the connection, which should be attached to this object, to a list of
     * attached connections so that elements connected to this object can be accessed
     * via this object.
     * @param conn The connection to add to the cached connection list.
     */
    public void addConnection(OSMXBasicConnection conn) {
        if (conn instanceof OSMXRelSetConnectionType) {
            relSetConnections.add(conn);
        } else if (conn instanceof OSMXGeneralizationConnection) {
            genConnections.add(conn);
        } else if (conn instanceof OSMXSpecializationConnection) {
            specConnections.add(conn);
        }
    }
    
    /**
     * Removes a connection from the connection list, presumably because it is no
     * longer attached to this object.
     * @param conn The connection to remove.
     */
    public void removeConnection(OSMXBasicConnection conn) {
        if (conn instanceof OSMXRelSetConnectionType) {
            relSetConnections.remove(conn);
        } else if (conn instanceof OSMXGeneralizationConnection) {
            genConnections.remove(conn);
        } else if (conn instanceof OSMXSpecializationConnection) {
            specConnections.remove(conn);
        }
    }
    
    /**
     * Registers an object to listen for position changes by this element.
     * @param pl The object to register for position change event notification.
     */
    public void addPositionListener(PositionListener pl) {
        positionListeners.add(pl);
    } // method addPositionListener
    
    /**
     * Un-registers an object currently listening for position change events.
     * @param pl The object to remove from the notification list for position changes.
     */
    public void removePositionListener(PositionListener pl) {
        positionListeners.remove(pl);
    } // method removePositionListener
    
    /**
     * Changes the x-coordinate position and causes the position listeners to be
     * notified.
     * @param value The new x-coordinate position of this element.
     */
    public void setX(int value) {
        super.setX(value);
        for (Iterator i = positionListeners.iterator(); i.hasNext();) {
            PositionListener pl = (PositionListener) i.next();
            pl.updateX(this, value);
        }
    } // method setX
    
    /**
     * Changes the y-coordinate position and causes the position listeners to be
     * notified.
     * @param value The new y-coordinate position of this element.
     */
    public void setY(int value) {
        super.setY(value);
        for (Iterator i = positionListeners.iterator(); i.hasNext();) {
            PositionListener pl = (PositionListener) i.next();
            pl.updateY(this, value);
        }
    } // method setY
    
    /**
     * Changes the value of the <CODE>order</CODE> attribute and notifies the position
     * change listeners.
     * @param value The new value for the z-order of this element.
     */
    public void setOrder(int value) {
        super.setOrder(value);
        for (Iterator i = positionListeners.iterator(); i.hasNext();) {
            PositionListener pl = (PositionListener) i.next();
            pl.updateOrder(this, value);
        }
    } // method setOrder
    
    /**
     * Indicates whether this element has explicitly been given an x and y coordinate
     * position.
     * @return <CODE>true</CODE> if both the <CODE>x</CODE> and <CODE>y</CODE> coordinate
     * attributes have been set; <CODE>false</CODE> otherwise.
     */
    public boolean isSetPosition() {
        return (isSetX() && isSetY());
    }
    
    /**
     * Overrides the inherited method in order to pass the call recursively down to its
     * child elements.
     * @param elem The new parent of this element.
     */
    void setParent(OSMXElement elem) {
        parent = elem;
        addToParentDocument();
        this.setAsParentOf((OSMXElement) getName());
    }
    
    /**
     * Stores a new value for the name of this object.  This does not affect the
     * object's <CODE>ID</CODE> or <CODE>value</CODE> attributes.
     * @param value The new value to store.
     */
    public void setName(PositionedText value) {
        this.replaceChild((OSMXElement) getName(), (OSMXElement) value);
        super.setName(value);
    }
    
    /**
     * Links this object to the object set to which it belongs.
     * @param value The ID of the object set of which this object is a member.
     */
    public void setObjectSet(String value) {
        String oldVal = getObjectSet();
        super.setObjectSet(value);
        OSMXDocument parentDoc = getParentDocument();
        if (parentDoc != null) {
            if (oldVal != null && oldVal != value) {
                OSMXObjectSetType objSet = (OSMXObjectSetType) parentDoc.getElement(oldVal);
                if (objSet != null) objSet.getWorkingInstance().remove(this);
            }
            OSMXObjectSetType objSet = (OSMXObjectSetType) parentDoc.getElement(value);
            if (objSet != null) {
                objSet.getWorkingInstance().add(this);
            }
        }
    }
    
    /**
     * Determines whether this object is equivalent to another.
     * @param o The <CODE>Object</CODE> element to compare to.
     * @return <CODE>true</CODE> if the objects belong to the same object set and have the same
     * ID (if nonlexical) or value (if lexical), and if they have the same value for
     * the <CODE>location</CODE> attribute.  Returns <CODE>false</CODE> otherwise.
     */
    public boolean equals(Object o) {
        if (o == null || !(o instanceof OSMXObjectSetType)) return false;
        OSMXObjectType obj = (OSMXObjectType) o;
        boolean result = true;
        String objSet = this.getObjectSet();
        String val = this.getValue();
        String loc = this.getLocation();
        result = result && (objSet == null ? obj.getObjectSet() == null : objSet.equals(obj.getObjectSet()));
        if (val != null) {
            result = result && val.equals(obj.getValue());
        } else {
            String id = getId();
            result = result && (id == null ? obj.getId() == null : id.equals(obj.getId()));
        }
        result = result && (loc == null ? obj.getLocation() == null : loc.equals(obj.getLocation()));
        return result;
    }
    
}
