/*
 * OSMXPositionedText.java
 *
 * Created on June 11, 2003, 11:43 AM
 */

package edu.byu.deg.osmx;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import edu.byu.deg.osmx.binding.StyleType;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

/**
 * Represents a <CODE>BasicConnection</CODE> element in OSMX.
 * @author Alan Wessman
 */
public class OSMXPositionedText extends edu.byu.deg.osmx.binding.impl.PositionedTextImpl
implements PositionedElement, StyledElement {
    
    protected List textChangeListeners;
    protected List positionListeners;
    protected PropertyChangeListener styleListener;
    
    /** Creates a new instance of OSMXPositionedText */
    public OSMXPositionedText() {
        textChangeListeners = new ArrayList();
        positionListeners = new ArrayList();
    } // constructor
    
    /**
     * Sets the content of the <CODE>PositionedText</CODE> element to the input string.
     * @param s The new text content for this element.
     */
    public void setContent(String s) {
        if (s != null) {
            s = s.replaceAll(NEWLINE, ENCODED_NEWLINE);
        }
        super.setContent(s);
        for (Iterator i = textChangeListeners.iterator(); i.hasNext();) {
            TextChangeListener tcl = (TextChangeListener) i.next();
            tcl.textChanged(s);
        }
    } // method setContent
    
    /**
     * Registers an object to listen for changes to the text content of this element.
     * @param tcl The <CODE>TextChangeListener</CODE> to register to receive text change
     * notifications.
     */
    public void addTextChangeListener(TextChangeListener tcl) {
        textChangeListeners.add(tcl);
    } // method addTextChangeListener
    
    /**
     * Un-registers an object so it no longer receives notice of changes to this
     * element's text content.
     * @param tcl The <CODE>TextChangeListener</CODE> to un-register.
     */
    public void removeTextChangeListener(TextChangeListener tcl) {
        textChangeListeners.remove(tcl);
    } // method removeTextChangeListener
    
    /**
     * Registers an object to listen for position changes by this element.
     * @param pl The object to register for position change event notification.
     */
    public void addPositionListener(PositionListener pl) {
        positionListeners.add(pl);
    } // method addPositionListener
    
    /**
     * Un-registers an object currently listening for position change events.
     * @param pl The object to remove from the notification list for position changes.
     */
    public void removePositionListener(PositionListener pl) {
        positionListeners.remove(pl);
    } // method removePositionListener
    
    /**
     * Changes the x-coordinate position and causes the position listeners to be
     * notified.
     * @param value The new x-coordinate position of this element.
     */
    public void setX(int value) {
        super.setX(value);
        for (Iterator i = positionListeners.iterator(); i.hasNext();) {
            PositionListener pl = (PositionListener) i.next();
            pl.updateX(this, value);
        }
    } // method setX
    
    /**
     * Changes the y-coordinate position and causes the position listeners to be
     * notified.
     * @param value The new y-coordinate position of this element.
     */
    public void setY(int value) {
        super.setY(value);
        for (Iterator i = positionListeners.iterator(); i.hasNext();) {
            PositionListener pl = (PositionListener) i.next();
            pl.updateY(this, value);
        }
    } // method setY
    
    /**
     * Changes the value of the <CODE>order</CODE> attribute and notifies the position
     * change listeners.
     * @param value The new value for the z-order of this element.
     */
    public void setOrder(int value) {
        super.setOrder(value);
        for (Iterator i = positionListeners.iterator(); i.hasNext();) {
            PositionListener pl = (PositionListener) i.next();
            pl.updateOrder(this, value);
        }
    } // method setOrder
    
    /**
     * Indicates whether this element has explicitly been given an x and y coordinate
     * position.
     * @return <CODE>true</CODE> if both the <CODE>x</CODE> and <CODE>y</CODE> coordinate
     * attributes have been set; <CODE>false</CODE> otherwise.
     */
    public boolean isSetPosition() {
        return (isSetX() && isSetY());
    }
    
    /**
     * Overrides the inherited method in order to pass the call recursively down to its
     * child elements.
     * @param elem The new parent of this element.
     */
    void setParent(OSMXElement elem) {
        parent = elem;
        if (parent != null) addToParentDocument();
        this.setAsParentOf((OSMXElement) getStyle());
    }
    
    /**
     * Gives this element a new <CODE>Style</CODE> element and notifies listeners of a
     * style change.
     * @param value The new <CODE>Style</CODE> element.
     */
    public void setStyle(StyleType value) {
        StyleType oldVal = getStyle();
        this.replaceChild((OSMXElement) getStyle(), (OSMXElement) value);
        super.setStyle(value);
        if (value instanceof OSMXStyleType) {
            if (styleListener == null) {
                styleListener = new PropertyChangeListener() {
                    public void propertyChange(PropertyChangeEvent evt) {
                        firePropertyChange(OSMXStyleType.STYLE_PROPERTY,
                        evt.getOldValue(), evt.getNewValue());
                    }
                };
            }
            ((OSMXStyleType)value).addPropertyChangeListener(styleListener);
        }
        firePropertyChange(OSMXStyleType.STYLE_PROPERTY, oldVal, value);
    }
    
    /**
     * Removes the <CODE>Style</CODE> element.
     */
    public void unsetStyle() {
        StyleType oldVal = getStyle();
        this.replaceChild((OSMXElement) oldVal, null);
        super.unsetStyle();
        
        firePropertyChange(OSMXStyleType.STYLE_PROPERTY, oldVal, null);
    }
    
    private static final String NEWLINE = "\n";
    private static final String ENCODED_NEWLINE = "\\\\n";
    
    /**
     * Returns the text content of the element.
     * @return The text content of the element.
     */
    public java.lang.String getContent() {
        java.lang.String retValue;
        
        retValue = super.getContent();
        if (retValue != null) {
            retValue = retValue.replaceAll(ENCODED_NEWLINE, NEWLINE);
        }
        return retValue;
    }
    
}
