/*
 * OSMXRelSetConnectionType.java
 *
 * Created on June 11, 2003, 11:43 AM
 */

package edu.byu.deg.osmx;

import edu.byu.deg.osmx.binding.ParticipationConstraintType;
import edu.byu.deg.osmx.binding.StyleType;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;

/**
 * Represents a <CODE>RelSetConnection</CODE> element in OSMX.
 * @author Alan Wessman
 */
public class OSMXRelSetConnectionType extends edu.byu.deg.osmx.binding.impl.RelSetConnectionTypeImpl implements StyledElement {

  public static final String FUNCTIONAL_PROPERTY = "functional";
  public static final String OPTIONAL_PROPERTY = "optional";
  public static final String PARTICIPATION_CONSTRAINT_PROPERTY = "participationConstraint";

  private ParticipationConstraintType suppressedParticipationConstraint;
  private Map objectBindings;

  /**
   * Creates a new instance of OSMXRelSetConnectionType
   */  
  public OSMXRelSetConnectionType() {
    suppressedParticipationConstraint = null;
    styleListener = null;
    objectBindings = new HashMap();
  }
  
  private static String[] getParticipationValues(String constraint) {
      String[] result = constraint.split(":");
      return result;
  }
  
  /**
   * Accessor for the unrefined participation constraint for this connection.  An
   * unrefined constraint is one that applies to relationships in the connection's
   * relationship set globally.
   * @return The unrefined participation constraint, in string format.
   */  
  public String getUnrefinedParticipationConstraint() {
      String[] constraints = splitParticipationRefinements();
      return constraints[0].trim();
  }
  
  /**
   * Accessor for the refined participation constraint for this connection.  The
   * refined constraint, if present, appears after the <CODE>-></CODE> symbol in the
   * participation constraint string.  If the refined constraint is not present, the
   * unrefined constraint is returned.
   * @return The refined participation constraint.
   */  
  public String getRefinedParticipationConstraint() {
      String[] constraints = splitParticipationRefinements();
      return constraints[constraints.length - 1].trim();
  }
  
  private String[] splitParticipationRefinements() {
      ParticipationConstraintType pc = getParticipationConstraint();
      if (pc == null) return new String[] {""};
      String rawConstraint = pc.getContent();
      if (rawConstraint == null) return new String[] {""};
      String[] constraints = rawConstraint.trim().split("->");
      return constraints;
  }

  /** Determines whether the connection represents a functional dependancy.
   * @return the participation constraints indicate a one-max relationship
   */
  public boolean isOneMax(){
    return getMax()==1;
  }
  
  /**
   * Determines whether the refined participation constraint (or unrefined if no
   * refined constraint is provided) has a maximum participation of one.
   * @return <CODE>true</CODE> if the constraint is a max-one; <CODE>false</CODE> otherwise.
   */  
  public boolean isRefinedOneMax() {
      return getRefinedMax() == 1;
  }

  /**
   * Returns the maximum of this connection's participation constraint
   * @return The maximum number of times an object may be bound to this connection.
   */
  public int getMax(){
      String[] pcVals = getParticipationValues(getUnrefinedParticipationConstraint());
      String maxString = pcVals[pcVals.length - 1];
      try{
          return Integer.parseInt(maxString);
      } catch(Exception e){
          return Integer.MAX_VALUE;
      }
  }
  
  /**
   * Returns the minimum participation constraint for this connection.
   * @return The minimum number of times an object participating in this relationship via
   * this connection must be bound under the connection.
   */  
  public int getMin() {
      String[] pcVals = getParticipationValues(getUnrefinedParticipationConstraint());
      String minString = pcVals[0];
      try {
          return Integer.parseInt(minString);
      } catch (Exception e) {
          return Integer.MAX_VALUE;
      }
  }
  
  /**
   * Returns the refined maximum participation constraint for this connection,
   * defaulting to the unrefined constraint if no refinement exists.
   * @return The maximum number of times that an object may be bound to this connection
   * together with a unique combination of other objects bound to different
   * connections in the relationship set.
   */  
  public int getRefinedMax() {
      String[] pcVals = getParticipationValues(getRefinedParticipationConstraint());
      String maxString = pcVals[pcVals.length - 1];
      try{
          return Integer.parseInt(maxString);
      } catch(Exception e){
          return Integer.MAX_VALUE;
      }
  }
  
  /**
   * Returns the refined minimum participation constraint for this connection,
   * defaulting to the unrefined minimum if no refinement exists.
   * @return The minimum number of times an object participating in the relationship set must be
   * bound to this connection when the relationship is refined by a unique
   * combination of bound objects in the relationship's other connections.
   */  
  public int getRefinedMin() {
      String[] pcVals = getParticipationValues(getRefinedParticipationConstraint());
      String minString = pcVals[0];
      try {
          return Integer.parseInt(minString);
      } catch (Exception e) {
          return Integer.MAX_VALUE;
      }
  }
  
  /**
   * Determines whether a refined participation constraint exists for this
   * connection.
   * @return <CODE>true</CODE> if there is a <CODE>-></CODE> symbol in the constraint string
   * followed by a constraint number or range; <CODE>false</CODE> otherwise.
   */  
  public boolean hasRefinedParticipationConstraint() {
      return (splitParticipationRefinements().length > 1);
  }

  /**
   * Overrides the inherited method in order to pass the call recursively down to its
   * child elements.
   * @param elem The new parent of this element.
   */  
  void setParent(OSMXElement elem) {
    parent = elem;
    addToParentDocument();
    this.setAsParentOf((OSMXElement) getParticipationConstraint());
    this.setAsParentOf(getAnchor());
    this.setAsParentOf((OSMXElement) getStyle());
  }

  /**
   * Gives this element a new <CODE>Style</CODE> element and notifies listeners of a
   * style change.
   * @param value The new <CODE>Style</CODE> element.
   */  
  public void setStyle(StyleType value) {
    StyleType oldVal = getStyle();
    this.replaceChild((OSMXElement) getStyle(), (OSMXElement) value);
    super.setStyle(value);
    if (value instanceof OSMXStyleType) {
      if (styleListener == null) {
        styleListener = new PropertyChangeListener() {
          public void propertyChange(PropertyChangeEvent evt) {
            firePropertyChange(OSMXStyleType.STYLE_PROPERTY,
                               evt.getOldValue(), evt.getNewValue());
          }
        };
          }
      ((OSMXStyleType)value).addPropertyChangeListener(styleListener);
    }
    firePropertyChange(OSMXStyleType.STYLE_PROPERTY, oldVal, value);

  }

  /**
   * Assigns a <CODE>ParticipationConstraint</CODE> element to this connection.
   * @param value The <CODE>ParticipationConstraint</CODE> to assign.
   */  
  public void setParticipationConstraint(ParticipationConstraintType value) {
    Object oldVal = getParticipationConstraint();
    this.replaceChild((OSMXElement) oldVal, (OSMXElement) value);
    super.setParticipationConstraint(value);
    //Reema modification
    /*if (value != null) {
      setFunctional(false);
      setOptional(false);
    }
     */
    firePropertyChange(PARTICIPATION_CONSTRAINT_PROPERTY, oldVal, value);
  }

  /**
   * Removes the <CODE>ParticipationConstraint</CODE> element.
   */  
  public void unsetParticipationConstraint() {
    Object oldVal = getParticipationConstraint();
    super.unsetParticipationConstraint();
    firePropertyChange(PARTICIPATION_CONSTRAINT_PROPERTY, oldVal, null);
  }

  /**
   * Removes the current <CODE>ParticipationConstraint</CODE> element from the
   * connection but keeps it in memory so it can be restored later if desired.
   */  
  public void suppressParticipationConstraint() {
    /*if (suppressedParticipationConstraint == null) {
      suppressedParticipationConstraint = getParticipationConstraint();
      setParticipationConstraint(null);
      firePropertyChange(PARTICIPATION_CONSTRAINT_PROPERTY, suppressedParticipationConstraint, null);
    }*/
    //if (suppressedParticipationConstraint == null) {
        suppressedParticipationConstraint = getParticipationConstraint();
        setParticipationConstraint(null);
        firePropertyChange(PARTICIPATION_CONSTRAINT_PROPERTY, suppressedParticipationConstraint, null);
      //}
  } // method suppressParticipationConstraint

  /**
   * Restores a <CODE>ParticipationConstraint</CODE> element that was previously
   * suppressed.
   */  
  public void restoreParticipationConstraint() {
    if (suppressedParticipationConstraint != null && !isSetParticipationConstraint()) {
      setParticipationConstraint(suppressedParticipationConstraint);
      suppressedParticipationConstraint = null;
      firePropertyChange(PARTICIPATION_CONSTRAINT_PROPERTY, null, getParticipationConstraint());
    }
  } // method restoreParticipationConstraint

  /**
   * Sets the state of the <CODE>functional</CODE> flag for this connection.
   * @param value The new value to set: <CODE>true</CODE> if the connection should be set to
   * functional; <CODE>false</CODE> if it should not.
   */  
  public void setFunctional(boolean value) {
    /*
    if (value) {
      suppressParticipationConstraint();
    } else if (!isOptional()) {
      restoreParticipationConstraint();
    }
    */
    // REEMA MODIFICATION   
    if (value) {
      restoreParticipationConstraint();
    } else if (!isOptional()) {
      restoreParticipationConstraint();
    } 
    //
    Boolean oldVal = new Boolean(isSetFunctional() && isFunctional());
    Boolean newVal = new Boolean(value);
    super.setFunctional(value);
    firePropertyChange(FUNCTIONAL_PROPERTY, oldVal, newVal);
  }

  /**
   * Clears the current setting for the <CODE>functional</CODE> attribute.
   */  
  public void unsetFunctional() {
    /*
    restoreParticipationConstraint();
    */
    //REEMA MODIFICATION
    restoreParticipationConstraint(); 
    //***************
    Boolean oldVal = new Boolean(isSetFunctional() && isFunctional());
    super.unsetFunctional();
    firePropertyChange(FUNCTIONAL_PROPERTY, oldVal, Boolean.TRUE);
  }

  /**
   * Sets the state of the <CODE>optional</CODE> attribute for this connection.
   * @param value The new value to set: <CODE>true</CODE> if the connection should be set to
   * optional; <CODE>false</CODE> if it should not.
   */  
  public void setOptional(boolean value) {
    /*
    if (value) {
      suppressParticipationConstraint();
    } else if (!isFunctional()) {
      restoreParticipationConstraint();
    }
    */
    //REEMA MODIFICATION
    if (value) {
      restoreParticipationConstraint(); 
    } else if (!isFunctional()) {
      restoreParticipationConstraint();
    }  
      
    //****************  
    Boolean oldVal = new Boolean(isSetOptional() && isOptional());
    Boolean newVal = new Boolean(value);
    super.setOptional(value);
    firePropertyChange(OPTIONAL_PROPERTY, oldVal, newVal);
  }

  /**
   * Clears the current setting for the <CODE>optional</CODE> attribute.
   */  
  public void unsetOptional() {
    /*
    restoreParticipationConstraint();
    */
    //REEMA MODIFICATION
    restoreParticipationConstraint();  
    //************************  
    Boolean oldVal = new Boolean(isSetOptional() && isOptional());
    super.unsetOptional();
    firePropertyChange(OPTIONAL_PROPERTY, oldVal, Boolean.FALSE);
  }
  
  /**
   * Provides access to the relationship set to which this connection belongs.
   * @return The relationship set that owns this connection.
   */  
  public OSMXRelationshipSetType getRelSet() {
      return (OSMXRelationshipSetType) getParent();
  }
  
  /**
   * Returns all connections beside this one that belong to this connection's
   * relationship set.
   * @return A <CODE>Set</CODE> containing all <CODE>RelSetConnection</CODE> elements that belong
   * to the parent relationship set, except this connection element.
   */  
  public Set getOppositeConnections() {
      Set result = new HashSet(getRelSet().getRelSetConnection());
      result.remove(this);
      return result;
  }
  
  /**
   * Stores a reference to an <CODE>ObjectBinding</CODE> that acts as an
   * instantiation of this connection.  This is done to enable access to all object
   * bindings governed by this connection, via this connection.
   * @param binding The object binding to add for future access purposes.
   */  
  public void addObjectBinding(OSMXObjectBindingType binding) {
      if (!binding.getConnectionRef().equals(this.getId())) return;
      Map bindMap = (Map) objectBindings.get(binding.getObjectRef());
      if (bindMap == null) bindMap = new HashMap();
      Set bindSet = (Set) bindMap.get(binding.getConnectionRef());
      if (bindSet == null) bindSet = new HashSet();
      bindSet.add(binding);
      bindMap.put(binding.getConnectionRef(), bindSet);
      objectBindings.put(binding.getObjectRef(), bindMap);
  }
  
  /**
   * Returns all object bindings associated with this connection and that bind to the
   * given object.
   * @param objectRef The object whose bindings are to be returned.
   * @return A <CODE>Set</CODE> of <CODE>ObjectBinding</CODE> elements.
   */  
  public Set getObjectBindings(String objectRef) {
     Map bindMap = (Map) objectBindings.get(objectRef);
     if (bindMap == null) return null;
     return (Set) bindMap.get(this.getId());
  }

}
