/*
 * OSMXRelationshipSetType.java
 *
 * Created on June 11, 2003, 11:43 AM
 */

package edu.byu.deg.osmx;

import java.util.*;
import edu.byu.deg.osmx.binding.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

/**
 * Represents a <CODE>RelationshipSet</CODE> element in OSMX.
 * @author Alan Wessman
 */
public class OSMXRelationshipSetType extends edu.byu.deg.osmx.binding.impl.RelationshipSetTypeImpl
implements PositionedElement, StyledElement, ModelContainer {
    
    protected List positionListeners;
    protected PropertyChangeListener styleListener;
    protected Set workingInstance;
    
    /** Creates a new instance of OSMXRelationshipSetType */
    public OSMXRelationshipSetType() {
        super();
        styleListener = null;
        positionListeners = new ArrayList();
        workingInstance = new HashSet();
        final OSMXElementList connectionList = this.getInnerList(this.getRelSetConnection());
        connectionList.addListObserver(
        new OSMXElementList.ListObserver() {
            public void elementAdded(OSMXElement elem) { /* do nothing */ }
            
            public void elementRemoved(OSMXElement elem) {
                if (connectionList.size() < 2) {
                    OSMXRelationshipSetType.this.delete();
                }
            }
        }
        );
    } // constructor
    
    /**
     * Registers an object to listen for position changes by this element.
     * @param pl The object to register for position change event notification.
     */
    public void addPositionListener(PositionListener pl) {
        positionListeners.add(pl);
    } // method addPositionListener
    
    /**
     * Un-registers an object currently listening for position change events.
     * @param pl The object to remove from the notification list for position changes.
     */
    public void removePositionListener(PositionListener pl) {
        positionListeners.remove(pl);
    } // method removePositionListener
    
    /**
     * Changes the x-coordinate position and causes the position listeners to be
     * notified.
     * @param value The new x-coordinate position of this element.
     */
    public void setX(int value) {
        super.setX(value);
        for (Iterator i = positionListeners.iterator(); i.hasNext();) {
            PositionListener pl = (PositionListener) i.next();
            pl.updateX(this, value);
        }
    } // method setX
    
    /**
     * Changes the y-coordinate position and causes the position listeners to be
     * notified.
     * @param value The new y-coordinate position of this element.
     */
    public void setY(int value) {
        super.setY(value);
        for (Iterator i = positionListeners.iterator(); i.hasNext();) {
            PositionListener pl = (PositionListener) i.next();
            pl.updateY(this, value);
        }
    } // method setY
    
    /**
     * Changes the value of the <CODE>order</CODE> attribute and notifies the position
     * change listeners.
     * @param value The new value for the z-order of this element.
     */
    public void setOrder(int value) {
        super.setOrder(value);
        for (Iterator i = positionListeners.iterator(); i.hasNext();) {
            PositionListener pl = (PositionListener) i.next();
            pl.updateOrder(this, value);
        }
    } // method setOrder
    
    /** Returns the connection within the relationship set to the given object set or
     * null if one does not exist.
     * @return the corresponding connection
     * @param os the object set to search for
     */
    public OSMXRelSetConnectionType getConnectionFor(OSMXObjectSetType os) {
        List connections = getRelSetConnection();
        if(os==null)
            return null;
        for (Iterator c_it = connections.iterator(); c_it.hasNext(); ) {
            OSMXRelSetConnectionType c = (OSMXRelSetConnectionType)c_it.next();
            if (c.getObjectSet().compareTo(os.getId())==0)
                return c;
        }
        return null;
    }
    
    /**
     * Indicates whether this element has explicitly been given an x and y coordinate
     * position.
     * @return <CODE>true</CODE> if both the <CODE>x</CODE> and <CODE>y</CODE> coordinate
     * attributes have been set; <CODE>false</CODE> otherwise.
     */
    public boolean isSetPosition() {
        return (isSetX() && isSetY());
    }
    
    /**
     * Returns the name of the element, if present; otherwise, returns the ID of the
     * element.
     * @return Either the name or the ID of the element.
     */
    public String getPrimaryName() {
        String result = this.getId();
        PositionedText name = getName();
        if (name == null) return result;
        return name.getContent();
    }
    
    /**
     * Represents the element as a string.
     * @return The string returned from {@link #getPrimaryName}.
     */
    public String toString() {
        return getPrimaryName();
    } // method toString
    
    /**
     * Overrides the inherited method in order to pass the call recursively down to its
     * child elements.
     * @param elem The new parent of this element.
     */
    void setParent(OSMXElement elem) {
        parent = elem;
        addToParentDocument();
        this.setAsParentOf(getRelSetConnection());
        this.setAsParentOf(getCoOccurrenceConstraint());
        this.setAsParentOf((OSMXElement) getStyle());
        this.setAsParentOf((OSMXElement) getOSM());
        this.setAsParentOf((OSMXElement) getName());
        initWorkingInstance();
    }
    
    /**
     * Caches the current set of relationships that belong to this relationship set and
     * that are stored within data instances in the ontology.  This provides a way to
     * access the relationships via the relationship set.
     */
    protected void initWorkingInstance() {
        if (getParentDocument() == null) return;
        for (Iterator i = getParentDocument().getElementsOfType(OSMXDataInstanceType.class).iterator(); i.hasNext();) {
            OSMXDataInstanceType di = (OSMXDataInstanceType) i.next();
            if (di.getDataInstanceElements() == null) continue;
            loadMemberRelationships(di.getDataInstanceElements());
        }
    }
    
    /**
     * Caches the relationships from the ontology's data instances that belong to this
     * relationship set, so that they can be accessed via this element.
     * @param dataInstanceElements The data instance elements in the ontology.
     */
    protected void loadMemberRelationships(List dataInstanceElements) {
        for (Iterator i = dataInstanceElements.iterator(); i.hasNext();) {
            Object o = i.next();
            if (o instanceof OSMXDataInstanceType) {
                OSMXDataInstanceType di = (OSMXDataInstanceType) o;
                if (di.getDataInstanceElements() == null) continue;
                loadMemberRelationships(di.getDataInstanceElements());
            } else if (o instanceof OSMXRelationshipType) {
                OSMXRelationshipType rel = (OSMXRelationshipType) o;
                if (rel.getRelationshipSet() == this.getId()) workingInstance.add(rel);
                for (Iterator j = rel.getObjectBinding().iterator(); j.hasNext();) {
                    OSMXObjectBindingType binding = (OSMXObjectBindingType) j.next();
                    String connId = binding.getConnectionRef();
                    OSMXElement elem = getParentDocument().getElement(connId);
                    OSMXRelSetConnectionType conn = (OSMXRelSetConnectionType) elem;
                    conn.addObjectBinding(binding);
                }
            }
        }
    }
    
    /**
     * Gives this element a new <CODE>Style</CODE> element and notifies listeners of a
     * style change.
     * @param value The new <CODE>Style</CODE> element.
     */
    public void setStyle(StyleType value) {
        StyleType oldVal = getStyle();
        this.replaceChild((OSMXElement) getStyle(), (OSMXElement) value);
        super.setStyle(value);
        if (value instanceof OSMXStyleType) {
            if (styleListener == null) {
                styleListener = new PropertyChangeListener() {
                    public void propertyChange(PropertyChangeEvent evt) {
                        firePropertyChange(OSMXStyleType.STYLE_PROPERTY,
                        evt.getOldValue(), evt.getNewValue());
                    }
                };
            }
            ((OSMXStyleType)value).addPropertyChangeListener(styleListener);
        }
        firePropertyChange(OSMXStyleType.STYLE_PROPERTY, oldVal, value);
    }
  
  /**
   * Assigns a sub-model to the relationship set.  This does not depend on or affect the
   * state of the relationship set's <CODE>highLevel</CODE> attribute.
   * @param value The <CODE>OSM</CODE> element that is the root of the sub-model to add to the
   * relationship set.
   */  
    public void setOSM(OSMType value) {
        this.replaceChild((OSMXElement) getOSM(), (OSMXElement) value);
        super.setOSM(value);
    }
    
    /**
     * Stores the given <CODE>PositionedText</CODE> element as the bearer of the
     * relationship set's name.
     * @param value The <CODE>PositionedText</CODE> element to store.
     */
    public void setName(PositionedText value) {
        this.replaceChild((OSMXElement) getName(), (OSMXElement) value);
        super.setName(value);
    }
    
    /**
     * Sets the relationship set's name to the given value.
     * @param value The new name for the relationship set.
     */
    public void setName(String value) {
        PositionedText nameElem = getName();
        if (nameElem == null) {
            nameElem = new OSMXPositionedText();
            setName(nameElem);
        }
        nameElem.setContent(value);
    }
  
  /**
   * Returns the set of relationships that currently belong to this relationship set.
   * @return A <CODE>Set</CODE> of <CODE>Relationship</CODE> elements.
   */  
    public Set getWorkingInstance() {
        return workingInstance;
    }
    
    /**
     * Returns the number of connections that belong to this relationship set.
     * @return The arity of the relationship set.
     */
    public int getArity() {
        return getRelSetConnection().size();
    }
    
}
