/*
 * OSMXStyleType.java
 *
 * Created on June 11, 2003, 11:43 AM
 */

package edu.byu.deg.osmx;

import edu.byu.deg.osmx.binding.LineType;
import edu.byu.deg.osmx.binding.FontType;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.awt.Color;

/**
 * Represents a <CODE>Style</CODE> element in OSMX.
 * @author Alan Wessman
 */
public class OSMXStyleType extends edu.byu.deg.osmx.binding.impl.StyleTypeImpl {
    
    public static final String STYLE_PROPERTY = "style";
    
    protected PropertyChangeListener propertyListener;
    
    /**
     * Creates a new instance of OSMXStyleType
     */
    public OSMXStyleType() {
        super();
        propertyListener = new PropertyChangeListener() {
            public void propertyChange(PropertyChangeEvent evt) {
                Object src = evt.getSource();
                //        String propName = evt.getPropertyName();
                if (src instanceof OSMXLineType || src instanceof OSMXFontType) {
                    firePropertyChange(STYLE_PROPERTY, evt.getOldValue(), evt.getNewValue());
                }
            }
        };
    }
    
    /**
     * Responds to notification of the deletion of an external element (one that isn't
     * <CODE>this</CODE>).  For this class, the deleted element is most likely to be a
     * child element such as <CODE>Line</CODE> or <CODE>Font</CODE>.
     * @param deletedElement The element that was deleted.
     */
    protected void handleDeletionEvent(OSMXElement deletedElement) {
        deletedElement.removePropertyChangeListener(propertyListener);
        super.handleDeletionEvent(deletedElement);
    }
    
    /** Replaces an element that is a child of <CODE>this</CODE> with another element.
     * When this method has finished, {@link OSMXElement#getParent} returns <CODE>null</CODE> for
     * the old child element and returns <CODE>this</CODE> for the new element.
     * @param oldChild The child element to replace.
     * @param newChild The element to set in place of the former child element.
     */
    protected void replaceChild(OSMXElement oldChild, OSMXElement newChild) {
        if (oldChild != null) {
            oldChild.removeElementDeletionListener(deleteListener);
            oldChild.removePropertyChangeListener(propertyListener);
        }
        super.replaceChild(oldChild, newChild);
        if (newChild != null) {
            newChild.addPropertyChangeListener(propertyListener);
            newChild.addElementDeletionListener(deleteListener);
        }
    }
    
    /**
     * Replaces the current <CODE>Line</CODE> child element with the given one.
     * @param value The <CODE>Line</CODE> element to set as child of <CODE>this</CODE>.
     */
    public void setLine(LineType value) {
        OSMXLineType oldVal = (OSMXLineType) getLine();
        replaceChild(oldVal, (OSMXElement) value);
        super.setLine(value);
        firePropertyChange(STYLE_PROPERTY, oldVal, value);
    }
    
    /**
     * Replaces the current <CODE>Font</CODE> element with the given one.
     * @param value The <CODE>Font</CODE> element to set as child of <CODE>this</CODE>.
     */
    public void setFont(FontType value) {
        OSMXFontType oldVal = (OSMXFontType) getFont();
        replaceChild(oldVal, (OSMXElement) value);
        super.setFont(value);
        firePropertyChange(STYLE_PROPERTY, oldVal, value);
        
    }
    
    /** Links this element to its parent element so that upward traversal of the object
     * tree may be possible.  JAXB does not natively create links from child to parent
     * so this method must be invoked before child<PRE>-></PRE>parent links may be
     * traversed.
     * <br>
     * The method also invokes {@link OSMXElement#addToParentDocument}.
     * @param elem The element to which this element belongs.
     */
    public void setParent(OSMXElement elem) {
        parent = elem;
        addToParentDocument();
        this.setAsParentOf((OSMXElement) getLine());
        if (getLine() != null) {
            ((OSMXElement) getLine()).addElementDeletionListener(deleteListener);
            ((OSMXElement) getLine()).addPropertyChangeListener(propertyListener);
        }
        this.setAsParentOf((OSMXElement) getFont());
        if (getFont() != null) {
            ((OSMXElement) getFont()).addElementDeletionListener(deleteListener);
            ((OSMXElement) getFont()).addPropertyChangeListener(propertyListener);
        }
    }
    
    /**
     * Convenience method for accessing the <CODE>width</CODE> attribute of the child
     * <CODE>Line</CODE> element.
     * @return The current line width for the style element, or 0.0 if no line width has been
     * set.
     */
    public float getLineWidth() {
        if (isSetLine())
            return getLine().getWidth();
        else return 0.0f;
    }
    
    /**
     * Convenience method for setting the <CODE>width</CODE> attribute of the child
     * <CODE>Line</CODE> element.
     * @param width The desired line width for the style element.
     */
    public void setLineWidth(float width) {
        if (!isSetLine()) {
            setLine(new OSMXLineType());
        }
        getLine().setWidth(width);
    }
    
    /**
     * Convenience method for accessing the <CODE>color</CODE> attribute of the child
     * <CODE>Line</CODE> element.
     * @return The current line color for the style element, or <CODE>Color.BLACK</CODE> if no
     * line color has been set.
     */
    public java.awt.Color getLineColor() {
        if (isSetLine())
            return getLine().getColor();
        else return Color.BLACK;
    }
    
    /**
     * Convenience method for setting the <CODE>color</CODE> attribute of the child
     * <CODE>Line</CODE> element.
     * @param c The desired line color for the style element.
     */
    public void setLineColor(java.awt.Color c) {
        Color oldVal = this.getLineColor();
        if (!isSetLine()) {
            setLine(new OSMXLineType());
        }
        getLine().setColor(c);
        firePropertyChange(STYLE_PROPERTY, oldVal, c);
    }
    
    /**
     * Convenience method for accessing the <CODE>family</CODE> attribute of the child
     * <CODE>Font</CODE> element.
     * @return The current font family for the style element, or <CODE>sans-serif</CODE> if no
     * font family has been set.
     */
    public String getFontFamily() {
        if (isSetFont())
            return getFont().getFamily();
        else return "sans-serif";
    }
    
    /**
     * Convenience method for setting the <CODE>family</CODE> attribute of the child
     * <CODE>Line</CODE> element.
     * @param value The desired font family for the style element.
     */
    public void setFontFamily(String value) {
        String oldVal = this.getFontFamily();
        if (!isSetFont()) {
            setFont(new OSMXFontType());
        }
        getFont().setFamily(value);
        firePropertyChange(STYLE_PROPERTY, oldVal, value);
    }
    
    /**
     * Convenience method for accessing the <CODE>size</CODE> attribute of the child
     * <CODE>Font</CODE> element.
     * @return The current font size for the style element, or <CODE>12.0</CODE> if no
     * font size has been set.
     */
    public float getFontSize() {
        if (isSetFont())
            return getFont().getSize();
        else return 12.0f;
    }
    
    /**
     * Convenience method for setting the <CODE>size</CODE> attribute of the child
     * <CODE>Line</CODE> element.
     * @param value The desired font size for the style element.
     */
    public void setFontSize(float value) {
        if (!isSetFont()) {
            setFont(new OSMXFontType());
        }
        getFont().setSize(value);
    }
    
    /**
     * Convenience method for accessing the <CODE>textStyle</CODE> attribute of the child
     * <CODE>Font</CODE> element.
     * @return The current text style for the style element, or <CODE>null</CODE> if no
     * text style has been set.
     */
    public String getFontStyle() {
        if (isSetFont())
            return getFont().getTextStyle();
        else return null;
    }
    
    /**
     * Convenience method for setting the <CODE>textStyle</CODE> attribute of the child
     * <CODE>Line</CODE> element.
     * @param value The desired font style for the style element.
     */
    public void setFontStyle(String value) {
        String oldVal = this.getFontStyle();
        if (!isSetFont()) {
            setFont(new OSMXFontType());
        }
        getFont().setTextStyle(value);
        firePropertyChange(STYLE_PROPERTY, oldVal, value);
    }
    
    /**
     * Convenience method for accessing the <CODE>color</CODE> attribute of the child
     * <CODE>Font</CODE> element.
     * @return The current font color for the style element, or <CODE>Color.BLACK</CODE> if no
     * font color has been set.
     */
    public Color getFontColor() {
        if (isSetFont())
            return getFont().getColor();
        else return Color.BLACK;
    }
    
    /**
     * Convenience method for setting the <CODE>color</CODE> attribute of the child
     * <CODE>Line</CODE> element.
     * @param value The desired font color for the style element.
     */
    public void setFontColor(Color value) {
        Color oldVal = this.getFontColor();
        if (!isSetFont()) {
            setFont(new OSMXFontType());
        }
        getFont().setColor(value);
        firePropertyChange(STYLE_PROPERTY,oldVal,value);
    }
    
    /**
     * Sets the fill color for this style element.
     * @param value The desired fill color for the style element.
     */
    public void setFillColor(Color value) {
        Color oldVal = getFillColor();
        super.setFillColor(value);
        firePropertyChange(STYLE_PROPERTY, oldVal, value);
    }
    
    /**
     * Clears the current setting for the <CODE>fillColor</CODE> attribute.
     */
    public void unsetFillColor() {
        Color oldVal = getFillColor();
        super.unsetFillColor();
        firePropertyChange(STYLE_PROPERTY, oldVal, null);
    }
    
}
