/*
 * OSMXTypeSpecification.java
 *
 * Created on June 11, 2003, 11:43 AM
 */

package edu.byu.deg.osmx;

import edu.byu.deg.osmx.binding.*;
import javax.xml.bind.JAXBException;

/**
 * Represents a <CODE>TypeSpecification</CODE> element in OSMX.
 * @author Alan Wessman
 */
public class OSMXTypeSpecification extends edu.byu.deg.osmx.binding.impl.TypeSpecificationImpl {
    
  public static final String TYPE_CHANGE_PROPERTY = "typeChanged";
  
  public static final int DATA_TYPE = 0;
  public static final int OBJECT_SET_REF = 1;
  
  private DataTypeType storedDataType;
  private ObjectSetReferenceType storedObjectSetRef;
  
  private boolean isSwitch = false;
  
  /** Creates a new instance of OSMXTypeSpecification */
  public OSMXTypeSpecification() {
    if (!isSetObjectSetReference() && !isSetDataType()) {
      try {
        setDataType(createDataTypeType());
        getDataType().setTypeName("int");
      } catch (JAXBException e) {
        e.printStackTrace();
      }
    }
  }
  
  private DataTypeType createDataTypeType() throws JAXBException {
    OSMXDocument doc = getParentDocument();
    ObjectFactory factory;
    if (doc != null) {
      factory = doc.getObjectFactory();
    } else {
      factory = new ObjectFactory();
    }
    return factory.createDataTypeType();
  }
  
  private ObjectSetReferenceType createObjectSetReferenceType() throws JAXBException {
    OSMXDocument doc = getParentDocument();
    ObjectFactory factory;
    if (doc != null) {
      factory = doc.getObjectFactory();
    } else {
      factory = new ObjectFactory();
    }
    return factory.createObjectSetReferenceType();
  }
  
  /**
   * Toggles the underlying data type between being defined as a primitive type with
   * possible units of measure defined, or a complex type represented by an object
   * set in the ontology.
   * @param newType The constant that represents the new type to switch to: either <CODE>DATA_TYPE</CODE>
   * (for primitive types) or <CODE>OBJECT_SET_REF</CODE>.
   */  
  public void switchType(int newType) {
    isSwitch = true;
    boolean didSwitch = false;
    try {
      if (this.isSetObjectSetReference() && newType == DATA_TYPE) {
        storedObjectSetRef = this.getObjectSetReference();
        this.setObjectSetReference(null);
        if (storedDataType == null) {
          storedDataType = createDataTypeType();
          storedDataType.setTypeName("int");
        }
        this.setDataType(storedDataType);
        didSwitch = true;
      } else if (this.isSetDataType() && newType == OBJECT_SET_REF) {
        storedDataType = this.getDataType();
        this.setDataType(null);
        if (storedObjectSetRef == null) {
          storedObjectSetRef = createObjectSetReferenceType();
        }
        this.setObjectSetReference(storedObjectSetRef);
        didSwitch = true;
      }
    } catch (JAXBException e) {
      e.printStackTrace();
    }
    isSwitch = false;
    if (didSwitch) firePropertyChange(TYPE_CHANGE_PROPERTY, null, null);
  }
  
  /**
   * Stores an <CODE>ObjectSetReference</CODE> element as the representation of the
   * underlying data type for this element.
   * @param value The <CODE>ObjectSetReference</CODE> to store.
   */  
  public void setObjectSetReference(ObjectSetReferenceType value) {
    ObjectSetReferenceType oldVal = getObjectSetReference();
    if (value == oldVal) return;
    this.replaceChild((OSMXElement) oldVal, (OSMXElement) value);
    super.setObjectSetReference(value);
    if (!isSwitch) {
      firePropertyChange(TYPE_CHANGE_PROPERTY, oldVal, value);
    }
  }
  
  /**
   * Stores a <CODE>DataType</CODE> element as a representation of the underlying
   * data type for this element.
   * @param value The <CODE>DataType</CODE> element to store.
   */  
  public void setDataType(DataTypeType value) {
    DataTypeType oldVal = getDataType();
    if (value == oldVal) return;
    this.replaceChild((OSMXElement) oldVal, (OSMXElement) value);
    super.setDataType(value);
    if (!isSwitch) {
      firePropertyChange(TYPE_CHANGE_PROPERTY, oldVal, value);
    }
  }
  
  /**
   * Represents this type specification as a string.
   * @return A string in one of two possible formats: if the underlying type is primitive,
   * the format is <CODE>type <i>&lt;units&gt;</CODE>; if it is an object set
   * reference, it is <CODE>object-set-ID</CODE>.
   */  
  public String toString() {
    if (isSetDataType()) {
      DataTypeType dt = getDataType();
      String result = dt.getTypeName();
      String units = dt.getUnitOfMeasure();
      if (units != null && units.length() > 0) {
        result += "<" + units + ">";
      }
      return result;
    } else if (isSetObjectSetReference()) {
      ObjectSetReferenceType objref = getObjectSetReference();
      String result = objref.getIdref();
      return result;
    } else {
      return null;
    }
  }
  
  /**
   * Convenience method for setting the underlying type to reference a particular
   * object set.
   * @param objSetRefId The ID of the object set to establish as the underlying data type for this type
   * specification.
   */  
  public void setObjectSetRefId(String objSetRefId) {
    switchType(OBJECT_SET_REF);
    getObjectSetReference().setIdref(objSetRefId);
  }
  
  /**
   * Convenience method for setting the underlying data type to a primitive type and
   * specifying the name of the primitive type to use.
   * @param dataTypeName The name of the primitive type; e.g., <CODE>String</CODE> or <CODE>int</CODE>.
   */  
  public void setDataTypeName(String dataTypeName) {
    switchType(DATA_TYPE);
    getDataType().setTypeName(dataTypeName);
  }
  
  /**
   * Convenience method for setting the underlying data type to a primitive type and
   * specifying the units of measure for the data type.
   * @param dataTypeUnits The units of measure for the primitive type.  Example: <CODE>miles per
   * hour</CODE>.
   */  
  public void setDataTypeUnits(String dataTypeUnits) {
    switchType(DATA_TYPE);
    getDataType().setUnitOfMeasure(dataTypeUnits);
  }
  
  /**
   * Accessor for the ID of the referenced object set.
   * @return The ID of the referenced object set, or <CODE>null</CODE> if the underlying data
   * type is a primitive type.
   */  
  public String getObjectSetRefId() {
    if (!isSetObjectSetReference()) return null;
    return getObjectSetReference().getIdref();
  }
  
  /**
   * Accessor for the name of the underlying primitive type.
   * @return The name of the underlying primitive type, or <CODE>null</CODE> if the
   * underlying type is not primitive.
   */  
  public String getDataTypeName() {
    if (!isSetDataType()) return null;
    return getDataType().getTypeName();
  }
  
  /**
   * Accessor for the units of measure for the underlying primitive type.
   * @return The units of measure for the underlying primitive type, or <CODE>null</CODE> if
   * the underlying type is not primitive.
   */  
  public String getDataTypeUnits() {
    if (!isSetDataType()) return null;
    return getDataType().getUnitOfMeasure();
  }
  
  /**
   * Overrides the inherited method in order to pass the call recursively down to its
   * child elements.
   * @param elem The new parent of this element.
   */  
  void setParent(OSMXElement elem) {
    parent = elem;
    addToParentDocument();
    this.setAsParentOf((OSMXElement) getObjectSetReference());
    this.setAsParentOf((OSMXElement) getDataType());
  }
  
}
