/*
 * DOMString.java
 *
 * Created on August 4, 2003, 2:30 PM
 */

package edu.byu.deg.util;

import org.w3c.dom.Node;
import org.w3c.dom.DOMException;


import java.util.ArrayList;
import java.lang.StringBuffer;

/** Wraps a dom tree so it can be matched with regular expressions like a string.
 * Still needs to implement getXPathFor() method.
 * @author troywalk
 */
public class DOMString implements CharSequence {
  int[] offsets;
  ArrayList nodes;
  ArrayList strings;
  int length;
  int nodesIndexed;
  int remainder;
  int lastIndex;
  int baseOffset;
  String lastString;

  /** Creates a new instance of DOMString
   * @param rootNode The DOM tree root to wrap
   */
  public DOMString(Node rootNode) {
    length = 0;
    nodesIndexed = 0;
    offsets = new int[1];
    remainder = -1;
    nodes = new ArrayList();
    strings = new ArrayList();
    lastString = "";
    lastIndex = 0;
    baseOffset = 0;
    calculateOffsets(rootNode, true);
  }
  private DOMString(DOMString domstr, int baseOffset, int length){
    this.length = length;
    nodesIndexed = domstr.nodesIndexed;
    offsets = domstr.offsets;
    remainder = -1;
    nodes = domstr.nodes;
    strings = domstr.strings;
    lastString = "";
    lastIndex = 0;
    this.baseOffset = baseOffset;
  }
  /**
   * Returns the starting character offset of the given DOM node.
   * @param node A node in the DOM tree.
   * @return The 0-based starting offset of the node in question,
   * relative to the start of the document.
   */  
  public int getStartOf(Node node){
    for(int i=0;i<nodes.size();i++){
      if(node==nodes.get(i)){
        return offsets[i];
      }
    }
    return -1;
  }
  /**
   * Returns the ending character offset of the given DOM node.
   * @param node A node in the DOM tree.
   * @return The 0-based ending offset of the node in question,
   * relative to the start of the document.
   */  
  public int getEndOf(Node node){
    for(int i=0;i<nodes.size();i++){
      if(node==nodes.get(i)){
        if(i+1==offsets.length)
          return length;
        return offsets[i+1];
      }
    }
    return -1;
  }

  private void insert(Node node, String str){
    if(str.length()==0)
      return;
    nodes.add(node);
    strings.add(str);
    nodesIndexed++;
    //resize offsets if it needs more room
    if(offsets.length < nodesIndexed){
      int[] newoffs = new int[offsets.length*2];
      System.arraycopy(offsets, 0, newoffs, 0, offsets.length);
      offsets = newoffs;
    }
    //put the offset into offsets
    offsets[nodesIndexed-1] = length;
    length = length + str.length();
  }

  private String getText(Node node){
    String str;
    try{
      str = node.getNodeValue();
    } catch(DOMException e){ System.err.println("This should be impossible");str = ""; }
    return str;
  }

  private void calculateOffsets(Node node, boolean isroot){
    if(node.getNodeType()==Node.TEXT_NODE){
      insert(node, getText(node));
    }
    TagInfo tag = TagInfo.getTagForNode(node);
    if(tag.getStartSpace()!=TagInfo.BL&&nodesIndexed!=0)
      insert((Node)nodes.get(nodesIndexed-1), tag.getStartSpace());
    Node next = node.getFirstChild();
    if(next!=null)
      calculateOffsets(next, false);
    if(tag.getEndSpace()!=TagInfo.BL&&nodesIndexed>0)
      insert((Node)nodes.get(nodesIndexed-1), tag.getEndSpace());
    next = node.getNextSibling();
    if(next!=null&&!isroot)
      calculateOffsets(next, false);
  }

  private int getIndexOf(int offset){
    int begin = 0;
    int end = nodesIndexed;
    int current = 0;
    //if(offset==length)
    //  return nodesIndexed-1;
    while(begin!=end){
      current = (begin + end) / 2;
      if(offsets[current]==offset)
        return current;
      if(offsets[current]<offset)
        begin = current + 1;
      else
        end = current;
    }
    if(offsets[current]<offset)
      return current;
    else
      return current - 1;
  }

  
  
  /** Method to retrieve a character from the virtual string
   * @param offset offset of the char to return
   * @return the char at the given offset
   */
  public char charAt(int offset) {
    int index;
    String str;
    offset = offset + baseOffset;
    if((offsets[lastIndex]<offset)&&((offsets[lastIndex]+lastString.length()) > offset)){
      index = lastIndex;
      str = lastString;
    } else {
      index = lastIndex = getIndexOf(offset);
      str = lastString = (String)strings.get(index);
    }
    return str.charAt(offset - offsets[index]);
  }

  /** Method to return the length of the virtual string
   * @return the length of the virtual string contained in the <CODE>DOMString</CODE>
   */
  public int length() {
    return length;
  }

  /** Returns a DOMString representing the given subsequence.
   * @param start starting offset of the new <CODE>DOMString</CODE>
   * @param end ending offset of the new <CODE>DOMString</CODE>
   * @return A new <CODE>DOMString</CODE> containing the given subsequence
   */
  public CharSequence subSequence(int start, int end) {
    return subString(start, end);
  }

  /** Returns a DOMString representing the given subsequence.
   * @param start starting offset of the new <CODE>DOMString</CODE>
   * @param end ending offset of the new <CODE>DOMString</CODE>
   * @return A new <CODE>DOMString</CODE> containing the given subsequence
   */
  public DOMString subString(int start, int end){
    return new DOMString(this, baseOffset+start, end-start);
  }

  /** Returns a <CODE>DOMString</CODE> beginning at the given offset and continuing to
   * the end of this <CODE>DOMString</CODE>
   * @param start starting offset of the new <CODE>DOMString</CODE>
   * @return A new <CODE>DOMString</CODE> containing the given subsequence
   */
  public DOMString subString(int start) {
    return new DOMString(this, baseOffset+start, length-start);
  }

  /** Returns the <CODE>Node</CODE> which contains the text at the given offset.
   * @param offset the offset
   * @return The <CODE>Node</CODE>
   */
  public Node getNodeFor(int offset){
    int index;
    offset = offset + baseOffset;
    if((offsets[lastIndex]<offset)&&((offsets[lastIndex]+lastString.length()) > offset)){
      index = lastIndex;
    } else {
      index = lastIndex = getIndexOf(offset);
      lastString = (String)strings.get(index);
    }
    remainder = offset - offsets[index];
    return (Node)nodes.get(index);
  }
  /** Calculates the XPath of the node at the given offset.  Not yet implemented.
   * @param offset the offset
   * @return A String representing the location of the offset in the DOM tree
   */
  public String getXPathFor(int offset){
//    Node node = getNodeFor(offset);
    //not yet implemented
    //need to do stuff
    return "";
  }
  /** Returns the index of the offset within the <CODE>String</CODE> or <CODE>Node</CODE>
   * returned by the last {@link #getNodeFor(int)} or {@link #getStringAt(int)} method
   * call.
   * @return Returns the last remainder
   */
  public int getLastRemainder(){
    return remainder;
  }

  /** Returns the <CODE>String</CODE> which contains the text at the given offset.
   * @param offset the offset
   * @return the string that contains the text at the offset
   */
  public String getStringAt(int offset){
    offset = offset + baseOffset;
    if((offsets[lastIndex]<offset)&&((offsets[lastIndex]+lastString.length()) > offset)){
      remainder = offset - offsets[lastIndex];
      return lastString;
    } else {
      lastIndex = getIndexOf(offset);
      remainder = offset - offsets[lastIndex];
      return lastString = (String)strings.get(lastIndex);
    }
  }
  /** Returns a string representation of the contents of the <CODE>DOMString</CODE>
   * @return the string
   */
  public String toString(){
    int startindex = getIndexOf(baseOffset);
    if(startindex==-1)
      return "";
    String str = (String)strings.get(startindex);
    int endindex = getIndexOf(baseOffset + length);

    if(startindex==endindex){
      return str.substring(baseOffset - offsets[startindex], baseOffset - offsets[startindex] + length);
    }
    StringBuffer b = new StringBuffer();
    str = str.substring(baseOffset - offsets[startindex]);
    b.append(str);
    for(int i=startindex+1;i<endindex;i++){
      b.append( (String) strings.get(i));
    }
    str = (String)strings.get(endindex);
    str = str.substring(0, length + baseOffset - offsets[endindex]);
    b.append(str);
    return b.toString();
  }
  
  /**
   * Returns the earliest DOM node in the document, or <CODE>null</CODE> if the
   * document is empty.
   * @return The first node in the document.
   */  
  public Node getFirstText(){
    if(this.length==0)
      return null;
    return (Node)nodes.get(0); 
  }
  
  /**
   *
   * @return The start of the last string relative to the base offset.
   */  
  public int getRelativeStartOfLastString(){
    //if(lastIndex==getIndexOf(baseOffset))
    //  return baseOffset - offsets[lastIndex];
    return offsets[lastIndex] - baseOffset;
  }

}
